package jp.takayuki.graphics;

/**
 * <p>Device Independent Bitmap Image.</p>
 * 
 * 
 * @author Takayuki Shoji
 */
public class DibImage {
	/**
	 * t@Cwb_(null̏ꍇ)
	 */
	private FileHeader fileHeader;

	/**
	 * wb_
	 */
	private InfoHeader infoHeader;

	/**
	 * pbg(null̏ꍇ)
	 */
	private Pallet pallet;

	/**
	 * 摜f[^
	 */
	byte[] imageData;

	/**
	 * oCĩRXgN^
	 * 
	 * @param bytes
	 */
	public DibImage(byte[] bytes) {
		boolean fromClipboard = !hasFileHeader(bytes);
		if (!fromClipboard) {
			this.fileHeader = new FileHeader(bytes);
		}

		this.infoHeader = new InfoHeader(bytes, fromClipboard);

		if (infoHeader.getPalletSize() > 0) {
			// pbg̓ǂݍ
			this.pallet = new Pallet(bytes, fromClipboard, infoHeader
					.getPalletSize());
		}

		//System.out.println("摜̃TCY(␳O):" + infoHeader.getImageDataSize());

		// wb_TCY
		int offset = fromClipboard ? 40 : 54;
		// kݒ̃TCY
		//offset += infoHeader.getCompressionSettingSize();
		// pbg̃TCY
		if (pallet != null) {
			offset += pallet.getBytes().length;
		}

		// 摜TCY߂ČvZ
		int imageSize = bytes.length - offset;
		infoHeader.setImageDataSize(bytes.length - offset
				- infoHeader.getCompressionSettingSize());

		this.imageData = new byte[imageSize];
		System.arraycopy(bytes, offset, imageData, 0, imageSize);
	}

	/**
	 * oCiɕϊ
	 * 
	 * @param toClipboard
	 *            Nbv{[hpȂtrue
	 * @return oCi
	 */
	public byte[] getBytes(boolean toClipboard) {

		byte[] fileHeaderBytes, infoHeaderBytes, palletBytes;

		// wb_
		infoHeaderBytes = infoHeader.getBytes();

		// pbg
		if (pallet == null) {
			palletBytes = null;
		} else {
			palletBytes = pallet.getBytes();
		}

		// t@Cwb_
		if (!toClipboard) {
			if (fileHeader != null) {
				fileHeaderBytes = fileHeader.getBytes();
			} else {
				// ItZbgƃt@CTCYvZ
				FileHeader h = new FileHeader();
				h.offset = 14 + 40
						+ (palletBytes == null ? 0 : palletBytes.length)
						+ infoHeader.getCompressionSettingSize();
				h.fileSize = h.offset + imageData.length;
				fileHeaderBytes = h.getBytes();
			}
		} else {
			fileHeaderBytes = null;
		}

		// t@CTCYvZ
		int fileSize = 0;
		if (fileHeaderBytes != null) {
			fileSize += 14;
		}
		fileSize += 40;

		// pbg
		if (palletBytes != null) {
			fileSize += palletBytes.length;
		}
		// 摜f[^
		fileSize += imageData.length;

		// f[^oCiɂ
		byte[] result = new byte[fileSize];
		int pos = 0;
		// t@Cwb_
		if (fileHeaderBytes != null) {
			System.arraycopy(fileHeaderBytes, 0, result, 0, 14);
			pos += 14;
		}
		// wb_
		System.arraycopy(infoHeaderBytes, 0, result, pos, 40);
		pos += 40;

		// pbg
		if (palletBytes != null) {
			System.arraycopy(palletBytes, 0, result, pos, palletBytes.length);
			pos += palletBytes.length;
		}

		// 摜f[^
		System.arraycopy(imageData, 0, result, pos, imageData.length);

		return result;
	}

	/**
	 * @return 
	 */
	public int getWidth() {
		return byteToInt(infoHeader.bytes, 4);
	}

	/**
	 * @return 
	 */
	public int getHeight() {
		return byteToInt(infoHeader.bytes, 8);
	}

	/**
	 * t@Cwb_ꍇtrueԂ܂B
	 * 
	 * @param bytes oCif[^
	 * @return t@Cwb_ꍇtrue
	 */
	private static boolean hasFileHeader(byte[] bytes) {
		if (bytes == null || bytes.length < 54) {
			return false;
		}
		if (bytes[0] == 'B' && bytes[1] == 'M') {
			return true;
		}
		return false;
	}

	/**
	 * gGfBAbyte[4]intɕϊ܂B
	 * 
	 * @param integer
	 *            4oCgȏbytez
	 * @param start
	 *            ϊJnintegerz񒆂̈ʒu
	 * @return ϊꂽint^̐l
	 */
	private static int byteToInt(byte[] integer, int start) {
		int result = 0;
		//System.out.println("start:"+start);
		for (int i = start + 3; i > start; i--) {
			//System.out.println("i:"+i+" integer[i]:"+integer[i]);
			result += unsignedByteToInt(integer[i]);
			result = result << 8;
		}
		result += unsignedByteToInt(integer[start]);
		//System.out.println("result:"+result+",
		// integer["+start+"]:"+integer[start]);
		return result;
	}

	/**
	 * byte𕄍ȂƂ݂Ȃintɂ܂B
	 * @param b ϊΏ
	 * @return ϊ
	 */
	private static int unsignedByteToInt(byte b) {
		int result = b;
		if (result < 0) {
			result += 256;
		}
		return result;
	}

	/**
	 * t@Cwb_
	 */
	private static class FileHeader {
		/**
		 * t@CTCY
		 */
		long fileSize;

		/**
		 * t@C̐擪摜f[^܂ł̃ItZbg
		 */
		long offset;

		/** ̃RXgN^ */
		FileHeader() {
			//
		}

		/**
		 * oCĩRXgN^ 
		 * @param bytes oCif[^ 
		 */
		FileHeader(byte[] bytes) {
			fileSize = byteToInt(bytes, 2);
			offset = byteToInt(bytes, 10);
			// E\Ȃ̂
			if (fileSize != bytes.length) {
				//System.out.println("t@CTCY␳O=" + fileSize + ", ␳="
				//        + bytes.length);
				fileSize = bytes.length;
			}
		}

		/**
		 * oCiɕϊ܂B
		 * @return oCif[^
		 */
		byte[] getBytes() {
			byte[] result = new byte[14];
			// t@C^Cv
			result[0] = 'B';
			result[1] = 'M';
			// TCY
			result[2] = (byte) (fileSize & 0x000000ff);
			result[3] = (byte) ((fileSize & 0x0000ff00) >> 8);
			result[4] = (byte) ((fileSize & 0x00ff0000) >> 16);
			result[5] = (byte) ((fileSize & 0xff000000) >> 24);

			// ItZbg
			result[10] = (byte) (offset & 0x000000ff);
			result[11] = (byte) ((offset & 0x0000ff00) >> 8);
			result[12] = (byte) ((offset & 0x00ff0000) >> 16);
			result[13] = (byte) ((offset & 0xff000000) >> 24);

			return result;
		}
	}

	/**
	 * rbg}bvwb_
	 */
	private static class InfoHeader {
		/** wb_̃f[^ */
		byte[] bytes;

		/**
		 * RXgN^
		 */
		InfoHeader() {
			//
		}

		/**
		 * oCĩRXgN^
		 * 
		 * @param bytes
		 *            f[^
		 * @param fromClipboard
		 *            t@Cwb_ꍇfalseBœǂݍ݂JnItZbg𒲐߂܂B
		 */
		InfoHeader(byte[] bytes, boolean fromClipboard) {
			int offset = fromClipboard ? 0 : 14;
			// wb_̃m
			this.bytes = new byte[40];
			// Rs[
			System.arraycopy(bytes, offset, this.bytes, 0, 40);
		}

		/**
		 * oCiԂ܂B
		 * @return oCif[^
		 */
		byte[] getBytes() {
			return bytes;
		}

		/**
		 * fԂ܂B
		 * @return f
		 */
		int getBitCount() {
			return byteToInt(bytes, 14);
		}

		/**
		 * 摜̃f[^TCYԂ܂B ۂ̃TCYƂ͈قȂ邱Ƃ̂ŐMpł܂B
		 * @return TCY
		 */
		int getImageDataSize() {
			return byteToInt(bytes, 20);
		}

		/**
		 * 摜̃f[^TCYݒ肵܂B
		 * 
		 * @param size
		 *            VTCY
		 */
		void setImageDataSize(int size) {
			bytes[20] = (byte) (size & 0x000000ff);
			bytes[21] = (byte) ((size & 0x0000ff00) >> 8);
			bytes[22] = (byte) ((size & 0x00ff0000) >> 16);
			bytes[23] = (byte) ((size & 0xff000000) >> 24);
		}

		/**
		 * FԂ܂B 0̏ꍇbitCountɑ΂ő̐FƂ݂ȂĂB
		 * @return F
		 */
		int getColorCount() {
			return byteToInt(bytes, 32);
		}

		/**
		 * rbgJEgƐFۂ̃pbg̃TCYԂ܂B
		 * 
		 * @return pbg̃TCY(byte)
		 */
		int getPalletSize() {
			int usedColor = getColorCount();
			if (usedColor == 0) {
				int bitCount = getBitCount();
				switch (bitCount) {
				case 1:
					usedColor = 2;
					break;
				case 4:
					usedColor = 16;
					break;
				case 8:
					usedColor = 256;
					break;
				default:
					// 1,4,8ȊO̓pbg݂ȂB
					usedColor = 0;
				}
			}
			return usedColor * 4;
		}

		/**
		 * rbgtB[hɂ鈳kݒ肪ꍇɃTCYԂ܂B
		 * 
		 * @return TCY
		 */
		int getCompressionSettingSize() {
			int comp = byteToInt(bytes, 16);
			return comp == 3 ? 12 : 0;
		}
	}

	/**
	 * pbg(RGBQUAD)
	 */
	private static class Pallet {
		/**
		 * 1rbg - 2 * 4 byte
		 * 4rbg - 16 * 4 byte
		 * 8rbg - 256 * 4 byte
		 * ̓pbg
		 */
		byte[] bytes;

		/**
		 * RXgN^
		 */
		Pallet() {
			//
		}

		/**
		 * oCĩRXgN^
		 * 
		 * @param bytes oCif[^
		 * @param fromClipboard t@Cwb_true
		 * @param palletSize pbgTCY
		 */
		Pallet(byte[] bytes, boolean fromClipboard, int palletSize) {
			// pbg݂ȂƂ
			if (palletSize == 0) {
				return;
			}
			// ItZbg߂
			int offset = fromClipboard ? 40 : 54;

			// Rs[
			this.bytes = new byte[palletSize];
			System.arraycopy(bytes, offset, this.bytes, 0, palletSize);
		}

		/**
		 * oCiԂ܂B 
		 * @return oCif[^
		 */
		byte[] getBytes() {
			return bytes;
		}
	}

}