package jp.takayuki.graphics;

import java.util.ArrayList;
import java.util.List;

import jp.takayuki.tools.Tool;

/**
 * Windows Metafile
 * @author Takayuki Shoji
 */
public class WindowsMetafile {

	/** Bitmap Record Type: META_STRETCHDIB = 0x0F43 */
	private static final int META_STRETCHDIB = 0x0F43;

	/**
	 * 摜f[^
	 */
	private byte[] bytes;

	/** MapMode */
	private final int mapMode;

	/**
	 * RXgN^
	 * @param bytes 摜f[^
	 * @param mapMode MapMode(1-8)
	 */
	public WindowsMetafile(byte[] bytes, int mapMode) {
		this.bytes = bytes;
		this.mapMode = mapMode;
	}

	/**
	 * wb_(META_HEADER)Ȃǂ؂܂B
	 * @return ̏ꍇtrueB
	 */
	public boolean isValidHeader() {
		if (bytes == null || bytes.length < 18 + 6) {
			// 18 - header size, 6 - EOF record
			return false;
		}
		if (1 > mapMode || mapMode > 8) {
			return false;
		}
		int metafileType = Tool.bytesToShort(bytes, 0);
		if (metafileType != 1 && metafileType != 2) {
			// 1 - in memory, 2 - on disk
			return false;
		}
		int headerSize = Tool.bytesToShort(bytes, 2) * 2;
		if (headerSize != 18) {
			return false;
		}
		int metafileVersion = Tool.bytesToShort(bytes, 4);
		if (metafileVersion != 0x0300 && metafileVersion != 0x0100) {
			// 0x0100DIBT|[gĂȂAeB
			return false;
		}
		int fileSize = Tool.byteToInt(bytes, 6) * 2;
		if (fileSize != bytes.length) {
			return false;
		}
		// NumberOfObjects Brush, Font, Palette, Pen, Region
		int maxRecordSize = Tool.byteToInt(bytes, 12) * 2;
		if (bytes.length - 18 - 6 < maxRecordSize) {
			return false;
		}
		return true;
	}

	/**
	 * ߍ܂ꂽDIB擾܂B
	 * @return DibImagẽXgBDIBꍇ͋̃XgԂ܂B
	 */
	public List getDibImages() {
		ArrayList result = new ArrayList();
		int offset = 18;
		while (offset + 6 <= bytes.length) {
			int recSize = Tool.byteToInt(bytes, offset + 0) * 2;
			if (offset + recSize > bytes.length) {
				break;
			}
			if (recSize > 28) {
				int recType = Tool.bytesToShort(bytes, offset + 4);
				if (recType == META_STRETCHDIB) {
					byte[] dib = new byte[recSize - 28];
					System.arraycopy(bytes, offset + 28, dib, 0, recSize - 28);
					DibImage image = new DibImage(dib);
					result.add(image);
				}
			}
			offset += recSize;
		}
		return result;
	}

	/**
	 * DIBWMFŃbv܂B
	 * 
	 * @param img ߍމ摜
	 * @return WMF̃oCgz
	 */
	public static byte[] wrapDibImage(DibImage img) {
		byte[] dib = img.getBytes(true);
		byte[] all = new byte[18 + 10 + 10 + 28 + dib.length + 6];

		int offset = 18;
		// 0x020B (META_SETWINDOWORG)[10]
		Tool.writeInt32LittleEndian(all, offset + 0, 10 / 2);//size
		Tool.writeInt16LittleEndian(all, offset + 4, 0x020B);//type
		// c4oCǵA(x,y)=(0,0)

		offset += 10;

		// 0x020C (META_SETWINDOWEXT)[10]
		Tool.writeInt32LittleEndian(all, offset + 0, 10 / 2);//size
		Tool.writeInt16LittleEndian(all, offset + 4, 0x020C);//type
		Tool.writeInt16LittleEndian(all, offset + 6, Tool
				.pixelToLogicalUnit(img.getHeight()));//vertical extent
		Tool.writeInt16LittleEndian(all, offset + 8, Tool
				.pixelToLogicalUnit(img.getWidth()));//horizontal extent

		offset += 10;

		// 0x0F43 (META_STRETCHDIB)[28]
		Tool.writeInt32LittleEndian(all, offset + 0, (dib.length + 28) / 2);//size
		Tool.writeInt16LittleEndian(all, offset + 4, 0x0F43);//type
		Tool.writeInt32LittleEndian(all, offset + 6, 0x00CC0020);//RasterOperation
		Tool.writeInt16LittleEndian(all, offset + 10, 0x0000);//ColorUsage
		Tool.writeInt16LittleEndian(all, offset + 12, img.getHeight());//srcHeight
		Tool.writeInt16LittleEndian(all, offset + 14, img.getWidth());//srcWidth
		// ySrc, xSrc = (0, 0)
		Tool.writeInt16LittleEndian(all, offset + 20, Tool
				.pixelToLogicalUnit(img.getHeight()));//destHeight
		Tool.writeInt16LittleEndian(all, offset + 22, Tool
				.pixelToLogicalUnit(img.getWidth()));//destWidth
		// yDest, xDest = (0, 0)

		offset += 28;
		System.arraycopy(dib, 0, all, offset, dib.length);

		offset += dib.length;

		// EOF[6]
		all[offset] = 0x03;

		// ĕԂ
		// header[18]
		offset = 0;
		Tool.writeInt16LittleEndian(all, offset + 0, 0x0001);//FileType
		Tool.writeInt16LittleEndian(all, offset + 2, 18 / 2);//HeaderSize
		Tool.writeInt16LittleEndian(all, offset + 4, 0x0300);//Version
		Tool.writeInt32LittleEndian(all, offset + 6, all.length / 2);//FileSize
		//NumOfObjects=0
		Tool.writeInt32LittleEndian(all, offset + 12, (28 + dib.length) / 2);//MaxRecordSize
		//NumOfParams=0

		return all;
	}
}
