/*
 * Based on The Seasar Foundation's Seasar2.
 * Copyright 2004-2006 the Seasar Foundation and the Others.
 * Copyright 2006-2008 Takayuki Shoji.
 */
/*
 * Copyright 2004-2006 the Seasar Foundation and the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package jp.takayuki.logging;

import java.text.MessageFormat;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

import org.apache.log4j.Level;

/**
 * Logger
 * <p>
 * The Seasar FoundationS2.4.0-beta-3
 * org.seasar.framework.log.Logger
 * org.seasar.framework.message.MessageFormatterɍ쐬܂B
 * <p>
 * log4jp1NXsĂ܂B
 * 
 * @author higa
 * 
 * @author Takayuki Shoji
 */
public class Logger {

	/**
	 * FQCN
	 */
	private static final String FQCN = Logger.class.getName();

	/**
	 * bZ[W\[X
	 */
	private static final ThreadLocal bundle = new ThreadLocal() {
		protected Object initialValue() {
			return new HashMap();
		}
	};

	/**
	 * Logger̃LbV
	 * <br>L[: Class
	 * <br>l: Logger
	 */
	private static Map map = new HashMap();

	/**
	 * ۂɏo͂sLogger
	 */
	private org.apache.log4j.Logger log;

	/**
	 * RXgN^
	 * @param clazz
	 */
	private Logger(Class clazz) {
		log = org.apache.log4j.Logger.getLogger(clazz);
	}

	/**
	 * Logger擾܂B
	 * @param clazz NX
	 * @return Logger
	 */
	public static final Logger getLogger(Class clazz) {
		Logger logger;
		synchronized (map) {
			logger = (Logger) map.get(clazz);
			if (logger == null) {
				logger = new Logger(clazz);
				map.put(clazz, logger);
			}

		}
		return logger;
	}

	/**
	 * Oo͂܂B
	 * @param messageCode bZ[WR[h(x VXeID R[h)
	 * @param args bZ[WR[hɑΉ
	 */
	public void log(String messageCode, Object[] args) {
		log(messageCode, args, null);
	}

	/**
	 * Oo͂܂B
	 * @param messageCode bZ[WR[h(x VXeID R[h)
	 * @param args bZ[WR[hɑΉ
	 * @param throwable
	 */
	public void log(String messageCode, Object[] args, Throwable throwable) {
		if (messageCode == null || messageCode.length() == 0) {
			return;
		}

		String message = getMessage(messageCode, args);

		Level level = getLevel(messageCode.charAt(0));

		log.log(FQCN, level, message, throwable);
	}

	/**
	 * bZ[WR[hɑΉp^[ŃtH[}bgbZ[W擾܂B
	 * @param messageCode bZ[WR[h
	 * @param args bZ[W
	 * @return tH[}bgς݃bZ[W
	 */
	private String getMessage(String messageCode, Object[] args) {
		StringBuffer buffer = new StringBuffer();
		buffer.append('[');
		buffer.append(messageCode);
		buffer.append(']');

		Map messageMap = (Map) bundle.get();
		String pattern = (String) messageMap.get(messageCode);
		if (pattern == null) {
			pattern = loadMessage(messageCode);
			if (pattern == null) {
				buffer.append(getMissingMessage(args));
				return buffer.toString();
			}
		}

		MessageFormat fmt = new MessageFormat(pattern);
		buffer.append(fmt.format(args));
		return buffer.toString();
	}

	/**
	 * p^[Ȃꍇ̃tH[}bgs܂B
	 * @param args bZ[W
	 * @return tH[}bgς݃bZ[W
	 */
	private static String getMissingMessage(Object[] args) {
		if (args == null || args.length == 0) {
			return ""; //$NON-NLS-1$
		}
		StringBuffer buffer = new StringBuffer();
		for (int i = 0; i < args.length; i++) {
			buffer.append(args[i]);
			buffer.append(", "); //$NON-NLS-1$
		}
		buffer.setLength(buffer.length() - 2);
		return buffer.toString();
	}

	/**
	 * ResourceBundleThreadLocal
	 * @param messageCode bZ[WR[h
	 * @return Ήp^[
	 */
	private String loadMessage(String messageCode) {
		String bundleName = messageCode.substring(1, Math.min(4, messageCode
				.length()));
		try {
			Map messageMap = (Map) bundle.get();

			ResourceBundle r = ResourceBundle.getBundle(bundleName);
			Enumeration en = r.getKeys();
			while (en.hasMoreElements()) {
				String key = (String) en.nextElement();
				String value = r.getString(key);
				messageMap.put(key, value);
			}
			return r.getString(messageCode);

		} catch (MissingResourceException e) {
			return null;
		}
	}

	/**
	 * bZ[WR[h1ڂ烍Ox擾܂B
	 * @param c bZ[WR[ḧꕶ(D,I,W,E,F̂ꂩ)
	 * @return Ox
	 */
	private Level getLevel(char c) {
		switch (c) {
		case 'D':
			return Level.DEBUG;
		case 'I':
			return Level.INFO;
		case 'W':
			return Level.WARN;
		case 'E':
			return Level.ERROR;
		case 'F':
			return Level.FATAL;
		default:
			throw new IllegalArgumentException(c + " is invalid level."); //$NON-NLS-1$
		}
	}

	/**
	 * fobOx̃bZ[Wo͂܂B
	 * @param message bZ[W
	 */
	public void debug(Object message) {
		log.debug(message);
	}

	/**
	 * fobOx̃bZ[Wo͂܂B
	 * @param message bZ[W
	 * @param th Throwable
	 */
	public void debug(Object message, Throwable th) {
		log.debug(message, th);
	}

	/**
	 * 񃌃x̃bZ[Wo͂܂B
	 * @param message bZ[W
	 */
	public void info(Object message) {
		log.info(message);
	}

	/**
	 * 񃌃x̃bZ[Wo͂܂B
	 * @param message bZ[W
	 * @param th Throwable
	 */
	public void info(Object message, Throwable th) {
		log.info(message, th);
	}

	/**
	 * xx̃bZ[Wo͂܂B
	 * @param message bZ[W
	 */
	public void warn(Object message) {
		log.warn(message);
	}

	/**
	 * xx̃bZ[Wo͂܂B
	 * @param message bZ[W
	 * @param th Throwable
	 */
	public void warn(Object message, Throwable th) {
		log.warn(message, th);
	}

	/**
	 * G[x̃bZ[Wo͂܂B
	 * @param message bZ[W
	 */
	public void error(Object message) {
		log.error(message);
	}

	/**
	 * G[x̃bZ[Wo͂܂B
	 * @param message bZ[W
	 * @param th Throwable
	 */
	public void error(Object message, Throwable th) {
		log.error(message, th);
	}

	/**
	 * vIx̃bZ[Wo͂܂B
	 * @param message bZ[W
	 */
	public void fatal(Object message) {
		log.fatal(message);
	}

	/**
	 * vIx̃bZ[Wo͂܂B
	 * @param message bZ[W
	 * @param th Throwable
	 */
	public void fatal(Object message, Throwable th) {
		log.fatal(message, th);
	}

	/**
	 * DEBUGx̏o͂LɂȂĂ邩ǂ𔻒肵܂B
	 * @return Lȏꍇtrue
	 */
	public boolean isDebugEnabled() {
		return log.isDebugEnabled();
	}

	/**
	 * INFOx̏o͂LɂȂĂ邩ǂ𔻒肵܂B
	 * @return Lȏꍇtrue
	 */
	public boolean isInfoEnabled() {
		return log.isInfoEnabled();
	}
}
