package jp.takayuki.swt.util;

/*
 * based on Snippet178.java.
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * Copyright (c) 2008 Takayuki Shoji.
 * 
 * Eclipse Public License v1.0
 * http://www.eclipse.org/legal/epl-v10.html
 */

/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/

/*
 * How to access About, Preferences and Quit menus on carbon.
 * NOTE: This snippet uses internal SWT packages that are
 * subject to change without notice.
 * 
 * For a list of all SWT example snippets see
 * http://www.eclipse.org/swt/snippets/
 */

import org.eclipse.swt.internal.Callback;
import org.eclipse.swt.internal.carbon.HICommand;
import org.eclipse.swt.internal.carbon.OS;
import org.eclipse.swt.widgets.Display;

import petaclone.swt.SWTMainWindow;
import petaclone.swt.SWTMainWindow.Actions;

/**
 * based on Snippet178.java.
 * <p>
 * Mac OS X̃AvP[Vj[\z܂B
 * 
 * @author Takayuki Shoji
 */
public class SWTMacOSXMenu implements Runnable {

	/**
	 * ݒ
	 */
	private static final int kHICommandPreferences = ('p' << 24) + ('r' << 16)
			+ ('e' << 8) + 'f';

	/**
	 * About
	 */
	private static final int kHICommandAbout = ('a' << 24) + ('b' << 16)
			+ ('o' << 8) + 'u';

	/**
	 * T[rX
	 */
	private static final int kHICommandServices = ('s' << 24) + ('e' << 16)
			+ ('r' << 8) + 'v';

	/**
	 * CxgXi[
	 */
	private Actions actions;

	/**
	 * R[obN
	 */
	private Callback commandCallback;

	/**
	 * AvP[Vj[ɃR[obNݒ肵܂B
	 * 
	 * @param display Display
	 * @param aboutName Aboutj[̃eLXg
	 * @param actions CxgXi[
	 */
	public void hookApplicationMenu(Display display, final String aboutName,
			SWTMainWindow.Actions actions) {
		this.actions = actions;

		commandCallback = new Callback(this, "commandProc", 3); //$NON-NLS-1$
		int commandProc = (int) /* cast for 64bit SWT Callback class */ commandCallback.getAddress();
		if (commandProc == 0) {
			commandCallback.dispose();
			return; // give up
		}

		// Install event handler for commands
		int[] mask = new int[] { OS.kEventClassCommand, OS.kEventProcessCommand };
		OS.InstallEventHandler(OS.GetApplicationEventTarget(), commandProc,
				mask.length / 2, mask, 0, null);

		// create About ... menu command
		int[] outMenu = new int[1];
		short[] outIndex = new short[1];
		if (OS.GetIndMenuItemWithCommandID(0, kHICommandPreferences, 1,
				outMenu, outIndex) == OS.noErr
				&& outMenu[0] != 0) {
			int menu = outMenu[0];

			int l = aboutName.length();
			char buffer[] = new char[l];
			aboutName.getChars(0, l, buffer, 0);
			int str = OS.CFStringCreateWithCharacters(OS.kCFAllocatorDefault,
					buffer, l);
			OS.InsertMenuItemTextWithCFString(menu, str, (short) 0, 0,
					kHICommandAbout);
			OS.CFRelease(str);

			// add separator between About & Preferences
			OS.InsertMenuItemTextWithCFString(menu, 0, (short) 1,
					OS.kMenuItemAttrSeparator, 0);

			// enable pref menu
			OS.EnableMenuCommand(menu, kHICommandPreferences);

			// disable services menu
			OS.DisableMenuCommand(menu, kHICommandServices);
		}

		// schedule disposal of callback object
		display.disposeExec(this);
	}

	/**
	 * disposeExecp
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		commandCallback.dispose();
	}

	/**
	 * R[obN\bhB
	 * <p>
	 * commandID𔻒肵ēK؂ȃ\bhĂяo܂B
	 * 
	 * @param nextHandler
	 * @param theEvent
	 * @param userData
	 * @return OS.noErr ܂ OS.eventNotHandledErr
	 */
	public int commandProc(int nextHandler, int theEvent, int userData) {
		if (OS.GetEventKind(theEvent) == OS.kEventProcessCommand) {
			HICommand command = new HICommand();
			OS.GetEventParameter(
					theEvent,//
					OS.kEventParamDirectObject, OS.typeHICommand, null,
					HICommand.sizeof, null, command);
			switch (command.commandID) {
			case kHICommandPreferences:
				actions.doMenuFileConfig(null);
				return OS.noErr;
			case kHICommandAbout:
				actions.doMenuHelpAbout(null);
				return OS.noErr;
			default:
				break;
			}
		}
		return OS.eventNotHandledErr;
	}
}
