package jp.takayuki.swt.util;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Tool;

import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTError;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.dnd.DND;
import org.eclipse.swt.dnd.ImageTransfer;
import org.eclipse.swt.dnd.TextTransfer;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.dnd.TransferData;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.ImageLoader;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.swt.widgets.IME;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MessageBox;
import org.eclipse.swt.widgets.Monitor;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.Widget;

/**
 * SWT̃[eBeB\bh
 * @author Takayuki Shoji
 */
public class SWTUtils {

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/**
	 * \ĂShell擾܂B
	 * 
	 * @param except ɏOShellBnullB
	 * @param exceptMinimized ŏꂽShellOꍇtrue
	 * @param exceptMaximized ő剻ꂽShellOꍇtrue
	 * @return Shell̔z
	 */
	public static Shell[] getVisibleShells(Shell except,
			boolean exceptMinimized, boolean exceptMaximized) {
		Display display = Display.getDefault();
		Shell[] shells = display.getShells();
		ArrayList list = new ArrayList();
		for (int i = 0; i < shells.length; i++) {
			if (!shells[i].getVisible()) {
				continue;
			}
			if (shells[i] == except) {
				continue;
			}
			if (exceptMinimized && shells[i].getMinimized()) {
				continue;
			}
			if (exceptMaximized && shells[i].getMaximized()) {
				continue;
			}
			list.add(shells[i]);
		}
		return (Shell[]) list.toArray(new Shell[list.size()]);
	}

	/**
	 * \ĂShell̋`擾܂B
	 * 
	 * @return Rectangle̔z
	 */
	public static Rectangle[] getVisibleShellRects() {
		Shell[] shells = getVisibleShells(null, true, true);
		Rectangle[] rects = new Rectangle[shells.length];
		for (int i = 0; i < shells.length; i++) {
			rects[i] = shells[i].getBounds();
		}
		return rects;
	}

	/**
	 * w肵`̈ɍWƃTCYƗ]w肵`̈悪܂邩ǂԂ܂B
	 * 
	 * @param rect `̈
	 * @param width eXg`̈̉
	 * @param height eXg`̈̍
	 * @param padding eXg`̗̈]
	 * @param x eXg`̈̍XW
	 * @param y eXg`̈̍YW
	 * @return ͈͓̏ꍇtrue
	 */
	public static boolean withinDisplay(Rectangle rect, int width, int height,
			int padding, int x, int y) {

		// _0,0ɕ␳
		x -= rect.x;
		y -= rect.y;

		if (!(padding <= x && x <= rect.width - width - padding)) {
			return false;
		}

		if (!(padding <= y && y <= rect.height - height - padding)) {
			return false;
		}

		return true;
	}

	/**
	 * bigsmallɊ܂܂邩ǂԂ܂B
	 * 
	 * @param big 傫`
	 * @param small `
	 * @return smallbigɊ܂܂ꍇtrueAꍇtrueB
	 */
	public static boolean containsRect(Rectangle big, Rectangle small) {
		return big.contains(small.x, small.y)
				&& big.contains(small.x + small.width, small.y + small.height);
	}

	/**
	 * ̋`̈悩w̋`̈ƏdȂŏ̗̈擾܂B
	 * 
	 * @param rect eXg`̈
	 * @param rects ̋`̈
	 * @return dȂ̈BdȂȂꍇnull
	 */
	public static Rectangle getIntersectRect(Rectangle rect, Rectangle[] rects) {
		for (int i = 0; i < rects.length; i++) {
			if (rect.intersects(rects[i])) {
				return rects[i];
			}
		}
		return null;
	}

	/**
	 * ׂĂShellɑ΂setEnabled(true)s܂B
	 */
	public static void enableAllShells() {
		Display display = Display.getDefault();
		Shell[] shells = display.getShells();
		for (int i = 0; i < shells.length; i++) {
			shells[i].setEnabled(true);
		}
	}

	/**
	 * _CAOJĂ邩ǂ𔻒肵܂B
	 * @return _CAȌꍇtrue
	 */
	public static boolean isDialogShellOpened() {
		Shell sh = Display.getDefault().getActiveShell();
		if (sh == null) {
			return false;
		}
		int style = sh.getStyle();
		return (style & SWT.APPLICATION_MODAL) != 0
				|| (style & SWT.PRIMARY_MODAL) != 0
				|| (style & SWT.SYSTEM_MODAL) != 0;
	}

	/**
	 * L[R[hL[ǂ𔻒肵܂B
	 * @param keyCode L[R[h
	 * @return L[̏ꍇtrue
	 */
	public static boolean isArrowKeyCode(int keyCode) {
		switch (keyCode) {
		case SWT.ARROW_UP:
		case SWT.ARROW_DOWN:
		case SWT.ARROW_LEFT:
		case SWT.ARROW_RIGHT:
			return true;
		default:
			return false;
		}
	}

	/**
	 * Nbv{[hɃeLXgRs[܂B
	 * @param text Nbv{[hɃRs[eLXg
	 * @return ꍇtrueAsꍇfalseB
	 */
	public static boolean setTextToClipboard(String text) {
		if (text == null || text.length() == 0) {
			return false;
		}
		TextTransfer textTransfer = TextTransfer.getInstance();
		Transfer[] transfers = new Transfer[] { textTransfer };
		Object[] data = new Object[] { text };
		Clipboard clip = new Clipboard(Display.getDefault());
		try {
			clip.setContents(data, transfers);
		} catch (SWTError e) {
			if (e.code == DND.ERROR_CANNOT_SET_CLIPBOARD) {
				// recoverable
				return false;
			}
			throw e;

		} finally {
			clip.dispose();
		}
		return true;
	}

	/**
	 * Nbv{[hɃf[^ݒ肵܂B
	 * 
	 * @param transferList Transfer̃Xg
	 * @param dataList f[^̃Xg
	 * @return ꍇtrueAsꍇfalseB
	 */
	public static boolean setClipboard(List transferList, List dataList) {
		Transfer[] trArray = (Transfer[]) transferList
				.toArray(new Transfer[transferList.size()]);
		Object[] dataArray = dataList.toArray(new Object[dataList.size()]);
		Clipboard clip = new Clipboard(Display.getDefault());
		try {
			clip.setContents(dataArray, trArray);
			return true;
		} catch (SWTError e) {
			if (e.code == DND.ERROR_CANNOT_SET_CLIPBOARD) {
				// recoverable
				return false;
			}
			throw e;

		} finally {
			clip.dispose();
		}
	}

	/**
	 * Nbv{[hɃT|[g摜݂邩ǂԂ܂B
	 * @param clipboard Nbv{[h
	 * @return p\ȏꍇtrue
	 */
	public static boolean isImageAvailable(Clipboard clipboard) {
		return isAvailableType(clipboard, ImageTransfer.getInstance());
	}

	/**
	 * Nbv{[hɃeLXg݂邩ǂԂ܂B
	 * @param clipboard Nbv{[h
	 * @return p\ȏꍇtrue
	 */
	public static boolean isTextAvailable(Clipboard clipboard) {
		return isAvailableType(clipboard, TextTransfer.getInstance());
	}

	/**
	 * w肵f[^^Nbv{[hɑ݂邩Ԃ܂B
	 * @param clipboard Nbv{[h
	 * @param tr f[^^ɉTransfer
	 * @return ݂ꍇtrue
	 */
	private static boolean isAvailableType(Clipboard clipboard, Transfer tr) {
		TransferData[] types = clipboard.getAvailableTypes();
		for (int i = 0; i < types.length; i++) {
			if (tr.isSupportedType(types[i])) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Nbv{[h̉摜擾܂B
	 * 
	 * @param clipboard Nbv{[h
	 * @return 擾łȂꍇnullB
	 */
	public static ImageData getClipboardImage(Clipboard clipboard) {
		ImageTransfer transfer = ImageTransfer.getInstance();
		return (ImageData) clipboard.getContents(transfer);
	}

	/**
	 * Nbv{[h̃eLXg擾܂B
	 * 
	 * @param clipboard Nbv{[h
	 * @return 擾łȂꍇnullB
	 */
	public static String getClipboardText(Clipboard clipboard) {
		TextTransfer transfer = TextTransfer.getInstance();
		return (String) clipboard.getContents(transfer);
	}

	/**
	 * ImageDatagqŎw肳ꂽ`ŕۑ܂B
	 * <p>
	 * fileɃfBNgw肵ꍇ͎Ńt@C쐬܂B
	 * nullw肵ꍇ́AVXẽe|fBNggp܂B
	 * 
	 * @param savePath ۑBt@C܂̓fBNg܂nullB
	 * @param img ۑC[W
	 * @return ۑ̃t@C
	 * @throws IOException ۑ̗O
	 */
	public static File saveImageAs(File savePath, ImageData img)
			throws IOException {
		File tmp;
		if (savePath == null || savePath.isDirectory()) {
			tmp = File.createTempFile(mm.get("SWTUtils.Clipboard"), //$NON-NLS-1$
					".png", savePath); //$NON-NLS-1$
		} else {
			tmp = savePath;
		}

		try {

			// tH[}bg
			int format;
			String name = tmp.getName().toLowerCase();
			if (name.endsWith(".bmp")) { //$NON-NLS-1$
				format = SWT.IMAGE_BMP_RLE;
			} else if (name.endsWith(".jpg") || name.endsWith(".jpeg")) { //$NON-NLS-1$ //$NON-NLS-2$
				format = SWT.IMAGE_JPEG;
			}
			// F[x̕ϊKv
			// else if (name.endsWith(".tiff")) {
			//	format = SWT.IMAGE_TIFF;
			//} else if (name.endsWith(".gif")) {
			//	format = SWT.IMAGE_GIF;
			//} else if (name.endsWith(".ico")) {
			//	format = SWT.IMAGE_ICO;
			//}
			else {
				format = SWT.IMAGE_PNG;
			}

			ImageLoader loader = new ImageLoader();
			loader.data = new ImageData[] { img };
			loader.save(tmp.getAbsolutePath(), format);
			return tmp;
		} catch (SWTException e) {
			throw Tool.newIOException("Could not save the ImageData.", e); //$NON-NLS-1$
		}
	}

	/**
	 * t@Cۑ_CAO\āApXԂ܂B
	 * [U[w肵fBNg̗LmFA
	 * ȏꍇ͍ēx_CAOJ܂B
	 * @param parent eShell
	 * @param text ^Cg
	 * @param filename ftHg̃t@C
	 * @param filterExts gqtB^(: *.png)BnullB
	 * @param warnOverwrite ㏑xꍇtrue
	 * @return FileBLZꍇnullB
	 */
	public static File openSaveFileDialog(Shell parent, String text,
			String filename, String[] filterExts, boolean warnOverwrite) {
		while (true) {
			FileDialog dialog = new FileDialog(parent, SWT.SAVE
					| SWT.APPLICATION_MODAL | SWT.SHEET);
			dialog.setFileName(filename);
			dialog.setText(text);
			if (filterExts != null) {
				dialog.setFilterExtensions(filterExts);
			}
			dialog.setOverwrite(warnOverwrite);
			String path = dialog.open();
			if (path == null) {
				return null;
			}
			File f = new File(path);
			File dir = f.getParentFile();
			if (!isExistingDirectory(parent, dir)) {
				continue;
			}
			return f;
		}
	}

	/**
	 * fBNg݂ꍇtrue,
	 * ݂Ȃꍇ̓bZ[W{bNX\false
	 * Ԃ܂B
	 * @param parent Shell
	 * @param dir fBNg
	 * @return ݂ꍇtrue
	 */
	public static boolean isExistingDirectory(Shell parent, File dir) {
		if (dir != null && dir.isDirectory()) {
			return true;
		}
		// _CAOł̓t@C\܂߂邱Ƃł
		// .URLt@CɊւoO
		String msg = mm.get("SWTUtils.invalidFilepath"); //$NON-NLS-1$
		if (dir != null) {
			msg += '\n' + dir.getAbsolutePath();
		}
		showError(parent, msg);
		return false;
	}

	/**
	 * 񂪃e[u̍牽ڂԂ܂B
	 * 
	 * @param col e[uJ
	 * @return CfbNX(0-)BȂꍇ-1B
	 */
	public static int getColumnIndex(TableColumn col) {
		if (col != null) {
			Table table = col.getParent();
			int size = table.getColumnCount();
			for (int index = 0; index < size; index++) {
				if (col == table.getColumn(index)) {
					return index;
				}
			}
		}
		return -1;
	}

	/**
	 * ӂڂĂ邩ǂԂ܂B
	 * 
	 * @param rect1 `(client area)
	 * @param rect2 `(shell)
	 * @return ӂڂĂꍇtrue
	 */
	public static boolean isSharedBorderUp(Rectangle rect1, Rectangle rect2) {
		return rect1.y == rect2.y;
	}

	/**
	 * ӂڂĂ邩ǂԂ܂B
	 * 
	 * @param rect1 `(client area)
	 * @param rect2 `(shell)
	 * @return ӂڂĂꍇtrue
	 */
	public static boolean isSharedBorderLeft(Rectangle rect1, Rectangle rect2) {
		return rect1.x == rect2.x;
	}

	/**
	 * EӂڂĂ邩ǂԂ܂B
	 * 
	 * @param rect1 `(client area)
	 * @param rect2 `(shell)
	 * @return EӂڂĂꍇtrue
	 */
	public static boolean isSharedBorderRight(Rectangle rect1, Rectangle rect2) {
		return rect1.x + rect1.width == rect2.x + rect2.width;
	}

	/**
	 * ӂڂĂ邩ǂԂ܂B
	 * 
	 * @param rect1 `(client area)
	 * @param rect2 `(shell)
	 * @return ӂڂĂꍇtrue
	 */
	public static boolean isSharedBorderBottom(Rectangle rect1, Rectangle rect2) {
		return rect1.y + rect1.height == rect2.y + rect2.height;
	}

	/**
	 * rect1rect2̉Eӂڂ悤rect2ړ܂B
	 * 
	 * @param rect1 `1
	 * @param rect2 `2
	 */
	public static void glueBorderRight(Rectangle rect1, Rectangle rect2) {
		rect2.x = rect1.x + rect1.width - rect2.width;
	}

	/**
	 * rect1rect2̉ӂڂ悤rect2ړ܂B
	 * 
	 * @param rect1 `1
	 * @param rect2 `2
	 */
	public static void glueBorderBotom(Rectangle rect1, Rectangle rect2) {
		rect2.y = rect1.y + rect1.height - rect2.height;
	}

	/**
	 * }EX|C^郂j^̃CfbNX擾܂B
	 * 
	 * @param display Display
	 * @param monitors ڑĂfBXvC
	 * @return monitors̃CfbNXBȂꍇ-1(ʏ͖Ǝv)B
	 */
	public static int getPointedMonitorIndex(Display display, Monitor[] monitors) {
		Point location = display.getCursorLocation();
		return getMonitorIndexByCoordinate(monitors, location.x, location.y);
	}

	/**
	 * }EX|C^郂j^擾܂B
	 * @param display Display
	 * @param monitors ڑĂfBXvC
	 * @return YMonitorB݂Ȃꍇ̓vC}fBXvCMonitorB
	 */
	public static Monitor getPointedMonitor(Display display, Monitor[] monitors) {
		int index = getPointedMonitorIndex(display, monitors);
		if (index == -1) {
			return display.getPrimaryMonitor();
		}
		return monitors[index];
	}

	/**
	 * }EX|C^fBXvC̃NCAg̈擾܂B
	 * @param display Display
	 * @return `B
	 */
	public static Rectangle getPointedMonitorClient(Display display) {
		Monitor[] monitors = display.getMonitors();
		Monitor monitor = getPointedMonitor(display, monitors);
		return monitor.getClientArea();
	}

	/**
	 * w肳ꂽW܂ރj^̃CfbNX擾܂B
	 * 
	 * @param monitors ڑĂfBXvC
	 * @param x XW
	 * @param y YW
	 * @return monitors̃CfbNXB܂܂Ȃꍇ-1B
	 */
	public static int getMonitorIndexByCoordinate(Monitor[] monitors, int x,
			int y) {
		for (int i = 0; i < monitors.length; i++) {
			Rectangle rect = monitors[i].getBounds();
			if (rect.contains(x, y)) {
				return i;
			}
		}
		return -1;
	}

	/**
	 * w肳ꂽW܂ރj^̃CfbNX擾܂B
	 * 
	 * @param monitors ڑĂfBXvC
	 * @param x XW
	 * @param y YW
	 * @param display 
	 * @return YMonitorB݂Ȃꍇ̓vC}fBXvCMonitorB
	 */
	public static Monitor getMonitorByCoordinate(Monitor[] monitors, int x,
			int y, Display display) {
		int index = getMonitorIndexByCoordinate(monitors, x, y);
		if (index == -1) {
			return display.getPrimaryMonitor();
		}
		return monitors[index];
	}

	/**
	 * w肳ꂽW܂ރj^擾܂B
	 * 
	 * @param x XW
	 * @param y YW
	 * @param display Display
	 * @return YMonitorB݂Ȃꍇ̓vC}fBXvCMonitorB
	 */
	public static Monitor getMonitorByCoordinate(int x, int y, Display display) {
		Monitor[] monitors = display.getMonitors();
		int index = getMonitorIndexByCoordinate(monitors, x, y);
		if (index == -1) {
			return display.getPrimaryMonitor();
		}
		return monitors[index];
	}

	/**
	 * w肳ꂽW܂ރj^̃NCAg̈擾܂B
	 * 
	 * @param x XW
	 * @param y YW
	 * @param display Display
	 * @return YMonitor̃NCAg̈B݂Ȃꍇ̓vC}fBXvCMonitor̃NCAg̈B
	 */
	public static Rectangle getMonitorClientByCoordinate(int x, int y,
			Display display) {
		Monitor[] monitors = display.getMonitors();
		int index = getMonitorIndexByCoordinate(monitors, x, y);
		if (index == -1) {
			return display.getPrimaryMonitor().getClientArea();
		}
		return monitors[index].getClientArea();
	}

	/**
	 * w肳ꂽW܂ރj^̃NCAg̈擾܂B
	 * 
	 * @param monitors ڑĂfBXvC
	 * @param x XW
	 * @param y YW
	 * @param display Display
	 * @return YMonitor̃NCAg̈B݂Ȃꍇ̓vC}fBXvCMonitor̃NCAg̈B
	 */
	public static Rectangle getMonitorClientByCoordinate(Monitor[] monitors,
			int x, int y, Display display) {
		int index = getMonitorIndexByCoordinate(monitors, x, y);
		if (index == -1) {
			return display.getPrimaryMonitor().getClientArea();
		}
		return monitors[index].getClientArea();
	}

	/**
	 * rectbase̓ɓ悤ɈړETCY܂B
	 * 
	 * @param base ̋`(NCAg̈)
	 * @param rect ړΏۂ̋`(EBhE)
	 */
	public static void moveAndPack(Rectangle base, Rectangle rect) {
		if (base.contains(rect.x, rect.y)
				&& base.contains(rect.x + rect.width, rect.y + rect.height)) {
			// łɓɂ
			return;
		}
		// 傫
		rect.width = Math.min(rect.width, base.width - 10);
		rect.height = Math.min(rect.height, base.height - 10);

		// ʒu
		if (base.y + base.height < rect.y + rect.height) {
			// 
			rect.y = base.y + base.height - rect.height - 5;
		} else if (base.y > rect.y) {
			// 
			rect.y = base.y + 5;
		}

		if (base.x + base.width < rect.x + rect.width) {
			// E
			rect.x = base.x + base.width - rect.width - 5;
		} else if (base.x > rect.x) {
			// 
			rect.x = base.x + 5;
		}
	}

	/**
	 * base̒rectړ܂B
	 * 
	 * @param base ̋`(NCAg̈)
	 * @param rect ړΏۂ̋`(EBhE)
	 */
	private static void moveToCenter(Rectangle base, Rectangle rect) {
		int center = base.x + base.width / 2;
		rect.x = center - rect.width / 2;

		center = base.y + base.height / 2;
		rect.y = center - rect.height / 2;
		//System.out.println(rect);
	}

	/**
	 * Shell|C^fBXvC̃NCAg̈̒Ɉړ܂B
	 * @param shell ړShell
	 */
	public static void moveToCenter(Shell shell) {
		Display display = Display.getDefault();
		Monitor[] monitors = display.getMonitors();
		moveToCenter(display, shell, monitors);
	}

	/**
	 * Shell|C^fBXvC̃NCAg̈̒Ɉړ܂B
	 * @param display Display
	 * @param shell ړShell
	 * @param monitors ڑĂfBXvC
	 */
	public static void moveToCenter(Display display, Shell shell,
			Monitor[] monitors) {
		Monitor monitor = getPointedMonitor(display, monitors);
		moveToCenter(shell, monitor);
	}

	/**
	 * EBhEw肳ꂽfBXvC̃NCAg̈̒Ɉړ܂B
	 * 
	 * @param shell EBhE
	 * @param monitor fBXvC
	 */
	public static void moveToCenter(Shell shell, Monitor monitor) {
		Rectangle client = monitor.getClientArea();
		Rectangle shellRect = shell.getBounds();
		moveToCenter(client, shellRect);
		shell.setBounds(shellRect);
	}

	/**
	 * o[ToolTipp̒_Ԃ܂B
	 * 
	 * @param base ̋`(NCAg̈)
	 * @param rect _`
	 * @return rect̂ꂩ̒_
	 */
	public static Point getVertexForBalloon(Rectangle base, Rectangle rect) {
		// o[͔zu\ȌEɕ\
		// o[̃TCY350,100Ɖ肵Ă
		// Windows XP {ł̓o[̃TCY275,93
		int x;
		if (base.x + base.width - 350 > rect.x + rect.width) {
			// E
			x = rect.x + rect.width;
		} else {
			x = rect.x;
		}

		int y;
		if (base.y + base.height - 100 > rect.y + rect.height) {
			// 
			y = rect.y + rect.height;
		} else {
			y = rect.y;
		}

		return new Point(x, y);
	}

	/**
	 * G[bZ[W\܂B
	 * 
	 * @param parent eShell
	 * @param msg \郁bZ[W
	 */
	public static void showError(Shell parent, String msg) {
		showMessageDialog(parent, mm.get("SWTUtils.error"), //$NON-NLS-1$
				msg, SWT.ICON_ERROR | SWT.OK);
	}

	/**
	 * xbZ[W\܂B
	 * 
	 * @param parent eShell
	 * @param msg \郁bZ[W
	 */
	public static void showWarning(Shell parent, String msg) {
		showMessageDialog(parent, mm.get("SWTUtils.warning"), //$NON-NLS-1$
				msg, SWT.ICON_WARNING | SWT.OK);
	}

	/**
	 * 񃁃bZ[W\܂B
	 * 
	 * @param parent eShell
	 * @param title ^Cg
	 * @param msg \郁bZ[W
	 */
	public static void showInformation(Shell parent, String title, String msg) {
		showMessageDialog(parent, title, msg, SWT.ICON_INFORMATION | SWT.OK);
	}

	/**
	 * \Čʂ擾܂B
	 * 
	 * @param parent eShell
	 * @param title ^Cg
	 * @param msg \郁bZ[W
	 * @return YEŜƂtrue
	 */
	public static boolean showConfirm(Shell parent, String title, String msg) {
		return showConfirmDialog(parent, title, msg, SWT.ICON_QUESTION
				| SWT.YES | SWT.NO, SWT.YES);
	}

	/**
	 * G[bZ[W\čĎs邩ǂ擾܂B
	 * 
	 * @param parent eShell
	 * @param title ^Cg
	 * @param msg \郁bZ[W
	 * @return RETRŶƂtrue
	 */
	public static boolean showErrorAndConfirm(Shell parent, String title,
			String msg) {
		return showConfirmDialog(parent, title, msg, SWT.ICON_ERROR | SWT.RETRY
				| SWT.CANCEL, SWT.RETRY);
	}

	/**
	 * \ SWT.YES, SWT.NO, SWT.CANCEL ̂ꂩ擾܂B
	 * 
	 * @param parent eShell
	 * @param title ^Cg
	 * @param msg \郁bZ[W
	 * @return SWT.YES, SWT.NO, SWT.CANCEL̂ꂩ
	 */
	public static int showCancellableConfirm(Shell parent, String title,
			String msg) {
		return showMessageDialog(parent, title, msg, SWT.ICON_QUESTION
				| SWT.YES | SWT.NO | SWT.CANCEL);
	}

	/**
	 * [U[ɐq˂`̃_CAO\܂B
	 * 
	 * @param parent eShell
	 * @param title ^Cg
	 * @param msg \郁bZ[W
	 * @param style bZ[W̃X^C
	 * @param trueStyle trueŕԂ{^X^C
	 * @return trueStyleŎw肵{^ꍇtrue
	 */
	private static boolean showConfirmDialog(Shell parent, String title,
			String msg, int style, int trueStyle) {
		MessageBox box = new MessageBox(parent, style);
		box.setText(title);
		box.setMessage(msg);
		return box.open() == trueStyle;
	}

	/**
	 * bZ[W\܂B
	 * @param parent eShell
	 * @param title ^Cg
	 * @param msg \郁bZ[W
	 * @param style bZ[W̃X^C(SWT.ICON_*)
	 * @return {^̃X^C萔
	 */
	private static int showMessageDialog(Shell parent, String title,
			String msg, int style) {
		MessageBox box = new MessageBox(parent, style);
		box.setText(title);
		box.setMessage(msg);
		return box.open();
	}

	/**
	 * Shell擾܂B
	 * 
	 * @param widget Shell܂Control
	 * @return Shell܂null
	 */
	public static Shell getShellFromWidget(Widget widget) {
		if (widget instanceof Shell) {
			return (Shell) widget;
		}
		if (widget instanceof Control) {
			return ((Control) widget).getShell();
		}

		return null;
	}

	/**
	 * vbgtH[CarbonǂԂ܂B
	 * 
	 * @return carbon̏ꍇtrue
	 */
	public static boolean isCarbon() {
		return "carbon".equals(SWT.getPlatform()); //$NON-NLS-1$
	}

	/**
	 * e[u̗𐶐܂B
	 * 񖼂͍񂹂łB
	 * 
	 * @param table e[u
	 * @param sortListener \[gp̃Xi[BnullB
	 * @param width B̒l͖܂B
	 * @param key 񖼂̃L[
	 * @return TableColumn
	 */
	public static TableColumn createTableColumn(Table table,
			SelectionListener sortListener, int width, String key) {
		TableColumn column = new TableColumn(table, SWT.LEFT);
		column.setText(mm.get(key));
		if (width >= 0) {
			column.setWidth(width);
		}
		if (sortListener != null) {
			column.addSelectionListener(sortListener);
		}
		return column;
	}

	/**
	 * IMEŕҏWǂԂ܂B
	 * @param widget EBWFbg
	 * @return IMEҏW̏ꍇtrue
	 */
	public static boolean isIMEEditing(Widget widget) {
		if (widget instanceof Canvas) {
			Canvas w = (Canvas) widget;
			IME ime = w.getIME();
			if (ime != null) {
				return ime.getCompositionOffset() >= 0;
			}
		}
		return false;
	}

	/**
	 * Rg[ɐݒ肳ꂽ|bvAbvj[Rg[̒ɕ\܂B
	 * @param event Rg[̑ICxg
	 */
	public static void showButtonPopupMenu(SelectionEvent event) {
		final Control control = (Control) event.widget;
		Menu menu = control.getMenu();
		if (menu != null) {
			Point size = control.getSize();
			Point point = control.toDisplay(0, size.y);
			menu.setLocation(point);
			menu.setVisible(true);
		}
	}

}
