package jp.takayuki.tools;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Properties;

/**
 * vpeB[Ɋւ郆[eBeBNXłB
 * @author Takayuki Shoji
 */
public class PropertiesUtils {
	/**
	 * vf0̔z
	 */
	private static final String[] STRING_ARRAY = new String[0];

	/**
	 * booleanƂĎ擾邽߂̃[eBeB\bhB
	 * @param prop vpeB
	 * @param key L[
	 * @param defaultValue ftHgl
	 * @return l&quot;0&quot;̂ƂfalseAȊOtrueԂ܂B
	 * null󕶎̏ꍇ̓ftHgԂ܂B
	 * true, false󂯕t܂BȊO̓̕ftHgԂ܂B
	 */
	public static boolean getBoolean(Properties prop, String key,
			boolean defaultValue) {
		String value = (String) prop.get(key);
		if (value == null || value.length() == 0) {
			return defaultValue;
		}
		if (Character.isDigit(value.charAt(0))) {
			return !value.equals("0"); //$NON-NLS-1$
		}

		if (value.equals("true")) { //$NON-NLS-1$
			return true;
		}
		if (value.equals("false")) { //$NON-NLS-1$
			return false;
		}
		return defaultValue;
	}

	/**
	 * booleanۑ邽߂̃[eBeB\bhB
	 * true'1'Afalse'0'Ƃĕۑ܂B
	 * @param prop vpeB
	 * @param key L[
	 * @param tf true܂false
	 */
	public static void setBoolean(Properties prop, String key, boolean tf) {
		prop.setProperty(key, tf ? "1" : "0"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * intƂĎ擾邽߂̃[eBeB\bhB
	 * @param prop vpeB
	 * @param key L[
	 * @param defaultValue ftHgl
	 * @return l̏ꍇ͂̒lAȊȌꍇ̓ftHglԂ܂B
	 */
	public static int getInt(Properties prop, String key, int defaultValue) {
		String value = (String) prop.get(key);
		if (value == null) {
			return defaultValue;
		}
		try {
			return Integer.parseInt(value);
		} catch (NumberFormatException e) {
			return defaultValue;
		}
	}

	/**
	 * intݒ肷邽߂̃[eBeB\bhB
	 * @param prop vpeB
	 * @param key L[
	 * @param value ݒ肷l
	 */
	public static void setInt(Properties prop, String key, int value) {
		prop.setProperty(key, Integer.toString(value));
	}

	/**
	 * longƂĎ擾邽߂̃[eBeB\bhB
	 * @param prop vpeB
	 * @param key L[
	 * @param defaultValue ftHgl
	 * @return l̏ꍇ͂̒lAȊȌꍇ̓ftHglԂ܂B
	 */
	public static long getLong(Properties prop, String key, long defaultValue) {
		String value = (String) prop.get(key);
		if (value == null) {
			return defaultValue;
		}
		try {
			return Long.parseLong(value);
		} catch (NumberFormatException e) {
			return defaultValue;
		}
	}

	/**
	 * longݒ肷邽߂̃[eBeB\bh
	 * @param prop vpeB
	 * @param key L[
	 * @param value ݒ肷l
	 */
	public static void setLong(Properties prop, String key, long value) {
		prop.setProperty(key, Long.toString(value));
	}

	/**
	 * vpeBURLƂĎ擾܂B
	 * @param prop vpeB
	 * @param key L[
	 * @return URL܂null
	 */
	public static URL getURL(Properties prop, String key) {
		String urlStr = prop.getProperty(key);
		if (urlStr == null) {
			return null;
		}
		try {
			return new URL(urlStr);
		} catch (MalformedURLException e) {
			return null;
		}
	}

	/**
	 * vpeBt@CƂĎ擾܂B
	 * @param prop vpeB
	 * @param key L[
	 * @param exists ݂mF邩ǂ
	 * @param dir ݂ꍇAfBNgǂ
	 * @return File
	 */
	public static File getFile(Properties prop, String key, boolean exists,
			boolean dir) {
		String fileStr = prop.getProperty(key);
		if (fileStr == null || fileStr.length() == 0) {
			return null;
		}
		File file = new File(fileStr);
		if (exists) {
			if (dir && file.isDirectory()) {
				return file;
			} else if (file.isFile()) {
				return file;
			}
			return null;
		}
		return file;
	}

	/**
	 * w肳ꂽFileProperties擾܂B
	 * @param file t@C
	 * @param defaults ftHgl
	 * @return PropertiesIuWFNgBfile݂Ȃ܂͓ǂ߂ȂꍇnullB
	 * @throws IOException 
	 */
	public static Properties getPropertiesFromFile(File file,
			Properties defaults) throws IOException {
		if (!file.canRead()) {
			return null;
		}

		InputStream in = new FileInputStream(file);
		return loadProperties(in, defaults);
	}

	/**
	 * \[XProperties擾܂B
	 * @param path \[XpX
	 * @param defaults ftHgl
	 * @return PropertiesIuWFNgBpath݂ȂꍇnullB
	 * @throws IOException
	 */
	public static Properties getPropertiesFromResource(String path,
			Properties defaults) throws IOException {
		InputStream in = PropertiesUtils.class.getResourceAsStream(path);
		if (in == null) {
			return new Properties(defaults);
		}
		return loadProperties(in, defaults);
	}

	/**
	 * Xg[Propertiesǂݍ݁AXg[܂B
	 * @param in ̓Xg[
	 * @param defaults ftHgl
	 * @return PropertiesIuWFNgB
	 * @throws IOException
	 */
	public static Properties loadProperties(InputStream in, Properties defaults)
			throws IOException {
		try {
			Properties prop = new Properties(defaults);
			prop.load(in);
			return prop;
		} finally {
			in.close();
		}
	}

	/**
	 * w肵FilePropertiesۑ܂B
	 * @param file ۑ
	 * @param prop ۑvpeB
	 * @param comments Rg
	 * @throws IOException 
	 */
	public static void saveProperties(File file, Properties prop,
			String comments) throws IOException {
		FileOutputStream fo = null;
		try {
			fo = new FileOutputStream(file);
			prop.store(fo, comments);
		} finally {
			if (fo != null) {
				fo.close();
			}
		}
	}

	/**
	 * vpeB̒lJ}ŋ؂ĔzŎ擾܂B
	 * 󔒂̗vf͏O܂B
	 * 
	 * @param prop vpeB
	 * @param key L[
	 * @param allowDuplication dꍇtrue
	 * @return ̔z
	 */
	public static String[] getArray(Properties prop, String key,
			final boolean allowDuplication) {
		String value = prop.getProperty(key, ""); //$NON-NLS-1$
		if (value.length() == 0) {
			return STRING_ARRAY;
		}
		// 
		String[] confIds = Tool.split(value, ',', 0);

		// 󔒂̗vfƏd폜
		HashSet set = allowDuplication ? null : new HashSet();
		ArrayList result = new ArrayList(confIds.length);
		for (int i = 0; i < confIds.length; i++) {
			confIds[i] = confIds[i].trim();
			if (confIds.length == 0) {
				continue;
			}
			boolean notExists = allowDuplication ? true : set.add(confIds[i]);
			if (notExists) {
				result.add(confIds[i]);
			}
		}
		if (set != null)
			set.clear();
		return (String[]) result.toArray(new String[result.size()]);
	}
}
