package jp.takayuki.tools;

import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.List;

/**
 * ؂t@C[_
 * @author Takayuki Shoji
 */
public class SeparatedValueReader extends Reader {
	/**
	 * ؂蕶
	 */
	private char separator;

	/**
	 * eReader
	 */
	private Reader r;

	/**
	 * ǂݖ߂pobt@
	 */
	private int unget = -2;

	/**
	 * RXgN^
	 * @param r Reader
	 * @param separator ؂蕶
	 */
	public SeparatedValueReader(Reader r, char separator) {
		this.r = r;
		this.separator = separator;
	}

	/**
	 * RXgN^B^u؂t@C
	 * @param r Reader
	 */
	public SeparatedValueReader(Reader r) {
		this(r, '\t');
	}

	/**
	 * T|[gĂ܂B
	 * @see java.io.Reader#read(char[], int, int)
	 */
	public int read(char[] cbuf, int off, int len) throws IOException {
		return -1;
	}

	/**
	 * ׂẴR[hǂ݂܂B
	 * @return XgBl: String[]B
	 * @throws IOException
	 */
	public List readAllRecords() throws IOException {
		ArrayList list = new ArrayList();

		while (true) {
			if (isEof()) {
				break;
			}
			String[] record = readRecord();
			list.add(record);
		}

		return list;
	}

	/**
	 * 1R[hǂ݂܂B
	 * @return J̔zBEOF̏ꍇ͋󕶎1̔zB
	 * @throws IOException 
	 */
	public String[] readRecord() throws IOException {

		ArrayList list = new ArrayList();

		while (true) {
			String col = readToken();
			if (col == null) {
				break;
			}
			list.add(col);
		}

		return (String[]) list.toArray(new String[list.size()]);
	}

	/**
	 * soꍇtrueɂĂ
	 */
	private boolean endOfLine = false;

	/**
	 * J1ǂݍ݂܂B
	 * @return Bs̏ꍇnullBEOF̏ꍇnull
	 * @throws IOException 
	 */
	public String readToken() throws IOException {
		StringBuffer sb = new StringBuffer();

		if (endOfLine) {
			endOfLine = false;
			return null;
		}

		boolean quote = false;
		int c = getc();

		if (c == separator) {
			return ""; //$NON-NLS-1$
		}
		switch (c) {
		case '"':
			quote = true;
			break;

		case '\r':
			int next = getc();
			if (next == '\n') {
				// CRLF̂Ƃ
				endOfLine = true;
				return ""; //$NON-NLS-1$
			}
			// CR̂
			ungetc(next);
			endOfLine = true;
			return ""; //$NON-NLS-1$

		case '\n':
			// LF̂
			endOfLine = true;
			return ""; //$NON-NLS-1$

		case -1:
			// EOF
			endOfLine = true;
			return ""; //$NON-NLS-1$

		default:
			sb.append((char) c);
			break;
		}

		while (true) {
			c = getc();

			if (quote && c == '"') {
				int next = getc();
				if (next != '"') {
					if (next != separator) {
						// p̃GXP[vȊÔƂ
						switch (next) {

						case '\r':
							next = getc();
							if (next != '\n') {
								// CR݂̂̂Ƃ͖߂
								ungetc(next);
							}
							endOfLine = true;
							break;

						case '\n':
							endOfLine = true;
							break;

						case -1:
							endOfLine = true;
							break;

						default:
							break;
						}

					}

					return sb.toString();
				}
			} else if (!quote) {

				if (c == separator) {
					break;
				} else if (c == '\r') {
					int next = getc();
					if (next != '\n') {
						ungetc(next);
					}
					endOfLine = true;
					break;
				} else if (c == '\n') {
					endOfLine = true;
					break;
				} else if (c == -1) {
					endOfLine = true;
					break;
				}
			}

			sb.append((char) c);
		}
		return sb.toString();
	}

	/**
	 * 1擾܂B
	 * @return 1BEOF-1ԂB
	 * @throws IOException
	 */
	private int getc() throws IOException {
		// -1EOFȂ̂-2ǂݍݍς݃tOƂăZbg܂B
		if (unget < -1) {
			return r.read();
		}
		int c = unget;
		unget = -2;
		return c;
	}

	/**
	 * t@C̏I[ɒB𔻒肵܂B
	 * @return EOF̏ꍇtrue
	 * @throws IOException
	 */
	public boolean isEof() throws IOException {
		int c = getc();
		ungetc(c);
		return c == -1;
	}

	/**
	 * 1ǂݖ߂܂B
	 * @param c 
	 */
	private void ungetc(int c) {
		unget = c;
	}

	/**
	 * ܂B
	 * @see java.io.Reader#close()
	 */
	public void close() throws IOException {
		if (r != null) {
			r.close();
			r = null;
		}
	}

	/**
	 * IuWFNgėp܂B
	 * RXgN^Ŏ󂯎Reader͕܂B
	 * @param r Reader
	 * @throws IOException 
	 */
	public void reuse(Reader r) throws IOException {
		close();
		this.r = r;
		this.unget = -2;
	}
}
