package jp.takayuki.tools;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.URL;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Enumeration;
import java.util.Properties;

import jp.takayuki.logging.Logger;

/**
 * o[W擾郆[eBeBNX
 * 
 * @author Takayuki Shoji
 */
public class Version {

	/** log */
	private static final Logger log = Logger.getLogger(Version.class);

	/**
	 * \[X̎擾ɎsꍇStringԂ\bh̖߂lƂĎgl
	 */
	private static final String UNKNOWN = "unknown"; //$NON-NLS-1$

	/**
	 * [hς݃\[X̃LbV
	 */
	private static Version CACHE;

	/**
	 * _E[ht@C
	 */
	public static final String FILE = "file"; //$NON-NLS-1$

	/**
	 * Stage: General Availability
	 */
	private static final String GA = "GA"; //$NON-NLS-1$

	/**
	 * JXe[W: GAȊO͊J
	 */
	public static final String VERSION_STAGE = "version.stage"; //$NON-NLS-1$

	/**
	 * o[W
	 */
	public static final String VERSION_TEXT = "version.text"; //$NON-NLS-1$

	/**
	 * rhԍ
	 */
	public static final String BUILD_NUMBER = "build.number"; //$NON-NLS-1$

	/**
	 * AppPropertiesƃ}[Wۂ̃vtBbNX
	 */
	public static final String PREFIX = "app.updateCache."; //$NON-NLS-1$

	/**
	 * [Xm[gURL
	 */
	public static final String RELEASE_NOTE = "releaseNote.url"; //$NON-NLS-1$

	/**
	 * vpeB[
	 */
	private Properties prop;

	/**
	 * RXgN^
	 * @param prop Properties
	 */
	public Version(Properties prop) {
		this.prop = prop;
	}

	/**
	 * @return Properties
	 */
	public Properties getProperties() {
		return prop;
	}

	/**
	 * @return rhԍ
	 */
	public String getBuildNumberStr() {
		return prop.getProperty(BUILD_NUMBER);
	}

	/**
	 * @return rhԍB[J\[X̃[hɎsꍇ -1B
	 */
	public int getBuildNumberAsInt() {
		return PropertiesUtils.getInt(prop, BUILD_NUMBER, -1);
	}

	/**
	 * @return o[W
	 */
	public String getVersionText() {
		return prop.getProperty(VERSION_TEXT);
	}

	/**
	 * @return Xe[W
	 */
	public String getStage() {
		return prop.getProperty(VERSION_STAGE, GA);
	}

	/**
	 * @return Jo[WǂԂ܂B
	 */
	public boolean isDevelopmentStage() {
		return !getStage().equalsIgnoreCase(GA);
	}

	/**
	 * @param file _E[ht@C
	 */
	public void setFile(File file) {
		prop.setProperty(FILE, file.getAbsolutePath());
	}

	/**
	 * @return _E[ht@C
	 */
	public File getFile() {
		return PropertiesUtils.getFile(prop, FILE, true, false);
	}

	/**
	 * @return [Xm[gURL
	 */
	public URL getReleaseNoteURL() {
		return PropertiesUtils.getURL(prop, RELEASE_NOTE);
	}

	/**
	 * \[Xversion.properties[h܂B
	 * @return 擾łȂꍇ̓_~[̃IuWFNgԂ܂B
	 */
	public static synchronized Version loadFromResource() {
		if (CACHE != null) {
			return CACHE;
		}
		try {
			Properties p = PropertiesUtils.getPropertiesFromResource(
					"/version.properties", null); //$NON-NLS-1$
			CACHE = new Version(p);
			return CACHE;
		} catch (IOException e) {
			Properties p = new Properties();
			p.setProperty(BUILD_NUMBER, UNKNOWN);
			p.setProperty(VERSION_STAGE, UNKNOWN);
			p.setProperty(VERSION_TEXT, UNKNOWN);
			p.setProperty(RELEASE_NOTE, UNKNOWN);
			return new Version(p);
		}
	}

	/**
	 * o[W擾܂B
	 * <p>
	 * JXe[WGȀꍇAversion.propertiesB
	 * ȊO̊JXe[ẂAꂼ̊JXe[W̏ƃhbg擪ɂB
	 * <p>
	 * : alpha.version.properties, beta.version.properties
	 * @param localVersion ݂̃o[W
	 * @return ǂݍProperties
	 * @throws IOException
	 */
	public static Version loadFromInternet(Version localVersion)
			throws IOException {
		ByteArrayOutputStream out = new ByteArrayOutputStream(1024 * 4);
		StringBuffer urlStr = new StringBuffer(96);
		urlStr.append("http://hp.vector.co.jp/authors/VA035113/petaclone/"); //$NON-NLS-1$
		if (localVersion.isDevelopmentStage()) {
			// JpXV`l
			urlStr.append(localVersion.getStage().toLowerCase()).append('.');
		}
		urlStr.append("version.properties"); //$NON-NLS-1$
		URL url = new URL(urlStr.toString());
		if (Tool.download(url, out) == HttpURLConnection.HTTP_OK) {
			ByteArrayInputStream in = new ByteArrayInputStream(
					out.toByteArray());
			Properties prop = PropertiesUtils.loadProperties(in, null);
			return new Version(prop);
		}
		return null;
	}

	/**
	 * @param postfix
	 * @return L[
	 */
	private String createKey(String postfix) {
		return "download." + Os.getPlatform() + postfix; //$NON-NLS-1$
	}

	/**
	 * @return URL܂null
	 */
	public URL getDownloadUrl() {
		return PropertiesUtils.getURL(prop, createKey(".url")); //$NON-NLS-1$
	}

	/**
	 * @return _E[h^CvB܂-1B
	 */
	public int getDownloadType() {
		return PropertiesUtils.getInt(prop, createKey(".type"), -1); //$NON-NLS-1$
	}

	/**
	 * _E[ht@C؂܂B
	 * @param file Ώۂ̃t@C
	 * @return ̏ꍇtrueȀُꍇfalse
	 * @throws IOException
	 */
	public boolean validateFile(File file) throws IOException {
		if (file == null || !file.isFile()) {
			return false;
		}
		try {
			MessageDigest md5 = MessageDigest.getInstance("MD5"); //$NON-NLS-1$
			MessageDigest sha1 = MessageDigest.getInstance("SHA1"); //$NON-NLS-1$

			MessageDigest[] mds = new MessageDigest[] { md5, sha1 };

			Tool.getDigest(file, mds);

			for (int j = 0; j < mds.length; j++) {
				StringBuffer sb = new StringBuffer(48);
				sb.append('{').append(mds[j].getAlgorithm()).append('}');
				sb.append(Tool.bytesToHexString(mds[j].digest()));

				// r
				String key = createKey('.' + mds[j].getAlgorithm()
						.toLowerCase());
				String hash = prop.getProperty(key);
				if (hash == null || !hash.equalsIgnoreCase(sb.toString())) {
					log.info("Invalid hash: key=" + key + ", expect=" + hash //$NON-NLS-1$ //$NON-NLS-2$
							+ ", actual=" + sb.toString()); //$NON-NLS-1$
					Enumeration names = prop.propertyNames();
					while (names.hasMoreElements()) {
						String name = (String) names.nextElement();
						log.info(name + "=" + prop.getProperty(name)); //$NON-NLS-1$
					}
					return false;
				}
			}
			return true;

		} catch (NoSuchAlgorithmException e) {
			throw new IOException("Failed to check hash code.", e); //$NON-NLS-1$
		}
	}

}
