package petaclone;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Iterator;
import java.util.Map.Entry;
import java.util.Properties;

import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Os;
import jp.takayuki.tools.PropertiesUtils;
import jp.takayuki.tools.Tool;
import jp.takayuki.tools.Version;

/**
 * y^N[̏ݒIuWFNgłB
 * t@CóAݒl̎擾Ɛݒs܂B
 * ܂AsIuWFNgێĂ܂B
 * @author Takayuki Shoji
 */
public class AppPreferences {

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap msg = MessageMap.getMessageMap();

	/** ݒt@C */
	private static final String PREFERENCE_FILENAME = "petaclone.properties"; //$NON-NLS-1$

	/** ݒt@C̃wb_ */
	private static final String PREFERENCE_HEADER = "Petaclone configuration file."; //$NON-NLS-1$

	/** petaclone.properties */
	private Properties properties;

	/** ݒfBNg */
	private final File basePath;

	/** 񃍁[hǂ */
	private boolean firsttime = false;

	/** s̒l */
	private RuntimeProperties runtime = new RuntimeProperties();

	/**
	 * application policy
	 */
	private Policy policy;

	/**
	 * Policyݒ肵܂B
	 * @param policy application policy
	 */
	public void setPolicy(Policy policy) {
		this.policy = policy;
	}

	/**
	 * Policy擾܂B
	 * @return application policy
	 */
	public Policy getPolicy() {
		return this.policy;
	}

	/**
	 * ݒ\l
	 * <p>
	 * Win32POINT\̂Cint
	 */
	public static final int IMPOSSIBLE = 0x8000;

	/**
	 * RXgN^
	 * @param basePath ݒfBNg
	 */
	public AppPreferences(File basePath) {
		if (basePath == null) {
			throw new IllegalArgumentException("this.basePath == null"); //$NON-NLS-1$
		}
		this.basePath = basePath;
	}

	/**
	 * PropertiesIuWFNgԂ܂B
	 * ̃\bh͏ݒ_CAÔ߂ɗpӂĂ܂B
	 * @return PropertiesIuWFNg
	 */
	public Properties getProperties() {
		return properties;
	}

	/**
	 * ݒtH_ւ̃pXԂ܂B
	 * @return ݒtH_
	 */
	public File getBasePath() {
		return this.basePath;
	}

	/**
	 * NǂԂ܂B
	 * load()̌ďoɐlԂ悤ɂȂ܂B
	 * @return NȂtrueԂB
	 */
	public boolean isFirsttime() {
		return firsttime;
	}

	/**
	 * sԂ܂B
	 * @return RuntimeProperties
	 */
	public RuntimeProperties getRuntime() {
		return runtime;
	}

	/**
	 * ݒt@Cݒ[h܂B
	 * @throws IOException o͂ŗOꍇ
	 */
	public void load() throws IOException {
		if (properties != null) {
			throw new IllegalStateException("this.properties != null"); //$NON-NLS-1$
		}

		// ftHg̃[h
		Properties defaults = PropertiesUtils.getPropertiesFromResource(
				"/petaclone_default.properties", null); //$NON-NLS-1$

		// t@C݂̑`FbN
		File file = new File(this.basePath, PREFERENCE_FILENAME);
		if (!file.exists()) {
			properties = new Properties(defaults);
			firsttime = true;
			return;
		}

		// t@C烍[h
		properties = PropertiesUtils.getPropertiesFromFile(file, defaults);
	}

	/**
	 * ݒݒt@Cɕۑ܂B
	 * @throws FileNotFoundException t@C쐬łȂꍇ
	 * @throws IOException o͂ŗOꍇ
	 */
	public void save() throws FileNotFoundException, IOException {
		// o̓Xg[p
		FileOutputStream fo = null;
		BufferedOutputStream out = null;
		try {
			File file = new File(basePath, PREFERENCE_FILENAME);
			fo = new FileOutputStream(file);
			out = new BufferedOutputStream(fo);
			this.properties.store(out, PREFERENCE_HEADER);
		} finally {
			if (out != null) {
				out.close();
			} else if (fo != null) {
				fo.close();
			}
		}
	}

	//
	// evpeB
	//

	/** bZW[̃[U[̃vpeB */
	public static final String USER_NAME = "app.userName"; //$NON-NLS-1$

	/** ftHg̃[U[(Rs[^ւ̃OC) */
	public static final String USER_NAME_DEFAULT = System
			.getProperty("user.name"); //$NON-NLS-1$

	/**
	 * [U擾܂B
	 * @return [U
	 */
	public String getUserName() {
		String user = properties.getProperty(USER_NAME, USER_NAME_DEFAULT);
		if (user.length() == 0) {
			return USER_NAME_DEFAULT;
		}
		return user;
	}

	/**
	 * [U擾܂B
	 * @param maxLen ő咷(byte)
	 * @param charset Zbg
	 * @return [U
	 */
	public String getUserName(int maxLen, String charset) {
		StringBuffer nb = new StringBuffer(getUserName());
		while (true) {
			byte[] bytes;
			try {
				bytes = nb.toString().getBytes(charset);
			} catch (UnsupportedEncodingException e) {
				// 肦Ȃ
				return null;
			}
			if (bytes.length <= maxLen) {
				break;
			}
			nb.deleteCharAt(nb.length() - 1);
		}
		return nb.toString();
	}

	/** O[ṽvpeB */
	public static final String GROUP_NAME = "app.group"; //$NON-NLS-1$

	/**
	 * O[v擾܂B
	 * @return O[v
	 */
	public String getGroupName() {
		return properties.getProperty(GROUP_NAME, ""); //$NON-NLS-1$
	}

	/** Nɕs݃[hɂ邩̃vpeB */
	public static final String IS_ABSENCE_AT_STARTUP = "app.isAbsenceAtStartup"; //$NON-NLS-1$

	/**
	 * @return Nɕs݃[hɂꍇtrueԂ܂B
	 */
	public boolean isAbsenceAtStartup() {
		return PropertiesUtils.getBoolean(this.properties,
				IS_ABSENCE_AT_STARTUP, false);
	}

	/** sݒʒm̃vpeB */
	public static final String ABSENCE_MESSAGE = "app.absenceMessage"; //$NON-NLS-1$

	/** ftHg̕sݒʒm */
	private static final String ABSENCE_MESSAGE_DEFAULT = msg
			.get("AppPreferences.absenceMessage"); //$NON-NLS-1$

	/**
	 * sݒʒm擾܂B
	 * @return sݒʒm
	 */
	public String getAbsenceMessage() {
		return properties.getProperty(ABSENCE_MESSAGE, ABSENCE_MESSAGE_DEFAULT);
	}

	/**
	 * sݒʒmݒ肵܂B
	 * @param message sݒʒm
	 */
	public void setAbsenceMessage(String message) {
		properties.setProperty(ABSENCE_MESSAGE, message);
	}

	/** MbZ[WIɕۑ̃vpeB */
	public static final String SAVE_SENT_MESSAGE = "app.saveSentMessage"; //$NON-NLS-1$

	/**
	 * @return MbZ[WIɕۑꍇtrueԂ܂B
	 */
	public boolean getSaveSentMessage() {
		return PropertiesUtils.getBoolean(this.properties, SAVE_SENT_MESSAGE,
				false);
	}

	/** MbZ[WIɕۑ̃vpeB */
	public static final String SAVE_RECEIVED_MESSAGE = "app.saveReceivedMessage"; //$NON-NLS-1$

	/**
	 * @return MbZ[WIɕۑꍇtrueԂ܂B
	 */
	public boolean getSaveReceivedMessage() {
		return PropertiesUtils.getBoolean(this.properties,
				SAVE_RECEIVED_MESSAGE, false);
	}

	/** mainWindowTop̃vpeB */
	public static final String MAIN_WINDOW_TOP = "app.mainWindowTop"; //$NON-NLS-1$

	/**
	 * @return CEBhE̕\ʒuYW擾܂B
	 */
	public int getMainWindowTop() {
		return PropertiesUtils.getInt(this.properties, MAIN_WINDOW_TOP, 50);
	}

	/**
	 * CEBhE̕\ʒuYWݒ肵܂B
	 * @param y YW
	 */
	public void setMainWindowTop(int y) {
		PropertiesUtils.setInt(this.properties, MAIN_WINDOW_TOP, y);
	}

	/** CEBhẺ̃vpeB */
	public static final String MAIN_WINDOW_WIDTH = "app.mainWindowWidth"; //$NON-NLS-1$

	/**
	 * @return CEBhẺ擾܂B
	 */
	public int getMainWindowWidth() {
		return PropertiesUtils.getInt(this.properties, MAIN_WINDOW_WIDTH, 200);
	}

	/**
	 * CEBhẺݒ肵܂B
	 * @param x 
	 */
	public void setMainWindowWidth(int x) {
		PropertiesUtils.setInt(this.properties, MAIN_WINDOW_WIDTH, x);
	}

	/** CEBhE̍̃vpeB */
	public static final String MAIN_WINDOW_HEIGHT = "app.mainWindowHeight"; //$NON-NLS-1$

	/**
	 * @return CEBhE̍擾܂B
	 */
	public int getMainWindowHeight() {
		return PropertiesUtils.getInt(this.properties, MAIN_WINDOW_HEIGHT, 50);
	}

	/**
	 * CEBhE̍ݒ肵܂B
	 * @param x 
	 */
	public void setMainWindowHeoght(int x) {
		PropertiesUtils.setInt(this.properties, MAIN_WINDOW_HEIGHT, x);
	}

	/** mainWindowLeft̃vpeB */
	public static final String MAIN_WINDOW_LEFT = "app.mainWindowLeft"; //$NON-NLS-1$

	/**
	 * @return CEBhE̕\ʒuXW擾܂B
	 */
	public int getMainWindowLeft() {
		return PropertiesUtils.getInt(this.properties, MAIN_WINDOW_LEFT, 50);
	}

	/**
	 * CEBhE̕\ʒuXWݒ肵܂B
	 * @param x XW
	 */
	public void setMainWindowLeft(int x) {
		PropertiesUtils.setInt(this.properties, MAIN_WINDOW_LEFT, x);
	}

	/** ۊǔYW */
	private static final String STORAGE_WINDOW_TOP = "app.storageWindowTop"; //$NON-NLS-1$

	/**
	 * @return ۊǔEBhE̕\ʒuYW擾܂B
	 */
	public int getStorageWindowTop() {
		return PropertiesUtils.getInt(this.properties, STORAGE_WINDOW_TOP,
				IMPOSSIBLE);
	}

	/**
	 * ۊǔEBhE̕\ʒuYWݒ肵܂B
	 * @param y YW
	 */
	public void setStorageWindowTop(int y) {
		PropertiesUtils.setInt(this.properties, STORAGE_WINDOW_TOP, y);
	}

	/** ۊǔXW */
	private static final String STORAGE_WINDOW_LEFT = "app.storageWindowLeft"; //$NON-NLS-1$

	/**
	 * @return ۊǔEBhE̕\ʒuXW擾܂B
	 */
	public int getStorageWindowLeft() {
		return PropertiesUtils.getInt(this.properties, STORAGE_WINDOW_LEFT,
				IMPOSSIBLE);
	}

	/**
	 * ۊǔEBhE̕\ʒuXWݒ肵܂B
	 * @param x XW
	 */
	public void setStorageWindowLeft(int x) {
		PropertiesUtils.setInt(this.properties, STORAGE_WINDOW_LEFT, x);
	}

	/** ۊǔ̉ */
	private static final String STORAGE_WINDOW_WIDTH = "app.storageWindowWidth"; //$NON-NLS-1$

	/**
	 * @return ۊǔEBhẺ擾܂B
	 */
	public int getStorageWindowWidth() {
		return PropertiesUtils.getInt(this.properties, STORAGE_WINDOW_WIDTH,
				540);
	}

	/**
	 * ۊǔEBhẺݒ肵܂B
	 * @param width 
	 */
	public void setStorageWindowWidth(int width) {
		PropertiesUtils.setInt(this.properties, STORAGE_WINDOW_WIDTH, width);
	}

	/** ۊǔ̍ */
	private static final String STORAGE_WINDOW_HEIGHT = "app.storageWindowHeight"; //$NON-NLS-1$

	/**
	 * @return ۊǔEBhE̍擾܂B
	 */
	public int getStorageWindowHeight() {
		return PropertiesUtils.getInt(this.properties, STORAGE_WINDOW_HEIGHT,
				300);
	}

	/**
	 * ۊǔEBhE̍ݒ肵܂B
	 * @param height 
	 */
	public void setStorageWindowHeight(int height) {
		PropertiesUtils.setInt(this.properties, STORAGE_WINDOW_HEIGHT, height);
	}

	/** ۊǔ\Ăǂ */
	private static final String STORAGE_WINDOW_SHOWN = "app.storageWindowShown"; //$NON-NLS-1$

	/**
	 * ۊǔEBhE\Ăǂ擾܂B
	 * @return \Ăꍇtrue
	 */
	public boolean getStorageWindowShown() {
		return PropertiesUtils.getBoolean(this.properties,
				STORAGE_WINDOW_SHOWN, false);
	}

	/**
	 * ۊǔEBhE\Ă邩ǂݒ肵܂B
	 * @param shown \Ăꍇtrue
	 */
	public void setStorageWindowShown(boolean shown) {
		PropertiesUtils
				.setBoolean(this.properties, STORAGE_WINDOW_SHOWN, shown);
	}

	/** őOʂɕ\̃vpeB */
	public static final String KEEP_ON_TOP = "app.keepOnTop"; //$NON-NLS-1$

	/**
	 * @return őOʂɕ\ꍇtrueԂ܂B
	 */
	public boolean getKeepOnTop() {
		return PropertiesUtils.getBoolean(this.properties, KEEP_ON_TOP, false);
	}

	/** bZ[WEBhẺ̃vpeB */
	public static final String MESSAGE_WIDTH = "app.messageWidth"; //$NON-NLS-1$

	/**
	 * @return bZ[WEBhẺ擾܂B
	 */
	public int getMessageWidth() {
		return PropertiesUtils.getInt(this.properties, MESSAGE_WIDTH, 300);
	}

	/** bZ[WEBhE̍̃vpeB */
	public static final String MESSAGE_HEIGHT = "app.messageHeight"; //$NON-NLS-1$

	/**
	 * @return bZ[WEBhE̍擾܂B
	 */
	public int getMessageHeight() {
		return PropertiesUtils.getInt(this.properties, MESSAGE_HEIGHT, 250);
	}

	/** N\̃vpeB */
	public static final String LAUNCH_CONFIG = "app.launchConfiguration"; //$NON-NLS-1$

	/**
	 * N\(NID̔z)擾܂BNID̏d͎菜܂B
	 * N\̓J}؂łBNID͐łB
	 * @return NID̔zBݒ肪ꍇ͒0̔zԂ܂B
	 */
	public String[] getLaunchConfigurations() {
		return PropertiesUtils.getArray(properties, LAUNCH_CONFIG, false);
	}

	/**
	 * N\̃Xgݒ肵܂B
	 * @param confIds NID̃J}؂蕶
	 */
	public void setLaunchConfigurations(String confIds) {
		properties.setProperty(LAUNCH_CONFIG, confIds);
	}

	/** u[hLXgAhX̃vpeB */
	public static final String BROADCAST = "app.broadcast"; //$NON-NLS-1$

	/**
	 * AvP[Ṽu[hLXgAhXXg擾܂B
	 * @return List l: String IPAhX܂̓zXg
	 */
	public String[] getBroadcastAddresses() {
		return PropertiesUtils.getArray(properties, BROADCAST, true);
	}

	/** Limitedu[hLXgAhX̃vpeB */
	public static final String LIMITED_BROADCAST = "app.limitedBroadcast"; //$NON-NLS-1$

	/**
	 * Limited broadcast sǂ擾܂B
	 * @return Limited broadcast sꍇtrue
	 */
	public boolean getLimitedBroadcast() {
		return PropertiesUtils.getBoolean(properties, LIMITED_BROADCAST, true);
	}

	/** Directedu[hLXgAhX̃vpeB */
	public static final String DIRECTED_BROADCAST = "app.directedBroadcast"; //$NON-NLS-1$

	/**
	 * Limited broadcast sǂ擾܂B
	 * @return Limited broadcast sꍇtrue
	 */
	public boolean getDirectedBroadcast() {
		return PropertiesUtils.getBoolean(properties, DIRECTED_BROADCAST, true);
	}

	/** o^_CAÕ[U[D悷̃vpeB */
	public static final String USE_REGISTERED_USER_NAME = "app.useRegisteredUserName"; //$NON-NLS-1$

	/**
	 * @return o^ς݃[U[D悷ꍇtrueԂ܂B
	 */
	public boolean useRegisteredUserName() {
		return PropertiesUtils.getBoolean(this.properties,
				USE_REGISTERED_USER_NAME, false);
	}

	/** o^_CAÕO[vD悷̃vpeB */
	public static final String USE_REGISTERED_GROUP_NAME = "app.useRegisteredGroup"; //$NON-NLS-1$

	/**
	 * @return o^ς݃O[vD悷ꍇtrueԂ܂B
	 */
	public boolean useRegisteredGroupName() {
		return PropertiesUtils.getBoolean(this.properties,
				USE_REGISTERED_GROUP_NAME, false);
	}

	/** bZ[WɃTEh炷̃vpeB */
	public static final String PLAY_SOUND = "app.playSound"; //$NON-NLS-1$

	/** bZ[W̒ʒm@̃ftHgl */
	public static final String PLAY_SOUND_DEFAULT = "0"; //$NON-NLS-1$

	/**
	 * bZ[W̃TEh^Cv擾܂B
	 * @return 
	 * <ul>
	 * 	<li>0 - TEhĐȂ</li>
	 * 	<li>beep - n[hEFÃr[v炷</li>
	 * 	<li>sound - ʓrw肷TEht@CĐ</li>
	 * </ul>
	 */
	public String getPlaySound() {
		return properties.getProperty(PLAY_SOUND, PLAY_SOUND_DEFAULT);
	}

	/** TEht@C̏ꏊ̃vpeB */
	public static final String SOUND_URL = "app.soundURL"; //$NON-NLS-1$

	/**
	 * @return TEht@C̏ꏊ擾܂B
	 */
	public String getSoundUrl() {
		return properties.getProperty(SOUND_URL, ""); //$NON-NLS-1$
	}

	/** Ytt@C̈ꎞۑfBNg̃vpeB */
	public static final String ATTACHMENT_STORAGE = "app.attachmentStorage"; //$NON-NLS-1$

	/**
	 * @return Ytt@C̈ꎞۑfBNgԂ܂B
	 */
	public String getAttachmentStorage() {
		String value = properties.getProperty(ATTACHMENT_STORAGE);
		if (value != null && value.length() != 0) {
			return value;
		}
		// ftHg
		File f = new File(basePath, "attachment"); //$NON-NLS-1$
		return f.getPath();
	}

	/** _E[hYtt@C̍őTCỸvpeB */
	public static final String MAX_AUTO_DOWNLOAD_SIZE = "app.maxAutoDownloadSize"; //$NON-NLS-1$

	/**
	 * @return _E[hYtt@C̍őTCY擾܂B
	 */
	public int getMaxAutoDownloadSize() {
		return PropertiesUtils.getInt(this.properties, MAX_AUTO_DOWNLOAD_SIZE,
				20 * 1024 * 1024);
	}

	/** CC摜^E[h邩ǂ */
	public static final String INLINE_IMAGE_AUTO_DOWNLOAD = "app.InlineImageAutoDownload"; //$NON-NLS-1$

	/**
	 * @return CC摜_E[hꍇtrueԂ܂B
	 */
	public boolean useInlineImageAutoDownload() {
		return PropertiesUtils.getBoolean(this.properties,
				INLINE_IMAGE_AUTO_DOWNLOAD, true);
	}

	/** tH__E[h邩ǂ */
	public static final String FOLDER_AUTO_DOWNLOAD = "app.FolderAutoDownload"; //$NON-NLS-1$

	/**
	 * @return tH__E[hꍇtrueԂ܂B
	 */
	public boolean useFolderAutoDownload() {
		return PropertiesUtils.getBoolean(this.properties,
				FOLDER_AUTO_DOWNLOAD, false);
	}

	/** p̃vpeB */
	public static final String QUOTE = "app.quote"; //$NON-NLS-1$

	/** p̃ftHgl */
	public static final String QUOTE_DEFAULT = ">"; //$NON-NLS-1$

	/**
	 * @return p̍s擾܂B
	 */
	public String getQuoteString() {
		return properties.getProperty(QUOTE, QUOTE_DEFAULT);
	}

	/** uSɑMvgpȂ̃vpeB */
	public static final String NEVER_USE_SEND_ALL = "app.neverSendAll"; //$NON-NLS-1$

	/**
	 * @return uSɑMvgpȂȂtrueԂ܂B
	 */
	public boolean getNeverUseSendAll() {
		return PropertiesUtils.getBoolean(this.properties, NEVER_USE_SEND_ALL,
				true);
	}

	/** bZ[WEBhEIMLɂ */
	public static final String USE_IM_MSG = "app.useIMAtMesssageWindow"; //$NON-NLS-1$

	/**
	 * @return bZ[WEBhEIMLɂȂtrueԂ܂B
	 */
	public boolean useImInMessageWindow() {
		return PropertiesUtils.getBoolean(this.properties, USE_IM_MSG, true);
	}

	/** M_CAOIMLɂ */
	public static final String USE_IM_SEND = "app.useIMAtSendDialog"; //$NON-NLS-1$

	/**
	 * @return M_CAOIMLɂꍇtrueԂ܂B
	 */
	public boolean useImInSendDialog() {
		return PropertiesUtils.getBoolean(this.properties, USE_IM_SEND, false);
	}

	/** IMLɂꍇ̃[h[}͂ɂ邩ǂ */
	public static final String IM_MODE_IS_ROMAN = "app.IMModeIsRoman"; //$NON-NLS-1$

	/**
	 * @return IMLɂꍇ̃[h[}͂ɂȂtrueԂ܂B
	 */
	public boolean getImModeIsRoman() {
		return PropertiesUtils.getBoolean(this.properties, IM_MODE_IS_ROMAN,
				true);
	}

	/** VbZ[W̃`FbN{bNX̏Ԃ߂vpeB */
	public static final String PRIVACY_PROTECTION = "app.privacyProtection"; //$NON-NLS-1$

	/** PRIVACY_PROTECTIOÑftHgl */
	public static final String PRIVACY_PROTECTION_NONE = ""; //$NON-NLS-1$

	/** PRIVACY_PROTECTION̕\l */
	public static final String PRIVACY_PROTECTION_SEALED = "sealed"; //$NON-NLS-1$

	/** PRIVACY_PROTECTION̗vpX[h\l */
	public static final String PRIVACY_PROTECTION_PASSWORD = "password"; //$NON-NLS-1$

	/**
	 * VKbZ[W̃vCoV[ی샌x(EpX[h)Ԃ܂B
	 * @return <ul>
	 * 	<li>PRIVACY_PROTECTION_NONE     - ʏ</li>
	 * 	<li>PRIVACY_PROTECTION_SEALED   - </li>
	 * 	<li>PRIVACY_PROTECTION_PASSWORD - pX[h</li>
	 * </ul>
	 */
	public String getPrivacyProtectionLevel() {
		return properties.getProperty(PRIVACY_PROTECTION,
				PRIVACY_PROTECTION_NONE);
	}

	/** bZ[WOL^邩ǂ\vpeB */
	public static final String MESSAGE_LOG_ENABLED = "app.logEnabled"; //$NON-NLS-1$

	/**
	 * @return bZ[WOL^ꍇtrueԂ܂B
	 */
	public boolean getLogEnabled() {
		return PropertiesUtils.getBoolean(this.properties, MESSAGE_LOG_ENABLED,
				false);
	}

	/** bZ[WOt@C̈ʒu\vpeB */
	public static final String MESSAGE_LOG_PATH = "app.logPath"; //$NON-NLS-1$

	/**
	 * @return bZ[WO̕ۑpX擾܂B
	 */
	public String getMessageLogPath() {
		return properties.getProperty(MESSAGE_LOG_PATH, ""); //$NON-NLS-1$
	}

	/** Or[̈ʒu\vpeB */
	public static final String LOG_VIEWER = "app.logViewer"; //$NON-NLS-1$

	/**
	 * @return Or[̃pXԂ܂B
	 */
	public String getLogViewerPath() {
		return properties.getProperty(LOG_VIEWER, ""); //$NON-NLS-1$
	}

	/** x̊Ǌ|V[t@CCXg[Ă邩 */
	public static final String CIPHER_IS_UNLIMITED = "app.cipherIsUnlimited"; //$NON-NLS-1$

	/**
	 * @return x̊Ǌ|V[t@CCXg[Ăꍇ
	 * trueԂ܂B
	 */
	public boolean getCipherIsUnlimited() {
		return PropertiesUtils.getBoolean(this.properties, CIPHER_IS_UNLIMITED,
				false);
	}

	/** o^o[̃Xg\vpeB */
	public static final String REGISTERED_MEMBERS = "app.registeredMember"; //$NON-NLS-1$

	/**
	 * @return o^o[̃XgԂ܂B
	 */
	public String getRegisteredMembers() {
		return properties.getProperty(REGISTERED_MEMBERS, ""); //$NON-NLS-1$
	}

	/**
	 * o^o[̃Xgݒ肵܂B
	 * @param members o^o[̃Xg
	 */
	public void setRegisteredMembers(String members) {
		properties.setProperty(REGISTERED_MEMBERS, members);
	}

	/** Jʒm̃[Uւ̒ʒm@(\)\vpeB */
	public static final String OPEN_NOTICE_DISPLAY = "app.opennotice.display"; //$NON-NLS-1$

	/**
	 * Jʒm̃ftHg(o[)
	 */
	public static final int OPEN_NOTICE_DISPLAY_BALLOON = 0;

	/**
	 * JʒmA[gŕ\B
	 */
	public static final int OPEN_NOTICE_DISPLAY_ALERT = 1;

	/**
	 * Jʒm͉ȂB
	 */
	public static final int OPEN_NOTICE_DISPLAY_NONE = 2;

	/**
	 * Jʒm̃[Uւ̒ʒm@(\)Ԃ܂B
	 * @return 
	 * <ul>
	 * 	<li>OPEN_NOTICE_DISPLAY_DEFAULT: o[\</li>
	 * 	<li>OPEN_NOTICE_DISPLAY_ALERT: A[g\</li>
	 * 	<li>OPEN_NOTICE_DISPLAY_NONE: Ȃ</li>
	 * </ul>
	 */
	public int getOpenNoticeDisplay() {
		return PropertiesUtils.getInt(properties, OPEN_NOTICE_DISPLAY,
				OPEN_NOTICE_DISPLAY_BALLOON);
	}

	/** ŏԂŋN邩ǂ\vpeB */
	public static final String IS_MINIMIZED_AT_STARTUP = "app.isMinimizedAtStartup"; //$NON-NLS-1$

	/**
	 * ŏԂŋN邩ǂԂ܂B
	 * @return ŏԂŋNȂtrue
	 */
	public boolean isMinimizedAtStartup() {
		return PropertiesUtils.getBoolean(this.properties,
				IS_MINIMIZED_AT_STARTUP, false);
	}

	/** ̃bZ[W𑗐M邩ǂÃvpeB */
	public static final String NEVER_SEND_EMPTY_MESSAGE = "app.neverSendEmptyMessage"; //$NON-NLS-1$

	/**
	 * ̃bZ[W𑗐M邩ǂԂ܂B
	 * @return ftHgfalse(M)Ԃ܂B
	 */
	public boolean getNeverSendEmptyMessage() {
		return PropertiesUtils.getBoolean(this.properties,
				NEVER_SEND_EMPTY_MESSAGE, false);
	}

	/** vpX[hbZ[W̊JppX[h̃vpeB */
	public static final String OPEN_PASSWORD = "app.openPassword"; //$NON-NLS-1$

	/**
	 * vpX[hbZ[W̊JppX[hԂ܂B
	 * @return pX[h(nbV`ƃnbVꂽ)
	 */
	public String getOpenPassword() {
		return properties.getProperty(OPEN_PASSWORD, ""); //$NON-NLS-1$
	}

	/** 1s\ɑI͈͂̍ŏ̍sgÃvpeB */
	public static final String USE_FIRST_LINE_OF_SELECTION = "app.useFirstLineOfSelection"; //$NON-NLS-1$

	/**
	 * 1s\ɑI͈͂̍ŏ̍sgǂԂ܂B
	 * @return ŏ̍sgꍇtrue
	 */
	public boolean getUseFirstLineOfSelection() {
		return PropertiesUtils.getBoolean(this.properties,
				USE_FIRST_LINE_OF_SELECTION, false);
	}

	/** M_CAÖʒuۑ */
	public static final String SAVE_SEND_DIALOG_LOCATION = "app.saveSendDialogLocation"; //$NON-NLS-1$

	/**
	 * @return M_CAÖʒuۑ/ꍇtrueԂ܂B
	 */
	public boolean getSaveSendDialogLocation() {
		return PropertiesUtils.getBoolean(this.properties,
				SAVE_SEND_DIALOG_LOCATION, false);
	}

	/**
	 * M_CAÖʒuۑ邩ǂݒ肵܂B
	 * @param save ۑꍇtrue
	 */
	public void setSaveSendDialogLocation(boolean save) {
		PropertiesUtils.setBoolean(this.properties, SAVE_SEND_DIALOG_LOCATION,
				save);
	}

	/** M_CAȎ傫ۑ */
	public static final String SAVE_SEND_DIALOG_SIZE = "app.saveSendDialogSize"; //$NON-NLS-1$

	/**
	 * @return M_CAÕTCYۑ/ꍇtrueԂ܂B
	 */
	public boolean getSaveSendDialogSize() {
		return PropertiesUtils.getBoolean(this.properties,
				SAVE_SEND_DIALOG_SIZE, true);
	}

	/**
	 * M_CAÕTCYۑ邩ǂݒ肵܂B
	 * @param save ۑꍇtrue
	 */
	public void setSaveSendDialogSize(boolean save) {
		PropertiesUtils
				.setBoolean(this.properties, SAVE_SEND_DIALOG_SIZE, save);
	}

	/** M_CAOYW */
	private static final String SEND_DIALOG_TOP = "app.sendDialogTop"; //$NON-NLS-1$

	/**
	 * @return M_CAO̕\ʒuYW擾܂B
	 */
	public int getSendDialogTop() {
		return PropertiesUtils.getInt(this.properties, SEND_DIALOG_TOP,
				IMPOSSIBLE);
	}

	/**
	 * M_CAO̕\ʒuYWݒ肵܂B
	 * @param y YW
	 */
	public void setSendDialogTop(int y) {
		PropertiesUtils.setInt(this.properties, SEND_DIALOG_TOP, y);
	}

	/** M_CAOXW */
	private static final String SEND_DIALOG_LEFT = "app.sendDialogLeft"; //$NON-NLS-1$

	/**
	 * @return M_CAO̕\ʒuXW擾܂B
	 */
	public int getSendDialogLeft() {
		return PropertiesUtils.getInt(this.properties, SEND_DIALOG_LEFT,
				IMPOSSIBLE);
	}

	/**
	 * M_CAO̕\ʒuXWݒ肵܂B
	 * @param x XW
	 */
	public void setSendDialogLeft(int x) {
		PropertiesUtils.setInt(this.properties, SEND_DIALOG_LEFT, x);
	}

	/** M_CAỎ */
	private static final String SEND_DIALOG_WIDTH = "app.sendDialogWidth"; //$NON-NLS-1$

	/**
	 * @return M_CAỎ擾܂B
	 */
	public int getSendDialogWidth() {
		return PropertiesUtils.getInt(this.properties, SEND_DIALOG_WIDTH, 540);
	}

	/**
	 * M_CAỎݒ肵܂B
	 * @param width 
	 */
	public void setSendDialogWidth(int width) {
		PropertiesUtils.setInt(this.properties, SEND_DIALOG_WIDTH, width);
	}

	/** M_CAO̍ */
	private static final String SEND_DIALOG_HEIGHT = "app.sendDialogHeight"; //$NON-NLS-1$

	/**
	 * @return M_CAO̍擾܂B
	 */
	public int getSendDialogHeight() {
		return PropertiesUtils.getInt(this.properties, SEND_DIALOG_HEIGHT, 300);
	}

	/**
	 * M_CAO̍ݒ肵܂B
	 * @param height 
	 */
	public void setSendDialogHeight(int height) {
		PropertiesUtils.setInt(this.properties, SEND_DIALOG_HEIGHT, height);
	}

	/** M_CAO̗̕ */
	private static final String SEND_DIALOG_COLUMN_WIDTHS = "app.sendDialogColumnWidths"; //$NON-NLS-1$

	/**
	 * M_CAO̗̕ݒ肵܂B
	 * @param csv J}؂̐l
	 */
	public void setSendDialogColumnWidths(String csv) {
		this.properties.setProperty(SEND_DIALOG_COLUMN_WIDTHS, csv);
	}

	/**
	 * ̕Ԃ܂B
	 * @return ̕
	 */
	public int[] getSendDialogColumnWidths() {
		return Tool.parseIntArray(this.properties
				.getProperty(SEND_DIALOG_COLUMN_WIDTHS));
	}

	/**
	 * ŏɃ^XNo[ɕ\Ȃ
	 */
	public static final String HIDE_ON_ICONIFIED = "app.hideOnIconified"; //$NON-NLS-1$

	/**
	 * ŏɃ^XNo[ɕ\邩ǂ
	 * @return ŏɃ^XNo[ɕ\Ȃꍇtrue
	 */
	public boolean hideOnIconified() {
		return PropertiesUtils.getBoolean(this.properties, HIDE_ON_ICONIFIED,
				false);
	}

	/**
	 * ۑ̎MbZ[WƂɌx
	 */
	public static final String WARN_CLOSING_UNSAVED_RECEIVED_MESSAGE = "app.warnClosingUnsavedReceivedMessage"; //$NON-NLS-1$

	/**
	 * ۑ̎MbZ[WƂɌx邩ǂ
	 * @return xꍇtrueBftHgfalse
	 */
	public boolean warnClosingUnsavedReceivedMessage() {
		return PropertiesUtils.getBoolean(this.properties,
				WARN_CLOSING_UNSAVED_RECEIVED_MESSAGE, false);
	}

	/**
	 * 莞ԃ}EXȂ玩Iɕs݂ɂ
	 */
	public static final String AUTO_ABSENCE = "app.autoAbsence"; //$NON-NLS-1$

	/**
	 * Iɕs݂ɂ邩ǂ
	 * @return Iɕs݂ɂꍇtrue
	 */
	public boolean useAutoAbsence() {
		return PropertiesUtils.getBoolean(this.properties, AUTO_ABSENCE, false);
	}

	/**
	 * Iɕs݂ɂ܂ł̎(b)
	 */
	public static final String AUTO_ABSENCE_LIMIT_SEC = "app.autoAbsenceLimit"; //$NON-NLS-1$

	/**
	 * Iɕs݂ɂ܂ł̎(b)擾܂B
	 * @return Iɕs݂ɂ܂ł̎(b)
	 */
	public int getAutoAbsenceLimit() {
		return PropertiesUtils.getInt(this.properties, AUTO_ABSENCE_LIMIT_SEC,
				600);
	}

	/**
	 * VvbZ[WEBhEŃJX^tHggp邩ǂ
	 */
	public static final String USE_CUSTOM_FONT_IN_SIMPLE_MESSAGE = "app.useCustomFontInSimpleMessage"; //$NON-NLS-1$

	/**
	 * VvbZ[WEBhEŃJX^tHggp邩ǂ擾܂B
	 * @return gpꍇtrue
	 */
	public boolean useCustomFontInSimpleMessage() {
		return PropertiesUtils.getBoolean(this.properties,
				USE_CUSTOM_FONT_IN_SIMPLE_MESSAGE, false);
	}

	/**
	 * VvbZ[WEBhẼJX^tHg̃vtBbNX
	 */
	public static final String FONT_IN_SIMPLE_MESSAGE = "app.fontInSimpleMessage"; //$NON-NLS-1$

	/**
	 * VvbZ[WEBhẼJX^tHg擾܂B
	 * @return tHg
	 */
	public String getFontNameInSimpleMessage() {
		return this.properties
				.getProperty(FONT_IN_SIMPLE_MESSAGE + ".name", ""); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * VvbZ[WEBhẼJX^tHgTCY擾܂B
	 * @return tHgTCY
	 */
	public int getFontSizeInSimpleMessage() {
		return PropertiesUtils.getInt(this.properties, FONT_IN_SIMPLE_MESSAGE
				+ ".size", 9); //$NON-NLS-1$
	}

	/**
	 * VvbZ[WEBhẼJX^tHgX^C擾܂B
	 * @return tHgX^C
	 */
	public int getFontStyleInSimpleMessage() {
		return PropertiesUtils.getInt(this.properties, FONT_IN_SIMPLE_MESSAGE
				+ ".style", 0); //$NON-NLS-1$
	}

	/**
	 * VvbZ[WEBhẼJX^tHgP[擾܂B
	 * @return tHgP[
	 */
	public String getFontLocaleInSimpleMessage() {
		return properties.getProperty(FONT_IN_SIMPLE_MESSAGE + ".locale"); //$NON-NLS-1$
	}

	/**
	 * ^u
	 */
	public static final String TAB_WIDTH = "app.tabWidthPx"; //$NON-NLS-1$

	/**
	 * ^u4̋󔒕łƉ肵ꍇ
	 * 4̋󔒕̉擾܂B
	 * @return ^u
	 */
	public int getTabWidthPx() {
		return PropertiesUtils.getInt(this.properties, TAB_WIDTH, 4);
	}

	/**
	 * 1s\Ńc[`bv\
	 */
	public static final String USE_TOOLTIP_AT_ONELINE = "app.useTooltipAtOneline"; //$NON-NLS-1$

	/**
	 * 1s\Ńc[`bv\邩ǂ擾܂B
	 * @return 1s\Ńc[`bv\ꍇtrue
	 */
	public boolean getUseTooltipAtOneline() {
		return PropertiesUtils.getBoolean(this.properties,
				USE_TOOLTIP_AT_ONELINE, true);
	}

	/**
	 * VXeProxygp邩ǂ
	 */
	public static final String USE_SYSTEM_PROXIES = "app.useSystemProxies"; //$NON-NLS-1$

	/**
	 * VXeProxyݒgp邩ǂ
	 * @return VXeProxyݒgpꍇtrue
	 */
	public boolean getUseSystemProxies() {
		return PropertiesUtils.getBoolean(this.properties, USE_SYSTEM_PROXIES,
				true);
	}

	/**
	 * VbZ[W̃EBhETCY͓eɂ킹
	 */
	public static final String FIT_CONTENT_OF_NEW_MESSAGE = "app.fitContent"; //$NON-NLS-1$

	/**
	 * VbZ[W̃EBhETCYeɂ킹邩ǂ
	 * @return eɍ킹ꍇtrue
	 */
	public boolean isFitContentOfNewMessage() {
		return PropertiesUtils.getBoolean(this.properties,
				FIT_CONTENT_OF_NEW_MESSAGE, false);
	}

	/**
	 * ŌɕۑtH_̃pX
	 */
	public static final String LAST_DOWNLOAD_FOLDER = "app.lastDownloadFolder"; //$NON-NLS-1$

	/**
	 * ŌɕۑtH_̃pXݒ肵܂B
	 * @param path pX
	 */
	public void setLastDownloadFolder(String path) {
		this.properties.setProperty(LAST_DOWNLOAD_FOLDER, path);
	}

	/**
	 * ŌɕۑtH_̃pX擾܂B
	 * @return pX
	 */
	public String getLastDownloadFolder() {
		return this.properties.getProperty(LAST_DOWNLOAD_FOLDER);
	}

	/**
	 * Nbv{[h̉摜Ytۂɕۑ_CAO\
	 */
	public static final String ATTACH_CLIPBOARD_CONTENT_WITH_DIALOG = "app.attachClipboardImageWithDialog"; //$NON-NLS-1$

	/**
	 * Nbv{[h̉摜Ytۂɕۑ_CAO\邩ǂ擾܂B
	 * @return \ꍇtrue
	 */
	public boolean getAttachClipboardContentWithDialog() {
		return PropertiesUtils.getBoolean(this.properties,
				ATTACH_CLIPBOARD_CONTENT_WITH_DIALOG, false);
	}

	/**
	 * bZ[WOUTF-8ŕۑ
	 */
	public static final String MESSAGE_LOG_UTF8 = "app.logEncodingUtf8"; //$NON-NLS-1$

	/**
	 * bZ[WOUTF-8ŕۑ邩ǂ擾܂B
	 * @return bZ[WOUTF-8ŕۑꍇtrue
	 */
	public boolean isUtf8MessageLog() {
		return PropertiesUtils.getBoolean(this.properties, MESSAGE_LOG_UTF8,
				false);
	}

	/**
	 * IPv6AhXD悷邩ǂ
	 */
	public static final String PREFER_IPV6_ADDRESSES = "app.preferIPv6Addresses"; //$NON-NLS-1$

	/**
	 * IPv6AhXD悷邩ǂ擾܂B
	 * @return IPv6AhXD悷ꍇtrueBftHgfalseB
	 */
	public boolean preferIPv6Addresses() {
		return PropertiesUtils.getBoolean(this.properties,
				PREFER_IPV6_ADDRESSES, false);
	}

	/**
	 * tH[JXzbgL[(Events.DisplayWindow)
	 */
	public static final String HOTKEY_DISPLAYWINDOW = "app.hotkey.DisplayWindow"; //$NON-NLS-1$

	/**
	 * tH[JXzbgL[(Events.DisplayWindow)擾܂B
	 * @return ݒ肳ĂȂꍇ͋󕶎B
	 */
	public String getHotkeyDisplayWindow() {
		return properties.getProperty(HOTKEY_DISPLAYWINDOW);
	}

	/**
	 * tH[JXzbgL[(Events.DisplayIndependentWindow)
	 */
	public static final String HOTKEY_SHOW_IN_INDEPENDENT_WINDOW = "app.hotkey.DisplayIndependentWindow"; //$NON-NLS-1$

	/**
	 * tH[JXzbgL[(Events.DisplayIndependentWindow)擾܂B
	 * @return ݒ肳ĂȂꍇ͋󕶎B
	 */
	public String getHotkeyShowInIndependentWindow() {
		return properties.getProperty(HOTKEY_SHOW_IN_INDEPENDENT_WINDOW);
	}

	/**
	 * VbZ[WEBhE\邷zbgL[(Events.OpenBlankMessageWindow)
	 */
	public static final String HOTKEY_OPENBLANKMESSAGEWINDOW = "app.hotkey.OpenBlankMessageWindow"; //$NON-NLS-1$

	/**
	 * VbZ[WEBhE\邷zbgL[(Events.OpenBlankMessageWindow)擾܂B
	 * @return ݒ肳ĂȂꍇ͋󕶎B
	 */
	public String getHotkeyOpenBlankMessageWindow() {
		return properties.getProperty(HOTKEY_OPENBLANKMESSAGEWINDOW);
	}

	/**
	 * VbZ[WEBhE\邷zbgL[(Events.OpenBlankMessageWindow)
	 */
	public static final String HOTKEY_OPENBLANKINDEPENDENTMESSAGEWINDOW = "app.hotkey.OpenBlankIndependentMessageWindow"; //$NON-NLS-1$

	/**
	 * VbZ[WEBhE\邷zbgL[(Events.OpenBlankIndependentMessageWindow)擾܂B
	 * @return ݒ肳ĂȂꍇ͋󕶎B
	 */
	public String getHotkeyOpenBlankIndependentMessageWindow() {
		return properties.getProperty(HOTKEY_OPENBLANKINDEPENDENTMESSAGEWINDOW);
	}

	/**
	 * Or[A\zbgL[
	 */
	public static final String HOTKEY_OPENLOGVIEWER = "app.hotkey.OpenLogViewer"; //$NON-NLS-1$

	/**
	 * Or[A\zbgL[(Events.OpenLogViewer)擾܂B
	 * @return ݒ肳ĂȂꍇ͋󕶎B
	 */
	public String getHotkeyOpenLogViewer() {
		return properties.getProperty(HOTKEY_OPENLOGVIEWER);
	}

	/**
	 * bZ[WEBhEƗEBhE[hŊJ
	 */
	public static final String USE_INDEPENDENT_MESSAGE_WINDOW = "app.useIndependentMessageWindow"; //$NON-NLS-1$

	/**
	 * bZ[WEBhEƗEBhE[hŊJǂ擾܂B
	 * @return ƗEBhE[h̏ꍇtrueBftHgfalseB
	 */
	public boolean getUseIndependentMessageWindow() {
		return PropertiesUtils.getBoolean(this.properties,
				USE_INDEPENDENT_MESSAGE_WINDOW, false);
	}

	/**
	 * EBhEꍇɌx\邩ǂ
	 */
	public static final String WARN_EXCESSIVE_WINDOWS = "app.warnExcessiveWindows"; //$NON-NLS-1$

	/**
	 * EBhEꍇɌx\邩ǂ擾܂B
	 * @return xKvȏꍇtrue
	 */
	public boolean warnExcessiveWindows() {
		return PropertiesUtils.getBoolean(this.properties,
				WARN_EXCESSIVE_WINDOWS, Os.isWindows());
	}

	/**
	 * IRSA 512 bitRC2 40 bitgB
	 */
	public static final String USE_SMALL_KEY = "app.useSmallKey"; //$NON-NLS-1$

	/**
	 * IRSA 512 bitRC2 40 bitgǂ擾܂B
	 * @return RSA 512 bitRC2 40 bitgꍇtrueB
	 */
	public boolean useSmallKey() {
		return PropertiesUtils
				.getBoolean(this.properties, USE_SMALL_KEY, false);
	}

	/**
	 * M_CAOmigemogp
	 */
	public static final String USE_MIGEMO_IN_SENDDIALOG = "app.migemo.SendDialog"; //$NON-NLS-1$

	/**
	 * M_CAOmigemogp邩ǂ擾܂B
	 * 
	 * @return ftHgtrue(gp)B
	 */
	public boolean getUseMigemoInSendDialog() {
		return PropertiesUtils.getBoolean(this.properties,
				USE_MIGEMO_IN_SENDDIALOG, true);
	}

	/**
	 * M_CAOmigemogp邩ǂݒ肵܂B
	 * @param use gpꍇtrueB
	 */
	public void setUseMigemoInSendDialog(boolean use) {
		PropertiesUtils.setBoolean(this.properties, USE_MIGEMO_IN_SENDDIALOG,
				use);
	}

	/**
	 * ۊǔmigemogp
	 */
	public static final String USE_MIGEMO_IN_MESSAGE_STORAGE = "app.migemo.MessageStorage"; //$NON-NLS-1$

	/**
	 * ۊǔmigemogp邩ǂ擾܂B
	 * 
	 * @return ftHgtrue(gp)B
	 */
	public boolean getUseMigemoInMessageStorage() {
		return PropertiesUtils.getBoolean(this.properties,
				USE_MIGEMO_IN_MESSAGE_STORAGE, true);
	}

	/**
	 * ۊǔmigemogp邩ǂݒ肵܂B
	 * 
	 * @param use gpꍇtrueB
	 */
	public void setUseMigemoInMessageStorage(boolean use) {
		PropertiesUtils.setBoolean(this.properties,
				USE_MIGEMO_IN_MESSAGE_STORAGE, use);
	}

	/**
	 * fXNgbvmigemogp
	 */
	public static final String USE_MIGEMO_IN_DESKTOP_SEARCH = "app.migemo.DesktopSearch"; //$NON-NLS-1$

	/**
	 * fXNgbvmigemogp邩ǂ擾܂B
	 * 
	 * @return ftHgtrue(gp)B
	 */
	public boolean getUseMigemoInDesktopSearch() {
		return PropertiesUtils.getBoolean(this.properties,
				USE_MIGEMO_IN_DESKTOP_SEARCH, true);
	}

	/**
	 * fXNgbvmigemogp邩ǂݒ肵܂B
	 * 
	 * @param use gpꍇtrueB
	 */
	public void setUseMigemoInDesktopSearch(boolean use) {
		PropertiesUtils.setBoolean(this.properties,
				USE_MIGEMO_IN_DESKTOP_SEARCH, use);
	}

	/**
	 * CEBhÉ~{^ŃACRꍇtrueB
	 */
	public static final String ICONIFY_INSTEDOF_QUITTING = "app.iconifyInstedOfQuitting"; //$NON-NLS-1$

	/**
	 * CEBhÉ~{^ŃACR邩ǂ擾܂B
	 * 
	 * @return CEBhÉ~{^ŃACRꍇtrueBftHgtrueB
	 */
	public boolean iconifyInstedOfQuitting() {
		return PropertiesUtils.getBoolean(this.properties,
				ICONIFY_INSTEDOF_QUITTING, false);
	}

	/**
	 * \tgEFA̍XVmF
	 */
	public static final String CHECK_FOR_UPDATES = "app.checkForUpdates.v2"; //$NON-NLS-1$

	/**
	 * \tgEFA̍XVmFB
	 * 
	 * @return 0:mFȂ, 1:_E[hOɒʒm, 2:_E[hɒʒm
	 */
	public int checkForUpdates() {
		return PropertiesUtils.getInt(this.properties, CHECK_FOR_UPDATES, 2);
	}

	/**
	 * ŌɎ擾\tgEFAXṼrhԍ
	 */
	public static final String LAST_BUILDNUMBER = "app.lastBuildNumber"; //$NON-NLS-1$

	/**
	 * ŌɎ擾\tgEFAXṼrhԍ擾܂B
	 * 
	 * @return rhԍ
	 */
	public String getLastBuildNumber() {
		return this.properties.getProperty(LAST_BUILDNUMBER, "0"); //$NON-NLS-1$
	}

	/**
	 * ŌɎ擾\tgEFAXṼrhԍ擾܂B
	 * 
	 * @return rhԍ
	 */
	public int getLastBuildNumberAsInt() {
		return PropertiesUtils.getInt(properties, LAST_BUILDNUMBER, -1);
	}

	/**
	 * ŌɎ擾\tgEFAXṼrhԍݒ肵܂B
	 * 
	 * @param buildNumber rhԍ
	 */
	public void setLastBuildNumber(String buildNumber) {
		this.properties.setProperty(LAST_BUILDNUMBER, buildNumber);
	}

	/**
	 * ŌɃo[W`FbN
	 */
	public static final String LAST_CHECK_TIME = "app.checkForUpdates.time"; //$NON-NLS-1$

	/**
	 * @return ŌɃo[W`FbN
	 */
	public long getLastCheckTime() {
		return PropertiesUtils.getLong(properties, LAST_CHECK_TIME, 0);
	}

	/**
	 * @param time ŌɃo[W`FbN
	 */
	public void setLastCheckTime(long time) {
		PropertiesUtils.setLong(properties, LAST_CHECK_TIME, time);
	}

	/**
	 * 摜ҏW\tg
	 */
	public static final String GRAPHIC_EDITOR = "app.graphicEditor"; //$NON-NLS-1$

	/**
	 * 摜ҏW\tg擾܂B
	 * 
	 * @return 摜ҏW\tgւ̃pX
	 */
	public String getGraphicEditor() {
		return this.properties.getProperty(GRAPHIC_EDITOR, ""); //$NON-NLS-1$
	}

	/** bZ[W_CAỎ */
	public static final String MESSAGE_SEARCH_WIDTH = "app.messageSearch.width"; //$NON-NLS-1$

	/** bZ[W_CAO̍ */
	public static final String MESSAGE_SEARCH_HEIGHT = "app.messageSearch.height"; //$NON-NLS-1$

	/**
	 * bZ[W_CAỎ擾܂B
	 * 
	 * @return bZ[W_CAỎ(ftHg320)
	 */
	public int getMessageSearchWidth() {
		return PropertiesUtils.getInt(this.properties, MESSAGE_SEARCH_WIDTH,
				320);
	}

	/**
	 * bZ[W_CAỎݒ肵܂B
	 * @param value 
	 */
	public void setMessageSearchWidth(int value) {
		PropertiesUtils.setInt(this.properties, MESSAGE_SEARCH_WIDTH, value);
	}

	/**
	 * bZ[W_CAO̍擾܂B
	 * 
	 * @return bZ[W_CAO̍(ftHg300)
	 */
	public int getMessageSearchHeight() {
		return PropertiesUtils.getInt(this.properties, MESSAGE_SEARCH_HEIGHT,
				300);
	}

	/**
	 * bZ[W_CAO̍ݒ肵܂B
	 * @param value 
	 */
	public void setMessageSearchHeight(int value) {
		PropertiesUtils.setInt(this.properties, MESSAGE_SEARCH_HEIGHT, value);
	}

	/** ׂĕ\ŃtH[JXƂɓx邩ǂ */
	public static String CANCEL_TRANSPARENCY_ON_FOCUS = "app.cancelTransparencyOnFocus"; //$NON-NLS-1$

	/**
	 * @return tH[JX瓧xꍇtrue
	 */
	public boolean getCancelTransParencyOnFocus() {
		return PropertiesUtils.getBoolean(this.properties,
				CANCEL_TRANSPARENCY_ON_FOCUS, false);
	}

	/**
	 * QRR[hG[\
	 */
	public static String QRCODE_ERROR_CORRECTION_LEVEL = "app.qrCodeErrorCorrectionLevel"; //$NON-NLS-1$

	/**
	 * QRR[hG[\͂擾܂B
	 * 
	 * @return QRR[hG[\('L','M','Q','H')
	 */
	public char getQRCodeErrorCorrectionLevel() {
		String str = properties.getProperty(QRCODE_ERROR_CORRECTION_LEVEL);
		if (str == null || str.length() != 1) {
			return 'M';
		}
		return str.charAt(0);
	}

	/**
	 * QRR[h1Z̃sNZ
	 */
	public static String QRCODE_MODULE_SIZE = "app.qrCodeModuleSize"; //$NON-NLS-1$

	/**
	 * @return QRR[h1Z̃sNZ(ftHg2px)
	 */
	public int getQRCodeModuleSize() {
		return PropertiesUtils.getInt(this.properties, QRCODE_MODULE_SIZE, 2);
	}

	/**
	 * QRR[h̕GR[fBO
	 */
	public static final String QRCODE_ENCODING = "app.qrEncoding"; //$NON-NLS-1$

	/**
	 * @return QRR[h̕GR[fBOBݒ̏ꍇnullB
	 */
	public String getQREncoding() {
		return properties.getProperty(QRCODE_ENCODING);
	}

	/**
	 * QRR[h̕o[W
	 */
	public static final String QRCODE_DIVISION = "app.qrDivision"; //$NON-NLS-1$

	/**
	 * @return QRR[h̕o[WB
	 */
	public int getQRDivision() {
		return PropertiesUtils.getInt(properties, QRCODE_DIVISION, 40);
	}

	/**
	 * bZ[W̔zu@
	 */
	public static final String MESSAGE_ARRANGEMENT = "app.messageArrangement"; //$NON-NLS-1$

	/**
	 * bZ[W̔zu@: ォ
	 */
	public static final String MESSAGE_ARRANGEMENT_TOP_LEFT = "fromTopLeft"; //$NON-NLS-1$

	/**
	 * bZ[W̔zu@: 
	 */
	public static final String MESSAGE_ARRANGEMENT_BOTTOM_CENTER = "fromBottomCenter"; //$NON-NLS-1$

	/**
	 * bZ[W̔zu@: }EX|C^̈ʒu
	 */
	public static final String MESSAGE_ARRANGEMENT_MOUSE = "mousePosition"; //$NON-NLS-1$

	/**
	 * @return bZ[W̔zu@
	 */
	public String getMessageArrangement() {
		return properties.getProperty(MESSAGE_ARRANGEMENT,
				MESSAGE_ARRANGEMENT_TOP_LEFT);
	}

	/** ł񂫗\NCAgLɂ邩ǂ */
	public static final String ENABLE_DENKI_YOHOU_CLIENT = "app.enableDenkiYohouClient"; //$NON-NLS-1$

	/**
	 * @return ł񂫗\NCAgLɂꍇtrue(ftHg: false)
	 */
	public boolean enableDenkiYohouClient() {
		return PropertiesUtils.getBoolean(this.properties,
				ENABLE_DENKI_YOHOU_CLIENT, false);
	}

	/** y^낤vgR(vgR)Lɂ邩ǂ */
	public static final String ENABLE_OLD_PETAROU_COMPONENT = "app.enableOldPetarouComponent"; //$NON-NLS-1$

	/**
	 * @return y^낤R|[lg(vgR)Lɂꍇtrue(ftHg: false)
	 */
	public boolean enableOldPetarouComponent() {
		return PropertiesUtils.getBoolean(this.properties,
				ENABLE_OLD_PETAROU_COMPONENT, false);
	}

	/** IP MessengerR|[lgLɂ邩ǂ */
	public static final String ENABLE_IPMESSENGER_COMPONENT = "app.enableIPMessengerComponent"; //$NON-NLS-1$

	/**
	 * @return IP MessengerR|[lgLɂꍇtrue(ftHg: true)
	 */
	public boolean enableIPMessengerComponent() {
		return PropertiesUtils.getBoolean(this.properties,
				ENABLE_IPMESSENGER_COMPONENT, true);
	}

	/** y^낤XR|[lgLɂ邩ǂ */
	public static final String ENABLE_PETAROUX_COMPONENT = "app.enablePetarouXComponent"; //$NON-NLS-1$

	/**
	 * @return y^낤XR|[lgLɂꍇtrue(ftHg: true)
	 */
	public boolean enablePetarouXComponent() {
		return PropertiesUtils.getBoolean(this.properties,
				ENABLE_PETAROUX_COMPONENT, true);
	}

	/** tⳎ̑KCh\邩ǂ */
	public static final String SHOW_STICKY_OPERATION_GUIDE = "app.showStickyOperationGuide"; //$NON-NLS-1$

	/**
	 * @return tⳎ̑KCh\ꍇtrue(ftHg: true)
	 */
	public boolean showStickyOperationGuide() {
		return PropertiesUtils.getBoolean(this.properties,
				SHOW_STICKY_OPERATION_GUIDE, true);
	}

	/** ŏ̃hbvŃCC摜ɂ邩ǂq˂ */
	public static final String ASK_WHETHER_MAKE_FIRST_DROPPED_ITEM_INLINE = "app.askWhetherMakeFirstDroppedItemInline"; //$NON-NLS-1$

	/**
	 * @return ŏ̃hbvŃCC摜ɂ邩ǂq˂ꍇtrue
	 */
	public boolean askWhetherMakeFirstDroppedItemInline() {
		return PropertiesUtils.getBoolean(this.properties,
				ASK_WHETHER_MAKE_FIRST_DROPPED_ITEM_INLINE, true);
	}

	/** ŌɑIĂݒ_CAÕ^u */
	public static final String LAST_PREFERENCE_TAB_INDEX = "app.lastPreferenceTabIndex"; //$NON-NLS-1$

	/**
	 * @return ŌɑIĂݒ_CAÕ^u
	 */
	public int getLastPreferenceTabIndex() {
		return PropertiesUtils.getInt(properties, LAST_PREFERENCE_TAB_INDEX, 0);
	}

	/**
	 * @param lastPreferenceTabIndex ŌɑIĂݒ_CAÕ^u
	 */
	public void setLastPreferenceTabIndex(int lastPreferenceTabIndex) {
		PropertiesUtils.setInt(properties, LAST_PREFERENCE_TAB_INDEX,
				lastPreferenceTabIndex);
	}

	/** M_CAÕ\[g */
	public static final String SEND_DIALOG_SORT_ORDER = "app.sendDialog.sortOrder"; //$NON-NLS-1$

	/**
	 * @return M_CAÕ\[g(ftHg: 0(\[gȂ))
	 */
	public int getSendDialogSortOrder() {
		return PropertiesUtils.getInt(properties, SEND_DIALOG_SORT_ORDER, 0);
	}

	/**
	 * @param sortOrder M_CAÕ\[g
	 */
	public void setSendDialogSortOrder(int sortOrder) {
		PropertiesUtils.setInt(properties, SEND_DIALOG_SORT_ORDER, sortOrder);
	}

	/** M_CAÕ\[gΏۗ */
	public static final String SEND_DIALOG_SORT_COLUMN = "app.sendDialog.sortColumn"; //$NON-NLS-1$

	/**
	 * @return M_CAÕ\[gΏۗ(ftHg: -1(\[gȂ))
	 */
	public int getSendDialogSortColumn() {
		return PropertiesUtils.getInt(properties, SEND_DIALOG_SORT_COLUMN, -1);
	}

	/**
	 * @param sortColumn M_CAÕ\[gΏۗ
	 */
	public void setSendDialogSortColumn(int sortColumn) {
		PropertiesUtils.setInt(properties, SEND_DIALOG_SORT_COLUMN, sortColumn);
	}

	/**
	 * J[\p̌Ɉړ
	 */
	public static final String MOVE_CURSOR_TO_LAST = "app.moveCursorToLast"; //$NON-NLS-1$

	/**
	 * J[\p̑O܂͌Ɉړ
	 * @return true: iftHgj, false: O
	 */
	public boolean getMoveCursorToLast() {
		return PropertiesUtils
				.getBoolean(properties, MOVE_CURSOR_TO_LAST, true);
	}

	/**
	 * c[̎spX̃|XgtBbNX
	 */
	public static final String TOOL_EXEC_FILE = ".execFile."; //$NON-NLS-1$

	/**
	 * c[̈̃|XgtBbNX
	 */
	public static final String TOOL_PARAM = ".param."; //$NON-NLS-1$

	/**
	 * c[̍ƃtH_̃|XgtBbNX
	 */
	public static final String TOOL_WORKING_DIR = ".workingDir."; //$NON-NLS-1$

	/**
	 * t@C}l[W(t@C)̃vpeB̃vtBbNX
	 */
	public static final String FILE_MANAGER_CONTAINING_FOLDER = "app.fileManager.containingFolder"; //$NON-NLS-1$

	/**
	 * t@C}l[W(tH_)̃vpeB̃vtBbNX
	 */
	public static final String FILE_MANAGER_FOLDER = "app.fileManager.folder"; //$NON-NLS-1$

	/**
	 * @param prefix vpeB̃vtBbNX
	 * @return vpeB̃L[
	 */
	public static String getToolCommandProperty(String prefix) {
		return prefix + TOOL_EXEC_FILE + Os.getName();
	}

	/**
	 * @param prefix vpeB̃vtBbNX
	 * @return vpeB̃L[
	 */
	public static String getToolParameterProperty(String prefix) {
		return prefix + TOOL_PARAM + Os.getName();
	}

	/**
	 * @param prefix vpeB̃vtBbNX
	 * @return vpeB̃L[
	 */
	public static String getToolWorkingFolderProperty(String prefix) {
		return prefix + TOOL_WORKING_DIR + Os.getName();
	}

	/**
	 * @param prefix vpeB̃vtBbNX
	 * @return st@C
	 */
	public String getToolCommand(String prefix) {
		return properties.getProperty(getToolCommandProperty(prefix), ""); //$NON-NLS-1$
	}

	/**
	 * @param prefix vpeB̃vtBbNX
	 * @return R}hC
	 */
	public String getToolParameter(String prefix) {
		return properties.getProperty(getToolParameterProperty(prefix), ""); //$NON-NLS-1$
	}

	/**
	 * @param prefix vpeB̃vtBbNX
	 * @return ƃtH_
	 */
	public String getToolWorkingFolder(String prefix) {
		return properties.getProperty(getToolWorkingFolderProperty(prefix), ""); //$NON-NLS-1$
	}

	/**
	 * @param prefix vpeB̃vtBbNX
	 * @return ƃtH_
	 */
	public File getToolWorkingFolderAsFile(String prefix) {
		String dir = getToolWorkingFolder(prefix);
		if (dir == null || dir.length() == 0) {
			return null;
		}
		return new File(dir);
	}

	/**
	 * Propertiesۑ܂B
	 * @param prefix }[WvpeB̃L[ɕtvtBbNX
	 * @param prop }[WProperties
	 */
	public void merge(String prefix, Properties prop) {
		Iterator itr = prop.entrySet().iterator();
		while (itr.hasNext()) {
			Entry entry = (Entry) itr.next();
			String key = (String) entry.getKey();
			String val = (String) entry.getValue();
			properties.setProperty(prefix + key, val);
		}
	}

	/**
	 * w肵Propertiesɒlݒ肵܂B
	 * @param p XgA
	 * @param prefix 擾pvtBbNX
	 * @param key 擾EݒpL[
	 */
	private void restore(Properties p, String prefix, String key) {
		String v = properties.getProperty(prefix + key);
		if (v != null) {
			p.setProperty(key, v);
		}
	}

	/**
	 * @return VersionɓnProperties
	 */
	public Properties getVersionProperties() {
		Properties p = new Properties();
		restore(p, Version.PREFIX, Version.FILE);
		restore(p, Version.PREFIX, Version.BUILD_NUMBER);
		restore(p, Version.PREFIX, Version.VERSION_STAGE);
		restore(p, Version.PREFIX, Version.VERSION_TEXT);
		return p;
	}

}
