package petaclone;

import java.io.File;
import java.io.IOException;
import java.io.Writer;

import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Tool;

/**
 * Ytt@C
 * @author Takayuki Shoji
 */
public class Attachment {

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap msg = MessageMap.getMessageMap();

	/** _E[hIDp̃JE^ */
	private static int count = (int) System.currentTimeMillis();

	/** _E[hID */
	public final int downloadId;

	/** vgRƎID */
	public int id;

	/** t@C */
	public String name;

	/** t@CTCY */
	public long size;

	/** ŏIXV(1bP) */
	public long lastModified;

	/** ǂݍݐp */
	public boolean isReadOnly;

	/** Bt@C */
	public boolean isHidden;

	/** t@C(TYPE_FILE)fBNg(TYPE_DIR)܂̓CC(TYPE_INLINE) */
	public int type;

	/**
	 * Ƀ[h摜f[^Ȃ
	 * <p>
	 * SWT UIł́A\ImageAUndoX^bNŕێꍇImageDataݒ肳܂B
	 */
	public Object data;

	/** type: t@C */
	public static final int TYPE_FILE = 0;

	/** type: fBNg */
	public static final int TYPE_DIR = 1;

	/** type: ЂƂ̃fBNg(IP Messengerp) */
	public static final int TYPE_PARENT = 2;

	/** type: CC摜 */
	public static final int TYPE_INLINE = 3;

	/** type: ̑(ps) */
	public static final int TYPE_OTHER = -1;

	/**
	 * _E[hꂽt@CEtH_܂͑Mp̐΃pXB
	 */
	public String localPath;

	/**
	 * CC摜̖ߍ݈ʒu
	 * <p>
	 * platformIndex̏ꍇ <code>position <= -2</code>
	 * <br>ҏWplatformIndexŐݒ肳ĂꍇB
	 * <code>(position*-1)-2</code> Ŏۂ̒lƑݕϊB
	 * <p>
	 * storableIndex̏ꍇ <code>position >= 0</code>
	 * <p>
	 * TYPE_INLINEȊO͖-1B
	 * @see #setPlatformIndex(int)
	 * @see #getPlatformIndex()
	 */
	public int position = -1;

	/**
	 * _E[hȂǂ̏: 0ȏ̓_E[his<br>
	 * eMessenger_E[hXi[ɒʒmۂɏށB
	 */
	public long status = STATUS_NONE;

	/** status: aborted */
	public final static long STATUS_ABORTED = -5;

	/** status: finished */
	public final static long STATUS_FINISHED = -4;

	/** status: ȂɂĂȂ */
	public final static long STATUS_NONE = -3;

	/** status: suspended */
	public final static long STATUS_SUSPENDED = -2;

	/** status: ڑ҂(queue`ڑs) */
	public final static long STATUS_WAIT = -1;

	/** status: ڑς */
	public final static long STATUS_CONNECTED = 0;

	/**
	 * c[`bvLbV
	 */
	private String toolTipCache;

	/**
	 * [J_E[hXi[
	 */
	public DownloadListener listener;

	/** _E[hx(byte/sec) */
	public long bytePerSec;

	/**
	 * RXgN^
	 */
	public Attachment() {
		synchronized (Attachment.class) {
			downloadId = count++;
		}
	}

	/**
	 * p̃RXgN^łB
	 * @param downloadId _E[hID
	 */
	private Attachment(int downloadId) {
		this.downloadId = downloadId;
	}

	/**
	 * CX^X𕡐܂B
	 * <p>
	 * inheritDownloadIdtruew肵ĕIuWFNg͓downloadId܂B
	 * 
	 * @param inheritDownloadId downloadIdRs[ꍇtrue
	 * @return 
	 */
	public Attachment duplicate(boolean inheritDownloadId) {
		Attachment att;
		if (inheritDownloadId) {
			att = new Attachment(downloadId);
		} else {
			att = new Attachment();
		}
		att.id = id;
		att.name = name;
		att.size = size;
		att.lastModified = lastModified;
		att.isReadOnly = isReadOnly;
		att.isHidden = isHidden;
		att.type = type;
		att.data = data;
		att.localPath = localPath;
		att.position = position;
		att.status = status;
		att.toolTipCache = toolTipCache;
		att.listener = listener;
		att.bytePerSec = bytePerSec;
		return att;
	}

	/**
	 * name`FbNÃ݂vbgtH[ŖȂ悤ɂ܂B(Mp)
	 */
	public void validateName() {
		name = Tool.validateFileName(name);
		// WISHLIST ݂Ă烊l[
	}

	/**
	 * localPathFileIuWFNg擾܂B
	 * @return localPathnull̂Ƃnull
	 */
	public File getFile() {
		if (localPath == null) {
			return null;
		}
		return new File(localPath);
	}

	/**
	 * localPatht@C̎擾
	 * @return t@CɃANZXłȂ(ǂ߂Ȃ)Ƃfalse
	 */
	public boolean getFileInfo() {
		File file = getFile();
		if (file == null) {
			return false;
		}

		if (!file.exists()) {
			return false;
		}
		// ǂ߂ȂΑȂ
		if (!file.canRead()) {
			return false;
		}
		// 
		if (file.isFile()) {
			if (this.type != TYPE_INLINE) {
				this.type = TYPE_FILE;
			}
		} else if (file.isDirectory()) {
			this.type = TYPE_DIR;
		} else {
			return false;
		}
		// ̑
		this.name = file.getName();
		this.size = file.length();
		this.lastModified = file.lastModified();
		this.isReadOnly = !file.canWrite();
		this.isHidden = file.isHidden();

		return true;
	}

	/**
	 * Attachment̑t@Cɔf<br>
	 * ŏIXVƃ[hI[̂݁B
	 * @return t@C݂true
	 */
	public boolean applyAttribute() {
		File file = getFile();
		if (file == null || !file.exists()) {
			return false;
		}
		file.setLastModified(lastModified);
		if (isReadOnly) {
			file.setReadOnly();
		}
		return true;
	}

	/**
	 * t@C폜܂B
	 * ł͂ȂfBNg͍폜ł܂B
	 * @return ݂ȂA폜ɐꍇtrue
	 */
	public boolean delete() {
		if (localPath == null) {
			return true;
		}
		File file = getFile();
		if (!file.exists()) {
			return true;
		}
		return file.delete();
	}

	/**
	 * t@CEfBNg𐶐܂B
	 * @param parentDir efBNg
	 * @return ꍇtrue
	 */
	public boolean create(File parentDir) {
		File f;
		switch (type) {
		case TYPE_FILE:
		case TYPE_INLINE:
			f = Tool.createUniqueFile(parentDir, name, 10000);
			break;
		case TYPE_DIR:
			f = Tool.createUniqueDierectory(parentDir, name, 10000);
			break;
		default:
			throw new IllegalStateException("type == " + this.type); //$NON-NLS-1$
		}
		if (f == null) {
			return false;
		}
		// XV
		this.name = f.getName();
		this.localPath = f.getAbsolutePath();
		return true;
	}

	/**
	 * t@Cړ܂Bړɓ̃Ggꍇ
	 * l[čĎs܂B
	 * localPathnameXV܂B
	 * łdestɃt@CꍇtrueԂ܂B
	 * _E[h̓l[܂B
	 * @param dest ړ̃fBNg
	 * @return l[ꍇtrueB
	 */
	public boolean move(File dest) {
		// _E[h̓l[Ȃ
		if (localPath == null || status >= 0 || status == STATUS_WAIT) {
			return false;
		}

		// ړ悪݂ƓtH_ȂړȂ
		File f = getFile();
		if (dest.equals(f.getParentFile())) {
			return true;
		}

		File destination = Tool.move(f, dest, 10000);
		if (destination == null) {
			return false;
		}
		this.localPath = destination.getAbsolutePath();
		this.name = destination.getName();
		return true;
	}

	/**
	 * c[`bveLXg擾܂B
	 * @param useCache LbVgꍇtrueA
	 * 	Vꍇfalsew肵܂B
	 * @return c[`bveLXg
	 */
	public synchronized String getToolTipText(boolean useCache) {
		if (useCache && toolTipCache != null) {
			return toolTipCache;
		}
		StringBuffer sb = new StringBuffer();
		sb.append(name).append('\n');

		// bZ[W
		Object[] arg = new Object[1];

		if (type == TYPE_FILE) {
			arg[0] = new Long(size);
			String msgSize = msg.get("Attachment.tooltip.size", arg); //$NON-NLS-1$
			sb.append(msgSize);

		} else if (type == TYPE_INLINE) {
			arg[0] = new Long(size);
			String msgSize = msg.get("Attachment.tooltip.inline", arg); //$NON-NLS-1$
			sb.append(msgSize);

		} else {
			String msgDirectory = msg.get("Attachment.tooltip.directory"); //$NON-NLS-1$
			sb.append(msgDirectory);
		}
		if (localPath != null) {
			arg[0] = localPath;
			String msgPath = msg.get("Attachment.tooltip.path", arg); //$NON-NLS-1$
			sb.append('\n').append(msgPath);
		}
		if (bytePerSec != 0) {
			arg[0] = new Long(bytePerSec);
			String msgBps = msg.get("Attachment.tooltip.bps", arg); //$NON-NLS-1$
			sb.append('\n').append(msgBps);
		}
		toolTipCache = sb.toString();
		return toolTipCache;
	}

	/**
	 * c[`bveLXg̃LbVɐݒ肵܂B
	 * @param text ݒ肷eLXg
	 */
	public synchronized void setToolTipCache(String text) {
		toolTipCache = text;
	}

	/**
	 * _E[hisǂԂ܂B
	 * @return _E[his̏ꍇtrueB
	 */
	public boolean isProgress() {
		return status >= STATUS_CONNECTED;
	}

	/**
	 * 񉻂B
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("Attachment{"); //$NON-NLS-1$
		sb.append("downloadId: ").append(downloadId); //$NON-NLS-1$
		sb.append(", id: ").append(id); //$NON-NLS-1$
		sb.append(", name: ").append(name); //$NON-NLS-1$
		sb.append(", size: ").append(size); //$NON-NLS-1$
		sb.append(", lastModified: ").append(lastModified); //$NON-NLS-1$
		sb.append(", isReadOnly: ").append(isReadOnly); //$NON-NLS-1$
		sb.append(", isHidden: ").append(isHidden); //$NON-NLS-1$
		sb.append(", position: ").append(position); //$NON-NLS-1$
		sb.append(", type: ").append(type); //$NON-NLS-1$
		switch (type) {
		case TYPE_FILE:
			sb.append("(TYPE_FILE)"); //$NON-NLS-1$
			break;
		case TYPE_INLINE:
			sb.append("(TYPE_INLINE)"); //$NON-NLS-1$
			break;
		case TYPE_DIR:
			sb.append("(TYPE_DIR)"); //$NON-NLS-1$
			break;
		case TYPE_PARENT:
			sb.append("(TYPE_PARENT)"); //$NON-NLS-1$
			break;
		case TYPE_OTHER:
			sb.append("(TYPE_OTHER)"); //$NON-NLS-1$
			break;
		default:
			sb.append("(?)"); //$NON-NLS-1$
			break;
		}
		sb.append(", localPath: ").append(localPath); //$NON-NLS-1$
		sb.append(", status: ").append(status); //$NON-NLS-1$
		if (status <= 0) {
			switch ((int) status) {
			case (int) STATUS_ABORTED:
				sb.append("(STATUS_ABORTED)"); //$NON-NLS-1$
				break;
			case (int) STATUS_FINISHED:
				sb.append("(STATUS_FINISHED)"); //$NON-NLS-1$
				break;
			case (int) STATUS_NONE:
				sb.append("(STATUS_NONE)"); //$NON-NLS-1$
				break;
			case (int) STATUS_SUSPENDED:
				sb.append("(STATUS_SUSPENDED)"); //$NON-NLS-1$
				break;
			case (int) STATUS_WAIT:
				sb.append("(STATUS_WAIT)"); //$NON-NLS-1$
				break;
			case (int) STATUS_CONNECTED:
				sb.append("(STATUS_CONNECTED)"); //$NON-NLS-1$
				break;
			default:
				sb.append("(?)"); //$NON-NLS-1$
				break;
			}
		}
		sb.append(", data(null?)=").append(data == null); //$NON-NLS-1$

		sb.append(", toolTipCache: ").append(toolTipCache); //$NON-NLS-1$
		sb.append("}"); //$NON-NLS-1$
		return sb.toString();
	}

	/**
	 * XML<br>
	 * \A^CvA[JpX̂<br>
	 * _E[hς݂OƂĂ
	 * @param w o͐
	 * @throws IOException 
	 */
	public void writeXml(Writer w) throws IOException {
		Tool.writeStartElement(w, "attacheditem"); //$NON-NLS-1$
		// \
		Tool.writeElement(w, "attachmentname", name); //$NON-NLS-1$
		// t@CAfBNg
		Tool.writeElement(w, "attachmenttype", Integer.toString(type)); //$NON-NLS-1$
		// [JpX
		Tool.writeElement(w, "attachmentpath", localPath); //$NON-NLS-1$
		// CC摜̖ߍ݈ʒu
		if (type == TYPE_INLINE) {
			Tool.writeElement(w,
					"attachmentposition", Integer.toString(position)); //$NON-NLS-1$
		}
		Tool.writeEndElement(w, "attacheditem"); //$NON-NLS-1$
	}

	/**
	 * positionplatformIndexݒ肵܂B
	 * @param platformIndex vbgtH[ˑʒu
	 */
	public void setPlatformIndex(int platformIndex) {
		if (platformIndex <= -2) {
			position = platformIndex;
		} else {
			position = (platformIndex * -1) - 2;
		}
	}

	/**
	 * positionplatformIndexƂĎ擾܂B
	 * storableIndex̏ꍇ-1Ԃ܂B
	 * @return platformIndex܂-1
	 */
	public int getPlatformIndex() {
		if (position <= -2) {
			return (position * -1) - 2;
		}
		return -1;
	}

	/**
	 * downloadIdCX^X𓙂Ƃ݂Ȃ܂B
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object obj) {
		if (obj instanceof Attachment) {
			return ((Attachment) obj).downloadId == downloadId;
		}
		return false;
	}
}
