package petaclone;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Tool;

import org.apache.log4j.PropertyConfigurator;

import petaclone.denki.DenkiYohouClient;
import petaclone.ipmsg.IPMessenger;
import petaclone.petarou.PetaMessenger;
import petaclone.petaroux.PetaxMessenger;

/**
 * <p>Maiñx[XNXłB</p>
 * <p>RXgNgAƂĎ̃\bhԂɌĂяo܂B</p>
 * <ol>
 * 	<li>getPolicy</li>
 * 	<li>makePreferenceDirectory</li>
 * 	<li>initLogging</li>
 * 	<li>loadAppPreferences</li>
 * 	<li>(initLogWriter)</li>
 * 	<li>initMembers</li>
 * 	<li>initMessengerFacade</li>
 * 	<li>loadLaunchConfigurations</li>
 * </ol>
 * <p>bZW[̊Jn͕ʓrs܂B</p>
 * @author Takayuki Shoji
 */
public class BaseMain {

	/**
	 * log4jev[gt@C̃\[XpX
	 */
	private static final String LOG4J_TEMPLATE = "/log4j_template.properties"; //$NON-NLS-1$

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/** AvP[V̖O */
	protected final static String APP_NAME = mm.get("BaseMain.appName"); //$NON-NLS-1$

	/**
	 * logger
	 */
	private static Logger logger;

	/**
	 * RXgN^B
	 */
	protected BaseMain() {
		//
	}

	/** ݒ */
	private AppPreferences pref;

	/**
	 * ݒǂݍ݂܂B
	 * @param base ݒ肪tH_
	 * @return ݒ
	 * @throws IOException 
	 */
	private AppPreferences loadAppPreferences(File base) throws IOException {
		AppPreferences p = new AppPreferences(base);
		p.load();
		return p;
	}

	/**
	 * @return AppPreferencesԂ܂B
	 */
	protected AppPreferences getAppPreferences() {
		return pref;
	}

	/** CxgǗ */
	private EventBroker eventBroker;

	/**
	 * EventBroker擾܂B
	 * @return EventBroker
	 */
	protected EventBroker getEventBroker() {
		return eventBroker;
	}

	/** o[Ǘ */
	public Members members;

	/**
	 * o[Ǘ܂B
	 * AppPreferences𐶐ɌĂяoKv܂B
	 */
	private void initMembers() {
		assert this.pref != null;
		members = new Members(this.pref);
	}

	/**
	 * bZW[
	 */
	protected MessengerFacade messengerFacade;

	/**
	 * MessengerFacade܂B
	 * AppPreferences𐶐ɌĂяoKv܂B
	 */
	private void initMessengerFacade() {
		assert this.pref != null;
		assert messengerFacade == null;

		this.messengerFacade = new MessengerFacade();

		// CXg[ς݃bZW[̓o^
		if (pref.enableIPMessengerComponent())
			installMessenger(new IPMessenger());

		if (pref.enablePetarouXComponent())
			installMessenger(new PetaxMessenger());

		if (pref.enableOldPetarouComponent())
			installMessenger(new PetaMessenger());

		if (pref.enableDenkiYohouClient())
			installMessenger(new DenkiYohouClient());

		this.messengerFacade.setAppPreferences(pref);
	}

	/**
	 * CXg[ς݂̃bZW[ǉ܂B
	 * @param msgr bZW[
	 */
	private void installMessenger(Messenger msgr) {
		messengerFacade.addInstalled(msgr.getProtocol().signature, msgr);
	}

	/**
	 * bZ[WL^
	 */
	private MessageLogger messageLogger;

	/**
	 * MessageLogger擾܂B
	 * @return MessageLogger
	 */
	protected MessageLogger getMessageLogger() {
		return messageLogger;
	}

	/**
	 * N\
	 */
	private LaunchConfigurations launchConfigutations;

	/**
	 * N\܂B
	 * @param base N\fBNg
	 * @param installed CXg[ς݃R|[lgID
	 */
	private void initLaunchConfigurations(File base, Set installed) {
		assert launchConfigutations == null;

		launchConfigutations = new LaunchConfigurations(base);
		launchConfigutations.loadLaunchConfigurations(pref.isFirsttime(),
				installed);
	}

	/**
	 * LaunchConfigurations擾܂B
	 * @return LaunchConfigurations
	 */
	protected LaunchConfigurations getLaunchConfigurations() {
		return launchConfigutations;
	}

	/**
	 * bZW[Jn܂B
	 * startupLaunchConfigurationsɓo^ꂽbZW[Jn܂B
	 * @param listener T[oXi[
	 * @throws LaunchException NG[
	 */
	protected void startMessengers(ServerListener listener)
			throws LaunchException {
		List startups = launchConfigutations.getStartups();
		Iterator itr = startups.iterator();
		while (itr.hasNext()) {
			LaunchConfiguration conf = (LaunchConfiguration) itr.next();
			messengerFacade.start(conf, listener, eventBroker);
		}
	}

	/**
	 * N̏s܂B
	 * @throws StartupException
	 */
	protected void initialize() {
		/* ****************************************************** */
		// policy̎擾
		// MȌ
		// ݒ̎擾
		// o[Ǘ̏
		// MessengerFacadȅ
		// N\̏
		/* ****************************************************** */

		// Policy̎擾
		Policy policy = getPolicy();

		// ݒptH_𑶍݂
		File prefPath = policy.getPreferenceLocation();
		makeDirectory(prefPath, mm.get("BaseMain.prefFolder")); //$NON-NLS-1$

		// VXevpeB petaclone.pref ̒`
		System.setProperty("petaclone.pref", prefPath.getAbsolutePath()); //$NON-NLS-1$

		// Ot@C֌W̏
		try {
			initLogging(prefPath);
		} catch (IOException e) {
			throw new StartupException(false, //
					mm.get("BaseMain.loggingInitializationFailure"), e); //$NON-NLS-1$
		}

		// logger̎擾
		logger = Logger.getLogger(BaseMain.class);
		logger.info("Logging system was initialized."); //$NON-NLS-1$

		// ݒ̎擾
		try {
			AppPreferences p = loadAppPreferences(prefPath);
			p.setPolicy(policy);

			this.pref = p;

		} catch (IOException e) {
			throw new StartupException(true, //
					mm.get("BaseMain.couldNotReadPreference"), e); //$NON-NLS-1$
		}

		// Proxyݒ
		initProxySetting();

		initIPv6Setting();

		// CxgǗ
		this.eventBroker = new EventBrokerImpl();

		// bZ[WL^
		this.messageLogger = new MessageLoggerImpl(pref);

		// o[Ǘ
		initMembers();

		// MessengerFacade
		initMessengerFacade();

		// N\
		// policyN\fBNg擾
		File launchPath = policy.getLaunchConfigurationPath();
		// N\̕ҏWɂ炸N\tH_̍쐬݂܂B
		makeDirectory(launchPath, mm.get("BaseMain.launchConfigurationFolder")); //$NON-NLS-1$
		// 
		Set signs = Tool.newSetFromArray(messengerFacade
				.getInstalledMessengerSignatures());
		initLaunchConfigurations(launchPath, signs);

		// migemofBNg
		File migemoPath = policy.getMigemoDictionariesPath();
		makeDirectory(migemoPath, mm.get("BaseMain.migemoDictionaryFolder")); //$NON-NLS-1$
	}

	/**
	 * IPv6ݒs܂B
	 */
	private void initIPv6Setting() {
		/*
		 * java.net.preferIPv6Addresses̃ftHgfalseB
		 */
		if (this.pref.preferIPv6Addresses()) {
			System.setProperty("java.net.preferIPv6Addresses", "true"); //$NON-NLS-1$ //$NON-NLS-2$
		}
	}

	/**
	 * Proxy̐ݒ܂B
	 */
	private void initProxySetting() {
		/*
		 * java.net.useSystemProxies̃ftHgfalse
		 */
		if (this.pref.getUseSystemProxies()) {
			System.setProperty("java.net.useSystemProxies", "true"); //$NON-NLS-1$ //$NON-NLS-2$
		}
	}

	/**
	 * MO܂B
	 * @param prefPath ݒt@CfBNg
	 * @throws IOException 
	 */
	private void initLogging(File prefPath) throws IOException {
		/*
		 * ݂Ȃ炻̂܂
		 * ݂ȂȂ烊\[XRs[
		 * PropertyConfiguratorŐݒǂݍށB
		 */
		File log4jProp = new File(prefPath, "log4j.properties"); //$NON-NLS-1$
		if (!log4jProp.exists()) {
			// \[XRs[
			InputStream is = null;
			OutputStream os = null;
			BufferedOutputStream bs = null;
			try {
				is = BaseMain.class.getResourceAsStream(LOG4J_TEMPLATE);
				if (is == null) {

					throw new StartupException(false, //
							mm.get("BaseMain.log4jTemplateIsNeeded")); //$NON-NLS-1$
				}
				os = new FileOutputStream(log4jProp);
				bs = new BufferedOutputStream(os);
				int b;
				while ((b = is.read()) != -1) {
					bs.write(b);
				}

			} finally {
				if (is != null) {
					try {
						is.close();
					} catch (IOException e) {
						//
					}
				}
				if (bs != null) {
					try {
						bs.close();
					} catch (IOException e) {
						//
					}
				} else if (os != null) {
					try {
						os.close();
					} catch (IOException e) {
						//
					}
				}
			}
		} else if (!log4jProp.isFile()) {
			throw new StartupException(false, mm.get(
					"BaseMain.nonFileEntryFound", //$NON-NLS-1$
					new Object[] { log4jProp.getAbsolutePath() }));
		}

		// ݒ̃[h
		PropertyConfigurator.configure(log4jProp.getAbsolutePath());
	}

	/**
	 * tH_̑݊mFƍ쐬
	 * @param f ݒtH_̃pX
	 * @param folderName ÕbZ[WŎgptH_
	 * @throws StartupException 쐬łȂꍇɔ܂B
	 */
	private void makeDirectory(File f, String folderName) {
		if (!Tool.makeDirectory(f)) {
			throw new StartupException(logger != null, //
					mm.get("BaseMain.couldNotMakeFolder", //$NON-NLS-1$
							new Object[] { folderName, f.getAbsolutePath() }));
		}
	}

	/**
	 * Policy擾܂B
	 * @return PolicyIuWFNg
	 */
	private Policy getPolicy() {
		// \[Xpolicy.propertiesǂ
		InputStream is = BaseMain.class
				.getResourceAsStream("/policy.properties"); //$NON-NLS-1$
		if (is == null) {
			is = BaseMain.class
					.getResourceAsStream("/policy_default.properties"); //$NON-NLS-1$
			if (is == null) {
				throw new StartupException(false, //
						mm.get("BaseMain.policyFileIsNeeded")); //$NON-NLS-1$
			}
		}
		Properties prop = new Properties();
		try {
			prop.load(is);
		} catch (IOException e) {
			throw new StartupException(false, //
					mm.get("BaseMain.couldNotLoadPolicyFile"), e); //$NON-NLS-1$
		} finally {
			try {
				is.close();
			} catch (IOException e) {
				throw new StartupException(false, //
						mm.get("BaseMain.couldNotLoadPolicyFile"), e); //$NON-NLS-1$
			}
		}

		Policy policy = new Policy(prop);

		return policy;
	}
}
