package petaclone;

import java.io.IOException;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import jp.takayuki.logging.Logger;

/**
 * IPAhX̕ύXĎ܂B
 * 
 * @author Takayuki Shoji
 */
public class IPAddressWatcher implements Runnable {

	/**
	 * logger
	 */
	private static final Logger logger = Logger
			.getLogger(IPAddressWatcher.class);

	/**
	 * XPW[
	 */
	private final ScheduledExecutorService scheduler;

	/**
	 * EventBroker
	 */
	private final EventBroker broker;

	/**
	 * List<InetAddress>
	 */
	private List lastAddressList = new ArrayList();

	/**
	 * ÕXL
	 */
	private long lastScanTime;

	/**
	 * XLԊu(1)
	 */
	private static final int SCAN_INTERVAL = 60 * 1000;

	/**
	 * @param scheduler XPW[(OŒ~邱)
	 * @param broker EventBroker
	 */
	public IPAddressWatcher(ScheduledExecutorService scheduler,
			EventBroker broker) {
		this.scheduler = scheduler;
		this.broker = broker;
	}

	/**
	 * AhX}bv\z܂B
	 * @param list ݒ
	 * @throws SocketException 
	 */
	private synchronized void build(List list) throws SocketException {
		list.clear();
		Enumeration en = NetworkInterface.getNetworkInterfaces();
		while (en.hasMoreElements()) {
			NetworkInterface ni = (NetworkInterface) en.nextElement();
			Enumeration addrs = ni.getInetAddresses();
			while (addrs.hasMoreElements()) {
				InetAddress addr = (InetAddress) addrs.nextElement();
				list.add(addr);
			}
		}
		lastScanTime = System.currentTimeMillis();
	}

	public void run() {
		try {
			if (lastScanTime != 0
					&& System.currentTimeMillis() - lastScanTime > 2 * SCAN_INTERVAL) {
				logger.debug("Resumed"); //$NON-NLS-1$
				build(lastAddressList);
				broker.notifyEvent(Events.Resumed, null);

			} else if (lastAddressList.isEmpty()) {
				build(lastAddressList);

			} else {
				Enumeration en = NetworkInterface.getNetworkInterfaces();
				scanloop: while (en.hasMoreElements()) {
					NetworkInterface ni = (NetworkInterface) en.nextElement();
					Enumeration addrs = ni.getInetAddresses();
					while (addrs.hasMoreElements()) {
						InetAddress addr = (InetAddress) addrs.nextElement();
						if (!lastAddressList.contains(addr)) {
							// AhXύX܂͒ǉ
							build(lastAddressList);
							broker.notifyEvent(Events.NetworkChanged, null);
							break scanloop;
						}
					}
				}
				lastScanTime = System.currentTimeMillis();
			}
		} catch (IOException e) {
			logger.error(e);
		}

		scheduler.schedule(this, SCAN_INTERVAL, TimeUnit.MILLISECONDS);
	}
}
