package petaclone;

import java.io.File;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.InetAddressUtils;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Tool;

/**
 * Messengerɑ΂鑀sNXłB
 * @author Takayuki Shoji
 */
public class MessengerFacade {

	/**
	 * logger
	 */
	private static final Logger logger = Logger
			.getLogger(MessengerFacade.class);

	/** fobOtO */
	private static final boolean DEBUG = false;

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/**
	 * CXg[Ă郁bZW[PrototypeCX^X
	 * <br>L[: vgR̃VOl`
	 * <br>l: Messenger
	 */
	private LinkedHashMap installed = new LinkedHashMap(4);

	/**
	 * p\ȃbZW[NXB
	 * livingMessengerCount()̌ďoɋNĂ郁bZW[c܂B
	 * <br>l: Messenger
	 */
	private ArrayList messengers = new ArrayList();

	/**
	 * AvP[Vݒ
	 */
	private AppPreferences pref;

	/**
	 * AvP[Vݒݒ肵܂B
	 * Messengerւ̃ZbgƃbZ[W̑MŎgp܂B
	 * @param pref AvP[Vݒ
	 */
	void setAppPreferences(AppPreferences pref) {
		this.pref = pref;
	}

	////////////////////////////////////////////
	// NƏI
	////////////////////////////////////////////

	/**
	 * CXg[ς݂̃bZW[ǉ܂B
	 * @param signature vgR
	 * @param msgr PrototypeCX^X
	 */
	void addInstalled(String signature, Messenger msgr) {
		assert signature != null;
		assert signature.length() != 0;
		assert msgr != null;

		this.installed.put(signature, msgr);
	}

	/**
	 * Nݒ̑Ó؂܂B
	 * <ul>
	 * 	<li>vgRݒ肳Ă邩</li>
	 * 	<li>CXg[ς݂</li>
	 * 	<li>IDݒ肳Ă邩</li>
	 * </ul>
	 * @param conf Nݒ
	 * @throws LaunchException G[ꍇɔ܂B
	 */
	private void validateLaunchConfiguration(LaunchConfiguration conf)
			throws LaunchException {
		String signature = conf.getProtocol();
		if (signature == null || signature.length() == 0) {

			throw new LaunchException(//
					mm.get("MessengerFacade.validate.protocolMustBeSet"), conf); //$NON-NLS-1$
		}
		// CXg[ς݂
		if (!installed.containsKey(signature)) {
			throw new LaunchException(
					//
					mm.get("MessengerFacade.validate.protocolNotInstalled"), conf); //$NON-NLS-1$
		}

		// IDݒ肳Ă邩
		String id = conf.getId();
		if (id == null || id.length() == 0) {
			throw new LaunchException(//
					mm.get("MessengerFacade.validate.launchIdMustBeSet"), conf); //$NON-NLS-1$
		}
	}

	/**
	 * N\̃XgɊÂăbZW[N܂B
	 * confs͎̏𖞂ƁB
	 * <ul>
	 * 	<li>vgR݂邱</li>
	 * 	<li>IDݒ肳Ă邱</li>
	 * 	<li>N邱</li>
	 * </ul>
	 * @param conf NNݒ
	 * @param listener T[oXi[
	 * @param broker EventBroker
	 * @throws LaunchException NG[
	 */
	public void start(LaunchConfiguration conf, ServerListener listener,
			EventBroker broker) throws LaunchException {

		// O̊mF
		validateLaunchConfiguration(conf);

		if (!conf.getIsValid()) {
			throw new IllegalStateException(//
					"Messenger(" + conf.getId() + ") is not configured to start."); //$NON-NLS-1$ //$NON-NLS-2$
		}

		// CX^X̐
		Messenger template = (Messenger) installed.get(conf.getProtocol());
		Messenger instance;
		if (template.getId() == null) {
			instance = template;
		} else {
			instance = template.newInstance();
		}
		// N
		setup(instance, conf, listener, broker);
	}

	/**
	 * XbhO[v(String key, MessengerThreadGroup value)
	 */
	private HashMap threadGroups = new HashMap();

	/**
	 * w肳ꂽbZW[̃CX^XŃT[oN܂B
	 * @param messenger bZW[
	 * @param conf Nݒ
	 * @param listener T[oXi[
	 * @param broker EventBroker
	 * @throws LaunchException NG[
	 */
	private void setup(Messenger messenger, LaunchConfiguration conf,
			ServerListener listener, EventBroker broker) throws LaunchException {

		// AvP[Vݒ
		messenger.setAppPreferences(pref);
		// Nݒݒ
		messenger.setConfiguration(conf);
		// Xi[ݒ
		messenger.setServerListener(listener);

		String name = messenger.getProtocol().protocolName + '('
				+ conf.getName() + ';' + conf.getId() + ')';
		MessengerThreadGroup gr = (MessengerThreadGroup) threadGroups.get(name);
		if (gr == null) {
			gr = new MessengerThreadGroup(name, broker);
			threadGroups.put(name, gr);
		}

		logger.info("Starting messenger. name=" + name); //$NON-NLS-1$

		// T[oJn
		messenger.begin(gr);
		// NXgɒǉ
		messengers.add(messenger);
	}

	/**
	 * NĂ郁bZW[ׂďI܂B
	 */
	public void quit() {
		Iterator iterator = this.messengers.iterator();
		while (iterator.hasNext()) {
			Messenger m = (Messenger) iterator.next();
			m.quit();
		}
		this.messengers.clear();
	}

	////////////////////////////////////////////
	// u[hLXgñ\bh
	////////////////////////////////////////////

	/**
	 * T[o삵Ă邷ׂẴbZW[
	 * sendFind()܂B
	 * @param addOnly truêƂÃo[XgɑMs܂B
	 */
	public void sendFind(boolean addOnly) {
		Iterator itr = this.messengers.iterator();
		while (itr.hasNext()) {
			Messenger msgr = (Messenger) itr.next();
			if (msgr.getAlive()) {
				msgr.sendFind(addOnly);
			}
		}
	}

	/**
	 * T[o삵Ă邷ׂẴbZW[
	 * sendRename()܂B
	 */
	public void sendRename() {
		Iterator itr = this.messengers.iterator();
		while (itr.hasNext()) {
			Messenger msgr = (Messenger) itr.next();
			if (msgr.getAlive()) {
				msgr.sendRename();
			}
		}
	}

	/**
	 * T[o삵Ă邷ׂẴbZW[
	 * sendAbsence()܂B
	 * @param isAbsence s݃[h̏ꍇtrue
	 */
	public void sendAbsenceMode(boolean isAbsence) {
		Iterator itr = this.messengers.iterator();
		while (itr.hasNext()) {
			Messenger msgr = (Messenger) itr.next();
			if (msgr.getAlive()) {
				msgr.sendAbsenceMode(isAbsence);
			}
		}
	}

	/**
	 * T[o삵Ă邷ׂẴbZW[
	 * sendRegister()܂B
	 */
	public void sendRegister() {
		Iterator itr = this.messengers.iterator();
		while (itr.hasNext()) {
			Messenger msgr = (Messenger) itr.next();
			if (msgr.getAlive()) {
				msgr.sendRegister();
			}
		}
	}

	////////////////////////////////////////////
	// _E[hXi[
	////////////////////////////////////////////

	/**
	 * w肵bZW[Ƀ_E[hXi[o^܂B
	 * null̏ꍇׂ͂ẴbZW[ɓo^܂B
	 * @param launchId NIDBnull̏ꍇׂ͂ẴbZW[
	 * @param listener o^郊Xi[
	 */
	public void addDownloadListener(String launchId, DownloadListener listener) {
		Iterator itr = this.messengers.iterator();
		while (itr.hasNext()) {
			Messenger msgr = (Messenger) itr.next();
			if (launchId == null || launchId.equals(msgr.getId())) {
				msgr.addDownloadListener(listener);
			}
		}
	}

	/**
	 * w肵bZW[̃_E[hXi[܂B
	 * @param launchId NID
	 * @param listener _E[hXi[
	 */
	public void removeDownloadListener(String launchId,
			DownloadListener listener) {
		Iterator itr = this.messengers.iterator();
		while (itr.hasNext()) {
			Messenger msgr = (Messenger) itr.next();
			if (launchId == null || launchId.equals(msgr.getId())) {
				msgr.removeDownloadListener(listener);
			}
		}
	}

	////////////////////////////////////////////
	// NID烁bZW[擾
	////////////////////////////////////////////

	/**
	 * <p>w肳ꂽNIDbZW[擾܂B
	 * płȂ΃ftHgԂ܂B</p>
	 * <p>NĂȂꍇnullԂ܂B</p>
	 * @param launchId N\IDBnull̏ꍇ̓ftHgbZW[B
	 * @return bZW[
	 */
	public Messenger getMessenger(String launchId) {
		// w肳ꂽNIDbZW[
		Messenger msgr = findMessenger(launchId);
		if (msgr != null) {
			return msgr;
		}
		if (messengers.isEmpty()) {
			return null;
		}
		// 擪(ftHg)bZW[Ԃ
		return (Messenger) messengers.get(0);
		/*
		 * ۑbZ[W̃XgAPɑMۂɗp
		 * bZW[߂B
		 * XgȀꍇIDvȂꍇAoChAhX
		 * bZW[]܂B
		 * PɑMۂ̃bZW[IDv̂I΂
		 * ƍlĂ悢B
		 */
	}

	/**
	 * Xg̒w肳ꂽNIDbZW[Ԃ܂B<br>
	 * bZW[Xg͋NĂ郁bZW[݂̂ł邱Ƃ
	 * OƂĂ܂B
	 * @param launchId NID
	 * @return bZW[BȂꍇnullB
	 */
	public Messenger findMessenger(String launchId) {
		if (launchId == null || messengers.isEmpty()) {
			return null;
		}
		Iterator itr = messengers.iterator();
		while (itr.hasNext()) {
			Messenger msgr = (Messenger) itr.next();
			if (launchId.equals(msgr.getId())) {
				return msgr;
			}
		}
		return null;
	}

	/**
	 * s̃ftHgbZW[̋NIDԂ܂B
	 * UniqueMemberȂǂŃvgR̉Ɏgp܂B
	 * @return NID
	 */
	private String getDefaultMessengerId() {
		if (messengers.isEmpty()) {
			return null;
		}
		Messenger msgr = (Messenger) messengers.get(0);
		return msgr.getId();
	}

	/**
	 * M\ȃbZW[邩ǂԂ܂B
	 * @return M\ȏꍇtrue
	 */
	public boolean isSendable() {
		return !messengers.isEmpty();
	}

	/**
	 * bZ[W𑗐M܂B
	 * ̃\bhĂяoOisSendable()
	 * M\ł邩ǂmFKv܂B
	 * @param msg bZ[W
	 * @param dests Đ惊Xg
	 * @param listener MʃXi[
	 */
	public void sendMessage(Message msg, UniqueMember[] dests,
			SendingResultListener listener) {

		assert dests != null;
		assert dests.length != 0;

		// }`LXg̃tO𗧂Ă
		if (dests.length > 1) {
			msg.isMulticast = true;
		} else {
			msg.isMulticast = false;
		}

		// M(IPAhXnull̂܂܂ɂĂ)
		RuntimeProperties rp = pref.getRuntime();

		msg.userName = pref.getUserName();
		msg.osUserName = rp.getLoginName();
		msg.computerName = InetAddressUtils.getHostName();

		// ML^
		msg.setTime(Calendar.getInstance());

		String defaultLaunchId = getDefaultMessengerId();
		if (defaultLaunchId == null) {
			throw new IllegalStateException(
					"defaultLaunchId == null, Check with isSendable() in advance."); //$NON-NLS-1$
		}

		// L[: NID, l: ArrayList(l: UniqueMember)
		HashMap map = new HashMap();
		// gpCX^X߁AMwo
		for (int i = 0; i < dests.length; i++) {
			// D悷郁bZW[
			String launchId = dests[i].resolveMessenger(msg.hasAttachment(),
					msg.launchId, defaultLaunchId);
			// o^
			ArrayList list;
			if (map.containsKey(launchId)) {
				list = (ArrayList) map.get(launchId);
			} else {
				list = new ArrayList();
				map.put(launchId, list);
			}
			list.add(dests[i]);
		}

		// bZ[W̒`FbN
		Iterator idIter = map.entrySet().iterator();
		while (idIter.hasNext()) {
			Entry entry = (Entry) idIter.next();

			String launchId = (String) entry.getKey();
			Messenger msgr = getMessenger(launchId);
			int length = msgr.getMessageLength(msg.message);

			if (length > msgr.getProtocol().maxMessageLength) {
				List list = (List) entry.getValue();
				UniqueMember member = (UniqueMember) list.get(0);

				listener.onTooLongMessage(msgr.getProtocol(), member, length);
				return;
			}
		}

		// ebZW[ɑM
		Iterator itr = map.entrySet().iterator();
		while (itr.hasNext()) {
			Entry entry = (Entry) itr.next();

			// NID
			String launchId = (String) entry.getKey();

			// Đ
			List list = (List) entry.getValue();
			UniqueMember[] to = (UniqueMember[]) list
					.toArray(new UniqueMember[list.size()]);
			// M
			Messenger msgr = getMessenger(launchId);
			msgr.sendMessage(msg, to, listener);
		}
	}

	/**
	 * msgɑ΂Jʒm𑗐M܂B
	 * @param msg bZ[W
	 */
	public void sendOpenNotice(Message msg) {
		// J̎MbZ[W
		if (msg.isOpened || msg.messageClass != Message.RECEIVED) {
			return;
		}
		msg.isOpened = true;
		// bZW[NIDŒTB
		if (msg.launchId == null) {
			logger.error("msg.launchId must not be null."); //$NON-NLS-1$
			return;
		}
		Messenger msgr = getMessenger(msg.launchId);
		if (msgr != null)
			msgr.sendOpenNotice(msg);
	}

	/**
	 * Ytt@Cۑ܂B<br>
	 * KvȂ_E[hB
	 * @param message attachment郁bZ[W
	 * @param attachment ۑYtt@C
	 * @param destDir ۑ̃fBNg(efBNg)
	 */
	public void saveAttachment(Message message, Attachment attachment,
			File destDir) {

		assert destDir != null;
		assert destDir.isDirectory();

		// TODO ԃ`FbNSTATUS̎擾Attachment̃\bhɂق

		// ԃ`FbN
		if (attachment.status == Attachment.STATUS_FINISHED) {
			if (attachment.localPath == null) {
				attachment.status = Attachment.STATUS_ABORTED;
				if (attachment.listener != null) {
					attachment.listener
							.onAbort(attachment,
									"loccalPath must not be null.(FINISHED && localPath == null)"); //$NON-NLS-1$
				}
				return;
			}
		}
		if (attachment.status == Attachment.STATUS_SUSPENDED) {
			if (attachment.localPath == null) {
				attachment.status = Attachment.STATUS_ABORTED;
				if (attachment.listener != null) {
					attachment.listener
							.onAbort(attachment,
									"loccalPath must not be null.(SUSPENDED && localPath == null)"); //$NON-NLS-1$
				}
				return;
			}
		}
		if (attachment.status == Attachment.STATUS_ABORTED) {
			if (attachment.localPath == null) {
				attachment.status = Attachment.STATUS_ABORTED;
				if (attachment.listener != null) {
					attachment.listener
							.onAbort(attachment,
									"loccalPath must not be null.(ABORTED && localPath == null)"); //$NON-NLS-1$
				}
				return;
			}
		}

		// ĊJȂtrue
		boolean resuming = false;

		// łɃ_E[h
		if (attachment.status >= Attachment.STATUS_WAIT) {
			// Ȃ
			return;
		}

		// _E[hς
		if (attachment.localPath != null
				&& attachment.status == Attachment.STATUS_FINISHED) {

			// t@Cړ邾
			if (DEBUG)
				logger.debug("Downloaded file(before move)=" //$NON-NLS-1$
						+ attachment.localPath);
			attachment.move(destDir);
			if (DEBUG)
				logger.debug("Downloaded file(after move)=" //$NON-NLS-1$
						+ attachment.localPath);
			return;
		}

		// ĊJ\
		if (attachment.localPath != null
				&& (attachment.status == Attachment.STATUS_ABORTED || attachment.status == Attachment.STATUS_SUSPENDED)) {

			if (DEBUG)
				logger.debug("Resumable file(before move)=" //$NON-NLS-1$
						+ attachment.localPath);

			// t@CړłĊJ
			resuming = attachment.move(destDir);

			if (DEBUG)
				logger.debug("Resumable file(after move)=" //$NON-NLS-1$
						+ attachment.localPath);
			// returnȂ
		}

		//--------------------------------------------
		// _E[h
		//--------------------------------------------

		// Messenger擾
		Messenger msgr = findMessenger(message.launchId);
		if (msgr == null) {
			// bZW[̂Ń_E[hłȂ
			attachment.status = Attachment.STATUS_ABORTED;
			if (attachment.listener != null) {
				String msg = mm.get("MessengerFacade.downloaderNotStarted"); //$NON-NLS-1$
				attachment.listener.onAbort(attachment, msg);
			}
			return;
		}

		if (attachment.type == Attachment.TYPE_FILE
				|| attachment.type == Attachment.TYPE_INLINE) {
			// t@C̃_E[h
			File destination;
			if (resuming) {
				// ĊJ
				destination = new File(destDir, attachment.name);
			} else {
				// t@C𐶐ė\
				destination = Tool.createUniqueFile(destDir, attachment.name,
						10000);
				if (destination == null) {
					attachment.status = Attachment.STATUS_ABORTED;
					if (attachment.listener != null) {
						attachment.listener.onAbort(attachment, //
								mm.get("MessengerFacade.failedToRenameFolder")); //$NON-NLS-1$
					}
					return;
				}
				attachment.name = destination.getName();
			}
			// _E[h
			msgr.saveAttachment(message, attachment, destination);
		} else {
			// fBNg̃_E[h
			// IP Messengerł͍ĊJ̏ꍇ͍ŏ蒼ɂȂ
			File destination;
			if (resuming) {
				destination = new File(destDir, attachment.name);
			} else {
				destination = Tool.createUniqueDierectory(destDir,
						attachment.name, 10000);
				if (destination == null) {
					attachment.status = Attachment.STATUS_ABORTED;
					if (attachment.listener != null) {
						attachment.listener.onAbort(attachment, //
								mm.get("MessengerFacade.failedToRenameFolder")); //$NON-NLS-1$
					}
					return;
				}
				attachment.name = destination.getName();
			}
			// _E[h
			msgr.saveAttachment(message, attachment, destination);
		}
	}

	/**
	 * _E[hLZ܂B
	 * @param launchId NID
	 * @param attachment Ytt@C
	 */
	public void cancelDownload(String launchId, Attachment attachment) {

		Messenger messenger = findMessenger(launchId);
		if (messenger == null) {
			attachment.status = Attachment.STATUS_ABORTED;
			if (attachment.listener != null) {
				attachment.listener.onAbort(attachment, //
						mm.get("MessengerFacade.downloaderNotStarted")); //$NON-NLS-1$
			}
			return;
		}

		// LZ
		messenger.cancelSavingAttachment(attachment);
	}

	/**
	 * _E[h
	 * @param message bZ[W
	 */
	public void autoDownload(Message message) {
		if (!message.hasAttachment()) {
			return;
		}

		// _E[h\ȍőTCY擾
		long max = pref.getMaxAutoDownloadSize();
		boolean downloadInlineImages = pref.useInlineImageAutoDownload();

		if (max <= 0 && !downloadInlineImages) {
			return;
		}
		boolean folderDownload = pref.useFolderAutoDownload();

		// x[XpX
		String base = pref.getAttachmentStorage();
		File basePath = new File(base);
		// ݂ۏ
		if (!Tool.makeDirectory(basePath)) {
			return;
		}
		// őTCYȉ̂̂Tă_E[h
		Iterator itr = message.attachments.iterator();
		while (itr.hasNext()) {
			Attachment att = (Attachment) itr.next();
			if (att.localPath != null) {
				// _E[hς
				continue;
			}
			switch (att.type) {
			case Attachment.TYPE_FILE:
				if (max > 0 && att.size <= max) {
					this.saveAttachment(message, att, basePath);
				}
				break;
			case Attachment.TYPE_INLINE:
				if (downloadInlineImages || att.size <= max) {
					this.saveAttachment(message, att, basePath);
				}
				break;
			case Attachment.TYPE_DIR:
				if (folderDownload) {
					this.saveAttachment(message, att, basePath);
				}
				break;
			}
		}
	}

	/**
	 * ݒs߂ɃbZW[̃CX^X擾܂B
	 * @param signature bZW[̃VOl`
	 * @return VOl`v郁bZW[
	 */
	public Messenger getInstalledMessenger(String signature) {
		return (Messenger) installed.get(signature);
	}

	/**
	 * CXg[Ă郁bZW[̖O̔zԂ܂B
	 * @return bZW[̖O̔z
	 */
	public String[] getInstalledMessengerNames() {
		String[] result = new String[installed.size()];
		int i = 0;
		for (Iterator itr = installed.values().iterator(); itr.hasNext();) {
			Messenger messenger = (Messenger) itr.next();
			result[i] = messenger.getProtocol().protocolName;
			i++;
		}
		return result;
	}

	/**
	 * CXg[Ă郁bZW[̃VOl`̔zԂ܂B
	 * @return bZW[̃VOl`̔z
	 */
	public String[] getInstalledMessengerSignatures() {
		String[] result = new String[installed.size()];
		int i = 0;
		for (Iterator itr = installed.values().iterator(); itr.hasNext();) {
			Messenger messenger = (Messenger) itr.next();
			result[i] = messenger.getProtocol().signature;
			i++;
		}
		return result;
	}

	/**
	 * oChIPAhXƃ|[gԍɋ
	 * Ȃ`FbN܂B
	 * @param confs LaunchConfiguratioñXg
	 * @return Nݒ̃CfbNXB
	 * ȂƂ-1B
	 */
	public int getConflictConfiguration(List confs) {
		if (confs.size() <= 1) {
			return -1;
		}

		/*
		 * L[: Integer
		 * l: List
		 */
		Map tcpAddrs = new HashMap();
		Map udpAddrs = new HashMap();

		for (int i = 0; i < confs.size(); i++) {
			LaunchConfiguration conf = (LaunchConfiguration) confs.get(i);
			if (!conf.getIsValid()) {
				continue;
			}
			String id = conf.getId();

			Messenger messenger = getInstalledMessenger(conf.getProtocol());
			try {
				// TCP
				InetSocketAddress[] addrs = messenger.getTcpBindAddress(conf);
				if (hasConflict(tcpAddrs, addrs, id, "TCP")) { //$NON-NLS-1$
					return i;
				}
			} catch (UnknownHostException e) {
				logger.warn("Could not resolve bind address.", e); //$NON-NLS-1$
			}

			try {
				// UDP
				InetSocketAddress[] addrs = messenger.getUdpBindAddress(conf);
				if (hasConflict(udpAddrs, addrs, id, "UDP")) { //$NON-NLS-1$
					return i;
				}
			} catch (UnknownHostException e) {
				logger.warn("Could not resolve bind address.", e); //$NON-NLS-1$
			}
		}

		return -1;
	}

	/**
	 * addrsportsrċȂmF܂B
	 * Ȃꍇportsɓo^܂B
	 * @param ports `FbNς݂̃|[gԍƃAhX(L[: Integer, l: List(InetAddress))
	 * @param addrs `FbNΏۂ̃AhX
	 * @param id NݒIDB
	 * @param type TCP܂UDPBOo͗pB
	 * @return ꍇtrueB
	 */
	private boolean hasConflict(Map ports, InetSocketAddress[] addrs,
			String id, String type) {
		for (int i = 0; i < addrs.length; i++) {
			Integer port = new Integer(addrs[i].getPort());
			if (!ports.containsKey(port)) {
				// Ώۂ̃|[gԍōŏ
				List list = new ArrayList();
				list.add(addrs[i].getAddress());
				ports.put(port, list);
			} else {
				// łɑ݂|[gԍ
				InetAddress addr = addrs[i].getAddress();
				if (addr == null) {
					logger.warn(mm.get("MessengerFacade.competitionDetected", //$NON-NLS-1$
							new Object[] { type, "all", port, id })); //$NON-NLS-1$
					return true;
				}
				// AhXv邩Anull΋
				List list = (List) ports.get(port);
				Iterator addrItr = list.iterator();
				while (addrItr.hasNext()) {
					InetAddress addr2 = (InetAddress) addrItr.next();
					if (addr2 == null) {
						logger.warn(mm.get(
								"MessengerFacade.competitionDetected", //$NON-NLS-1$
								new Object[] { type, "all", port, id })); //$NON-NLS-1$
						return true;
					} else if (addr2.equals(addr)) {
						logger.warn(mm.get(
								"MessengerFacade.competitionDetected", //$NON-NLS-1$
								new Object[] { type, addr2, port, id }));
						return true;
					}
				}
				list.add(addr);
			}
		}
		return false;
	}

}
