package petaclone;

import java.io.File;
import java.util.Properties;

import jp.takayuki.tools.MacOSXPathResolver;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Os;
import jp.takayuki.tools.PathResolver;
import jp.takayuki.tools.PropertiesUtils;
import jp.takayuki.tools.UnixPathResolver;
import jp.takayuki.tools.WindowsPathResolver;

/**
 * AvP[V̓|V[
 * @author Takayuki Shoji
 */
public class Policy {

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/**
	 * policy.properties
	 */
	private Properties prop;

	/**
	 * fBNg̉
	 */
	private PathResolver pathResolver;

	/**
	 * RXgN^
	 * @param prop policy.properties
	 */
	public Policy(Properties prop) {
		if (prop == null) {
			throw new IllegalArgumentException("prop must not be null."); //$NON-NLS-1$
		}
		this.prop = prop;
		switch (Os.getId()) {
		case Os.WINDOWS:
			pathResolver = new WindowsPathResolver();
			break;
		case Os.MACOSX:
			pathResolver = new MacOSXPathResolver();
			break;
		default:
			pathResolver = new UnixPathResolver();
			break;
		}
	}

	/**
	 * PathResolver Ԃ܂B
	 * @return PathResolver
	 */
	public PathResolver getPathResolver() {
		return pathResolver;
	}

	/**
	 * ݒfBNg擾܂B
	 * ݒfBNg͕K[Jt@CVXeƂ
	 * Fʒuɂ܂A݂Ȃꍇ܂B
	 * @return ݒfBNgBۂɂ݂͑ȂꍇB
	 */
	public File getPreferenceLocation() {

		return getDirectory("pref.location", getPreferenceFolderName(), //$NON-NLS-1$
				"pref.location.custom"); //$NON-NLS-1$
	}

	/**
	 * C^[lbgɐڑT[rX̗p邩ǂB
	 * @return ꍇtrue
	 */
	public boolean canUseInternet() {
		return PropertiesUtils.getBoolean(prop, "internet.enable", true); //$NON-NLS-1$
	}

	/**
	 * N\̕ҏW邩ǂB
	 * @return ꍇtrue
	 */
	public boolean canEditLaunchConfiguration() {
		return PropertiesUtils.getBoolean(prop, "launchConfiguration.enable", //$NON-NLS-1$
				true);
	}

	/**
	 * N\̃fBNg擾܂B
	 * @return N\fBNg
	 */
	public File getLaunchConfigurationPath() {
		String path = getPreferenceFolderName()
				+ System.getProperty("file.separator") + "launch"; //$NON-NLS-1$ //$NON-NLS-2$

		return getDirectory("launchConfiguration.location", path, //$NON-NLS-1$
				"launchConfiguration.location.custom"); //$NON-NLS-1$
	}

	/**
	 * migemofBNg擾܂B
	 * @return migemofBNg
	 */
	public File getMigemoDictionariesPath() {
		String path = getPreferenceFolderName()
				+ System.getProperty("file.separator") + "migemo"; //$NON-NLS-1$ //$NON-NLS-2$

		return getDirectory("migemoDictionaries.location", path, //$NON-NLS-1$
				"migemoDictionaries.location.custom"); //$NON-NLS-1$
	}

	/**
	 * updatefBNg擾܂B
	 * @return updatet@C̃_E[h
	 */
	public File getUpdateDirectoryPath() {
		String path = getPreferenceFolderName()
				+ System.getProperty("file.separator") + "update"; //$NON-NLS-1$ //$NON-NLS-2$

		return getDirectory("updateDirectory.location", path, //$NON-NLS-1$
				"updateDirectory.location.custom"); //$NON-NLS-1$
	}

	/**
	 * <p>ݒtH_擾܂B</p>
	 * <p>ex.) petaclone</p>
	 * @return tH_
	 */
	private String getPreferenceFolderName() {
		switch (Os.getId()) {
		case Os.WINDOWS:
			return "petaclone"; //$NON-NLS-1$
		case Os.MACOSX:
			return "petaclone"; //$NON-NLS-1$
		default:
			return ".petaclone"; //$NON-NLS-1$
		}
	}

	/**
	 * w肵vpeBɑΉfBNgԂ܂B
	 * @param name Application DataAMy DocumentsAProgram DirectoryACustom DirectorŷꂩB
	 * @param child Custom DirectoryProgram DirectoryȊȌꍇɎq̃fBNgw肵܂B
	 * @param customName nameCustom Directoryw肵ꍇɎgpJX^fBNg̃vpeBB
	 * @return fBNgB݂ȂꍇB
	 */
	private File getDirectory(String name, String child, String customName) {
		String location = prop.getProperty(name, "Application Data"); //$NON-NLS-1$
		if (location == null || location.length() == 0) {
			location = "Application Data"; //$NON-NLS-1$
		}

		if (location.equals("Program Directory")) { //$NON-NLS-1$
			return pathResolver.getProgramDirectory();

		} else if (location.equals("Custom Directory")) { //$NON-NLS-1$
			String custom = prop.getProperty(customName);
			if (custom == null || custom.length() == 0) {
				throw new IllegalPolicyException(//
						mm.get("Policy.dir.custom", //$NON-NLS-1$
								new Object[] { name, customName }));
			}
			return new File(custom);
		}

		if (!location.equals("Application Data") //$NON-NLS-1$
				&& !location.equals("My Documents")) { //$NON-NLS-1$

			throw new IllegalPolicyException(//
					mm.get("Policy.dir.constant", //$NON-NLS-1$
							new Object[] { name }));
		}

		File f;
		if (location.equals("My Documents")) { //$NON-NLS-1$
			f = pathResolver.getMyDocuments();
		} else {
			// ftHgApplication Data
			f = pathResolver.getApplicationData();
		}

		// qt
		if (child == null || child.length() == 0) {
			return f;
		}
		return new File(f, child);
	}
}
