package petaclone.ipmsg;

import java.math.BigInteger;
import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.PublicKey;
import java.security.interfaces.RSAPrivateKey;
import java.security.interfaces.RSAPublicKey;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.RSAPrivateKeySpec;
import java.security.spec.RSAPublicKeySpec;
import java.util.StringTokenizer;

import javax.crypto.Cipher;
import javax.crypto.KeyGenerator;
import javax.crypto.SecretKey;
import javax.crypto.spec.IvParameterSpec;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.Tool;

/**
 * 
 * @author Takayuki Shoji
 */
public class IPMCipher {

	/**
	 * logger
	 */
	private static final Logger logger = Logger.getLogger(IPMCipher.class);

	/** fobOtO */
	private static final boolean DEBUG = logger.isDebugEnabled();

	/** [U[ʎq */
	private String identifier;

	/** ÍgtO(ÍΉ\) */
	private long encryptionFlag;

	/** J */
	private PublicKey publickey;

	/**
	 * RXgN^
	 * @param identifier [Uʎq
	 * @param encryptionFlag ÍgtO(ÍΉ\)
	 * @param publickey J
	 */
	IPMCipher(String identifier, long encryptionFlag, PublicKey publickey) {

		this.identifier = identifier;
		this.encryptionFlag = encryptionFlag;
		this.publickey = publickey;
	}

	/**
	 * \(ANSPUBKEY̊g)\z܂B<br>
	 * ÍΉ\:J
	 * @param identifier [Uʎq
	 * @param str \(ANSPUBKEY̊g)
	 * @throws NoSuchAlgorithmException ASYpłȂƂ
	 * @throws InvalidKeySpecException dlȂƂ
	 */
	IPMCipher(String identifier, String str) throws NoSuchAlgorithmException,
			InvalidKeySpecException {
		StringTokenizer token = new StringTokenizer(str, ":"); //$NON-NLS-1$
		long cap = Long.parseLong(token.nextToken(), 16);
		KeyFactory kf = KeyFactory.getInstance("RSA"); //$NON-NLS-1$
		RSAPublicKeySpec keyspec = makePublicKeySpec(token.nextToken());
		PublicKey key = kf.generatePublic(keyspec);

		// tB[hɕۑ
		this.identifier = identifier;
		this.encryptionFlag = cap;
		this.publickey = key;
	}

	/**
	 * @return [U[ʎq擾܂B
	 */
	String getUserIdentifier() {
		return identifier;
	}

	/**
	 * @return ÍgptO擾܂B
	 */
	long getEncryptionFlag() {
		return encryptionFlag;
	}

	/**
	 * @return J擾܂B
	 */
	PublicKey getPublicKey() {
		return publickey;
	}

	/**
	 * M̌JŃbZ[WÍ܂B
	 * <p>
	 * useSmallKeyuseRC2LŁAM悪RC2T|[gĂꍇ̂݁AtrueLłB
	 * 
	 * @param message 
	 * @param encoding GR[fBO
	 * @param useRC2 RC2gp\ɂ(x̊Ǌ|V[t@CKv)
	 * @param useSmallKey IRSA 512 bit with RC2 40 bit[hgꍇtrueB
	 * @return ÍꂽBmessageƓ̏ꍇ͕̂܂܂ňÍĂȂB
	 */
	String encrypt(String message, String encoding, boolean useRC2,
			boolean useSmallKey) {

		try {
			// Í̎
			long flag = 0L;
			int keyLength = this.publickey.getEncoded().length;
			if (keyLength > 200)
				flag |= IPMSG.RSA_2048;
			else if (keyLength > 100)
				flag |= IPMSG.RSA_1024;
			else
				flag |= IPMSG.RSA_512;

			//// tO
			//if ((this.encryptionFlag & IPMSG.SIGN_MD5) != 0) {
			//	flag |= IPMSG.SIGN_MD5;
			//}

			// 閧ƈÍASY擾
			// ʌ̐
			SecretKey commonKey = null;
			String cipherName = null;
			int bit;

			if (useSmallKey && useRC2
					&& (this.encryptionFlag & IPMSG.RC2_40) != 0) {
				flag |= IPMSG.RC2_40;
				KeyGenerator kg = KeyGenerator.getInstance("RC2"); //$NON-NLS-1$
				kg.init(40);
				commonKey = kg.generateKey();
				cipherName = IPMSG.CIPHER_RC2;
				bit = 40;
			} else if ((this.encryptionFlag & IPMSG.BLOWFISH_128) != 0) {
				flag |= IPMSG.BLOWFISH_128;
				KeyGenerator kg = KeyGenerator.getInstance("Blowfish"); //$NON-NLS-1$
				kg.init(128);
				commonKey = kg.generateKey();
				cipherName = IPMSG.CIPHER_BLOWFISH;
				bit = 128;
			} else if (useRC2 && (this.encryptionFlag & IPMSG.RC2_128) != 0) {
				flag |= IPMSG.RC2_128;
				KeyGenerator kg = KeyGenerator.getInstance("RC2"); //$NON-NLS-1$
				kg.init(128);
				commonKey = kg.generateKey();
				cipherName = IPMSG.CIPHER_RC2;
				bit = 128;
			} else if (useRC2 && (this.encryptionFlag & IPMSG.RC2_40) != 0) {
				flag |= IPMSG.RC2_40;
				KeyGenerator kg = KeyGenerator.getInstance("RC2"); //$NON-NLS-1$
				kg.init(40);
				commonKey = kg.generateKey();
				cipherName = IPMSG.CIPHER_RC2;
				bit = 40;
			} else {
				logger.warn("There is no common encryption algorithm. useRC2=" //$NON-NLS-1$
						+ useRC2 + ", encryptionFlag=" + this.encryptionFlag //$NON-NLS-1$
						+ ", rsa key length=" + keyLength); //$NON-NLS-1$
				// ԂB
				return message;
			}

			StringBuffer sb = new StringBuffer();
			sb.append(Long.toString(flag, 16).toUpperCase());
			sb.append(':');

			// Í
			//Jɂ鋤ʌ̈Í
			Cipher c = Cipher.getInstance(IPMSG.CIPHER_RSA);
			c.init(Cipher.ENCRYPT_MODE, this.publickey);
			byte[] keydata = c.doFinal(commonKey.getEncoded());

			sb.append(Tool.bytesToHexString(keydata));
			sb.append(':');
			// ʌɂ郁bZ[ẄÍ
			c = Cipher.getInstance(cipherName);
			IvParameterSpec iv = new IvParameterSpec(new byte[8]);
			c.init(Cipher.ENCRYPT_MODE, commonKey, iv);
			byte[] binaryMessage = c.doFinal(message.getBytes(encoding));
			sb.append(Tool.bytesToHexString(binaryMessage));

			//// 
			//if ((this.encryptionFlag & IPMSG.SIGN_MD5) != 0) {
			//	sb.append(':');
			//	MessageDigest md5 = MessageDigest.getInstance("MD5");
			//	md5.update(message.getBytes(IPMessenger.ENCODING));
			//	byte digest[] = md5.digest();
			//	// TODO : ̔閧ňÍ
			//	String d = Tool.bytesToHexString(digest);
			//	System.out.println("܂:" + d);
			//	sb.append(d);
			//}

			if (DEBUG)
				logger.debug("Encrypted using " + cipherName //$NON-NLS-1$
						+ ". Strength of common key(bit)=" + bit //$NON-NLS-1$
						+ ", Encryption flag of destination=0x" //$NON-NLS-1$
						+ Long.toHexString(this.encryptionFlag)
						+ ", Encryption flag of send data=0x" //$NON-NLS-1$
						+ Long.toHexString(flag)
						+ ", Key length of destination(byte)=" + keyLength); //$NON-NLS-1$

			return sb.toString();

		} catch (Exception e) {
			logger.error("Failed to encrypt. " //$NON-NLS-1$
					+ "InvalidAlgorithmParameterException̏ꍇ" //$NON-NLS-1$
					+ "x̊Ǌ|V[t@CCXg[ĂȂ\܂B" //$NON-NLS-1$
					+ "RSA512bitARC2 40bit[ĥƂBadPaddingExceptionN肦܂B" //$NON-NLS-1$
					+ "useRC2=" + useRC2 + ", M̈ÍΉ\=0x" //$NON-NLS-1$ //$NON-NLS-2$
					+ Long.toHexString(this.encryptionFlag), e);
		}
		// Ԃ
		return message;
	}

	/**
	 * 
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();

		sb.append("<cipher>\n"); //$NON-NLS-1$

		sb.append("\t<user>"); //$NON-NLS-1$
		sb.append(this.identifier);
		sb.append("</user>\n"); //$NON-NLS-1$

		sb.append("\t<encflag>"); //$NON-NLS-1$
		sb.append(this.encryptionFlag);
		sb.append("</encflag>\n"); //$NON-NLS-1$

		sb.append("\t<pubkey>"); //$NON-NLS-1$
		sb.append(this.publickey.toString());
		sb.append("</pubkey>\n"); //$NON-NLS-1$

		sb.append("</cipher>\n"); //$NON-NLS-1$

		return sb.toString();
	}

	/**
	 * RSAPublicKeySpecɕϊ܂B
	 * @param src ̕\
	 * @return RSAJdl
	 * @throws NumberFormatException l̕񂪐ȂƂ
	 */
	static RSAPublicKeySpec makePublicKeySpec(String src)
			throws NumberFormatException {

		StringTokenizer token = new StringTokenizer(src, "-"); //$NON-NLS-1$
		BigInteger exponent = new BigInteger(token.nextToken(), 16);
		BigInteger modulus = new BigInteger(token.nextToken(), 16);
		return new RSAPublicKeySpec(modulus, exponent);
	}

	/**
	 * RSAPrivateKeySpecɕϊ܂B
	 * @param src ̕\
	 * @return RSA閧dl
	 * @throws NumberFormatException l̕񂪐ȂƂ
	 */
	static RSAPrivateKeySpec makePrivateKeySpec(String src)
			throws NumberFormatException {

		StringTokenizer token = new StringTokenizer(src, "-"); //$NON-NLS-1$
		BigInteger exponent = new BigInteger(token.nextToken(), 16);
		BigInteger modulus = new BigInteger(token.nextToken(), 16);
		return new RSAPrivateKeySpec(modulus, exponent);
	}

	/**
	 * RSAPublicKey𕶎ɕϊ܂B
	 * @param key RSAJ
	 * @return J̕\
	 */
	static String makeRSAPublicKeyString(RSAPublicKey key) {
		return key.getPublicExponent().toString(16).toUpperCase() + "-" //$NON-NLS-1$
				+ key.getModulus().toString(16).toUpperCase();
	}

	/**
	 * RSAPrivateKey𕶎ɕϊ܂B
	 * @param key RSA閧
	 * @return 閧̕\
	 */
	static String makeRSAPrivateKeyString(RSAPrivateKey key) {
		return key.getPrivateExponent().toString(16).toUpperCase() + "-" //$NON-NLS-1$
				+ key.getModulus().toString(16).toUpperCase();
	}

	/**
	 * ÍbZ[WÍ̎ނ擾܂B
	 * @param src ÍbZ[W
	 * @return Í̎
	 */
	static long getCapability(String src) {
		int index = src.indexOf(':');
		if (index == -1) {
			return 0;
		}
		String c = src.substring(0, index);
		try {
			return Long.parseLong(c, 16);
		} catch (NumberFormatException e) {
			logger.warn("Invalid number. str=" + c, e); //$NON-NLS-1$
			return 0;
		}
	}
}
