package petaclone.ipmsg;

import java.io.UnsupportedEncodingException;
import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Iterator;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.InetAddressUtils;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Tool;
import petaclone.AppPreferences;
import petaclone.Attachment;
import petaclone.InvalidDataFormatException;
import petaclone.Member;
import petaclone.Message;
import petaclone.RuntimeProperties;

/**
 * IP MessengerUDPpPbgB
 * @author Takayuki Shoji
 */
public class IPMPacket {

	/**
	 * logger
	 */
	private static final Logger logger = Logger.getLogger(IPMPacket.class);

	/**
	 * x[XpPbgԍ
	 */
	private static final long BASE_ID = System.currentTimeMillis() & 0xffffffffL;

	/**
	 * OpbZ[W
	 */
	private static final String TOO_SHORT = "Packet data is too short."; //$NON-NLS-1$

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/** g: CC摜̑}ʒu */
	private static final String POS = "8="; //$NON-NLS-1$

	/** ݒȂ */
	IPMessenger messenger;

	/** vgR̃o[W */
	int version;

	/** pPbgԍ */
	long id;

	/** OSւ̃OC */
	String loginName;

	/** Rs[^ */
	String hostName;

	/** R}h */
	long command;

	/**
	 * sR[h
	 */
	private static final String EOL = System.getProperty("line.separator"); //$NON-NLS-1$

	/**
	 * ǉB<br>
	 * BR_*,ANSENTRY:jbNl[ASENDMSG:bZ[W{
	 */
	String data;

	/**
	 * BR_*,ANSENTRY:O[vASENDMSG:t@CYtA
	 * TCP̂ƂpacketID:fileID:offset(tH_̏ꍇ:offset͂Ȃ)
	 */
	String extension;

	/** MpPbg̑MłBMɂ͎g܂B */
	InetSocketAddress address;

	/**
	 * MpRXgN^łB
	 * @param messenger bZW[
	 */
	IPMPacket(IPMessenger messenger) {
		this.messenger = messenger;
		this.version = 1;
	}

	/**
	 * MessageNX\zRXgN^łB<br>
	 * bZ[WMp<br>
	 * idAcommand͐ݒ肵܂B
	 * @param messenger bZW[
	 * @param msg bZ[W
	 */
	IPMPacket(IPMessenger messenger, Message msg) {
		this.messenger = messenger;
		this.version = 1;
		this.data = msg.getMessageWithoutPlaceholder();

		// Mɂ
		this.loginName = msg.osUserName;
		this.hostName = msg.computerName;
	}

	/**
	 * sR[h\nɓꂵ܂B
	 */
	void convertReturnCode() {
		StringBuffer sb = new StringBuffer(data);
		Tool.replaceAll(sb, "\r\n", "\n"); //$NON-NLS-1$ //$NON-NLS-2$
		Tool.replaceAll(sb, "\r", "\n"); //$NON-NLS-1$ //$NON-NLS-2$
		this.data = sb.toString();
	}

	/**
	 * oCi\̃RXgN^łB<br>
	 * Ytt@C͉͂܂B
	 * @param messenger bZW[
	 * @param buffer oCif[^
	 * @param address MAhXƃ|[g
	 * @throws InvalidDataFormatException
	 */
	IPMPacket(IPMessenger messenger, byte[] buffer, InetSocketAddress address)
			throws InvalidDataFormatException {
		this.messenger = messenger;
		this.address = address;
		unpack(buffer);
	}

	/**
	 * [U[ƃO[vdataɐݒ肵܂B
	 */
	void setUserGroupName() {
		AppPreferences pref = messenger.getAppPreferences();
		this.data = pref.getUserName();

		if (pref.getRuntime().isAbsence()) {
			this.data += mm.get("IPMPacket.absent"); //$NON-NLS-1$
		}

		this.extension = pref.getGroupName();
	}

	/**
	 * pPbgԍ쐬܂BMO1񂾂ĂяoƁB
	 */
	void setPacketNumber() {
		synchronized (IPMPacket.class) {
			// IP Messenger͕t32bit܂
			id = (BASE_ID + IPMessenger.count++) & 0x7fffffff;
		}
	}

	/**
	 * w肳ꂽo[FnR}hɎI
	 * IvVݒ肵܂B
	 * @param command R}h
	 * @param encryptOption ÍIvV(ÍȂꍇ0)
	 * @param destSpec 
	 */
	void setOrganizationCommand(long command, long encryptOption,
			IPMPacket destSpec) {
		assert (command & IPMSG.ANSENTRY) != 0
				|| (command & IPMSG.BR_ABSENCE) != 0
				|| (command & IPMSG.BR_ENTRY) != 0
				|| (command & IPMSG.BR_EXIT) != 0 : command;

		this.command = command;
		if (messenger.getAppPreferences().getRuntime().isAbsence()) {
			this.command |= IPMSG.ABSENCEOPT;
		}
		// t@CYt\
		this.command |= IPMSG.FILEATTACHOPT;
		// Í\
		this.command |= encryptOption;
		// UTF-8Ή
		this.command |= IPMSG.CAPUTF8OPT;
		// CC摜M\
		this.command |= IPMSG.CLIPBOARDOPT;

		// _CAbvł
		if (messenger.isOptionDialup()) {
			this.command |= IPMSG.DIALUPOPT;
		}

		// 悪UTF-8ɑΉĂ
		if (destSpec != null && destSpec.isUTF8Supported()) {
			this.command |= IPMSG.UTF8OPT;
		}
	}

	/**
	 * Mɍ킹SENDMSGp̃R}hƃIvVݒ肵܂B
	 * @param msg M悤ƂĂ郁bZ[W
	 * @param pack M̃[U[
	 */
	void setSendMsgCommand(Message msg, IPMPacket pack) {
		this.command = IPMSG.SENDMSG | IPMSG.SENDCHECKOPT;
		if (msg.isMulticast) {
			this.command |= IPMSG.MULTICASTOPT;
		}

		if (pack.isUTF8Supported()) {
			logger.debug("Added UTF-8 option."); //$NON-NLS-1$
			this.command |= IPMSG.UTF8OPT;
			this.command |= IPMSG.CAPUTF8OPT;
		}
		// t@CYt
		if (msg.hasAttachment()) {
			this.command |= IPMSG.FILEATTACHOPT;
		}
		// 
		if (!msg.isOpened) {
			this.command |= IPMSG.SECRETOPT | IPMSG.READCHECKOPT;
			// pX[h
			if (msg.isPasswordRequired) {
				this.command |= IPMSG.PASSWORDOPT;
			}
		}
	}

	/**
	 * CAPUTF8OPTtO̗LԂ܂B
	 * @return UTF-8̎MT|[gĂꍇtrue
	 */
	boolean isUTF8Supported() {
		return Tool.hasFlag(command, IPMSG.CAPUTF8OPT);
	}

	/**
	 * CLIPBOARDOPTtO̗LԂ܂B
	 * @return CC摜̎MT|[gĂꍇtrue
	 */
	boolean isInlineImageSupported() {
		return Tool.hasFlag(command, IPMSG.CLIPBOARDOPT);
	}

	/**
	 * FILEATTACHOPTtO̗LԂ܂B
	 * @return t@CYtT|[gĂꍇtrue
	 */
	private boolean isAttachmentSupported() {
		return Tool.hasFlag(command, IPMSG.FILEATTACHOPT);
	}

	/**
	 * @return ݒ肳ĂR}h
	 * BR_ABSENCEABR_ENTRYAANSENTRYABR_EXITȂtrue
	 */
	private boolean isOrganizationCommand() {
		long c = getPacketClass();
		return c == IPMSG.BR_ENTRY || c == IPMSG.BR_EXIT || c == IPMSG.ANSENTRY
				|| c == IPMSG.BR_ABSENCE;
	}

	/**
	 * UTF8OPT1ǂ擾܂B
	 * @return UTF8OPT̏ꍇAtrueB
	 */
	private boolean hasUtf8Option() {
		return (this.command & IPMSG.UTF8OPT) != 0;
	}

	/**
	 * commandR}h擾܂B
	 * @return R}hIvVԂ܂B
	 */
	long getPacketClass() {
		return this.command & IPMSG.MASK_COMMAND;
	}

	/**
	 * commandIvV擾܂B
	 * @return R}hR}hԂ܂B
	 */
	long getOption() {
		return this.command & IPMSG.MASK_OPTION;
	}

	/**
	 * ɂ܂B
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("version: ").append(version); //$NON-NLS-1$
		sb.append(", id: ").append(id); //$NON-NLS-1$
		sb.append(", loginName: ").append(loginName); //$NON-NLS-1$
		sb.append(", hostName: ").append(hostName); //$NON-NLS-1$
		sb.append(", command(PacketClass): ").append( //$NON-NLS-1$
				Long.toString(getPacketClass(), 16));
		sb.append(", command(Option): ").append(Long.toString(getOption(), 16)); //$NON-NLS-1$
		sb.append(", data: ").append(data); //$NON-NLS-1$
		sb.append(", group: ").append(extension); //$NON-NLS-1$
		return sb.toString();
	}

	/**
	 * Member擾܂B
	 * @return o[
	 */
	Member toMember() {
		assert (this.command & IPMSG.ANSENTRY) != 0
				|| (this.command & IPMSG.BR_ABSENCE) != 0
				|| (this.command & IPMSG.BR_ENTRY) != 0
				|| (this.command & IPMSG.BR_EXIT) != 0;

		Member member = new Member();
		if (this.data != null && this.data.length() != 0) {
			member.userName = this.data;
		} else {
			member.userName = this.loginName;
		}
		member.osUserName = this.loginName;
		member.computerName = this.hostName;
		member.group = this.extension == null ? "" : this.extension; //$NON-NLS-1$
		member.setLocation(this.address);

		// IvV
		if ((this.getOption() & IPMSG.ABSENCEOPT) != 0) {
			// s
			member.isAbsence = true;
		}
		if ((this.getOption() & IPMSG.FILEATTACHOPT) != 0) {
			// t@CM\
			member.canReceiveFile = true;
		}
		// vgR
		member.launchId = messenger.getId();

		return member;
	}

	/**
	 * Message擾܂B
	 * <p>
	 * bZW[̃[U[(jbNl[)̓bZ[W̃pPbg
	 * Ȃ̂łƂŐݒ肷邱ƁB
	 * <p>
	 * UDPő`̓Ytt@CAttachmentɕ܂B
	 * <p>
	 * CC摜܂ޏꍇAbZ[W{ɂ̓v[Xz_}܂B
	 * 
	 * @return bZ[W
	 */
	Message toMessage() {
		assert (this.command & IPMSG.SENDMSG) != 0
				|| (this.command & IPMSG.READMSG) != 0;

		Message message = new Message();

		// bZ[W^Cv
		if ((getPacketClass() & IPMSG.SENDMSG) != 0) {
			// ʏ̃bZ[W
			message.messageType = Message.MESSAGE;
			if ((getOption() & IPMSG.AUTORETOPT) != 0) {
				// ԐM
				message.setForegroundBlue();
			}
			if ((getOption() & IPMSG.FILEATTACHOPT) != 0) {
				// Ytt@C̉
				message.attachments = unpackAttachmentInfo(this.extension);
			}
		} else if ((getPacketClass() & IPMSG.READMSG) != 0) {
			// ̊Jʒm
			message.messageType = Message.OPEN_NOTICE;
		} else {
			// ̑
			message.messageType = Message.TYPE_OTHER;
		}
		message.id = (int) this.id;
		// ݂̎ݒ
		message.setTime(Calendar.getInstance());
		message.computerName = this.hostName;
		// IPAhX
		message.ipAddress = this.address.getAddress().getHostAddress();

		message.osUserName = this.loginName;
		message.port = this.address.getPort();
		//bZ[Wɂ̓jbNl[̃f[^ĂȂI
		// jbNl[͌Őݒ
		message.userName = this.loginName;

		long option = getOption();

		// {̕
		String msg;
		if ((option & IPMSG.ENCRYPTOPT) != 0) {

			long start = System.currentTimeMillis();
			msg = messenger.decryptMessage(this.data, getEncodingForUTF8OPT());
			if (logger.isInfoEnabled())
				logger.info("Decryption=" //$NON-NLS-1$
						+ (System.currentTimeMillis() - start));

			// Í̎
			long cap = IPMCipher.getCapability(this.data);

			// ÍASY̖O擾
			//String cipher_name;
			if (cap == 0) {
				message.encryptMethod = IPMSG.CIPHER_PLAIN;
			} else if ((cap & (IPMSG.BLOWFISH_128 | IPMSG.BLOWFISH_256)) == 0) {
				message.encryptMethod = IPMSG.CIPHER_RC2;
			} else {
				message.encryptMethod = IPMSG.CIPHER_BLOWFISH;
			}
		} else {
			msg = this.data;
		}

		// CC摜̃v[Xz_}
		if (message.hasAttachment()) {
			Collection atts = message.getInlineAttachments(true);
			if (atts != null && !atts.isEmpty()) {
				StringBuffer sb = new StringBuffer(msg);
				Iterator itr = atts.iterator();
				while (itr.hasNext()) {
					Attachment att = (Attachment) itr.next();
					if (att.position < 0 || att.position >= sb.length()) {
						sb.append('\uFFFC');
						att.position = sb.length() - 1;
					} else {
						sb.insert(att.position, '\uFFFC');
					}
				}
				msg = sb.toString();
			}
		}

		// sR[hvbgtH[Ƃɕϊ
		if (EOL.equals("\r\n")) { //$NON-NLS-1$
			message.message = Tool.replaceAll(msg, "\n", "\r\n"); //$NON-NLS-1$ //$NON-NLS-2$
		} else if (EOL.equals("\r")) { //$NON-NLS-1$
			message.message = msg.replace('\n', '\r');
		} else {
			message.message = msg;
		}

		message.protocol = messenger.getProtocol().signature;
		message.launchId = messenger.getId();

		if ((option & IPMSG.SECRETOPT) != 0) {
			// 
			message.initialSealed();
			if ((option & IPMSG.PASSWORDOPT) != 0) {
				message.isPasswordRequired = true;
			}
		} else {
			// ʏ̃bZ[W
			message.initialOpened();
		}

		// }`LXg
		if ((option & IPMSG.MULTICASTOPT) != 0
				|| (option & IPMSG.BROADCASTOPT) != 0) {
			message.isMulticast = true;
		}

		return message;
	}

	/**
	 * o[̎ʎq擾܂B
	 * (OCzXg)
	 * @return `̃o[ʎq
	 */
	String getIdentifierLong() {
		StringBuffer sb = new StringBuffer(this.loginName);
		String ipAddress = address.getAddress().getHostAddress();
		sb.append("@").append(ipAddress); //$NON-NLS-1$
		return sb.toString().toUpperCase();
	}

	/**
	 * Mf[^쐬܂B
	 * R}hApPbgIDAf[^Ag͂炩ߐݒ肵ĂƁB
	 * @return oCȋMf[^
	 */
	byte[] getBytes() {
		StringBuffer sb = new StringBuffer();
		sb.append(version).append(':');
		sb.append(id).append(':');

		RuntimeProperties rp = messenger.getAppPreferences().getRuntime();
		sb.append(rp.getLoginNameEscaped()).append(':');
		sb.append(InetAddressUtils.getHostNameEscaped()).append(':');

		sb.append(command).append(':');
		// {܂̓jbNl[܂̓_E[hvf[^
		sb.append(data);
		if (command != IPMSG.GETFILEDATA && command != IPMSG.GETDIRFILES) {
			// TCP̂Ƃ\0sv
			sb.append('\0');
		}
		// O[v܂̓t@CYt
		if (extension != null && extension.length() != 0) {
			assert (command & IPMSG.BR_ABSENCE) != 0
					|| (command & IPMSG.BR_ENTRY) != 0
					|| (command & IPMSG.ANSENTRY) != 0
					|| (command & IPMSG.SENDMSG) != 0;
			sb.append(extension).append('\0');
		}
		// byte[]ɕϊ
		try {
			byte[] bytes = sb.toString().getBytes(getEncodingForUTF8OPT());
			if (!isOrganizationCommand() || hasUtf8Option()) {
				return bytes;
			}

			StringBuffer utf8 = new StringBuffer();
			utf8.append('\n').append("UN:").append(rp.getLoginNameEscaped()); //$NON-NLS-1$
			utf8.append('\n').append("HN:").append( //$NON-NLS-1$
					InetAddressUtils.getHostNameEscaped());
			utf8.append('\n').append("NN:").append(data); //$NON-NLS-1$

			utf8.append('\n').append("GN:"); //$NON-NLS-1$
			if (extension != null && extension.length() != 0) {
				utf8.append(extension);
			}
			utf8.append('\n');
			byte[] utf8Bytes = utf8.toString().getBytes(
					IPMessenger.ENCODING_UTF8);
			byte[] result = new byte[bytes.length + utf8Bytes.length];
			System.arraycopy(bytes, 0, result, 0, bytes.length);
			System.arraycopy(utf8Bytes, 0, result, bytes.length,
					utf8Bytes.length);
			return result;
		} catch (UnsupportedEncodingException e) {
			logger.error("Cannot encode to byte array.", e); //$NON-NLS-1$
			return null;
		}
	}

	/**
	 * oCi\tB[h𕜌܂B
	 * @param buffer oCif[^
	 * @throws InvalidDataFormatException
	 */
	private void unpack(byte[] buffer) throws InvalidDataFormatException {

		if (buffer == null) {
			throw new InvalidDataFormatException("Target data is null."); //$NON-NLS-1$
		}

		// o[W
		int destIndex = Tool.indexOf(buffer, 0, ':');

		String str;
		if (destIndex == 1) {
			this.version = buffer[0] - '0';
		}

		if (this.version != 1) {
			logger.info("Unsupported packet version. version=" + version); //$NON-NLS-1$
			return;
		}

		int srcIndex = destIndex + 1;
		if (buffer.length <= srcIndex) {
			throw new InvalidDataFormatException(TOO_SHORT, buffer);
		}

		// pPbgID
		destIndex = Tool.indexOf(buffer, srcIndex, ':');
		str = new String(buffer, srcIndex, destIndex - srcIndex);
		try {
			this.id = Long.parseLong(str);
		} catch (NumberFormatException e) {
			throw new InvalidDataFormatException("Invalid packet id.", //$NON-NLS-1$
					e, buffer);
		}

		srcIndex = destIndex + 1;
		if (buffer.length <= srcIndex) {
			throw new InvalidDataFormatException(TOO_SHORT, buffer);
		}

		// OC
		destIndex = Tool.indexOf(buffer, srcIndex, ':');
		int loginSrc = srcIndex;
		int loginLen = destIndex - srcIndex;

		srcIndex = destIndex + 1;
		if (buffer.length <= srcIndex) {
			throw new InvalidDataFormatException(TOO_SHORT, buffer);
		}

		// zXg
		destIndex = Tool.indexOf(buffer, srcIndex, ':');
		int hostSrc = srcIndex;
		int hostLen = destIndex - srcIndex;

		srcIndex = destIndex + 1;
		if (buffer.length <= srcIndex) {
			throw new InvalidDataFormatException(TOO_SHORT, buffer);
		}

		// R}h
		destIndex = Tool.indexOf(buffer, srcIndex, ':');
		str = new String(buffer, srcIndex, destIndex - srcIndex);
		try {
			this.command = Long.parseLong(str);
		} catch (NumberFormatException e) {
			throw new InvalidDataFormatException("Invalid command.", //$NON-NLS-1$
					e, buffer);
		}

		// GR[fBO擾
		String encoding = getEncodingForUTF8OPT();

		// OCƃzXg𕶎
		try {
			this.loginName = new String(buffer, loginSrc, loginLen, encoding);
			this.hostName = new String(buffer, hostSrc, hostLen, encoding);
			this.hostName = this.hostName.toUpperCase();
		} catch (UnsupportedEncodingException e) {
			throw new InvalidDataFormatException(
					"Cannot get login name or host name.", //$NON-NLS-1$
					e, buffer);
		}

		srcIndex = destIndex + 1;
		if (buffer.length <= srcIndex) {
			// o[W擾Ȃ
			return;
		}

		// ǉ(g)([UEbZ[W{)
		destIndex = Tool.indexOf(buffer, srcIndex, '\0');
		try {
			this.data = new String(buffer, srcIndex, destIndex - srcIndex,
					encoding);
		} catch (UnsupportedEncodingException e) {
			throw new InvalidDataFormatException("Cannot get data part.", //$NON-NLS-1$
					e, buffer);
		}

		// [Uݒ肳ĂȂꍇɃOCŕ₤
		if (isOrganizationCommand()
				&& (this.data == null || this.data.length() == 0)) {
			this.data = this.loginName;
		}

		srcIndex = destIndex + 1;
		if (buffer.length <= srcIndex) {
			return;
		}

		if (buffer[srcIndex] != '\n') {
			// O[vEYt
			destIndex = Tool.indexOf(buffer, srcIndex, '\0');
			try {
				this.extension = new String(buffer, srcIndex, destIndex
						- srcIndex, encoding);
			} catch (UnsupportedEncodingException e) {
				throw new InvalidDataFormatException(
						"Cannot decode extention part.", //$NON-NLS-1$
						e, buffer);
			}
			srcIndex = destIndex + 1;
			if (buffer.length <= srcIndex + 1/*\n*/) {
				return;
			}
		}

		// UTF-8̃OCEzXgEjbNl[EO[v
		try {
			if (Tool.hasFlag(this.command, IPMSG.CAPUTF8OPT)
					&& this.isOrganizationCommand() && buffer[srcIndex] == '\n') {
				// ŏ\nXLbv
				srcIndex++;
				while (true) {
					destIndex = Tool.indexOf(buffer, srcIndex, ':');
					if (buffer.length == destIndex) {
						throw new InvalidDataFormatException(
								"Invalid UTF-8 part: premature end of key.", //$NON-NLS-1$
								buffer);
					}
					String key = new String(buffer, srcIndex, destIndex
							- srcIndex, IPMessenger.ENCODING_UTF8);

					srcIndex = destIndex + 1;
					if (buffer.length <= srcIndex) {
						// O[vȂǒlݒ肵ĂȂꍇ
						break;
					}
					destIndex = Tool.indexOf(buffer, srcIndex, '\n');
					String val = new String(buffer, srcIndex, destIndex
							- srcIndex, IPMessenger.ENCODING_UTF8);
					if (key.equals("UN")) { //$NON-NLS-1$
						this.loginName = val;
					} else if (key.equals("HN")) { //$NON-NLS-1$
						this.hostName = val;
					} else if (key.equals("NN")) { //$NON-NLS-1$
						this.data = val;
					} else if (key.equals("GN")) { //$NON-NLS-1$
						this.extension = val;
					}
					if (buffer.length == destIndex) {
						break;
					}
					srcIndex = destIndex + 1;
					if (srcIndex == buffer.length) {
						break;
					}
					if (buffer[srcIndex] == '\0'
							&& buffer.length - 1 == srcIndex) {
						// Ō\0tĂꍇɑΉ(IP Messenger for Mac OS X v0.11.1)
						break;
					}
				}
			}
		} catch (UnsupportedEncodingException e) {
			throw new InvalidDataFormatException("UTF-8", e, buffer); //$NON-NLS-1$
		}
	}

	/**
	 * UTF8OPTtOɃR}hGR[fBO擾܂B
	 * @return GR[fBOBWindows-31J܂UTF-8B
	 */
	String getEncodingForUTF8OPT() {
		return IPMessenger.getEncoding(hasUtf8Option());
	}

	/**
	 * UTF8OPTtOɃR}hGR[fBO擾܂B
	 * @return GR[fBOBWindows-31J܂UTF-8B
	 */
	String getEncodingForCAPUTF8OPT() {
		return IPMessenger.getEncoding(isUTF8Supported());
	}

	/**
	 * Ytt@C͂AArrayListŕԂ܂B(UDPM)
	 * @param info IPMPacket.extension
	 * @return AttachmentArrayList
	 */
	private ArrayList unpackAttachmentInfo(String info) {
		if (info == null) {
			return null;
		}
		ArrayList list = new ArrayList();
		ArrayList elements = new ArrayList();

		int i = 0;
		// Attachment
		//ATTACHMENT: 
		while (i < info.length()) {
			if (info.charAt(i) == ':') {
				// eAttachment̊Ԃ:
				i++;
			}

			// vf (\7oĂ܂œǂ)
			ELEMENT: while (i < info.length()) {
				//  (:oĂ܂œǂݍ)
				StringBuffer sb = new StringBuffer();
				//CHARACTER: 
				while (i < info.length()) {
					char c = info.charAt(i);
					if (c == ':') {
						// t@C:GXP[v
						if (elements.size() == 1 && i + 1 < info.length()
								&& info.charAt(i + 1) == ':') {
							// ̕:̂Ƃ:obt@ɒǉ
							// 2ڂ̃Řɐi߂
							i += 2;
						} else {
							// ̗vf
							i++;
							break;
						}
					} else if (c == '\7') {
						// Attachment̏I
						i++;
						elements.add(sb.toString());
						break ELEMENT;
					}
					sb.append(c);
					i++;
				}
				// vfۑ
				elements.add(sb.toString());
			}
			if (elements.size() < 5) {
				logger.warn("Invalid attachment information."); //$NON-NLS-1$
				return null;
			}

			Attachment att = new Attachment();
			// vfAttachmentɑ
			att.name = (String) elements.get(1);
			// t@C𐳂
			att.validateName();
			long attr = 0;
			try {
				att.id = Integer.parseInt((String) elements.get(0), 10);
				att.size = Long.parseLong((String) elements.get(2), 16);
				att.lastModified = Long.parseLong((String) elements.get(3), 16);
				attr = Long.parseLong((String) elements.get(4), 16);
			} catch (NumberFormatException e) {
				logger.warn("Invalid attachment information.", e); //$NON-NLS-1$
				return null;
			}

			// 
			if ((attr & IPMSG.FILE_REGULAR) != 0) {
				att.type = Attachment.TYPE_FILE;
			} else if ((attr & IPMSG.FILE_DIR) != 0) {
				att.type = Attachment.TYPE_DIR;
			} else if ((attr & IPMSG.FILE_CLIPBOARD) != 0) {
				att.type = Attachment.TYPE_INLINE;
			} else {
				att.type = Attachment.TYPE_OTHER;
			}
			// t@C
			if ((attr & IPMSG.FILE_RONLYOPT) != 0) {
				att.isReadOnly = true;
			}
			if ((attr & IPMSG.FILE_HIDDENOPT) != 0) {
				att.isHidden = true;
			}
			// g
			for (int p = 5; p < elements.size(); p++) {
				String str = (String) elements.get(p);
				if (str.startsWith(POS)) {
					try {
						String val = str.substring(POS.length());
						att.position = Integer.parseInt(val, 16);
					} catch (NumberFormatException e) {
						logger.warn("Invalid inline image position.", e); //$NON-NLS-1$
						return null;
					}
				}
			}
			elements.clear();
			list.add(att);
		}
		return list;
	}

	/**
	 * t@CYt܂B
	 * @param msg bZ[W(Ytt@C)
	 * @param spec M
	 */
	void setAttachments(Message msg, IPMPacket spec) {
		if (!spec.isAttachmentSupported()) {
			return;
		}

		if (!msg.hasAttachment()) {
			return;
		}
		final boolean inline = spec.isInlineImageSupported();
		// Ytt@C𕶎
		StringBuffer sb = new StringBuffer();
		Iterator itr = msg.attachments.iterator();
		while (itr.hasNext()) {
			Attachment att = (Attachment) itr.next();
			// ŐVɂ
			// TODO MO1񂾂ɂ
			att.getFileInfo();

			if (att.type == Attachment.TYPE_OTHER) {
				continue;
			}

			// ID (M10i)
			sb.append(att.id).append(':');
			// t@C(:GXP[v)
			StringBuffer fileName = new StringBuffer(att.name);
			Tool.replaceAll(fileName, ":", "::"); //$NON-NLS-1$//$NON-NLS-2$
			sb.append(fileName).append(':');
			// TCY
			sb.append(Long.toHexString(att.size)).append(':');
			// ŏIXV
			sb.append(Long.toHexString(att.lastModified / 1000)).append(':');
			// t@C
			long attribute = 0;
			// 
			switch (att.type) {
			case Attachment.TYPE_FILE:
				attribute |= IPMSG.FILE_REGULAR;
				break;
			case Attachment.TYPE_INLINE:
				attribute |= inline ? IPMSG.FILE_CLIPBOARD : IPMSG.FILE_REGULAR;
				break;
			case Attachment.TYPE_DIR:
				attribute |= IPMSG.FILE_DIR;
				break;
			}
			// 
			if (att.isReadOnly) {
				attribute |= IPMSG.FILE_RONLYOPT;
			}
			if (att.isHidden) {
				attribute |= IPMSG.FILE_HIDDENOPT;
			}
			sb.append(Long.toHexString(attribute));

			// g
			if (inline && att.type == Attachment.TYPE_INLINE
					&& att.position >= 0) {
				sb.append(":8=").append(Integer.toHexString(att.position)); //$NON-NLS-1$
			}
			// t@CA؂Ɩ[
			sb.append(":\7"); //$NON-NLS-1$
			if (itr.hasNext()) {
				sb.append(':');
			}
		}
		this.extension = sb.toString();
	}

	/**
	 * pPbgBʒmKvȃpPbgł邩ǂ擾܂B
	 * @return Kvȏꍇtrue
	 */
	boolean isResponseRequired() {
		long packetClass = getPacketClass();
		long option = getOption();

		/*
		 * u[hLXgIvVAopPbgɂ
		 * BʒmȂ
		 */
		if (Tool.hasFlag(option, IPMSG.BROADCASTOPT)
				|| Tool.hasFlag(option, IPMSG.AUTORETOPT)) {
			return false;
		}

		/*
		 * *CHECKOPTĂꍇɌA
		 * Bʒm
		 */
		if (packetClass == IPMSG.SENDMSG) {
			return Tool.hasFlag(option, IPMSG.SENDCHECKOPT);
		} else if (packetClass == IPMSG.READMSG) {
			return Tool.hasFlag(option, IPMSG.READCHECKOPT);
		}

		return false;
	}

	/**
	 * pPbgʒlԂ܂B
	 * @return pPbgʒl
	 */
	Long getPacketValue() {
		long result = InetAddressUtils.getAddressValue(address.getAddress());
		result += (address.getPort() << 16);
		result += this.id;
		return new Long(result);
	}
}
