package petaclone.ipmsg;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.BindException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.ByteList;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Tool;
import petaclone.Attachment;
import petaclone.InvalidDataFormatException;

/**
 * IP MessengerTCPT[o
 * @author Takayuki Shoji
 */
public class IPMTCPServer implements Runnable {

	/**
	 * logger
	 */
	private static final Logger logger = Logger.getLogger(IPMTCPServer.class);

	/** fobOtO */
	private static final boolean DEBUG = logger.isDebugEnabled();

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/** bZW[ */
	private IPMessenger messenger;

	/** T[õXCb` */
	private boolean isAlive;

	/** \Pbg */
	private ServerSocket server;

	/**
	 * t@CM<br>
	 * L[: packetID (String) 16iŁB
	 * l: AttachmentList
	 */
	private HashMap permission = new HashMap();

	/**
	 * Unicodet@C𑗐M邩ǂ
	 */
	private final boolean sendUnicodeFilename;

	/**
	 * Xbhv[<br>
	 * l: Server
	 */
	private ThreadPool pool;

	/**
	 * RXgN^
	 * @param messenger bZW[
	 */
	IPMTCPServer(IPMessenger messenger) {
		this.messenger = messenger;
		pool = new ThreadPool();
		sendUnicodeFilename = messenger.isOptionSendUnicodeFilename();
	}

	/**
	 * T[o
	 */
	public void run() {

		try {
			while (isAlive) {
				try {
					Socket socket = server.accept();
					startHandler(socket);
					continue;
				} catch (IOException e) {
					if (isAlive) {
						logger
								.error(
										"IO exception occurred during waiting.", e); //$NON-NLS-1$
						// TODO ʒm`ɂ
						messenger.notifyError(//
								mm.get("IPMTCPServer.ioExceptionOccurred"), //$NON-NLS-1$
								e);
						continue;
					}
					break;
				}
			}
		} finally {
			close();
		}
	}

	/**
	 * oCh܂B
	 * @param bindAddress oChAhX(null̂ƂAׂẴ[JAhX)
	 * @throws IOException
	 * @throws SocketException
	 * @throws BindException |[ggp܂̓[JAhX݂Ȃ
	 */
	void bind(InetAddress bindAddress) throws BindException, SocketException,
			IOException {
		server = new ServerSocket(messenger.port, -1, bindAddress);
		isAlive = true;
		logger.info("bind: addr=" + bindAddress + ", port=" + messenger.port); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * T[o\PbgN[Y܂B
	 */
	private synchronized void close() {
		if (server != null && !server.isClosed()) {
			try {
				isAlive = false;
				server.close();
			} catch (IOException e) {
				logger.error("Failed to close ServerSocket.", e); //$NON-NLS-1$
			}
		}
	}

	/**
	 * nhɏs킹܂B
	 * @param socket \Pbg
	 */
	private void startHandler(Socket socket) {
		// T[o擾
		Server server = pool.getServer();
		// s킹
		server.handleSocket(socket);
	}

	/**
	 * T[oĂ邩𒲂ׂ܂B
	 * @return Ăꍇtrue
	 */
	synchronized boolean getAlive() {
		//System.out.println("getAlive");
		return isAlive;
	}

	/**
	 * T[oɏIʒm܂B
	 */
	void quit() {
		isAlive = false;
		close();
		pool.quitServers();
	}

	/**
	 * _E[h܂B
	 * (UDPT[ogp)
	 * @param packetId pPbgID
	 * @param attachments Ytt@C̃Xg(l: Attachment)B
	 * List͕ύX邽߁AMessage.attachmentsƂ͕ʂ̃CX^Xw肷KvB
	 */
	void addPermission(long packetId, List attachments) {
		synchronized (permission) {
			// o^
			permission.put(Long.toHexString(packetId), attachments);
			if (DEBUG) {
				logger
						.debug("Adding download permission. packetId=" + packetId); //$NON-NLS-1$
				Iterator itr = attachments.iterator();
				while (itr.hasNext()) {
					Attachment att = (Attachment) itr.next();
					logger.debug(att.toString());
				}
			}
		}
	}

	/**
	 * _E[h폜܂B
	 * (UDPT[ogp)
	 * @param packetId pPbgID
	 */
	void removePermission(long packetId) {
		synchronized (permission) {
			permission.remove(Long.toHexString(packetId));
			if (DEBUG)
				logger.debug("Removing download permission:" + packetId); //$NON-NLS-1$
		}
	}

	/**
	 * _E[h폜܂B
	 * (Servergp)
	 * @param req Ytt@Cv
	 */
	private void removePermission(AttachmentRequest req) {
		if (DEBUG)
			logger.debug("Removing download permission: packetId=" //$NON-NLS-1$
					+ req.packId + ",fileId=" + req.fileId); //$NON-NLS-1$

		List attachments;
		synchronized (permission) {
			attachments = (List) permission.get(Long.toHexString(req.packId));
			if (attachments == null) {
				return;
			}
			if (attachments.size() <= 1) {
				removePermission(req.packId);
				return;
			}
		}
		synchronized (attachments) {
			// fileIDƍ
			Iterator itr = attachments.iterator();
			while (itr.hasNext()) {
				Attachment att = (Attachment) itr.next();
				if (att.id == req.fileId) {
					attachments.remove(att);
					return;
				}
			}
		}
	}

	/**
	 * pPbgԍƃt@CIDAttachment擾܂B
	 * @param req Ytt@Cv
	 * @return ĂȂꍇnull
	 */
	private Attachment getAttachment(AttachmentRequest req) {
		List attachments;
		synchronized (permission) {
			attachments = (List) permission.get(Long.toHexString(req.packId));
			if (attachments == null) {
				return null;
			}
		}
		synchronized (attachments) {
			// fileIdƍ
			Iterator itr = attachments.iterator();
			while (itr.hasNext()) {
				Attachment att = (Attachment) itr.next();
				if (att.id == req.fileId) {
					return att;
				}
			}
		}
		return null;
	}

	/**
	 * Server̃Xbhv[
	 */
	private class ThreadPool {
		/**
		 * ҂Xbh̃X^bN
		 * l: Server
		 */
		private final LinkedList threads = new LinkedList();

		/** Xbh̃JE^ */
		private int count;

		/**
		 * Server擾܂B
		 * ҂T[oꍇ͂̃T[oԂ܂B
		 * ΐVĕԂ܂B
		 * @return Server
		 */
		synchronized Server getServer() {
			if (threads.isEmpty()) {
				if (DEBUG)
					logger.debug("new Server()"); //$NON-NLS-1$
				count++;
				Server server = new Server(count);
				synchronized (server) {
					//FIXME 
					server.start();
					try {
						// server̋N҂
						server.wait();
					} catch (InterruptedException e) {
						logger.error(e);
					}
				}
				return server;
			}
			if (DEBUG)
				logger.debug("threads.removeLast()"); //$NON-NLS-1$
			return (Server) threads.removeLast();
		}

		/**
		 * v[ɖ߂܂B
		 * @param server T[o
		 */
		synchronized void addServer(Server server) {
			threads.addLast(server);
		}

		/**
		 * v[폜܂B
		 * @param server 폜T[o
		 */
		synchronized void removeServer(Server server) {
			threads.remove(server);
		}

		/**
		 * T[oI܂B
		 */
		synchronized void quitServers() {
			Iterator itr = threads.iterator();
			while (itr.hasNext()) {
				Server server = (Server) itr.next();
				synchronized (server) {
					server.notifyAll();
				}
			}
		}
	}

	/**
	 * Ytt@Cv
	 */
	private static class AttachmentRequest {
		/**
		 * pPbgID
		 */
		long packId;

		/**
		 * t@CID
		 */
		int fileId;

		/**
		 * ItZbg
		 */
		long offset;

		/**
		 * wb_̃GR[fBO
		 */
		String encoding;

		/**
		 * RXgN^
		 * @param data pPbgf[^̓Ytt@Cv
		 * @param encoding Mwb_̃GR[fBO
		 */
		AttachmentRequest(String data, String encoding) {
			this.encoding = encoding;
			parseAttachmentRequest(data);
		}

		/**
		 * t@Cv͂܂B
		 * @param data pPbgf[^̓Ytt@Cv
		 */
		private void parseAttachmentRequest(String data) {
			String[] info = Tool.split(data, ':', 0);
			if (info.length < 2) {
				throw new InvalidDataFormatException(
						"Invalid attachment request."); //$NON-NLS-1$
			}
			try {
				packId = Long.parseLong(info[0], 16);
				fileId = Integer.parseInt(info[1], 16);

				// ItZbg
				offset = 0;
				if (info.length >= 3 && info[2].length() > 0) {
					// fBNg_E[ĥƂ̓ItZbg͂Ȃ
					offset = Long.parseLong(info[2], 16);
				}
			} catch (NumberFormatException e) {
				throw new InvalidDataFormatException(
						"Invalid number format in attachment request.", e, data); //$NON-NLS-1$
			}
		}

		/**
		 * CX^X̏Ԃ𕶎ɂ܂B
		 * @see java.lang.Object#toString()
		 */
		public String toString() {
			StringBuffer sb = new StringBuffer("AttachmentRequest@"); //$NON-NLS-1$
			sb.append(hashCode());
			sb.append('{');
			sb.append("packId=").append(packId); //$NON-NLS-1$
			sb.append(", fileId=").append(fileId); //$NON-NLS-1$
			sb.append(", offset=").append(offset); //$NON-NLS-1$
			sb.append(", encoding=").append(encoding); //$NON-NLS-1$
			sb.append('}');
			return sb.toString();
		}
	}

	/**
	 * t@C𑗐MXbh
	 */
	private class Server extends Thread {

		/** \Pbg */
		private Socket socket;

		/** ʒmꂽ */
		private boolean notified;

		/**
		 * o̓obt@
		 */
		private byte[] buffer;

		/**
		 * RXgN^ 
		 * @param count XbhɂID
		 */
		Server(int count) {
			super(messenger.getThreadGroup(), "IPMTCPServer$Server" + count); //$NON-NLS-1$
			this.setDaemon(true);
		}

		/**
		 * Xbh
		 */
		public void run() {
			synchronized (this) {
				while (true) {
					try {
						if (DEBUG)
							logger.debug("IPMTCPServer.Server.wait() start" //$NON-NLS-1$
									+ this.getName());
						notifyAll();
						// bN
						this.wait(5 * 60 * 1000); //Ƃ肠5
						if (DEBUG)
							logger.debug("IPMTCPServer.Server.wait() end" //$NON-NLS-1$
									+ this.getName());
						// handleSocket()Ă΂Ƃ甲o
					} catch (InterruptedException e) {
						logger.error(e);
						expire();
						break;
					}
					if (!notified) {
						// ^CAEg
						expire();
						break;
					}
					// ߂Ă
					notified = false;
					// \Pbgf[^ǂݍ݁A
					serve();
					// Xbhv[ɓ
					waiting();
				}
			}
		}

		/**
		 * Xbhv[ɓ܂B
		 */
		private void waiting() {
			pool.addServer(this);
			logger.debug("Added thread to pool."); //$NON-NLS-1$
		}

		/**
		 * Xbhv[玩폜܂B
		 */
		private void expire() {
			pool.removeServer(this);
			logger.debug("Removed thread from pool."); //$NON-NLS-1$
		}

		/**
		 * w肵\Pbg̏Jn܂B
		 * @param socket \Pbg
		 */
		synchronized void handleSocket(Socket socket) {
			assert socket != null;
			this.socket = socket;
			this.notified = true;
			this.notify();
		}

		/**
		 * \Pbgf[^ǂŉ͂A
		 * t@CNCAgɑM܂B
		 * <p>
		 * Ƀ\bhŃ\Pbg܂B
		 */
		private void serve() {
			InputStream is = null;
			try {
				// v̓ǂݍ
				is = socket.getInputStream();

				ByteList list = new ByteList();
				byte[] rBuffer = new byte[64];

				while (true) {
					int readSize = is.read(rBuffer);
					if (readSize == -1) {
						// EOF
						break;
					}
					list.add(rBuffer, readSize);
					if (list.size() > 1024 * 8) {
						logger.warn("Input data is too big."); //$NON-NLS-1$
						return;
					}
					/*
					 * NCAgłEOF𑗂Ȃ߁A
					 * ǂƂƃbNĂ܂B
					 * Oɒ킩Ȃ̂ŁAobt@
					 * ̏ꍇ͖B
					 * NGXg͏̂ŁAא؂ɂȂ邱Ƃ
					 * ȂƎvB
					 */
					if (is.available() == 0) {
						break;
					}
				}

				socket.shutdownInput();

				// f[^̉
				byte[] result = list.toArray();
				IPMPacket pack = new IPMPacket(messenger, result, null);

				if (pack.version == 1) {
					AttachmentRequest req = new AttachmentRequest(pack.data,
							pack.getEncodingForUTF8OPT());
					// ]
					sendAttachment(req);
				}

			} catch (InvalidDataFormatException e) {
				logger.error("Invalid received data.", e); //$NON-NLS-1$
			} catch (IOException e) {
				logger.error("IO exception has occurred.", e); //$NON-NLS-1$
			} finally {
				try {
					if (is != null)
						is.close();
				} catch (IOException e) {
					logger.error("Failed to close input stream.", e); //$NON-NLS-1$
				}
				try {
					socket.close();
				} catch (IOException e) {
					logger.error("Failed to close socket.", e); //$NON-NLS-1$
				}
				socket = null;
			}
		}

		/**
		 * Ytt@Co͂܂B
		 * @param req Ytt@Cv
		 * @throws IOException o͗O
		 */
		private void sendAttachment(AttachmentRequest req) throws IOException {

			// Ă邩
			Attachment att = getAttachment(req);
			if (att == null) {
				logger.info("download request: refuse" + req); //$NON-NLS-1$
				return;
			}
			logger.info("download request: allow" + req); //$NON-NLS-1$

			this.buffer = new byte[1024 * 8];
			if (att.type == Attachment.TYPE_FILE
					|| att.type == Attachment.TYPE_INLINE) {
				// t@C
				sendFile(att, req.offset);
			} else {
				// fBNg
				sendDirectory(att, req.encoding);
			}
			this.buffer = null;
			// ̍폜
			removePermission(req);
			logger.info("sendAttachment() end."); //$NON-NLS-1$
		}

		/**
		 * fBNgƃt@Co͂܂B
		 * (fBNgM)
		 * @param att Ytt@C
		 * @param encoding wb_̃GR[fBO
		 * @throws IOException o͗O
		 */
		private void sendDirectory(Attachment att, String encoding)
				throws IOException {
			File root = att.getFile();
			if (!root.isDirectory()) {
				// fBNgȊO͂f
				logger
						.warn("Attachment.localPath must be a directory when its type is directory."); //$NON-NLS-1$
				return;
			}
			OutputStream ss = null;
			BufferedOutputStream os = null;
			try {
				// o̓Xg[𓾂
				ss = socket.getOutputStream();
				os = new BufferedOutputStream(ss);

				// fBNgƃt@Co
				LinkedList stack = new LinkedList();
				stack.addLast(new Entry(root, false));
				// stack
				while (!stack.isEmpty()) {
					// t@CɏقXgZςނƂ肠pop
					Entry entry = (Entry) stack.removeLast();

					long size;
					if (entry.file.isDirectory()) {
						size = 0;
					} else {
						size = entry.file.length();
					}
					long lastModified = entry.file.lastModified() / 1000;

					// wb_
					byte[] buffer = makeHeader(entry.file,
							entry.isReturnParent, size, lastModified, encoding);
					if (buffer == null) {
						continue;
					}
					// wb_f[^M
					os.write(buffer);

					if (entry.isReturnParent)
						continue;
					if (entry.file.isDirectory()) {
						// RETPARENTpush
						entry.isReturnParent = true;
						stack.addLast(entry);
						// ɓăXg𓾂Bpush
						File[] files = entry.file.listFiles();
						for (int i = 0; i < files.length; i++) {
							stack.addLast(new Entry(files[i], false));
						}
					} else if (entry.file.isFile()) {
						// f[^𑗐M
						sendFileData(os, entry.file, 0, size);
					} else {
						// ԂȂ
						continue;
					}
				}
			} finally {
				if (os != null) {
					os.close();
				} else if (ss != null) {
					ss.close();
				}
			}
		}

		/**
		 * w肳ꂽt@Co͂܂B
		 * (1t@CM)
		 * @param att Ytt@C
		 * @param offset ItZbg
		 * @throws IOException o͗O
		 */
		private void sendFile(Attachment att, long offset) throws IOException {
			OutputStream ss = null;
			OutputStream os = null;
			try {
				if (DEBUG)
					logger.debug("Creating output stream."); //$NON-NLS-1$

				ss = socket.getOutputStream();
				os = new BufferedOutputStream(ss);

				if (DEBUG)
					logger.debug("sending file."); //$NON-NLS-1$
				sendFileData(os, new File(att.localPath), offset, -1);

			} finally {
				if (os != null) {
					os.close();
				} else if (ss != null) {
					ss.close();
				}
			}
		}

		/**
		 * w肳ꂽo̓Xg[Ɏw肳ꂽItZbg
		 * t@Cf[^o͂܂B
		 * @param os o̓Xg[
		 * @param file t@C
		 * @param offset ItZbg
		 * @param maxLength o͂TCY(ItZbg0琔ꍇ)Bۂ̃TCYȂꍇ0x00Ŗ߂B-1̏ꍇ͓ǂ߂邾o͂B
		 * @throws IOException o͗O
		 */
		private void sendFileData(OutputStream os, File file, long offset,
				long maxLength) throws IOException {
			BufferedInputStream is = null;
			FileInputStream fin = null;

			try {
				fin = new FileInputStream(file);
				is = new BufferedInputStream(fin);

				// XLbv
				if (offset != 0) {
					long skip = offset;
					while (skip > 0) {
						long skipped = is.skip(skip);
						if (skipped <= 0) {
							break;
						}
						skip -= skipped;
					}
				}

				if (maxLength == -1) {
					// őTCYlȂꍇ
					int n;
					while ((n = is.read(buffer)) != -1) {
						os.write(buffer, 0, n);
					}
				} else {
					// őTCYlꍇ
					long remain = maxLength - offset;
					int n;
					while (remain > 0) {
						n = is.read(buffer, 0, (int) Math.min(remain,
								buffer.length));
						if (n == -1) {
							break;
						}
						os.write(buffer, 0, n);
						remain -= n;
					}

					if (remain > 0) {
						// Ă܂0Ŗ߂
						Arrays.fill(buffer, (byte) 0x00);
						while (remain > 0) {
							os.write(buffer, 0, (int) Math.min(remain,
									buffer.length));
							remain -= buffer.length;
						}
					}
				}

			} finally {
				if (is != null) {
					is.close();
				} else if (fin != null) {
					fin.close();
				}
			}
		}

		/**
		 * tH_Mp̃wb_쐬܂B<br>
		 * MłȂꍇnullԂ܂B
		 * @param file t@C
		 * @param returnParent ̃fBNgɖ߂ꍇtrue
		 * @param size t@CTCYBfBNg̏ꍇ0B
		 * @param lastModified ŏIXVB
		 * @param encoding wb_̃GR[fBO
		 * @return MpoCif[^
		 */
		private byte[] makeHeader(File file, boolean returnParent, long size,
				long lastModified, String encoding) {
			// 
			long kind;
			if (returnParent) {
				kind = IPMSG.FILE_RETPARENT;
			} else if (file.isDirectory()) {
				kind = IPMSG.FILE_DIR;
			} else if (file.isFile()) {
				kind = IPMSG.FILE_REGULAR;
			} else {
				return null;
			}
			// 
			if (!file.canWrite()) {
				kind |= IPMSG.FILE_RONLYOPT;
			}
			if (file.isHidden()) {
				kind |= IPMSG.FILE_HIDDENOPT;
			}
			// t@C
			String fileName;
			if (returnParent) {
				fileName = "."; //$NON-NLS-1$
			} else {
				// t@C:::ɕϊ
				StringBuffer sb = new StringBuffer(file.getName());
				Tool.replaceAll(sb, ":", "::"); //$NON-NLS-1$ //$NON-NLS-2$
				fileName = Tool.normalize(sb);
			}
			// wb_̍쐬
			StringBuffer sb = new StringBuffer();
			sb.append(':').append(fileName).append(':');
			sb.append(Long.toHexString(size)).append(':');
			sb.append(Long.toHexString(kind)).append(':');
			sb.append(Long.toHexString(IPMSG.FILE_MTIME)).append('=');
			sb.append(Long.toHexString(lastModified)).append(':');

			// Unicodet@C
			if (!returnParent && sendUnicodeFilename) {
				sb.append(Long.toHexString(IPMSG.FILE_UNICODEFNAME))
						.append('=');
			}

			byte[] data = null;
			try {
				data = sb.toString().getBytes(encoding);

				if (!returnParent && sendUnicodeFilename) {
					ByteList buf = new ByteList(data.length + 256);
					buf.add(data);
					buf.add(fileName.getBytes(IPMessenger.ENCODING_UTF8));
					buf.add((byte) ':');
					data = buf.toArray();
				}
			} catch (UnsupportedEncodingException e) {
				logger.error("Cannot convert to byte array.", e); //$NON-NLS-1$
				return null;
			}

			/*
			 * ݊̂߂̕ύX:
			 * IP Messenger for Mac OS X 4ŒŃwb_
			 * 擾悤Ƃ邽߁A[l߂4ŏo͂B
			 */
			int length = data.length + 4;
			byte[] lData;
			try {
				String hex = "000" + Long.toHexString(length); //$NON-NLS-1$
				hex = hex.substring(hex.length() - 4, hex.length());
				lData = hex.getBytes(IPMessenger.ENCODING);
			} catch (UnsupportedEncodingException e) {
				logger.error("Cannot convert to byte array. ENCODING=" //$NON-NLS-1$
						+ IPMessenger.ENCODING, e);
				return null;
			}

			byte[] result = new byte[length];
			System.arraycopy(lData, 0, result, 0, lData.length);
			System.arraycopy(data, 0, result, lData.length, data.length);
			return result;
		}
	}

	/**
	 * t@CAfBNgȂǂ\܂B
	 */
	private static class Entry {
		/** 1ɖ߂ */
		boolean isReturnParent;

		/** Ώۂ̃t@C */
		File file;

		/**
		 * RXgN^
		 * @param file t@C
		 * @param isReturnParent 1ɖ߂Ȃtrue
		 */
		Entry(File file, boolean isReturnParent) {
			this.file = file;
			this.isReturnParent = isReturnParent;
		}
	}

}
