package petaclone.ipmsg;

import java.io.IOException;
import java.net.BindException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.MulticastSocket;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.security.interfaces.RSAPublicKey;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.InetAddressUtils;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Task;
import jp.takayuki.tools.Tool;
import jp.takayuki.tools.Version;
import petaclone.AppPreferences;
import petaclone.Attachment;
import petaclone.Destination;
import petaclone.InvalidDataFormatException;
import petaclone.Message;
import petaclone.RuntimeProperties;
import petaclone.SendingResultListener;
import petaclone.UniqueMember;

/**
 * IP MessengerUDPT[o
 * @author Takayuki Shoji
 */
public class IPMUDPServer implements Runnable {

	/**
	 * logger
	 */
	private static final Logger logger = Logger.getLogger(IPMUDPServer.class);

	/** fobOtO */
	private static final boolean DEBUG = logger.isDebugEnabled();

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/** T[õXCb` */
	private boolean isAlive;

	/** \Pbg */
	private DatagramSocket socket = null;

	/**
	 * pPbgBʒm҂<br>
	 * L[: SENDMSĜƂ̓pPbgԍAGETPUBKEŶƂ̓[U[ʎq<br>
	 * l: Sender
	 */
	private HashMap checkList = new HashMap();

	/** MXbh̍ő吔 */
	private final static int MAX_SEND_THREAD = 4;

	/**
	 * dpPbg邽߂̃Xg(queue)<br>
	 * l: pPbgԍ
	 */
	private LinkedList duplicationList = new LinkedList();

	/** bZW[C^[tF[X */
	private IPMessenger messenger;

	/**
	 * o[ǗpB<br>
	 * L[: IPMPacket#getIdentifierLong()<br>
	 * l: IPMPacket
	 */
	private HashMap members;

	/**
	 * _CAbv<br>
	 * L[: IPMPacket#getIdentifierLong()<br>
	 * l: InetSocketAddress
	 */
	private HashMap dialup;

	/**
	 * J<br>
	 * L[: IPMCipher#getIdentifierLong()<br>
	 * l: IPMCipher
	 */
	private HashMap publicKeys;

	/**
	 * ÍIvVB0܂IPMSG.ENCRYPTOPT
	 */
	private long encryptOption;

	/** bZ[W̍ő咷(UDP̍ő咷) */
	private static final int MAX_MESSAGE_LENGTH = 65516;

	/**
	 * ̌J̃LbV
	 */
	private String publicKeyCache;

	/** RSA 512 bit with RC2 40 bit[h */
	private boolean useSmallKey;

	/**
	 * RXgN^
	 * @param messenger bZW[
	 * @param encryptOption ÍIvV
	 */
	IPMUDPServer(IPMessenger messenger, long encryptOption) {
		this.messenger = messenger;
		this.encryptOption = encryptOption;

		this.members = new HashMap();
		this.dialup = new HashMap();
		this.publicKeys = new HashMap();

		this.useSmallKey = messenger.getAppPreferences().useSmallKey();

		// J̕LbV쐬
		if (!messenger.isOptionUsePlain()) {
			StringBuffer sb = new StringBuffer();
			long cap = this.messenger.getCipherCapability();
			sb.append(Long.toString(cap, 16).toUpperCase());
			sb.append(':');
			RSAPublicKey publicKey = (RSAPublicKey) this.messenger.publicKey;
			sb.append(IPMCipher.makeRSAPublicKeyString(publicKey));
			publicKeyCache = sb.toString();
		}
	}

	/**
	 * T[oĂ邩𒲂ׂ܂B
	 * @return Ătrue
	 */
	synchronized boolean getAlive() {
		return isAlive;
	}

	/**
	 * T[oI܂B
	 */
	void quit() {
		this.isAlive = false;
		close();
	}

	/**
	 * @see java.lang.Runnable#run()
	 */
	public void run() {

		// obt@m
		byte[] buffer = new byte[MAX_MESSAGE_LENGTH];

		DatagramPacket packet = new DatagramPacket(buffer, buffer.length);
		try {
			while (isAlive) {
				try {
					// pPbg邩^CAEg܂ŃubN
					socket.receive(packet);
				} catch (IOException e) {
					if (isAlive) {
						logger.error("IO exception occurred " //$NON-NLS-1$
								+ "while receiving or waiting.", //$NON-NLS-1$
								e);
						// TODO ʒm`
						messenger.notifyError(//
								mm.get("IPMUDPServer.ioExceptionOccurred"), e); //$NON-NLS-1$

						continue;
					}
					break;
				}

				// ̃u[hLXgɔȂ悤ɂ
				if (!messenger.configuration.acceptSelfBroadcast()
						&& InetAddressUtils.isLocalAddress(packet.getAddress())) {
					if (DEBUG)
						logger.debug("Packet from local address is blocked."); //$NON-NLS-1$
					continue;
				}

				// obt@̒𒲐߂AIPMPacket쐬
				byte[] buffer2 = new byte[packet.getLength()];
				System.arraycopy(buffer, 0, buffer2, 0, buffer2.length);

				// pPbg̑M
				InetSocketAddress source = new InetSocketAddress(packet
						.getAddress(), packet.getPort());

				if (DEBUG)
					logger.debug("Received: source=" + source + ",data=" //$NON-NLS-1$ //$NON-NLS-2$
							+ Tool.escapeControl(buffer2));

				// oCgNXɕϊ
				IPMPacket pack;
				try {
					pack = new IPMPacket(messenger, buffer2, source);
				} catch (InvalidDataFormatException e) {
					logger.warn("Invalid packet data.", e); //$NON-NLS-1$
					continue;
				}

				// R}hƂɏ
				dispatchCommand(pack);
			}
		} finally {
			isAlive = false;
			close();
		}
	}

	/**
	 * \Pbg܂B
	 */
	private void close() {
		if (socket != null && !socket.isClosed()) {
			if (socket instanceof MulticastSocket) {
				MulticastSocket sock = (MulticastSocket) socket;
				try {
					sock.leaveGroup(InetAddress.getByName("ff02::1")); //$NON-NLS-1$
				} catch (IOException e) {
					logger.error("Exception occurred " //$NON-NLS-1$
							+ "while leaving multicast group.", e); //$NON-NLS-1$
				}
			}
			socket.close();
		}
	}

	/**
	 * oCh܂B
	 * @param bindAddress oChAhX(null̂ƂAׂẴ[JAhX)
	 * @throws BindException |[ggp܂̓[JAhX݂Ȃ
	 * @throws SocketException
	 * @throws IOException
	 */
	void bind(InetAddress bindAddress) throws BindException, SocketException,
			IOException {
		int v = InetAddressUtils.getIpVersion(bindAddress);

		switch (v) {
		case 6:
			InetSocketAddress addr = new InetSocketAddress(bindAddress,
					messenger.port);
			MulticastSocket s = new MulticastSocket(addr);
			s.joinGroup(InetAddress.getByName("ff02::1")); //$NON-NLS-1$
			this.socket = s;
			break;
		case 4:
		default:
			this.socket = new DatagramSocket(messenger.port, bindAddress);
			break;
		}

		isAlive = true;
		logger.info("bind: addr=" + bindAddress + ", port=" //$NON-NLS-1$ //$NON-NLS-2$
				+ messenger.port);
	}

	/**
	 * pPbgdĂȂׁAo^܂B
	 * @param pack pPbg
	 * @return dpPbĝƂtrue
	 */
	private boolean isDuplicationPacket(IPMPacket pack) {
		int count;
		synchronized (members) {
			count = members.size();
		}

		Long id = pack.getPacketValue();
		synchronized (this.duplicationList) {
			if (this.duplicationList.contains(id)) {
				return true;
			}
			// o^
			this.duplicationList.addFirst(id);
			while (this.duplicationList.size() > Math.max(20, count * 3)) {
				this.duplicationList.removeLast();
			}
		}
		return false;
	}

	/**
	 * R}hƂɏ܂B
	 * @param pack pPbg
	 * @return true p false continue
	 */
	private boolean dispatchCommand(IPMPacket pack) {

		if (pack.version != 1) {
			return true;
		}

		/*
		 * BʒmKvȃpPbgȊȌdpPbg
		 * ubN
		 */
		boolean duplication = isDuplicationPacket(pack);
		boolean responseRequired = pack.isResponseRequired();
		if (!responseRequired && duplication) {
			return false;
		}
		String key = pack.getIdentifierLong();

		// R}hɉď
		switch ((int) pack.getPacketClass()) {
		//////////////////////////////////////////////
		// o[Fn
		//
		case (int) IPMSG.NOOPERATION:
			break;
		case (int) IPMSG.BR_ENTRY:
			// o^ԓēo^
			sendReply(pack);
			dispatchANSENTRY(key, pack);
			break;
		case (int) IPMSG.ANSENTRY:
			// ԓo^
			dispatchANSENTRY(key, pack);
			break;
		case (int) IPMSG.BR_EXIT:
			// 폜
			dispatchBR_EXIT(key, pack);
			break;
		case (int) IPMSG.BR_ABSENCE:
			// ̍XVʒm
			dispatchBR_ABSENCE(key, pack);
			break;
		//////////////////////////////////////////////
		// zXgXgn
		//
		case (int) IPMSG.ANSLIST:
			dispatchANSLIST(key, pack);
			break;
		case (int) IPMSG.GETLIST:
			//TODO zXgXg̑o
			break;
		//////////////////////////////////////////////
		// bZ[WMn
		//
		case (int) IPMSG.SENDMSG:
			// bZ[WM
			dispatchSENDMSG(key, pack, duplication);
			break;
		case (int) IPMSG.READMSG:
			// Jʒm
			dispatchREADMSG(key, pack, duplication);
			break;
		case (int) IPMSG.DELMSG:
			// ̔jʒm
			dispatchDELMSG(pack);
			break;
		//////////////////////////////////////////////
		// pPbgBmF
		//
		case (int) IPMSG.RECVMSG:
		case (int) IPMSG.ANSREADMSG:
			// ʏ̃bZ[WƊJʒm̎Mʒm
			Sender send;
			synchronized (this.checkList) {
				send = (Sender) checkList.get(pack.data);
			}
			if (send != null) {
				send.setVerified();
			}
			break;
		//
		// obNOEh
		//
		case (int) IPMSG.GETINFO:
			// o[W̑Mv
			dispatchGETINFO(pack);
			break;
		case (int) IPMSG.SENDINFO:
			// o[W̎M
			break;
		case (int) IPMSG.GETABSENCEINFO:
			// sݒʒm̑Mv
			dispatchGETABSENCEINFO(pack);
			break;
		case (int) IPMSG.SENDABSENCEINFO:
			// sݒʒm̎M
			dispatchSENDABSENCEINFO(pack);
			break;
		case (int) IPMSG.GETPUBKEY:
			// J̑Mv
			dispatchGETPUBKEY(pack);
			break;
		case (int) IPMSG.ANSPUBKEY:
			// J̎M
			dispatchANSPUBKEY(key, pack);
			break;
		case (int) IPMSG.RELEASEFILES:
			dispatchRELEASEFILES(pack);
			break;
		default:
			logger.error("Unsupported command:" + pack.getPacketClass()); //$NON-NLS-1$
			break;
		}
		return true;
	}

	/**
	 * IPMSG_RELEASEFILES(Ytt@C̔j)
	 * @param pack pPbg
	 */
	private void dispatchRELEASEFILES(IPMPacket pack) {
		long id;
		try {
			id = Long.parseLong(pack.data);
		} catch (NumberFormatException e) {
			return;
		}
		messenger.releaseAttachment(id);
	}

	/**
	 * IPMSG_ANSLIST(zXgXg̑o)
	 * @param key
	 * @param pack
	 */
	private void dispatchANSLIST(String key, IPMPacket pack) {
		// 
		String[] elements = Tool.split(pack.data, '\7', 0);
		// Ō\7ŏÎŗvf1
		if (elements.length < 9 || (elements.length - 2) % 7 != 1) {
			logger.warn("ANSLIST is short: " + elements.length); //$NON-NLS-1$
			return;
		}
		// ̔ԍƃzXg擾
		int next, count;
		try {
			next = Integer.parseInt(elements[0].trim());
			count = Integer.parseInt(elements[1].trim());
		} catch (NumberFormatException e) {
			logger.warn("ANSLIST: Next index or number of hosts is invalid.", e); //$NON-NLS-1$
			return;
		}
		// JEg̃`FbN
		if (count != (elements.length - 2) / 7) {
			logger.warn("count: " + count); //$NON-NLS-1$
			return;
		}
		InetAddress[] localAddresses = InetAddressUtils.getLocalAddresses();

		// 
		for (int i = 0; i < count; i++) {
			IPMPacket member = new IPMPacket(messenger);
			// OC
			member.loginName = elements[2 + i * 7];
			// zXg
			member.hostName = elements[2 + i * 7 + 1];
			// R}h,|[gԍ
			int membersPort;
			try {
				member.command = Long.parseLong(elements[2 + i * 7 + 2]);
				membersPort = Integer.parseInt(elements[2 + i * 7 + 4]);
			} catch (NumberFormatException e) {
				continue;
			}
			// |[gԍ
			membersPort = Tool.exchangeEndian(membersPort);

			// IPAhX
			try {
				String addrStr = elements[2 + i * 7 + 3];
				if (addrStr == null || addrStr.length() == 0) {
					continue;
				}
				InetAddress iaddr = InetAddress.getByName(addrStr);

				// [JAhX̏ꍇ͖
				if (InetAddressUtils
						.isLocalAddressCached(iaddr, localAddresses)) {
					continue;
				}
				member.address = new InetSocketAddress(iaddr, membersPort);
			} catch (UnknownHostException e) {
				continue;
			}

			if (pack.hostName.equals(member.hostName)
					&& pack.address.getPort() == membersPort) {
				// T[o͎œo^Ă̂ŏO
				continue;
			}

			// jbNl[
			if (!elements[2 + i * 7 + 5].equals("\b")) { //$NON-NLS-1$
				member.data = elements[2 + i * 7 + 5];
			} else {
				// OCŕ₤
				member.data = member.loginName;
			}

			// O[v
			if (!elements[2 + i * 7 + 6].equals("\b")) { //$NON-NLS-1$
				member.extension = elements[2 + i * 7 + 6];
			}
			// o^
			dispatchANSENTRY(member.getIdentifierLong(), member);
		}
		// v
		if (next > 0) {
			IPMPacket nextPack = new IPMPacket(messenger);
			nextPack.command = IPMSG.GETLIST;
			nextPack.data = Integer.toString(next);
			nextPack.setPacketNumber();
			try {
				sendRaw(nextPack.getBytes(), pack.address);
			} catch (IOException e) {
				logger.error("Exception occurred while " //$NON-NLS-1$
						+ "sending request of GETLIST.", e); //$NON-NLS-1$
			}
		}
	}

	/**
	 * IPMSG_GETINFO(o[W̑Mv)
	 * @param pack pPbg
	 */
	private void dispatchGETINFO(IPMPacket pack) {
		IPMPacket info = new IPMPacket(messenger);
		info.command = IPMSG.SENDINFO;
		Version v = Version.loadFromResource();
		info.data = "Petaclone " + v.getVersionText() //$NON-NLS-1$
				+ " (Build " + v.getBuildNumberStr() + ")"; //$NON-NLS-1$ //$NON-NLS-2$
		info.setPacketNumber();
		try {
			sendRaw(info.getBytes(), pack.address);
		} catch (IOException e) {
			logger.error("Exception occurred " //$NON-NLS-1$
					+ "during responding GETINFO.", e); //$NON-NLS-1$
		}
	}

	/**
	 * IPMSG_GETABSENCEINFO(sݒʒm̑Mv)
	 * @param pack pPbg
	 */
	private void dispatchGETABSENCEINFO(IPMPacket pack) {
		IPMPacket info = new IPMPacket(messenger);
		info.command = IPMSG.SENDABSENCEINFO;
		info.setPacketNumber();
		//BaseMain main = messenger.getMain();
		AppPreferences pref = messenger.getAppPreferences();
		RuntimeProperties rp = pref.getRuntime();
		if (rp.isAbsence()) {
			info.data = rp.getAbsenceMessage();
			info.convertReturnCode();
		} else {
			info.data = mm.get("IPMUDPServer.iAmHere"); //$NON-NLS-1$
		}
		try {
			sendRaw(info.getBytes(), pack.address);
		} catch (IOException e) {
			logger.error("Exception occurred while " //$NON-NLS-1$
					+ "sending absent message.", e); //$NON-NLS-1$
		}
	}

	/**
	 * IPMSG_SENDABSENCEINFO(sݒʒm̎M)
	 * @param pack pPbg
	 */
	private void dispatchSENDABSENCEINFO(IPMPacket pack) {
		//TODO s݃bZ[W̎MT[oXi[ɒʒm
	}

	/**
	 * IPMSG_GETPUBKEY(J̑Mv)
	 * @param pack pPbg
	 */
	private void dispatchGETPUBKEY(IPMPacket pack) {
		if (publicKeyCache == null) {
			return;
		}
		IPMPacket info = new IPMPacket(messenger);
		info.command = IPMSG.ANSPUBKEY;
		info.setPacketNumber();
		info.data = publicKeyCache;

		try {
			// M
			sendRaw(info.getBytes(), pack.address);
		} catch (IOException e) {
			logger.error("Exception occurred " //$NON-NLS-1$
					+ "while sending public key.", e); //$NON-NLS-1$
		}
	}

	/**
	 * J̎M
	 * @param key o[ʎq(`)
	 * @param pack pPbg
	 */
	private void dispatchANSPUBKEY(String key, IPMPacket pack) {
		IPMCipher cipher;
		// Xgɉ
		try {
			cipher = new IPMCipher(key, pack.data);
			synchronized (this.publicKeys) {
				this.publicKeys.put(key, cipher);
			}
		} catch (Exception e) {
			logger.error("Exception occurred while " //$NON-NLS-1$
					+ "processing received public key.", e); //$NON-NLS-1$
			return;
		}
		// M҂̃Xbhɒʒm
		MessageSender sender;
		synchronized (this.checkList) {
			sender = (MessageSender) this.checkList.get(key);
		}
		if (sender != null) {
			sender.setPublicKey(cipher);
		}
	}

	/**
	 * IPMSG_DELMSG(̍폜ʒm)
	 * @param pack pPbg
	 */
	private void dispatchDELMSG(IPMPacket pack) {
		logger.info("Sealed message got dumped: " + pack); //$NON-NLS-1$
	}

	/**
	 * IPMSG_READMSG(Jʒm̏)
	 * @param key o[ʎq(`)
	 * @param pack pPbg
	 * @param duplication dpPbg̏ꍇtrue
	 */
	private void dispatchREADMSG(String key, IPMPacket pack, boolean duplication) {
		// pPbgBʒm
		sendArrivalNotification(pack, IPMSG.ANSREADMSG);
		if (duplication) {
			return;
		}
		Message msg = pack.toMessage();
		// jbNl[ǉ
		IPMPacket ipmp;
		synchronized (this.members) {
			ipmp = (IPMPacket) this.members.get(key);
		}

		if (ipmp != null && ipmp.data != null) {
			msg.userName = ipmp.data;
		}
		messenger.notifyOpenNotice(msg);
	}

	/**
	 * IPMSG_SENDMSG(bZ[WM)̏
	 * @param key o[ʎq(`)
	 * @param pack pPbg
	 * @param duplication dpPbg̏ꍇtrue
	 */
	private void dispatchSENDMSG(String key, IPMPacket pack, boolean duplication) {
		// pPbgBʒm
		sendArrivalNotification(pack, IPMSG.RECVMSG);
		if (duplication) {
			return;
		}
		// ϊ
		Message msg = pack.toMessage();
		msg.messageClass = Message.PENDING;
		// jbNl[擾
		IPMPacket ipmp;
		synchronized (this.members) {
			ipmp = (IPMPacket) this.members.get(key);
		}

		if (ipmp != null && ipmp.data != null) {
			msg.userName = ipmp.data;
		}
		// ʒm
		messenger.notifyMessage(msg);

		AppPreferences pref = messenger.getAppPreferences();
		RuntimeProperties rp = pref.getRuntime();

		// s݃[h̏
		if (!rp.isAbsence()) {
			return;
		}
		IPMPacket absenceMessage = new IPMPacket(messenger);
		absenceMessage.setPacketNumber();
		absenceMessage.setUserGroupName();
		absenceMessage.command = IPMSG.SENDMSG | IPMSG.AUTORETOPT;

		absenceMessage.data = rp.getAbsenceMessage();
		absenceMessage.convertReturnCode();

		// MɕԂ
		try {
			sendRaw(absenceMessage.getBytes(), pack.address);
		} catch (IOException e) {
			logger.error("Cannot reply absent message automatically.", e); //$NON-NLS-1$
		}
	}

	/**
	 * BmFpPbg𑗐M܂B
	 * 
	 * @param pack MpPbg
	 * @param command IPMSG.ANSREADMSG ܂ IPMSG.RECVMSG
	 */
	private void sendArrivalNotification(IPMPacket pack, long command) {
		if (pack.isResponseRequired()) {
			IPMPacket checkPacket = new IPMPacket(messenger);
			checkPacket.command = command;
			checkPacket.data = Long.toString(pack.id);
			checkPacket.setPacketNumber();
			try {
				sendRaw(checkPacket.getBytes(), pack.address);
			} catch (IOException e) {
				logger.error("Cannot send arrival notification.", e); //$NON-NLS-1$
			}
		}
	}

	/**
	 * IPMSG_BR_ABSENCE(XV)̏
	 * @param key o[ʎq(`)
	 * @param pack pPbg
	 */
	private void dispatchBR_ABSENCE(String key, IPMPacket pack) {
		// ̍XV
		synchronized (this.members) {
			this.members.put(key, pack);
		}

		// _CAbvo
		if ((pack.command & IPMSG.DIALUPOPT) != 0) {
			synchronized (this.dialup) {
				this.dialup.put(key, pack.address);
			}
		}
		// Xi[
		messenger.notifyUpdateMember(pack.toMember());
	}

	/**
	 * IPMSG_ANSENTRY̏
	 * @param key o[ʎq(`)
	 * @param pack pPbg
	 */
	private void dispatchANSENTRY(String key, IPMPacket pack) {

		synchronized (members) {
			// ԓo^
			if (members.containsKey(key)) {
				// IP Messengerł̓jbNl[̕ύXȂǂʂ̃R}h
				// ŝłł̓L[`FbN邾ł悢
				return;
			}
			// o^
			members.put(key, pack);
		}

		// _CAbvo
		if ((pack.command & IPMSG.DIALUPOPT) != 0) {
			synchronized (dialup) {
				dialup.put(key, pack.address);
			}
		}
		// Xi[ɓ`
		messenger.notifyAddMember(pack.toMember());
	}

	/**
	 * IPMSG_BR_EXIT̏
	 * @param key o[ʎq(`)
	 * @param pack pPbg
	 */
	private void dispatchBR_EXIT(String key, IPMPacket pack) {

		synchronized (this.members) {
			if (!members.containsKey(key)) {
				return;
			}
			members.remove(key);
		}

		synchronized (this.dialup) {
			dialup.remove(key);
		}

		synchronized (this.publicKeys) {
			this.publicKeys.remove(key);
		}

		//System.out.println("o^(폜):" + members.size());
		messenger.notifyDeleteMember(pack.toMember());
	}

	/**
	 * o[FnR}h̑M̃Xg擾܂B
	 * @return MIPAhXƃ|[gԍ̔z
	 */
	private InetSocketAddress[] getBroadcastAddresses() {
		AppPreferences pref = messenger.getAppPreferences();

		InetSocketAddress[] addrs = messenger.configuration
				.getBroadcastAddresses(pref, (InetSocketAddress) socket
						.getLocalSocketAddress());

		synchronized (this.dialup) {
			InetSocketAddress[] result = new InetSocketAddress[addrs.length
					+ this.dialup.size()];
			System.arraycopy(addrs, 0, result, 0, addrs.length);

			// _CAbv
			if (!dialup.isEmpty()) {
				int i = addrs.length;
				Iterator itr = dialup.values().iterator();
				for (; itr.hasNext(); i++) {
					result[i] = (InetSocketAddress) itr.next();
				}
			}
			return result;
		}
	}

	/**
	 * o[FnR}h𑗐M܂B(u[hLXg)
	 * @param commands IvVȂR}h̔z
	 * @param delays xԂ̔z
	 */
	private void sendOrganizationPacket(long[] commands, long[] delays) {
		Task.schedule(messenger.timer, new SendOrganizationPacketTask(commands,
				delays), delays[0]);

	}

	/**
	 * o[FnR}hM^XN
	 */
	private class SendOrganizationPacketTask implements Runnable {

		/** ݂̎s */
		private int count;

		/** u[hLXgAhX */
		private InetSocketAddress[] addrs;

		/** pPbg */
		private IPMPacket packet;

		/** R}h */
		private final long[] commands;

		/** x */
		private final long[] delays;

		/**
		 * RXgN^
		 * @param commands IvVȂR}h̔z
		 * @param delays xԂ̔z
		 */
		public SendOrganizationPacketTask(long[] commands, long[] delays) {
			this.commands = commands;
			this.delays = delays;
		}

		public void run() {
			try {
				switch (count) {
				case 0:
					addrs = getBroadcastAddresses();

					packet = new IPMPacket(messenger);
					packet.setUserGroupName();
				default:
					packet.setOrganizationCommand(commands[count],
							encryptOption, null);
					break;
				}

				// M
				for (int i = 0; i < addrs.length; i++) {
					try {
						packet.setPacketNumber();
						byte[] rawData = packet.getBytes();
						sendRaw(rawData, addrs[i]);
					} catch (IOException e) {
						logger.error("Failed to send organizing command.", e); //$NON-NLS-1$
					}
				}
				count++;
				if (count < delays.length) {
					Task.schedule(messenger.timer, this, delays[count]);
				}
			} catch (Exception e) {
				logger.error("Exception occurred " //$NON-NLS-1$
						+ "while sending organization packet.", e); //$NON-NLS-1$
			}
		}
	}

	/**
	 * Ytt@Cj܂B
	 * @param message j郁bZ[W
	 */
	void releaseAttachments(Message message) {
		if (!message.hasAttachment()) {
			logger.warn("No file is attached on this message."); //$NON-NLS-1$
			return;
		}
		// M擾
		InetSocketAddress addr = new InetSocketAddress(message.ipAddress,
				message.port);
		// pPbg
		IPMPacket pack = new IPMPacket(messenger);
		pack.command = IPMSG.RELEASEFILES;
		pack.setPacketNumber();
		pack.data = Integer.toString(message.id);
		byte[] rawData = pack.getBytes();
		try {
			sendRaw(rawData, addr);
		} catch (IOException e) {
			logger.error("Failed to send the destruction " //$NON-NLS-1$
					+ "notification of attachment.", e); //$NON-NLS-1$
		}
	}

	/**
	 * w肳ꂽf[^𑗐M܂B
	 * @param rawData Mf[^
	 * @param addr MIPAhXƃ|[g
	 * @throws IOException o͗O
	 */
	private void sendRaw(byte[] rawData, InetSocketAddress addr)
			throws IOException {
		if (DEBUG) {
			String s = "Sending packet:dest=" + addr.toString(); //$NON-NLS-1$
			s += ", data=" + Tool.escapeControl(rawData); //$NON-NLS-1$
			logger.debug(s);
		}
		DatagramPacket packet = new DatagramPacket(rawData, rawData.length,
				addr);
		socket.send(packet);
	}

	/**
	 * bZ[W𑗐M܂B<br>
	 * Jʒmł͂Ȃ<br>
	 * @param msg ꂩ瑗M郁bZ[W
	 * @param dest M
	 * @param listener ʂ󂯎郊Xi
	 */
	void sendMessage(Message msg, UniqueMember[] dest,
			SendingResultListener listener) {
		//
		// fileId蓖Ă(0`)
		if (msg.hasAttachment()) {
			List attachments = msg.attachments;
			for (int i = 0; i < attachments.size(); i++) {
				Attachment att = (Attachment) attachments.get(i);
				att.id = i;
			}
		}

		// M
		for (int i = 0; i < dest.length; i++) {
			MessageSender sender = new MessageSender(msg, dest[i], listener);
			sender.send();
		}
	}

	/**
	 * bZ[W̑MUTF-8T|[gĂ邩ǂԂ܂B
	 * 
	 * @param message
	 * @return UTF-8T|[gĂꍇtrueBo[łɑ݂ȂꍇfalseB
	 */
	boolean isMessageFromUTF8Member(Message message) {
		String id = message.getIdentifierLong();
		IPMPacket packet = (IPMPacket) members.get(id);
		if (packet != null) {
			return packet.isUTF8Supported();
		}
		return false;
	}

	/**
	 * Sender
	 */
	private interface Sender {
		/** ʒmƂSenderɓ`܂B*/
		void setVerified();
	}

	/**
	 * Đ1ɑ΂đMXbh
	 */
	private class MessageSender implements Runnable, Sender {
		/** setVerified()ꂽtrue */
		private boolean notified;

		/** MbZ[W */
		private Message msg;

		/** M */
		private UniqueMember dest;

		/** Xi[ */
		private SendingResultListener listener;

		/** M̌J */
		private IPMCipher cipher;

		/** s */
		private final static int limit = 2;

		/** pPbgBmF̍Œ҂(~b) */
		private final static long timeout = 2500;

		/**
		 *  RXgN^<br>
		 * @param msg M郁bZ[W
		 * @param dest M
		 * @param listener null̏ꍇ͌ʂʒmȂ
		 */
		MessageSender(Message msg, UniqueMember dest,
				SendingResultListener listener) {

			this.msg = msg;
			this.dest = dest;
			this.listener = listener;
		}

		/**
		 * MJn܂B
		 */
		void send() {
			Thread th = new Thread(messenger.getThreadGroup(), this,
					"IPMUDPServer$MessageSender"); //$NON-NLS-1$
			th.start();
		}

		/**
		 * `̃[U[ʎq擾܂B
		 * @return 擾łȂꍇnull
		 */
		private String getDestId() {
			String destId = dest.getIdentifierLong();
			synchronized (IPMUDPServer.this.members) {
				if (!IPMUDPServer.this.members.containsKey(destId)) {
					// TODO OCo^ĂȂƏƍɎs
					// FĂȂƂȂ
					return null;
				}
			}
			return destId;
		}

		/**
		 * o擾܂B
		 * @param destId o[ʎq(`)
		 * @return o[(pPbgf[^)
		 * @throws UnknownHostException Rs[^IPAhXɕϊłȂꍇ
		 */
		private IPMPacket getMember(String destId) throws UnknownHostException {
			IPMPacket member;
			synchronized (IPMUDPServer.this.members) {
				member = (IPMPacket) IPMUDPServer.this.members.get(destId);
			}
			// oXgɍڂĂȂƂ
			// M̃Rs[^畜݂
			if (member == null) {
				member = new IPMPacket(messenger);
				member.address = new InetSocketAddress(InetAddress
						.getByName(dest.getComputerName()), messenger.port);
			}
			return member;
		}

		/**
		 * J擾܂B
		 * (this.cipherɃZbg)
		 * @param destId o[ʎq(`)
		 * @param address IPAhXƃ|[gԍ
		 * @throws IOException o͗O
		 * @throws InterruptedException 
		 */
		private void getPublicKey(String destId, InetSocketAddress address)
				throws IOException, InterruptedException {

			synchronized (IPMUDPServer.this.publicKeys) {
				this.cipher = (IPMCipher) IPMUDPServer.this.publicKeys
						.get(destId);
			}

			if (this.cipher != null) {
				return;
			}
			// J̑MvpPbg쐬
			IPMPacket getPubkeyPacket = new IPMPacket(
					IPMUDPServer.this.messenger);

			getPubkeyPacket.loginName = this.msg.osUserName;
			getPubkeyPacket.hostName = this.msg.computerName;
			getPubkeyPacket.command = IPMSG.GETPUBKEY;
			getPubkeyPacket.setPacketNumber();
			String capability = Long.toString(
					IPMUDPServer.this.messenger.getCipherCapability(), 16)
					.toUpperCase();
			getPubkeyPacket.data = capability;

			// oCi擾
			byte[] rawData = getPubkeyPacket.getBytes();
			// M
			//String key = getKey(getPubkeyPacket);
			// L[͑̃[U[ʎq FIXME 2ʑȂ
			String key = destId;

			try {
				addSendThread(key, this);

				for (int i = 0; i < 5; i++) {
					IPMUDPServer.this.sendRaw(rawData, address);

					//ԐM҂
					synchronized (this) {
						wait(timeout);
						if (notified) {
							notified = false;
							logger.info("Got public key."); //$NON-NLS-1$
							return;
						}
					}
				}

			} finally {
				// o^
				removeSendThread(key);
			}
			logger.info("Failed to get public key."); //$NON-NLS-1$
			this.cipher = null;
		}

		/**
		 * Đ
		 */
		private void resolveDestination() {
			// IPAhXnull̂ƂA݂
			if (dest.getIpAddress() == null) {
				try {
					InetAddress addr = InetAddress.getByName(dest
							.getComputerName());
					dest.setIpAddress(addr.getHostAddress());

					// FĂ邱Ƃɂ
					IPMPacket pack = new IPMPacket(messenger);
					pack.address = new InetSocketAddress(dest.getIpAddress(),
							messenger.port);
					pack.hostName = dest.getComputerName();
					pack.loginName = dest.getOsUserName();
					pack.data = dest.getUserName();
					members.put(dest.getIdentifierLong(), pack);

				} catch (UnknownHostException e) {
					// Ȃ(㑱̏Œ~ɂȂ)
				}
			}
		}

		/**
		 * ̃\bh̑void send()gpB
		 */
		public void run() {
			resolveDestination();

			// [U[ʎq𓾂
			String destId = getDestId();
			if (destId == null) {
				if (DEBUG) {
					synchronized (members) {
						logger.debug(dest.getIdentifierLong()
								+ " is not recognized. members=" + members); //$NON-NLS-1$
					}
				}
				String str = mm.get("IPMUDPServer.isNotRecognized", //$NON-NLS-1$
						new Object[] { dest.getUserName() });
				listener.onSendResult(createDestination(dest, false),
						new Exception(str));
				return;
			}

			// M擾
			IPMPacket member;
			try {
				member = getMember(destId);
			} catch (UnknownHostException e) {
				listener.onSendResult(createDestination(dest, false), e);
				return;
			}

			if (DEBUG)
				logger.debug("Selected member: " + member); //$NON-NLS-1$

			// J̎擾
			if (encryptOption != 0
					&& (member.getOption() & IPMSG.ENCRYPTOPT) != 0) {
				try {
					getPublicKey(destId, member.address);
				} catch (Exception e) {
					logger.error("Exception occurred " //$NON-NLS-1$
							+ "while getting public key.", e); //$NON-NLS-1$
					listener.onSendResult(createDestination(dest, false), e);
					this.cipher = null;
				}
			}

			// bZ[W̃pPbg쐬AM
			// pPbgɕϊB
			IPMPacket pack = new IPMPacket(messenger, msg);
			// M悲ƂɃR}h쐬
			pack.setSendMsgCommand(msg, member);
			// Ytt@C
			pack.setAttachments(msg, member);

			// M̌Jꍇ̓bZ[WÍ
			if (encryptOption != 0 && this.cipher != null
					&& (member.command & IPMSG.ENCRYPTOPT) != 0) {
				boolean useRC2 = messenger.getAppPreferences()
						.getCipherIsUnlimited();

				String encrypted = this.cipher.encrypt(pack.data, pack
						.getEncodingForUTF8OPT(), useRC2, useSmallKey);
				if (!encrypted.equals(pack.data)) {
					pack.command |= IPMSG.ENCRYPTOPT;
					pack.data = encrypted;
					logger.info("Encrypt successful."); //$NON-NLS-1$
				} else {
					logger.warn("Failed to encrypt."); //$NON-NLS-1$
				}

			} else {
				logger.info("No encryption."); //$NON-NLS-1$
			}

			pack.setPacketNumber();
			// t@CYtĂ_E[hTCPT[oɒǉ
			if (msg.hasAttachment()) {
				/*
				 * addDownloadPermissionɎw肷YtXg
				 * ŕύX邽߁AŕB
				 */
				messenger.addDownloadPermission(pack.id, (List) msg.attachments
						.clone());
			}
			// T[oɓo^
			String packetId = getKey(pack);
			addSendThread(packetId, this);

			// Mf[^쐬
			byte[] rawData = pack.getBytes();

			if (DEBUG)
				logger.debug("send data:" + Tool.escapeControl(rawData)); //$NON-NLS-1$

			try {
				synchronized (this) {
					for (int i = 0; i < limit; i++) {
						// M
						sendRaw(rawData, member.address);
						// mF҂
						wait(timeout);
						if (notified) {
							if (listener != null) {
								listener.onSendResult(createDestination(dest,
										true), null);
							}
							return;
						}
					}
				}
				// OŒʒm
				throw new Exception(mm.get("IPMUDPServer.failedToSendMessage", //$NON-NLS-1$
						new Object[] { member.address }));
			} catch (Exception e) {
				logger.error("Failed to send message.", e); //$NON-NLS-1$
				if ((pack.command & IPMSG.FILEATTACHOPT) != 0) {
					// L
					messenger.releaseAttachment(pack.id);
				}
				if (listener != null) {
					listener.onSendResult(createDestination(dest, false), e);
				}
				return;
			} finally {
				// o^
				removeSendThread(packetId);
			}
		}

		/**
		 * pPbgԍ𕶎ŕԂ܂B 
		 * @param packet pPbg
		 * @return pPbgԍ
		 */
		String getKey(IPMPacket packet) {
			return Long.toString(packet.id);
		}

		/** ʒmƂSenderɓ`܂B */
		synchronized public void setVerified() {
			this.notified = true;
			this.notify();
		}

		/**
		 * J̎MSenderɒʒm܂B
		 * @param cipher 
		 */
		synchronized void setPublicKey(IPMCipher cipher) {
			this.cipher = cipher;
			this.notified = true;
			this.notify();
		}
	}

	/**
	 * MessagȇM̃[U[ʎq(IPMPacket`)擾܂B
	 * @param msg bZ[W
	 * @return o[ʎq(`)
	 */
	private static String getIdentifierLong(Message msg) {
		StringBuffer sb = new StringBuffer();
		sb.append(msg.osUserName).append('@').append(msg.ipAddress);
		return sb.toString().toUpperCase();
	}

	/**
	 * messageɑ΂Jʒm𑗐M܂B
	 * (񓯊AG[Ȃ)
	 * @param message JꂽbZ[W
	 */
	void sendOpenNotice(Message message) {
		IPMPacket pack = new IPMPacket(messenger);
		pack.setPacketNumber();
		// ̃pPbgԍ
		pack.data = Integer.toString(message.id);
		pack.command = IPMSG.READMSG | IPMSG.READCHECKOPT;
		// M̃AhXƃ|[g擾
		String key = getIdentifierLong(message);

		IPMPacket info;
		synchronized (members) {
			info = (IPMPacket) this.members.get(key);
			if (info == null) {
				logger.warn("Destination of open notification " //$NON-NLS-1$
						+ "is not recognized. key=" + key); //$NON-NLS-1$
				return;
			}
		}

		OpenNoticeSender sender = new OpenNoticeSender(pack, info.address,
				null, null);
		sender.send();
	}

	/**
	 * IPMSG_BR_ENTRY(o^)IPMSG_ANSENTRY(ԓ)܂B
	 * @param pack BR_ENTRỸpPbg
	 */
	private void sendReply(IPMPacket pack) {

		int count;
		synchronized (members) {
			count = members.size();
		}
		long delay;
		if (count < 50) {
			delay = (long) (1023 * Math.random());
		} else if (count < 300) {
			delay = (long) (2047 * Math.random());
		} else {
			delay = (long) (4095 * Math.random());
		}

		// x
		Task.schedule(messenger.timer, //
				new SendReplyPacketTask(pack), delay);
	}

	/**
	 * ANSENTRY̒x
	 */
	private class SendReplyPacketTask implements Runnable {

		/** BR_ENTRỸpPbg */
		private final IPMPacket dest;

		/**
		 * RXgN^
		 * @param dest BR_ENTRỸpPbg
		 */
		public SendReplyPacketTask(IPMPacket dest) {
			this.dest = dest;
		}

		public void run() {
			IPMPacket data = new IPMPacket(messenger);
			data.setOrganizationCommand(IPMSG.ANSENTRY, encryptOption, dest);
			data.setUserGroupName();
			data.setPacketNumber();

			try {
				byte[] rawData = data.getBytes();
				sendRaw(rawData, dest.address);
			} catch (IOException e) {
				logger.warn("Reply failed.", e); //$NON-NLS-1$
			}
		}
	}

	/**
	 * IPMSG_BR_ENTRY(o^)u[hLXg܂B
	 */
	void sendRegister() {
		long[] commands = new long[] { IPMSG.NOOPERATION, IPMSG.BR_EXIT,
				IPMSG.BR_EXIT, IPMSG.BR_ENTRY, IPMSG.BR_ENTRY };
		long[] delays = new long[] { 0, 500, 500, 500, 5000 };
		sendOrganizationPacket(commands, delays);
	}

	/**
	 * IPMSG_BR_EXIT(폜)u[hLXg܂B
	 */
	void sendDelete() {
		long[] commands = new long[] { IPMSG.BR_EXIT, IPMSG.BR_EXIT };
		long[] delays = new long[] { 0, 500 };
		sendOrganizationPacket(commands, delays);
	}

	/**
	 * IPMSG_BR_ABSENCE(XV)u[hLXg܂B
	 */
	void sendUpdate() {
		long[] commands = new long[] { IPMSG.BR_ABSENCE };
		long[] delays = new long[] { 0 };
		sendOrganizationPacket(commands, delays);
	}

	/**
	 * o[XgXV܂B
	 * @param addOnly truêƂ̃o[XgNA܂B
	 */
	void sendFind(boolean addOnly) {
		if (!addOnly) {
			synchronized (this.members) {
				members.clear();
			}
			synchronized (this.dialup) {
				dialup.clear();
			}
			synchronized (this.publicKeys) {
				publicKeys.clear();
			}
		}
		long[] commands = new long[] { IPMSG.NOOPERATION, IPMSG.BR_ENTRY,
				IPMSG.BR_ENTRY };
		long[] delays = new long[] { 0, 500, 5000 };
		sendOrganizationPacket(commands, delays);
	}

	/**
	 * Sendo^܂B
	 * @param key L[BSENDMSĜƂ̓pPbgԍAGETPUBKEŶƂ̓[U[ʎqB
	 * @param send Sender
	 */
	private void addSendThread(String key, Sender send) {
		synchronized (checkList) {
			while (checkList.size() > MAX_SEND_THREAD) {
				try {
					// TODO H
					checkList.wait();
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			this.checkList.put(key, send);
		}
	}

	/**
	 * Send̓o^܂B
	 * @param key L[
	 */
	private void removeSendThread(String key) {
		synchronized (checkList) {
			this.checkList.remove(key);
			checkList.notify();
		}
	}

	/**
	 * Jʒm̑MEmF҂EĎssXbh<br>
	 * SendResultListenerŌʂʒmB
	 */
	private class OpenNoticeSender implements Runnable, Sender {
		/** setVerified()ꂽtrue */
		private boolean notified;

		/** MΏۂ̃f[^ */
		private IPMPacket packet;

		/** MAhX */
		private InetSocketAddress address;

		/** M */
		private UniqueMember dest;

		/** Xi[ */
		private SendingResultListener listener;

		/** s */
		private final static int limit = 2;

		/** pPbgBmF̍Œ҂(~b) */
		private final static long timeout = 2500;

		/**
		 *  RXgN^<br>
		 * @param packet MpPbg
		 * @param addr M̃AhX
		 * @param dest M
		 * @param listener null̏ꍇ͌ʂʒmȂ
		 */
		OpenNoticeSender(IPMPacket packet, InetSocketAddress addr,
				UniqueMember dest, SendingResultListener listener) {

			this.packet = packet;
			this.address = addr;
			this.dest = dest;
			this.listener = listener;
			assert this.packet.id != 0;
		}

		/**
		 * pPbgԍ𕶎ŕԂ܂B 
		 * @return pPbgID
		 */
		String getKey() {
			return Long.toString(packet.id);
		}

		/**
		 * MJn܂B
		 */
		void send() {
			Thread th = new Thread(messenger.getThreadGroup(), this,
					"IPMUDPServer$OpenNoticeSender"); //$NON-NLS-1$
			th.start();
		}

		/** Ms܂B */
		public void run() {
			// bZ[W̑M

			String packetId = this.getKey();
			// T[oɓo^
			addSendThread(packetId, this);
			byte[] rawData = packet.getBytes();
			try {
				synchronized (this) {
					for (int i = 0; i < limit; i++) {
						// M
						sendRaw(rawData, address);
						// mF҂
						wait(timeout);
						if (notified) {
							if (listener != null) {
								listener.onSendResult(createDestination(dest,
										true), null);
							}
							return;
						}
					}
				}
				// OŒʒm
				throw new Exception(mm.get(
						"IPMUDPServer.failedToSendOpenNotification", //$NON-NLS-1$
						new Object[] { address }));
			} catch (Exception e) {
				logger.error("Failed to send open notification.", e); //$NON-NLS-1$
				if (listener != null) {
					listener.onSendResult(createDestination(dest, false), e);
				}
			} finally {
				// o^
				removeSendThread(packetId);
			}
		}

		/** ʒmƂSenderɓ`܂B */
		synchronized public void setVerified() {
			this.notified = true;
			this.notify();
		}
	}

	/**
	 * Destination쐬܂B
	 * @param dest Destination̐UniqueMember
	 * @param result MʁBtrue: , false: s
	 * @return Destination M
	 */
	private Destination createDestination(UniqueMember dest, boolean result) {
		String launchId = messenger.configuration.getId();
		//TODO |[gԍ
		return new Destination(launchId, dest.getOsUserName(), dest
				.getComputerName(), dest.getUserName(), dest.getIpAddress(), 0,
				result);
	}
}
