package petaclone.petaroux;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.BindException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketException;
import java.net.UnknownHostException;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.InetAddressUtils;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Tool;
import petaclone.AppPreferences;
import petaclone.Attachment;
import petaclone.ControlCreator;
import petaclone.Destination;
import petaclone.DownloadListener;
import petaclone.LaunchConfiguration;
import petaclone.LaunchException;
import petaclone.Member;
import petaclone.Message;
import petaclone.Messenger;
import petaclone.Protocol;
import petaclone.RuntimeProperties;
import petaclone.SendingResultListener;
import petaclone.ServerListener;
import petaclone.UniqueMember;

/**
 * y^낤XvgRMessengerNXB
 * 
 * @author Takayuki Shoji
 */
public class PetaxMessenger implements Messenger {

	/**
	 * logger
	 */
	private static final Logger logger = Logger.getLogger(PetaxMessenger.class);

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/** GR[fBO */
	static final String ENCODING = "UTF-8"; //$NON-NLS-1$

	/**
	 * vgRB
	 */
	private static Protocol protocol;
	static {
		protocol = new Protocol();
		protocol.maxMessageLength = 500;
		protocol.maxUserNameLength = 8;//8
		protocol.protocolName = mm.get("PetaxMessenger.protocolName"); //$NON-NLS-1$
		protocol.signature = "petax"; //$NON-NLS-1$
	}

	/**
	 * AvP[Vݒ
	 */
	private AppPreferences pref;

	/**
	 * Nݒ
	 */
	LaunchConfiguration configuration;

	/**
	 * T[oXi[
	 */
	private ServerListener serverListener;

	/**
	 * UDPT[o
	 */
	PetaxUDPServer udp;

	/**
	 * TCPT[o
	 */
	PetaxTCPServer tcp;

	/**
	 * oChAhX
	 */
	InetAddress bindAddress;

	/** TCPT[õXbh */
	private Thread tcpThread;

	/** UDPT[õXbh */
	private Thread udpThread;

	/**
	 * oChAhX擾܂B
	 * \PbgɎw肷ׂInetAddressƃoChAhXԍtB[hɐݒ肳܂B
	 * @param conf Nݒ
	 * @return oChAhX
	 */
	private InetAddress getBindAddress(LaunchConfiguration conf) {
		int bindAddressNumber = conf.getBindAddress();
		return InetAddressUtils.getLocalAddress(bindAddressNumber);
	}

	/**
	 * @see petaclone.Messenger#createConfigControl(petaclone.ControlCreator)
	 */
	public void createConfigControl(ControlCreator creator) {
		// \
		final String faces[] = {
		//
				mm.get("PetaxMessenger.expression.normal"), //$NON-NLS-1$
				mm.get("PetaxMessenger.expression.smiling"), //$NON-NLS-1$
				mm.get("PetaxMessenger.expression.angry"), //$NON-NLS-1$
				mm.get("PetaxMessenger.expression.crying"), //$NON-NLS-1$
				mm.get("PetaxMessenger.expression.puzzled"), //$NON-NLS-1$
				mm.get("PetaxMessenger.expression.surprised"), //$NON-NLS-1$
				mm.get("PetaxMessenger.expression.sleeping"), //$NON-NLS-1$
				"" }; //$NON-NLS-1$
		final String facesValues[] = { "normal", "smile", "anger", "cry", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$
				"lost", "surprise", "quiet", "special" }; //$NON-NLS-1$ //$NON-NLS-2$//$NON-NLS-3$//$NON-NLS-4$
		creator.createComboBox("face", //$NON-NLS-1$
				mm.get("PetaxMessenger.expression.label"), faces, facesValues, //$NON-NLS-1$
				"normal", true); //$NON-NLS-1$

		creator.createSeparator();
		creator.createLabel(mm.get("PetaxMessenger.portInfo.label")); //$NON-NLS-1$

		creator.createSeparator();
		creator.createLabel(mm.get("PetaxMessenger.original.label")); //$NON-NLS-1$
	}

	public void begin(ThreadGroup gr) throws LaunchException {

		checkGizmoid();

		this.bindAddress = getBindAddress(this.configuration);

		boolean result = false;
		try {
			tcp = new PetaxTCPServer(this);
			tcp.bind();

			udp = new PetaxUDPServer(this);
			udp.bind();

			result = true;
		} catch (BindException e) {
			throw new LaunchException(mm
					.get("PetaxMessenger.serverAddressIsUnavailable"), e, //$NON-NLS-1$
					this.configuration);

		} catch (SocketException e) {
			throw new LaunchException(mm
					.get("PetaxMessenger.underlyingProtocolError"), e, //$NON-NLS-1$
					this.configuration);

		} catch (IOException e) {
			throw new LaunchException(mm.get("PetaxMessenger.IOException"), e, //$NON-NLS-1$
					this.configuration);

		} finally {
			if (!result) {
				if (tcp != null) {
					tcp.quit();
				}
				if (udp != null) {
					udp.quit();
				}
			}
		}

		// TCPT[o̊Jn
		tcpThread = new Thread(gr, tcp);
		tcpThread.setName("PetarouX TCP"); //$NON-NLS-1$
		tcpThread.start();
		// UDPT[o̊Jn
		udpThread = new Thread(gr, udp);
		udpThread.setName("PetarouX UDP"); //$NON-NLS-1$
		udpThread.start();
	}

	/**
	 * gizmoid̃`FbNƐ
	 */
	private void checkGizmoid() {
		String gizmoid = configuration.getProperty("gizmoid", ""); //$NON-NLS-1$ //$NON-NLS-2$
		if (!PetaxPacket.isGizmoid(gizmoid)) {
			String id = PetaxPacket.generateGizmoid();
			configuration.setProperty("gizmoid", id); //$NON-NLS-1$
		}
	}

	public boolean getAlive() {
		if (udp == null || tcp == null) {
			return false;
		}
		return tcp.getAlive() && udp.getAlive();
	}

	public String getId() {
		if (configuration == null) {
			return null;
		}
		return configuration.getId();
	}

	public int getMessageLength(String message) {
		return message.length();
	}

	/**
	 * @see petaclone.Messenger#getProtocol()
	 */
	public Protocol getProtocol() {
		return protocol;
	}

	public void getStatusMessage(UniqueMember member) {
		// vgRT|[gĂȂ
	}

	public InetSocketAddress[] getTcpBindAddress(LaunchConfiguration conf)
			throws UnknownHostException {
		return getAddressAndPort(conf, PetaxTCPServer.port);
	}

	public InetSocketAddress[] getUdpBindAddress(LaunchConfiguration conf)
			throws UnknownHostException {
		return getAddressAndPort(conf, PetaxUDPServer.port);
	}

	/**
	 * Nݒ肩MessengeroChUDP̃AhXƃ|[gԍ擾܂B
	 * oChȂꍇ͗vfȂ̔z񂪕Ԃ܂B
	 * @param conf Nݒ
	 * @param port |[gԍ
	 * @return oChAhXƃ|[gԍ̔zB
	 */
	private InetSocketAddress[] getAddressAndPort(LaunchConfiguration conf,
			int port) {
		InetAddress addr = getBindAddress(conf);
		InetSocketAddress[] addrs = new InetSocketAddress[1];
		addrs[0] = new InetSocketAddress(addr, port);
		return addrs;
	}

	/**
	 * @see petaclone.Messenger#newInstance()
	 */
	public Messenger newInstance() {
		return new PetaxMessenger();
	}

	public void quit() {
		udp.sendDelete();
		tcp.quit();
		udp.quit();
	}

	public void sendAbsenceMode(boolean isAbsence) {
		//TODO s݃[h
	}

	public void sendFind(boolean addOnly) {
		udp.sendFind(addOnly);
	}

	public void sendMessage(Message message, UniqueMember[] destinations,
			SendingResultListener listener) {
		Send send = new Send(message, destinations, listener);
		Thread th = new Thread(send);
		th.start();
	}

	public void sendRegister() {
		udp.sendRegister();
	}

	public void sendRename() {
		udp.sendRename();
	}

	/**
	 * @see petaclone.Messenger#setAppPreferences(petaclone.AppPreferences)
	 */
	public void setAppPreferences(AppPreferences pref) {
		this.pref = pref;
	}

	/**
	 * @see petaclone.Messenger#setConfiguration(petaclone.LaunchConfiguration)
	 */
	public void setConfiguration(LaunchConfiguration conf) {
		this.configuration = conf;
	}

	public void setServerListener(ServerListener listener) {
		serverListener = listener;
	}

	// vgRł̃T|[g@\

	public void sendOpenNotice(Message message) {
		//Jʒm̋@\
	}

	public void addDownloadListener(DownloadListener listener) {
		//
	}

	public void cancelSavingAttachment(Attachment attachment) {
		//
	}

	public void releaseAttachments(Message message) {
		//
	}

	public void removeDownloadListener(DownloadListener listener) {
		//
	}

	public void saveAttachment(Message message, Attachment attachment, File dest) {
		//
	}

	//

	/**
	 * AvP[Vݒ擾܂B
	 * @return AvP[Vݒ
	 */
	AppPreferences getAppPreferences() {
		return pref;
	}

	// ʒmn

	/**
	 * T[oXi[ɃG[ʒm܂B
	 * @param message G[bZ[W
	 * @param t ThrowableIuWFNg
	 */
	void notifyError(String message, Throwable t) {
		serverListener.onError(message, t);
	}

	/**
	 * T[oXi[ɃbZ[W̓ʒm܂B
	 * @param message bZ[W
	 */
	void notifyMessage(Message message) {
		serverListener.onMessage(message);
	}

	/**
	 * T[oXi[ɊJʒm̓ʒm܂B
	 * @param message Jʒm
	 */
	void notifyOpenNotice(Message message) {
		serverListener.onOpenNotice(message);
	}

	/**
	 * T[oXi[Ƀo[̍폜ʒm܂B
	 * @param member 폜郁o[
	 */
	void notifyDeleteMember(Member member) {
		serverListener.deleteMember(member);
	}

	/**
	 * T[oXi[Ƀo[̒ǉʒm܂B
	 * @param member ǉ郁o[
	 */
	void notifyAddMember(Member member) {
		serverListener.addMember(member);
	}

	/**
	 * T[oXi[Ƀo[̍XVʒm܂B
	 * @param member XV郁o[
	 */
	void notifyUpdateMember(Member member) {
		serverListener.updateMember(member);
	}

	/**
	 * f[^𑗐M܂B
	 * @param addr MAhX
	 * @param rawData Mf[^
	 * @throws IOException o͗O
	 */
	private void sendRaw(InetSocketAddress addr, byte[] rawData)
			throws IOException {
		if (addr == null || rawData == null) {
			return;
		}
		Socket sock = null;
		OutputStream os = null;
		try {
			logger.info("Send packet: addr=" + addr); //$NON-NLS-1$
			sock = new Socket();
			// 5bȓɃzXgȂƂ͗O
			sock.connect(addr, 5 * 1000);
			try {
				os = sock.getOutputStream();
				os.write(rawData);
				os.flush();

				// ̊mF
				InputStream is = null;
				try {
					is = sock.getInputStream();
					byte[] reply = tcp.readStream(is);
					if (reply == null) {
						//TODO O𓊂悤ɂ
					}
					// TODO RETURN XML̊mF

				} finally {
					if (is != null)
						is.close();
				}

			} finally {
				if (os != null)
					os.close();
			}
		} finally {
			if (sock != null)
				sock.close();
		}
	}

	/**
	 * MXbh<br>
	 * 1ʂM܂B
	 */
	private class Send implements Runnable {

		/** bZ[W */
		Message message;

		/** Đ */
		UniqueMember[] dest;

		/** MʃXi[ */
		SendingResultListener listener;

		/**
		 * RXgN^
		 * @param message M郁bZ[W
		 * @param destinations Đ
		 * @param listener MʃXi[
		 */
		Send(Message message, UniqueMember[] destinations,
				SendingResultListener listener) {
			this.message = message;
			this.dest = destinations;
			this.listener = listener;
		}

		public void run() {
			for (int i = 0; i < dest.length; i++) {
				InetSocketAddress addr = null;
				try {
					// f[^(idς̂Ŗ)
					byte[] buffer = makeData(dest[i]);
					// M
					/*
					 * InetAddress̎擾
					 * 1. UDPT[oInetAddress擾
					 * 2. Rs[^InetAddress擾
					 * 
					 * 2̏ꍇ̓TulbgقȂƒʏ͎擾łȂB
					 */
					addr = udp.getIPAddress(dest[i].getIdentifier());
					if (addr == null) {
						// Rs[^
						InetAddress iaddr = InetAddress.getByName(dest[i]
								.getComputerName());
						// ło^Ă
						dest[i].setIpAddress(iaddr.getHostAddress());

						addr = new InetSocketAddress(iaddr, PetaxTCPServer.port);

					} else {
						addr = new InetSocketAddress(addr.getAddress(),
								PetaxTCPServer.port);
					}
					sendRaw(addr, buffer);
					Destination d = new Destination(configuration.getId(),
							dest[i].getOsUserName(), dest[i].getComputerName(),
							dest[i].getUserName(), dest[i].getIpAddress(), 0,
							true);
					listener.onSendResult(d, null);
				} catch (Exception e) {
					logger.warn("Failed to send. " + addr, e); //$NON-NLS-1$
					// IOException, UnknownHostException, SocketTimeoutException
					Destination d = new Destination(configuration.getId(),
							dest[i].getOsUserName(), dest[i].getComputerName(),
							dest[i].getUserName(), dest[i].getIpAddress(), 0,
							false);
					listener.onSendResult(d, e);
				}
			}
			// MI
		}

		/**
		 * messagepPbgf[^쐬܂B
		 * @param um M惁o[
		 * @return oCi
		 */
		private byte[] makeData(UniqueMember um) {
			// Mgizmoid擾
			String gizmoid = udp.getGizmoid(um.getIdentifier());
			if (gizmoid == null) {
				logger.info("Could not get destination gizmoid."); //$NON-NLS-1$
				gizmoid = "undefined"; //$NON-NLS-1$
			}
			String uid = configuration.getProperty("gizmoid", ""); //$NON-NLS-1$ //$NON-NLS-2$
			uid += gizmoid;
			uid += System.currentTimeMillis();
			message.setExt("uid", uid); //$NON-NLS-1$

			RuntimeProperties rp = pref.getRuntime();

			PetaxPacket packet = new PetaxPacket();
			packet.gmversion = PetaxPacket.GMVERSION;
			packet.msgseq = 1;
			packet.id = PetaxPacket.TYPE_SENDMSG;
			packet.gizmoid = configuration.getProperty("gizmoid", ""); //$NON-NLS-1$ //$NON-NLS-2$
			packet.username = rp.getLoginName();
			packet.hostname = InetAddressUtils.getHostName();
			// \[UUIł8܂
			packet.displayname = Tool.substring(pref.getUserName(), 0, 8);
			if (bindAddress == null) {
				try {
					packet.ipaddress = InetAddress.getLocalHost()
							.getHostAddress();
				} catch (UnknownHostException e) {
					logger.error("Failed to get local ip address.", e); //$NON-NLS-1$
					return null;
				}
			} else {
				packet.ipaddress = bindAddress.getHostAddress();
			}
			packet.msg = message;
			byte[] xml;
			try {
				String xmlStr = packet.toXml();
				logger.debug(xmlStr);
				xml = xmlStr.getBytes(ENCODING);

			} catch (UnsupportedEncodingException e) {
				logger
						.error(
								"Failed to convert from XML string into byte array.", e); //$NON-NLS-1$
				return null;
			}
			byte[] whole = new byte[4 + xml.length];
			Tool.writeInt32BigEndian(whole, 0, xml.length);
			System.arraycopy(xml, 0, whole, 4, xml.length);

			return whole;
		}
	}
}
