package petaclone.petaroux;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.Tool;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import petaclone.InvalidDataFormatException;
import petaclone.Message;

/**
 * y^낤XXMLp[T[
 * @author Takayuki Shoji
 */
public class PetaxPacketHandler extends DefaultHandler {

	/**
	 * logger
	 */
	private static final Logger logger = Logger
			.getLogger(PetaxPacketHandler.class);

	/**
	 * sR[h
	 */
	private static final String EOL = System.getProperty("line.separator"); //$NON-NLS-1$

	/**
	 * ̃pPbg
	 */
	private PetaxPacket packet;

	/**
	 * ǂݍݒ̕
	 */
	private String buffer = ""; //$NON-NLS-1$

	/**
	 * ݂̗vf
	 */
	private int current = NONE;

	/** current̒萔:  */
	private static final int NONE = -1;

	/** current̒萔:  */
	private static final int GMVERSION = 0;

	/** current̒萔:  */
	private static final int MSGSEQ = 1;

	/** current̒萔:  */
	private static final int ID = 2;

	/** current̒萔:  */
	private static final int GIZMOID = 3;

	/** current̒萔:  */
	private static final int USERNAME = 4;

	/** current̒萔:  */
	private static final int HOSTNAME = 5;

	/** current̒萔:  */
	private static final int DISPLAYNAME = 6;

	/** current̒萔:  */
	private static final int IPADDRESS = 7;

	/** current̒萔:  */
	private static final int MSG = 8;

	/** current̒萔:  */
	private static final int TXT = 10;

	/**
	 * @see org.xml.sax.ContentHandler#characters(char[], int, int)
	 */
	public void characters(char[] ch, int start, int length)
			throws SAXException {

		if (current == NONE) {
			return;
		}
		bufferCharactors(new String(ch, start, length));
	}

	/**
	 * ׂēǂݏI܂Ńobt@O܂B
	 * @param string 
	 */
	private void bufferCharactors(String string) {
		if (this.buffer == null || this.buffer.length() == 0) {
			this.buffer = string;
		} else {
			this.buffer += string;
		}
	}

	/**
	 * ܂B
	 */
	private void handleString() {

		buffer = buffer.trim();

		switch (current) {
		case GMVERSION:
			packet.gmversion = buffer;
			break;
		case MSGSEQ:
			packet.msgseq = Integer.parseInt(buffer);
			break;
		case ID:
			packet.id = buffer;
			break;
		case GIZMOID:
			packet.gizmoid = buffer;
			break;
		case USERNAME:
			packet.username = buffer;
			break;
		case HOSTNAME:
			packet.hostname = buffer.toUpperCase();
			break;
		case DISPLAYNAME:
			packet.displayname = buffer;
			break;
		case IPADDRESS:
			packet.ipaddress = buffer;
			break;
		case MSG:
			packet.msg = parseMsg(buffer);
			break;
		default:
			break;
		}
	}

	/**
	 * msg^OURLGR[hꂽ܂B
	 * @param msg URLGR[hꂽbZ[WB
	 * @return msgnull܂͋󕶎̏ꍇnullAȊOMessageB
	 * @throws SAXException 
	 * @throws ParserConfigurationException 
	 * @throws IOException 
	 */
	private Message parseMsg(String msg) {
		if (msg == null || msg.length() == 0) {
			return null;
		}

		try {
			SAXParserFactory spf = SAXParserFactory.newInstance();
			InputSource src = new InputSource(new StringReader(msg));
			SAXParser parser = spf.newSAXParser();
			PetaxMessageHandler hndl = new PetaxMessageHandler();
			parser.parse(src, hndl);
			return hndl.msg;
		} catch (Exception e) {
			logger.error("Exception occurred while parsing xml.", e); //$NON-NLS-1$
			return null;
		}
	}

	/**
	 * msgXML͂nh
	 */
	private class PetaxMessageHandler extends DefaultHandler {

		/**
		 * ͌
		 */
		private Message msg;

		/**
		 * obt@
		 */
		private String msgBuffer = ""; //$NON-NLS-1$

		/**
		 * @see org.xml.sax.helpers.DefaultHandler#characters(char[], int, int)
		 */
		public void characters(char[] ch, int start, int length)
				throws SAXException {
			String str = new String(ch, start, length);
			if (this.msgBuffer == null || this.msgBuffer.length() == 0) {
				this.msgBuffer = str;
			} else {
				this.msgBuffer += str;
			}
		}

		/**
		 * @see org.xml.sax.helpers.DefaultHandler#startElement(java.lang.String, java.lang.String, java.lang.String, org.xml.sax.Attributes)
		 */
		public void startElement(String namespaceURI, String localName,
				String qName, Attributes atts) throws SAXException {
			if (qName.equalsIgnoreCase("message")) { //$NON-NLS-1$
				this.msg = new Message();
				this.msg.initialSealed();
				this.msg.id = (int) (System.currentTimeMillis() / 1000);
				this.msg.computerName = packet.hostname;
				this.msg.ipAddress = packet.ipaddress;
				this.msg.osUserName = packet.username;
				this.msg.userName = packet.displayname;
				this.msg.setCurrentTime();
				handleMsgAttributes(atts);

			} else if (qName.equalsIgnoreCase("txt")) { //$NON-NLS-1$
				current = TXT;
				msgBuffer = null;
				parseColor(msg, atts.getValue("color")); //$NON-NLS-1$

			} else {
				current = NONE;
			}
		}

		/**
		 * msg̑܂B
		 * @param atts 
		 */
		private void handleMsgAttributes(Attributes atts) {
			setExt("uid", atts); //$NON-NLS-1$
			setExt("face", atts); //$NON-NLS-1$
			setExt("skinurl", atts); //$NON-NLS-1$
			setExt("skin", atts); //$NON-NLS-1$
			setExt("displayname", atts); //$NON-NLS-1$
			setExt("gizmoid", atts); //$NON-NLS-1$
		}

		/**
		 * l擾Agɐݒ肵܂B
		 * @param key L[
		 * @param atts 
		 */
		private void setExt(String key, Attributes atts) {
			msg.setExt(key, atts.getValue(key));
		}

		/**
		 * @see org.xml.sax.helpers.DefaultHandler#endElement(java.lang.String, java.lang.String, java.lang.String)
		 */
		public void endElement(String namespaceURI, String localName,
				String qName) throws SAXException {

			if (current == TXT) {

				if (msgBuffer != null) {
					StringBuffer sb = new StringBuffer();
					sb.append(msgBuffer);
					Tool.replaceAll(sb, "\r\n", "\n"); //$NON-NLS-1$ //$NON-NLS-2$
					Tool.replaceAll(sb, "\r", "\n"); //$NON-NLS-1$ //$NON-NLS-2$
					Tool.replaceAll(sb, "\n", EOL); //$NON-NLS-1$
					msg.message = sb.toString();
				}

			} else if (qName.equals("message")) { //$NON-NLS-1$
				PetaxPacketHandler.this.packet.msg = msg;
			}

			this.msgBuffer = ""; //$NON-NLS-1$
			current = NONE;
		}

		/**
		 * Frgb
		 * @param target Fݒ肳郁bZ[W
		 * @param str RGB256K16i(RGBe2)
		 */
		private void parseColor(Message target, String str) {

			if (str == null || str.length() != 6) {
				return;
			}
			if (str.equals("000000")) { //$NON-NLS-1$
				return;
			}

			int[] color = new int[3];
			color[0] = Integer.parseInt(str.substring(0, 1), 16);
			color[1] = Integer.parseInt(str.substring(2, 3), 16);
			color[2] = Integer.parseInt(str.substring(4, 5), 16);
			target.color = color;
		}
	}

	/**
	 * @see org.xml.sax.ContentHandler#endElement(String, String, String)
	 */
	public void endElement(String namespaceURI, String localName, String qName)
			throws SAXException {

		// ǂݍݍς݂̏̕
		handleString();
		this.buffer = ""; //$NON-NLS-1$
		current = NONE;
	}

	/**
	 * @see org.xml.sax.ContentHandler#startElement(String, String, String, Attributes)
	 */
	public void startElement(String namespaceURI, String localName,
			String qName, Attributes atts) throws SAXException {
		if (qName.equalsIgnoreCase("IPMSGXmlData")) { //$NON-NLS-1$
			this.packet = new PetaxPacket();
		} else if (qName.equalsIgnoreCase("gmversion")) { //$NON-NLS-1$
			current = GMVERSION;
		} else if (qName.equalsIgnoreCase("msgseq")) { //$NON-NLS-1$
			current = MSGSEQ;
		} else if (qName.equalsIgnoreCase("id")) { //$NON-NLS-1$
			current = ID;
		} else if (qName.equalsIgnoreCase("gizmoid")) { //$NON-NLS-1$
			current = GIZMOID;
		} else if (qName.equalsIgnoreCase("username")) { //$NON-NLS-1$
			current = USERNAME;
		} else if (qName.equalsIgnoreCase("hostname")) { //$NON-NLS-1$
			current = HOSTNAME;
		} else if (qName.equalsIgnoreCase("displayname")) { //$NON-NLS-1$
			current = DISPLAYNAME;
		} else if (qName.equalsIgnoreCase("ipaddress")) { //$NON-NLS-1$
			current = IPADDRESS;
		} else if (qName.equalsIgnoreCase("msg")) { //$NON-NLS-1$
			current = MSG;
		} else {
			current = NONE;
		}
	}

	/**
	 * Returns the result.
	 * @return List l: Message
	 */
	public PetaxPacket getResult() {
		return packet;
	}

	/**
	 * ̓Xg[XML͂܂B
	 * @param is Xg[
	 * @return pPbgf[^
	 */
	static PetaxPacket parse(InputStream is) {
		SAXParserFactory spf = SAXParserFactory.newInstance();

		InputSource src = new InputSource(is);

		SAXParser parser;
		try {
			parser = spf.newSAXParser();
			PetaxPacketHandler hndl = new PetaxPacketHandler();
			parser.parse(src, hndl);
			return hndl.getResult();

		} catch (Exception e) {
			throw new InvalidDataFormatException("Could not parse packet.", e); //$NON-NLS-1$
		}
	}

	/**
	 * eXg
	 * @param args
	 * @throws FileNotFoundException 
	 */
	public static void main(String[] args) throws FileNotFoundException {
		FileInputStream is = new FileInputStream("sendmsg.xml"); //$NON-NLS-1$
		PetaxPacket packet = parse(is);
		logger.debug(packet);
		logger.debug(packet.toXml());

		is = new FileInputStream("return.xml"); //$NON-NLS-1$
		packet = parse(is);
		logger.debug(packet);
		logger.debug(packet.toXml());

		is = new FileInputStream("sendbroad.xml"); //$NON-NLS-1$
		packet = parse(is);
		logger.debug(packet);
		logger.debug(packet.toXml());

		is = new FileInputStream("recvbroad.xml"); //$NON-NLS-1$
		packet = parse(is);
		logger.debug(packet);
		logger.debug(packet.toXml());

		String id = PetaxPacket.generateGizmoid();
		logger.debug(id);
		logger.debug("isGizmoid:" + PetaxPacket.isGizmoid(id)); //$NON-NLS-1$
	}
}
