package petaclone.petaroux;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.BindException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NoRouteToHostException;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.HashMap;

import jp.takayuki.logging.Logger;
import jp.takayuki.tools.InetAddressUtils;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Tool;
import petaclone.AppPreferences;
import petaclone.InvalidDataFormatException;
import petaclone.Member;
import petaclone.RuntimeProperties;

/**
 * y^낤XvgRUDPT[o
 * 
 * @author Takayuki Shoji
 */
public class PetaxUDPServer implements Runnable {

	/**
	 * logger
	 */
	private static final Logger logger = Logger.getLogger(PetaxUDPServer.class);

	/** fobOtO */
	private static final boolean DEBUG = logger.isDebugEnabled();

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/** f[^O\Pbg */
	private DatagramSocket socket = null;

	/** |[gԍ */
	static final int port = 20205;

	/** bZW[ */
	private PetaxMessenger messenger;

	/** NȂtrue */
	private boolean isAlive;

	/**
	 * UDPbZ[W̎Mpobt@TCY
	 */
	private final static int MAX_MESSAGE_LENGTH = 1024;

	/**
	 * o[ǗpB<br>
	 * L[: Member#getIdentifier()<br>
	 * l: PetaxPacket
	 */
	private HashMap members;

	/**
	 * w肵o[gizmoid擾܂B
	 * 
	 * @param id {@link Member#getIdentifier()}
	 * @return gizmoidBo[ȂꍇnullB
	 */
	String getGizmoid(String id) {
		PetaxPacket packet;
		synchronized (members) {
			packet = (PetaxPacket) members.get(id);
		}
		if (packet == null) {
			return null;
		}
		return packet.gizmoid;
	}

	/**
	 * RXgN^
	 * @param messenger bZW[
	 */
	PetaxUDPServer(PetaxMessenger messenger) {
		this.messenger = messenger;
		this.members = new HashMap();
	}

	/**
	 * T[oĂ邩𒲂ׂ܂B
	 * @return Ăꍇtrue
	 */
	synchronized boolean getAlive() {
		return isAlive;
	}

	/**
	 * o[NĂ邩ׂ܂B
	 * @param key o[ʎq
	 * @return NĂꍇtrue
	 */
	boolean isAliveMember(String key) {
		return members.containsKey(key);
	}

	/**
	 * T[oJn
	 * @see java.lang.Runnable#run()
	 */
	public void run() {

		byte[] buffer = new byte[MAX_MESSAGE_LENGTH];
		DatagramPacket packet = new DatagramPacket(buffer, buffer.length);
		try {
			while (isAlive) {
				try {
					socket.receive(packet);
				} catch (IOException e) {
					if (isAlive) {
						logger.error("I/O exception occurred " //$NON-NLS-1$
								+ "while waiting or receiving.", e); //$NON-NLS-1$
						messenger
								.notifyError(
										//
										mm
												.get("PetaxUDPServer.IOException.receive"), e); //$NON-NLS-1$
						continue;
					}
					break;
				}

				// ̃u[hLXgɔȂ悤ɂ
				if (!messenger.configuration.acceptSelfBroadcast()
						&& InetAddressUtils.isLocalAddress(packet.getAddress())) {
					if (DEBUG)
						logger.debug("Broadcast from local address."); //$NON-NLS-1$
					continue;
				}

				// obt@̒𒲐߂AUDPMessage쐬
				byte[] buffer2 = new byte[packet.getLength()];
				System.arraycopy(buffer, 0, buffer2, 0, buffer2.length);

				ByteArrayInputStream is = new ByteArrayInputStream(buffer2);
				PetaxPacket msg = null;
				try {
					msg = PetaxPacketHandler.parse(is);
				} catch (InvalidDataFormatException e) {
					logger.warn("Format of received data is invalid.", e); //$NON-NLS-1$
					continue;
				}
				msg.address = new InetSocketAddress(packet.getAddress(), packet
						.getPort());

				if (msg.id.equals(PetaxPacket.TYPE_SENDBROAD)) {
					// o^
					registerMember(msg);
					// ԓ
					InetSocketAddress[] addrs = new InetSocketAddress[] { msg.address };
					send(REPLY, addrs);

				} else if (msg.id.equals(PetaxPacket.TYPE_RECVBROAD)) {
					// o^
					registerMember(msg);

				} else if (msg.id.equals(PetaxPacket.TYPE_GIZMOEND)) {
					// 폜
					deleteMember(msg);

				} else {
					// Ή()
					logger.warn("Unsupported format=" + msg.toString()); //$NON-NLS-1$
				}
			}
		} finally {
			isAlive = false;
			close();
		}
	}

	/**
	 * o[Ǘ烁o[폜܂B
	 * @param msg pPbgf[^
	 */
	private void deleteMember(PetaxPacket msg) {
		Member member = getMember(msg);

		String id = member.getIdentifier();

		synchronized (members) {
			members.remove(id);
		}

		messenger.notifyDeleteMember(member);
	}

	/**
	 * o[ǗɃo[ǉ܂B
	 * @param msg pPbgf[^
	 */
	private void registerMember(PetaxPacket msg) {

		String id = msg.getIdentifier();

		boolean update = false;
		synchronized (members) {

			PetaxPacket registered = (PetaxPacket) members.get(id);
			if (registered != null) {
				if (registered.isSameMember(msg)) {
					return;
				}
				update = true;
			}
			// o^Ēʒm
			members.put(id, msg);
		}

		Member member = getMember(msg);
		if (update) {
			messenger.notifyUpdateMember(member);
		} else {
			messenger.notifyAddMember(member);
		}
	}

	/**
	 * PetaxPacketMemberɕϊ܂B
	 * @param packet pPbg
	 * @return Member
	 */
	private Member getMember(PetaxPacket packet) {
		String launchId = messenger.getId();
		return packet.toMember(launchId);
	}

	/**
	 * oCh܂B
	 * @throws BindException
	 * @throws SocketException
	 * @throws IOException
	 */
	void bind() throws BindException, SocketException, IOException {
		this.socket = new DatagramSocket(port, messenger.bindAddress);
		isAlive = true;
		logger.info("Bound: addr=" + messenger.bindAddress + ", port=" + port); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * \PbgN[Y܂B
	 */
	private void close() {
		if (socket != null && !socket.isClosed())
			socket.close();
	}

	/**
	 * Method send.
	 * @param messageType bZ[W^CvB
	 * 	DELETE, 
	 * 	REGISTER, 
	 * 	REPLY
	 * @param addrs M
	 */
	private void send(int messageType, InetSocketAddress[] addrs) {
		// M镶p
		byte[] msg = getString(messageType);
		if (msg == null) {
			return;
		}
		for (int i = 0; i < addrs.length; i++) {
			// UDPőM
			try {
				DatagramPacket packet = new DatagramPacket(msg, msg.length,
						addrs[i]);
				socket.send(packet);
			} catch (NoRouteToHostException e) {
				e.printStackTrace();
				continue;
			} catch (IOException e) {
				e.printStackTrace();
				continue;
			}
		}
	}

	/**
	 * messageType萔
	 * NAČ
	 */
	static final int REGISTER = 1;

	/**
	 * messageType萔
	 * I
	 */
	static final int DELETE = 2;

	/**
	 * messageType萔
	 * ԓ
	 */
	static final int REPLY = 3;

	/**
	 * Method getString.Mp̕擾܂B
	 * @param mType bZ[W^Cv 
	 * 	DELETE, 
	 * 	REGISTER, 
	 * 	REPLY
	 * @return byte[] Mp̃f[^
	 */
	private byte[] getString(int mType) {
		AppPreferences pref = messenger.getAppPreferences();
		RuntimeProperties rp = pref.getRuntime();

		PetaxPacket packet = new PetaxPacket();
		packet.gmversion = PetaxPacket.GMVERSION;
		packet.msgseq = 0;
		packet.gizmoid = messenger.configuration.getProperty("gizmoid", "bug"); //$NON-NLS-1$ //$NON-NLS-2$
		packet.username = rp.getLoginName();
		packet.hostname = InetAddressUtils.getHostName();
		// \[UUIł8܂
		packet.displayname = Tool.substring(pref.getUserName(), 0, 8);
		if (messenger.bindAddress == null) {
			try {
				packet.ipaddress = InetAddress.getLocalHost().getHostAddress();
			} catch (UnknownHostException e) {
				logger.error("Cannot get local IP address.", e); //$NON-NLS-1$
				return null;
			}
		} else {
			packet.ipaddress = messenger.bindAddress.getHostAddress();
		}

		switch (mType) {
		case REGISTER:
			packet.id = PetaxPacket.TYPE_SENDBROAD;
			break;
		case REPLY:
			packet.id = PetaxPacket.TYPE_RECVBROAD;
			break;
		case DELETE:
			packet.id = PetaxPacket.TYPE_GIZMOEND;
			break;
		default:
			throw new IllegalArgumentException("Invalid mType=" + mType); //$NON-NLS-1$
		}
		try {
			return packet.toXml().getBytes(PetaxMessenger.ENCODING);
		} catch (UnsupportedEncodingException e) {
			logger.error(e);
			return null;
		}
	}

	/**
	 * o[ʎqɑ΂AhXԂ܂B
	 * ݂ȂꍇnullԂ܂B
	 * @param id o[ʎq
	 * @return AhX
	 */
	InetSocketAddress getIPAddress(String id) {
		PetaxPacket udpMsg = null;
		synchronized (members) {
			if (!members.containsKey(id)) {
				return null;
			}
			udpMsg = (PetaxPacket) members.get(id);
		}

		return udpMsg.address;
	}

	/**
	 * o^
	 */
	void sendRegister() {
		AppPreferences pref = messenger.getAppPreferences();
		InetSocketAddress[] addrs = messenger.configuration
				.getBroadcastAddresses(pref, (InetSocketAddress) socket
						.getLocalSocketAddress());

		send(REGISTER, addrs);
	}

	/**
	 * 폜
	 */
	void sendDelete() {
		AppPreferences pref = messenger.getAppPreferences();
		InetSocketAddress[] addrs = messenger.configuration
				.getBroadcastAddresses(pref, (InetSocketAddress) socket
						.getLocalSocketAddress());
		send(DELETE, addrs);
	}

	/**
	 * Č
	 * @param addOnly ̃XgNAȂꍇtrue
	 */
	void sendFind(boolean addOnly) {
		AppPreferences pref = messenger.getAppPreferences();
		InetSocketAddress[] addrs = messenger.configuration
				.getBroadcastAddresses(pref, (InetSocketAddress) socket
						.getLocalSocketAddress());
		if (!addOnly) {
			synchronized (members) {
				members.clear();
			}
		}

		send(REGISTER, addrs);
	}

	/**
	 * [U[̕ύXʒmB
	 * 폜Ƃɂ͌ÂO͕svȂ̂łłɐݒ肳ꂽOœo^sB
	 * ČƓBAǉ݂̂ŌB
	 */
	void sendRename() {
		sendFind(true);
	}

	/**
	 * T[oɏIʒm܂B
	 */
	void quit() {
		this.isAlive = false;
		// Iɕ
		close();
	}
}
