package petaclone.swt;

import java.io.File;
import java.util.Properties;

import jp.takayuki.swt.util.SWTUtils;
import jp.takayuki.tools.InetAddressUtils;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Os;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Shell;

import petaclone.AppPreferences;
import petaclone.ControlCreator;
import petaclone.LaunchConfiguration;
import petaclone.LaunchConfigurations;
import petaclone.Messenger;
import petaclone.MessengerFacade;

/**
 * N\̐ݒ̕ҏW_CAO
 * @author Takayuki Shoji
 */
public class SWTLaunchConfigEditDialog {

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/** bZW[ */
	private MessengerFacade facade;

	/** ݒ */
	private AppPreferences pref;

	/** CX^XƂ̐ݒ */
	private LaunchConfiguration conf;

	/** Rg[ */
	private SWTControlCreator creator;

	/** Shell */
	private Shell shell;

	/** ݒ_CAO */
	private LaunchConfigCallback configDialog;

	/** vgRIpComposite */
	private Composite composite;

	/** ǉ[h */
	private boolean addMode;

	/**
	 * RXgN^
	 * @param pref AvP[Vݒ
	 * @param facade bZW[t@T[h
	 * @param configDialog ݒ_CAO
	 * @param conf 삷NݒBnull̏ꍇ͒ǉ[hB
	 * @param parent eEBWFbg
	 */
	SWTLaunchConfigEditDialog(AppPreferences pref, MessengerFacade facade,
			LaunchConfigCallback configDialog, Shell parent,
			LaunchConfiguration conf) {

		this.facade = facade;
		this.pref = pref;
		if (pref.getPolicy() == null) {
			throw new IllegalArgumentException(
					"policy in pref must not be null."); //$NON-NLS-1$
		}
		this.conf = conf;
		this.configDialog = configDialog;

		int onTop = 0;
		if (parent == null && Os.isWindows()) {
			onTop = SWT.ON_TOP;
		}
		shell = new Shell(parent, SWT.DIALOG_TRIM | SWT.APPLICATION_MODAL
				| onTop);
		if (conf == null) {
			initProtocolSelecter(shell);
			addMode = true;
		} else {
			if (parent != null)
				shell.setImage(parent.getImage());
			// ҏW̏ꍇ
			initWidgets(shell);
		}

		if (parent == null) {
			shell.setText(mm.get("SWTLaunchConfigEditDialog.wizard")); //$NON-NLS-1$
		} else {
			shell.setText(mm.get("SWTLaunchConfigEditDialog.launchConfig")); //$NON-NLS-1$
		}

		shell.pack(true);
		SWTUtils.moveToCenter(shell);
		shell.open();
	}

	/**
	 * @return ̃_CAOShell
	 */
	public Shell getShell() {
		return shell;
	}

	/**
	 * vgRIp̃EBWFbg܂B
	 * @param parent shell
	 */
	private void initProtocolSelecter(Composite parent) {
		parent.setLayout(new GridLayout(1, false));

		// LaunchConfiguration̐
		String id = LaunchConfigurations.generateId();

		// policyN\tH_擾LaunchConfiguration̈ʒu肷
		File path = LaunchConfiguration.makePath(pref.getPolicy()
				.getLaunchConfigurationPath(), id);

		this.conf = new LaunchConfiguration(new Properties(), path);
		this.conf.setId(id);

		// EBWFbg̐
		this.composite = new Composite(parent, SWT.NONE);
		GridLayout layout = new GridLayout(1, false);
		this.composite.setLayout(layout);

		// ControlCreatorpcomposite
		Composite creatorArea = new Composite(this.composite, SWT.NONE);
		GridLayout creatorAreaLayout = new GridLayout(2, false);
		creatorArea.setLayout(creatorAreaLayout);

		// ControlCreator̐
		this.creator = new SWTControlCreator(conf.getProperties());
		creator.setParent(creatorArea);

		// R|[lgI
		creator
				.createLabel(mm
						.get("SWTLaunchConfigEditDialog.selectComponent")); //$NON-NLS-1$
		String[] values = facade.getInstalledMessengerNames();
		String[] propertyValues = facade.getInstalledMessengerSignatures();

		creator.createComboBox(
				LaunchConfiguration.PROTOCOL,//
				mm.get("SWTLaunchConfigEditDialog.component"), //$NON-NLS-1$
				values, propertyValues,
				LaunchConfigurations.DEFAULT_MESSENGER_SIGNATURE, true);
		creator.createSeparator();

		createButtonsArea(composite, "SWTLaunchConfigEditDialog.next", //$NON-NLS-1$
				new NextButton(), "SWTLaunchConfigEditDialog.cancel", //$NON-NLS-1$
				new CancelButton(), values.length > 0);
	}

	/**
	 * ҏWpEBWFbg܂B
	 * @param parent eEBWFbg
	 */
	private void initWidgets(Composite parent) {
		parent.setLayout(new GridLayout(2, true));

		// ControlCreator̐(ҏŴƂnull)
		if (this.creator == null) {
			this.creator = new SWTControlCreator(conf.getProperties());
		}

		// ʐݒ
		Composite commonArea = new Composite(parent, SWT.NONE);
		commonArea
				.setLayoutData(new GridData(GridData.VERTICAL_ALIGN_BEGINNING));
		commonArea.setLayout(new GridLayout(2, false));
		creator.setParent(commonArea);
		createCommonConfiguration(creator);

		// 𓯂ɂ
		commonArea.pack();
		int width = commonArea.getSize().x;
		creator.setWidthHint(width);

		// vgRݒ
		Group protocolGroup = new Group(parent, SWT.SHADOW_ETCHED_IN);
		protocolGroup.setText(mm
				.get("SWTLaunchConfigEditDialog.componentConfiguration")); //$NON-NLS-1$
		protocolGroup.setLayoutData(new GridData(GridData.FILL_BOTH));
		protocolGroup.setLayout(new GridLayout(2, false));
		creator.setParent(protocolGroup);
		createProtocolConfiguration(creator);

		createButtonsArea(parent, "SWTLaunchConfigEditDialog.ok", //$NON-NLS-1$
				new OkButton(), "SWTLaunchConfigEditDialog.cancel", //$NON-NLS-1$
				new CancelButton(), true);
	}

	/**
	 * _CAÕ{^𐶐܂B
	 * 
	 * @param parent eEBWFbg
	 * @param label1 {^1̃x̃L[
	 * @param listener1 {^1̃Xi[
	 * @param label2 {^2̃x̃L[
	 * @param listener2 {^2̃Xi[
	 * @param enableOK {^1Lɂꍇtrue
	 */
	private void createButtonsArea(Composite parent, String label1,
			SelectionListener listener1, String label2,
			SelectionListener listener2, boolean enableOK) {
		// {^
		Composite buttonsArea = new Composite(parent, SWT.NONE);
		// E
		GridData parentAreaData = new GridData(GridData.HORIZONTAL_ALIGN_END);
		parentAreaData.horizontalSpan = 2;
		buttonsArea.setLayoutData(parentAreaData);

		// {^̈̃CAEg
		RowLayout parentLayout = new RowLayout(SWT.HORIZONTAL);
		parentLayout.spacing = 5;
		parentLayout.pack = false;
		buttonsArea.setLayout(parentLayout);

		// OK
		Button okButton = new Button(buttonsArea, SWT.PUSH);
		okButton.setText(mm.get(label1));
		okButton.addSelectionListener(listener1);
		okButton.setEnabled(enableOK);

		// LZ
		Button cancelButton = new Button(buttonsArea, SWT.PUSH);
		cancelButton.setText(mm.get(label2));
		cancelButton.addSelectionListener(listener2);

		// ftHg̃{^ݒ
		shell.setDefaultButton(okButton);

	}

	/**
	 * vgRݒ𐶐܂B
	 * @param creator Rg[NG[^
	 */
	private void createProtocolConfiguration(ControlCreator creator) {
		// bZW[擾
		Messenger msgr = facade.getInstalledMessenger(conf.getProtocol());
		msgr.createConfigControl(creator);
	}

	/**
	 * ʐݒ𐶐܂B
	 * @param creator Rg[NG[^
	 */
	private void createCommonConfiguration(SWTControlCreator creator) {
		// ݒ̖O
		creator.createText(LaunchConfiguration.NAME,//
				mm.get("SWTLaunchConfigEditDialog.configurationName"), //$NON-NLS-1$
				mm.get("SWTLaunchConfigEditDialog.defaultConfigurationName"), //$NON-NLS-1$
				false, null);
		// N
		creator.createCheckBox(LaunchConfiguration.IS_VALID,//
				mm.get("SWTLaunchConfigEditDialog.enable"), //$NON-NLS-1$
				true);
		// oChAhX
		// \l
		String[] local = InetAddressUtils.getLocalAddressesString();
		String[] values = new String[local.length + 1];
		values[0] = mm.get("SWTLaunchConfigEditDialog.allLocalAddresses"); //$NON-NLS-1$
		for (int i = 0; i < local.length; i++) {
			values[i + 1] = mm.get("SWTLaunchConfigEditDialog.localAddresses", //$NON-NLS-1$
					new Object[] { local[i], Integer.toString(i) });
		}
		// ۑl
		String[] propertyValues = new String[local.length + 1];
		propertyValues[0] = ""; //$NON-NLS-1$
		for (int i = 0; i < local.length; i++) {
			propertyValues[i + 1] = Integer.toString(i);
		}
		creator.createComboBox(LaunchConfiguration.BIND_ADDRESS,//
				mm.get("SWTLaunchConfigEditDialog.bindAddress"), //$NON-NLS-1$
				values, propertyValues, "", false); //$NON-NLS-1$

		//
		creator.createCheckBox(LaunchConfiguration.ACCEPT_SELF_BROADCAST,//
				mm.get("SWTLaunchConfigEditDialog.showMeInSendDialog"), //$NON-NLS-1$
				true);

		// u[hLXg
		creator.createBroadcastGroup(LaunchConfiguration.USE_DEFAULT_BROADCAST,
				LaunchConfiguration.BROADCAST,
				AppPreferences.LIMITED_BROADCAST,
				AppPreferences.DIRECTED_BROADCAST);
	}

	/**
	 * 
	 */
	private class NextButton extends SelectionAdapter {
		public void widgetSelected(SelectionEvent event) {
			// ۑ
			creator.saveProperties();
			shell.setVisible(false);
			// Rg[̃[X
			creator.releaseControls();
			// vgRIpEBWFbg̔j
			composite.dispose();
			//ҏWpEBWFbg
			initWidgets(shell);
			shell.pack();
			SWTUtils.moveToCenter(shell);
			shell.setVisible(true);
			shell.setActive();
		}
	}

	/**
	 * OK
	 */
	private class OkButton extends SelectionAdapter {
		public void widgetSelected(SelectionEvent event) {
			// vpeB[̕ۑ
			creator.saveProperties();
			// ꗗXV(VK̏ꍇ͒ǉ)
			if (addMode) {
				configDialog.addLaunchConfiguration(conf);
			} else {
				configDialog.updateLaunchConfiguration(conf);
			}
			shell.close();
		}
	}

	/**
	 * LZ
	 */
	private class CancelButton extends SelectionAdapter {
		public void widgetSelected(SelectionEvent event) {
			shell.close();
		}
	}

	/**
	 * Nݒ_CAǑʂ󂯎C^[tF[X
	 */
	public interface LaunchConfigCallback {
		/**
		 * Nݒǉ܂B
		 * @param conf Nݒ
		 */
		void addLaunchConfiguration(LaunchConfiguration conf);

		/**
		 * NݒXV܂B
		 * @param conf Nݒ
		 */
		void updateLaunchConfiguration(LaunchConfiguration conf);
	}
}
