package petaclone.swt;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;
import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import jp.takayuki.logging.Logger;
import jp.takayuki.swt.util.CocoaUIEnhancer;
import jp.takayuki.swt.util.ListenerDelegate;
import jp.takayuki.swt.util.SWTMacOSXMenu;
import jp.takayuki.swt.util.SWTMenuUtils;
import jp.takayuki.swt.util.SWTUtils;
import jp.takayuki.tools.GtkRestartHelper;
import jp.takayuki.tools.InetAddressUtils;
import jp.takayuki.tools.MacOSXRestartHelper;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Os;
import jp.takayuki.tools.RestartHelper;
import jp.takayuki.tools.Tool;
import jp.takayuki.tools.Version;
import jp.takayuki.tools.WindowsRestartHelper;

import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTError;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.ShellAdapter;
import org.eclipse.swt.events.ShellEvent;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.program.Program;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.Monitor;
import org.eclipse.swt.widgets.Shell;
import org.monazilla.migemo.Migemo;

import petaclone.AppPreferences;
import petaclone.ApplicationException;
import petaclone.ApplicationUpdateDownloaded;
import petaclone.BaseMain;
import petaclone.DeathGazer;
import petaclone.EventBroker;
import petaclone.Events;
import petaclone.HotkeyEventBridge;
import petaclone.IPAddressWatcher;
import petaclone.LaunchConfiguration;
import petaclone.LaunchConfigurations;
import petaclone.LaunchException;
import petaclone.Message;
import petaclone.MessageStorageIOException;
import petaclone.Policy;
import petaclone.RuntimeProperties;
import petaclone.SoundPlayer;
import petaclone.SoundPlayerImpl;
import petaclone.StartupException;
import petaclone.ToolExecutor;
import petaclone.logic.EndAutoAbsence;
import petaclone.logic.NewMessageReceived;
import petaclone.logic.StartAutoAbsence;
import petaclone.swt.SWTLaunchConfigEditDialog.LaunchConfigCallback;
import petaclone.swt.logic.SWTCheckForUpdates;
import petaclone.swt.logic.SWTDisplayWindow;
import petaclone.swt.logic.SWTFocusMainWindow;
import petaclone.swt.logic.SWTMessageSearch;
import petaclone.swt.logic.SWTMinimizeAllWindows;
import petaclone.swt.logic.SWTNewMessageReceived;
import petaclone.swt.logic.SWTOpenBlankMessageWindow;
import petaclone.swt.logic.SWTOpenLogViewer;
import petaclone.swt.logic.SWTOpenNewStickiesSet;
import petaclone.swt.logic.SWTOpenNoticeReceived;
import petaclone.swt.logic.SWTQualityFeedbackSendAction;
import petaclone.swt.logic.UnexpectedErrorOccured;

/**
 * CEBhE
 * @author Takayuki Shoji
 */
public class SWTMainWindow extends BaseMain implements ImageProvider,
		LaunchConfigCallback {

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/**
	 * logger
	 */
	private static Logger logger;

	/**
	 * Shutdown hookXbhmainXbh̏IĎ邽߂̃tOB
	 * mainXbh̏IIꍇtrueɂB
	 */
	static volatile boolean exit;

	/**
	 * IȏI\tO
	 * <p>
	 * j[̏IAOIt̂Ƃtrueɐݒ肷B
	 * shelĺ~{^̓CxgXi[ڌĂ΂邽߁ÃftHglݒ肵ĂB
	 */
	private boolean explicitlyQuit;

	/**
	 * Mac OS XCmd+QŏIDisplay SWT.Close2xĂ΂錻ۂɑΉ邽߂̃tOB
	 */
	private boolean shellDisposed;

	/**
	 * DisplayIuWFNg
	 */
	private Display display;

	/**
	 * CEBhEShell
	 */
	private Shell shell;

	/**
	 * M_CAO
	 */
	SWTSendDialog sendDialog;

	/**
	 * ۑbZ[WEBhE
	 */
	SWTMessageStorage savedMessageWindow;

	/**
	 * \̃bZ[WEBhE
	 */
	private SWTMessageWindowManager desktopMessages = new SWTMessageWindowManagerImpl();

	/**
	 * 摜\[X
	 */
	private SharedProgramIcons icons;

	/**
	 * TEh
	 */
	private SoundPlayer soundPlayer;

	/**
	 * ^XNgC
	 */
	private SWTTaskTray tray;

	/**
	 * s݃[hĎ
	 */
	private SWTPresenceWatcher absenceWatcher;

	/**
	 * T[oXi[
	 */
	private SWTServerListener serverListener;

	/**
	 * zbgL[Ǘ
	 */
	private HotkeyEventBridge hotkeyEventBridge;

	/**
	 * Cxg
	 */
	private Actions actions;

	/**
	 * EBhE؂ւ
	 */
	private SWTWindowSwitcher switcher;

	/**
	 * bZ[W_CAO
	 */
	private SWTMessageSearch messageSearch;

	/**
	 * O[oV[gJbgnh
	 */
	private SWTAppGlobalShortcut appGlobalShortcut;

	/**
	 * xLZ
	 */
	private SWTTransparencyCanceller transparencyCanceller;

	/**
	 * IPAhX̕ύXƃX[v̕AĎXPW[
	 */
	private ScheduledExecutorService scheduler;

	/**
	 * RXgN^BAopen()Kv܂B
	 */
	private SWTMainWindow() {
		super();
	}

	/**
	 * 摜\[X[h܂B
	 */
	private void loadImages() {
		assert icons == null;
		assert display != null;
		icons = new SharedProgramIcons(this.display);
	}

	/**
	 * 炩߃LbVĂ摜o܂B
	 */
	public Image getImage(String key) {
		return icons.getIcon(key);
	}

	/**
	 * DisplaỹCxgXi[
	 */
	private class DisplayListener implements Listener {
		public void handleEvent(Event event) {
			switch (event.type) {
			case SWT.Close:
				logger.debug("Display SWT.Close start."); //$NON-NLS-1$
				explicitlyQuit = true;
				if (!shellDisposed) {
					shell.close();
				} else {
					logger.info("Display SWT.Close: shell is already disposed by previous event."); //$NON-NLS-1$
				}
				logger.debug("Display SWT.Close end."); //$NON-NLS-1$
				break;
			case SWT.Dispose:
				logger.debug("Display SWT.Dispose Event."); //$NON-NLS-1$
				break;
			default:
				logger.error("Unknown event." + event.type); //$NON-NLS-1$
				break;
			}
		}
	}

	/**
	 * EBhEJAT[oJn܂B
	 */
	private void open() {
		AppPreferences pref = getAppPreferences();

		// EBhȄ
		Display.setAppName(BaseMain.APP_NAME);
		//DeviceData data = new DeviceData();
		//data.tracking = true;
		//this.display = new Display(data);
		//Sleak sleak = new Sleak();
		//sleak.open();
		this.display = Display.getDefault();

		// VXeIEOIt
		DisplayListener listener = new DisplayListener();
		display.addListener(SWT.Close, listener);
		display.addListener(SWT.Dispose, listener);

		// 摜\[X̃[h
		loadImages();

		this.soundPlayer = new SoundPlayerImpl(pref);

		// CEBhE̐
		shell = new Shell(display, SWT.TITLE | SWT.CLOSE | SWT.MIN | SWT.RESIZE);
		if (!Os.isMacOSX())
			shell.setImage(getImage("appl.png")); //$NON-NLS-1$

		// shell.close()̃Xi[
		ShellAction sa = new ShellAction();
		shell.addShellListener(sa);
		shell.addDisposeListener(sa);

		// j[\
		actions = new Actions();
		switcher = new SWTWindowSwitcher(desktopMessages);
		createMenuBar(actions);

		// EBhETCY肷
		setMainWindowSize();

		// EBhËʒuݒ
		restoreWindowLocation(pref);

		// EBhEJ
		setWindowTitle(mm.get("SWTMainWindow.title.startup")); //$NON-NLS-1$
		shell.setEnabled(false);
		shell.open();

		// O[oV[gJbgnh
		appGlobalShortcut = new SWTAppGlobalShortcut(getEventBroker());

		// xLZ
		transparencyCanceller = new SWTTransparencyCanceller(desktopMessages);

		// M_CAȌ
		sendDialog = new SWTSendDialog(this, pref, messengerFacade,
				getMessageLogger(), getEventBroker());

		// ۊǔ̏ƊJĂbZ[WEBhE𕜌
		savedMessageWindow = new SWTMessageStorage(this, pref, messengerFacade,
				getMessageLogger(), getEventBroker());

		// OIOɕۗڂꍇ
		boolean hasPending = (savedMessageWindow.getPendingCount() > 0);

		// ^XNgC̏
		tray = new SWTTaskTray(getEventBroker(), display, shell,
				createTaskTrayMenu(actions), hasPending);

		// bZ[W
		messageSearch = new SWTMessageSearch(this, shell, pref,
				desktopMessages, getEventBroker());

		/*
		 * bZW[̊Jn
		 * NO1łdNƂ݂Ȃ
		 * bZ[W\ɕۑŏI
		 */
		this.serverListener = new SWTServerListener(this, pref,
				messengerFacade, super.getEventBroker());

		// EventBrokerɓo^
		setupEventBroker(pref);

		try {
			super.startMessengers(serverListener);
		} catch (LaunchException e) {
			logger.warn(mm.get("SWTMainWindow.quitBecauseOfStartupException"), //$NON-NLS-1$
					e);
			messengerFacade.quit();

			SWTUtils.showError(shell,
					mm.get("SWTMainWindow.quitBecauseOfPortCompetition")); //$NON-NLS-1$

			// Vbg_EtbNLZ
			getAppPreferences().getRuntime().setQuit(true);
			shell.dispose();
			return;
		}

		// Nʒm
		super.messengerFacade.sendRegister();

		// EBhẼ^Cgʏɂ
		setWindowTitle();

		// Nɍŏ
		if (pref.isMinimizedAtStartup()) {
			getEventBroker().notifyEvent(Events.MinimizeAllWindows, null);
		} else {
			shell.setActive();
		}

		// ~{^ŏIȂ
		explicitlyQuit = !pref.iconifyInstedOfQuitting();

		// \ɂ
		shell.setEnabled(true);

		// s݃[h̏
		controlAbsence();

		// o[WmF̗\
		new SWTCheckForUpdates(getAppPreferences(), serverListener,
				getEventBroker()).scheduleVersionCheck();

		// IPAhXƃX[v̊Ď
		scheduleWatcher();

		// TODO ďö̊Ď

		// zbgL[
		setupHotkeys();
	}

	/**
	 * eĎ^XN
	 */
	private void scheduleWatcher() {
		if (!Tool.javaVersion1_5_orHigher) {
			return;
		}
		//MessengerThreadGroup gr = new MessengerThreadGroup("Scheduler", //$NON-NLS-1$
		//		getEventBroker());
		// 1.4ŎsłȂȂ
		//SchedulerThreadFactory threadFactory = new SchedulerThreadFactory(gr);
		//scheduler = Executors.newScheduledThreadPool(1, threadFactory);
		scheduler = Executors.newScheduledThreadPool(1);
		scheduler.schedule(new IPAddressWatcher(scheduler, getEventBroker()),
				5, TimeUnit.SECONDS);
	}

	/**
	 * bZW[ċN܂B
	 */
	void restartMessengers() {
		logger.info("Restarting messengers: start"); //$NON-NLS-1$

		// T[oXi[̃Cxg`B~
		this.serverListener.setEventTransmission(false);

		// bZW[I
		logger.debug("Quit messengers."); //$NON-NLS-1$
		super.messengerFacade.quit();

		sendDialog.reset();
		// startupĎ擾
		getLaunchConfigurations().reloadStartups();

		// T[oXi[̃Cxg`BĊJ
		this.serverListener.setEventTransmission(true);

		// bZW[Jn
		try {
			super.startMessengers(serverListener);
		} catch (LaunchException e) {
			LaunchConfiguration launchConf = e.getLaunchConfiguration();
			String msg = mm.get("SWTMainWindow.restartException", //$NON-NLS-1$
					new Object[] { e.getMessage(), launchConf.getProtocol(),
							launchConf.getName() });
			SWTUtils.showError(shell, msg);
		}
		// Nʒm
		super.messengerFacade.sendRegister();

		logger.info("Restarting messengers: end"); //$NON-NLS-1$
	}

	/**
	 * zbgL[̐ݒs܂B
	 */
	void setupHotkeys() {
		if (!Os.isWindows()) {
			return;
		}

		if (hotkeyEventBridge != null) {
			hotkeyEventBridge.removeAllHotkeys();
		}

		AppPreferences appPref = getAppPreferences();

		try {
			if (hotkeyEventBridge == null) {
				hotkeyEventBridge = new HotkeyEventBridge(getEventBroker());
			}

			hotkeyEventBridge.addHotkey(Events.DisplayWindow,
					appPref.getHotkeyDisplayWindow());
			hotkeyEventBridge.addHotkey(Events.ShowInIndependentWindow,
					appPref.getHotkeyShowInIndependentWindow());
			hotkeyEventBridge.addHotkey(Events.OpenBlankMessageWindow,
					appPref.getHotkeyOpenBlankMessageWindow());
			hotkeyEventBridge.addHotkey(
					Events.OpenBlankIndependentMessageWindow,
					appPref.getHotkeyOpenBlankIndependentMessageWindow());
			hotkeyEventBridge.addHotkey(Events.OpenLogViewer,
					appPref.getHotkeyOpenLogViewer());

		} catch (ApplicationException e) {
			logger.error(e);
			if (hotkeyEventBridge != null) {
				hotkeyEventBridge.dispose();
				hotkeyEventBridge = null;
			}
			SWTUtils.showError(shell,
					mm.get("SWTMainWindow.JIntellitypeNotFound")); //$NON-NLS-1$
		} catch (NoClassDefFoundError e) {
			String msg = mm.get("SWTMainWindow.broken"); //$NON-NLS-1$
			logger.fatal(msg, e);

			SWTUtils.showError(shell, msg);
		}
	}

	/**
	 * EBhËʒuOÏʒuɐݒ肵܂B
	 * <p>
	 * ̍Wꂩ̃fBXvC̃NCAg̈ɂ
	 * ꍇ̓VXeftHg̈ʒuɕ\܂B
	 * 
	 * @param pref ݒ
	 */
	private void restoreWindowLocation(AppPreferences pref) {
		int x = pref.getMainWindowLeft();
		int y = pref.getMainWindowTop();
		int width = pref.getMainWindowWidth();
		int height = pref.getMainWindowHeight();

		Monitor[] monitors = display.getMonitors();
		if (SWTUtils.getMonitorIndexByCoordinate(monitors, x, y) != -1) {
			shell.setLocation(x, y);
		}
		// TODO EBhETCY̕
		shell.setSize(width, height);
	}

	/**
	 * EventBrokerւ̓o^
	 * @param pref AvP[Vݒ
	 */
	private void setupEventBroker(AppPreferences pref) {
		EventBroker broker = getEventBroker();

		// NewMessageReceived
		broker.addListener(Events.NewMessageReceived, new NewMessageReceived(
				pref, getMessageLogger(), savedMessageWindow, soundPlayer,
				messengerFacade));

		broker.addListener(Events.NewMessageReceived,
				new SWTNewMessageReceived(tray, savedMessageWindow));

		// OpenNoticeReceived
		broker.addListener(Events.OpenNoticeReceived,
				new SWTOpenNoticeReceived(pref, tray));

		SWTDisplayWindow displayWindow = new SWTDisplayWindow(this, pref,
				messengerFacade, savedMessageWindow, tray, getMessageLogger());
		// DisplayWindow
		broker.addListener(Events.DisplayWindow, displayWindow);

		// ShowInIndependentWindow
		broker.addListener(Events.ShowInIndependentWindow, displayWindow);

		SWTOpenBlankMessageWindow openBlankMessageWindow = new SWTOpenBlankMessageWindow(
				this, pref, messengerFacade, getMessageLogger());
		// OpenBlankMessageWindow
		broker.addListener(Events.OpenBlankMessageWindow,
				openBlankMessageWindow);
		// OpenBlankIndependentMessageWindow
		broker.addListener(Events.OpenBlankIndependentMessageWindow,
				openBlankMessageWindow);

		// OpenNewStickiesSet
		broker.addListener(Events.OpenNewStickiesSet,
				new SWTOpenNewStickiesSet(this, pref, messengerFacade,
						getMessageLogger()));

		// UnexpectedMessengerError
		broker.addListener(Events.UnexpectedMessengerError,
				new UnexpectedErrorOccured());

		// UnexpectedErrorOccured
		broker.addListener(Events.UnexpectedErrorOccured,
				new UnexpectedErrorOccured());

		// FocusMainWindow
		broker.addListener(Events.FocusMainWindow, //
				new SWTFocusMainWindow(this));

		// MinimizeAllWindows
		broker.addListener(Events.MinimizeAllWindows, //
				new SWTMinimizeAllWindows(this));

		// OpenLogViewer
		broker.addListener(Events.OpenLogViewer, //
				new SWTOpenLogViewer(this));

		// ShowDesktopSearch
		broker.addListener(Events.ShowDesktopSearch, //
				messageSearch);

		// s݃[h
		if (pref.useAutoAbsence()) {
			// StartAutoAbsence
			broker.addListener(Events.StartAutoAbsence, //
					new StartAutoAbsence(pref, messengerFacade));

			// EndAutoAbsence
			broker.addListener(Events.EndAutoAbsence, //
					new EndAutoAbsence(pref, messengerFacade));
		}
		// XṼXi[
		broker.addListener(Events.UpdateDownloaded,
				new ApplicationUpdateDownloaded(this.getShell(), this,
						getEventBroker()));

		if (Tool.javaVersion1_5_orHigher) {
			// Resume
			broker.addListener(Events.Resumed, sendDialog);
			broker.addListener(Events.Resumed, new SWTCheckForUpdates(
					getAppPreferences(), serverListener, getEventBroker()));

			// NetworkChanged
			broker.addListener(Events.NetworkChanged, sendDialog);
		}
	}

	/**
	 * CEBhEœKȃTCYɂ܂B
	 */
	private void setMainWindowSize() {
		Composite blank = new Composite(shell, SWT.NULL);
		Color menu = display.getSystemColor(SWT.COLOR_WIDGET_BACKGROUND);
		blank.setBackground(menu);
		blank.setSize(180, 1);
		//shell.pack();// TODO TCY̎w
	}

	/**
	 * s݃[h̐
	 */
	private void controlAbsence() {
		AppPreferences pref = super.getAppPreferences();

		// s݃[hĎ
		if (pref.useAutoAbsence()) {
			this.absenceWatcher = new SWTPresenceWatcher(display,
					getEventBroker(), pref.getAutoAbsenceLimit() * 1000);
			absenceWatcher.startWatching();
		}

		// Ns݃[h
		if (!pref.isAbsenceAtStartup()) {
			return;
		}

		// s݃[hɂ
		new SWTAbsenceDialog(this.shell, messengerFacade, pref, true);
	}

	/**
	 * @return RestartHelper𐶐܂
	 */
	RestartHelper createRestartHelper() {
		String pf = Os.getPlatform();
		//System.out.println(pf);
		if (pf.startsWith("win32.")) { //$NON-NLS-1$
			// win32.32 ̂悤ȒlԂ
			return new WindowsRestartHelper(getAppPreferences());

		} else if (pf.startsWith("gtk.")) { //$NON-NLS-1$
			return new GtkRestartHelper(getAppPreferences());

		} else if (pf.startsWith("cocoa.") || pf.startsWith("carbon.")) { //$NON-NLS-1$ //$NON-NLS-2$
			// cocoa.64̂悤ȒlԂ
			return new MacOSXRestartHelper(getAppPreferences());

		}
		logger.warn("Unsupported platform:" + pf); //$NON-NLS-1$
		return null;
	}

	/**
	 * EBhE^Cgݒ肵܂B
	 */
	void setWindowTitle() {
		AppPreferences pref = getAppPreferences();

		String userName = pref.getUserName();
		StringBuffer sb = new StringBuffer(userName);

		String group = pref.getGroupName();
		if (group != null && group.length() != 0) {
			sb.append('(').append(group).append(')');
		}

		setWindowTitle(sb.toString());
	}

	/**
	 * EBhE^Cgݒ肵܂B
	 * 
	 * @param title AvP[V^Cg
	 */
	private void setWindowTitle(final String title) {
		if (Os.isOther()) {
			display.asyncExec(new Runnable() {
				public void run() {
					if (!shell.isDisposed()) {
						setWindowTitle2(title);
					}
				}
			});
		} else {
			setWindowTitle2(title);
		}
	}

	/**
	 * EBhE^Cgݒ肵܂B
	 * 
	 * @param title AvP[V^Cg
	 */
	private void setWindowTitle2(final String title) {
		shell.setText(title + " - " + APP_NAME); //$NON-NLS-1$
	}

	/**
	 * CEBhẼj[o[\z܂B
	 * @param actions CxgXi[
	 */
	private void createMenuBar(Actions actions) {

		// j[o[
		Menu menuBar = SWTMenuUtils.addMenuBar(shell);

		// t@Cj[
		Menu menuFile = SWTMenuUtils.addCascade(menuBar,//
				mm.get("SWTMainWindow.menu.file")); //$NON-NLS-1$
		createMenuFile(actions, menuFile, true);

		if (!Os.isMacOSX()) {

			createMenuWindow(menuFile, true, true, true);

			SWTMenuUtils.addSeparator(menuFile);

			createMenuFileRestart(actions, menuFile, true);
			// Ij[ACe
			createMenuFileQuit(actions, menuFile, true);

		} else {

			createMenuFileRestart(actions, menuFile, true);
			// Macł͓Ɨڂɂ
			createMenuWindow(menuBar, true, true, false);
		}

		// wvj[
		Menu menuHelp = SWTMenuUtils.addCascade(menuBar,
				mm.get("SWTMainWindow.menu.help")); //$NON-NLS-1$
		createMenuHelp(actions, menuHelp, true);

		// Mac OS X AvP[Vj[
		if (Os.isMacOSX()) {
			if (SWTUtils.isCarbon()) {
				new SWTMacOSXMenu().hookApplicationMenu(display,//
						mm.get("SWTMainWindow.menu.app.about", //$NON-NLS-1$
								new Object[] { BaseMain.APP_NAME }), actions);
			} else {
				CocoaUIEnhancer enhancer = new CocoaUIEnhancer(actions);
				enhancer.earlyStartup();
			}
		}
	}

	/**
	 * ^XNgC̃ReLXgj[𐶐܂B
	 * @param actions CxgXi[
	 * @return j[
	 */
	private Menu createTaskTrayMenu(Actions actions) {
		Menu menu = new Menu(shell, SWT.POP_UP);

		// t@Cj[
		createMenuFile(actions, menu, false);

		SWTMenuUtils.addSeparator(menu);

		// wvj[
		createMenuHelp(actions, menu, false);

		// Ij[ACe
		if (!Os.isMacOSX()) {
			SWTMenuUtils.addSeparator(menu);

			createMenuFileQuit(actions, menu, false);
		}

		return menu;
	}

	/**
	 * j[t@C
	 * @param actions CxgXi[
	 * @param menu ẽj[
	 * @param showKey V[gJbgL[\ꍇtrueB
	 */
	private void createMenuFile(Actions actions, Menu menu, boolean showKey) {

		createMenuFileMessage(menu, showKey, false, false);

		SWTMenuUtils.addMenuItem(menu, //
				mm.get("SWTMainWindow.menu.file.openStorage"), //$NON-NLS-1$
				"Ctrl+O", SWT.MOD1 | 'O', //$NON-NLS-1$
				getImage("messagebox.gif"), actions, //$NON-NLS-1$
				"doMenuFileOpenStorage", showKey); //$NON-NLS-1$

		SWTMenuUtils.addMenuItem(menu,//
				mm.get("SWTMainWindow.menu.file.openLog"), //$NON-NLS-1$
				"Ctrl+Shift+O", SWT.MOD1 | SWT.MOD2 | 'O', //$NON-NLS-1$
				getImage("message.gif"), actions, //$NON-NLS-1$
				"doMenuFileOpenLog", showKey); //$NON-NLS-1$

		SWTMenuUtils.addMenuItem(menu,//
				mm.get("SWTMainWindow.menu.file.absenceMode"), //$NON-NLS-1$
				"Ctrl+A", SWT.MOD1 | 'A', //$NON-NLS-1$
				null, actions, "doMenuFileAbsence", showKey); //$NON-NLS-1$

		SWTMenuUtils.addMenuItem(menu,//
				mm.get("SWTMainWindow.menu.file.preference"), //$NON-NLS-1$
				"Ctrl+D", SWT.MOD1 | 'D', //$NON-NLS-1$
				getImage("preference.gif"), actions, //$NON-NLS-1$
				"doMenuFileConfig", showKey); //$NON-NLS-1$

	}

	/**
	 * bZ[W֘Ãj[ڂ𐶐܂B
	 * 
	 * @param menu ẽj[
	 * @param showKey V[gJbgL[\ꍇtrue
	 * @param createSeparator 擪ɃZp[^ǉꍇtrue
	 * @param cascade JXP[hj[ɂ
	 */
	void createMenuFileMessage(Menu menu, boolean showKey,
			boolean createSeparator, boolean cascade) {

		if (cascade) {
			menu = SWTMenuUtils.addCascade(menu,//
					mm.get("SWTMainWindow.menu.file.message")); //$NON-NLS-1$
		}

		if (createSeparator) {
			SWTMenuUtils.addSeparator(menu);
		}

		SWTMenuUtils.addMenuItem(menu,//
				mm.get("SWTMainWindow.menu.file.message.new"), //$NON-NLS-1$
				"Ctrl+N", SWT.MOD1 | 'N', //$NON-NLS-1$
				getImage("message.gif"), actions, //$NON-NLS-1$
				"doMenuFileNew", showKey); //$NON-NLS-1$

		SWTMenuUtils.addMenuItem(menu,//
				mm.get("SWTMainWindow.menu.file.message.newStickiesSet"), //$NON-NLS-1$
				"Ctrl+Shift+N", SWT.MOD1 | SWT.MOD2 | 'N', //$NON-NLS-1$
				getImage("stickies16.gif"), actions, //$NON-NLS-1$
				"doMenuFileNewStickiesSet", showKey); //$NON-NLS-1$

		SWTMenuUtils.addMenuItem(menu,//
				mm.get("SWTMainWindow.menu.file.message.display"), //$NON-NLS-1$
				"Ctrl+M", SWT.MOD1 | 'M', //$NON-NLS-1$
				getImage("display_new_message.gif"), actions, //$NON-NLS-1$
				"doMenuFileDisplay", showKey); //$NON-NLS-1$

		SWTMenuUtils.addMenuItem(menu,//
				mm.get("SWTMainWindow.menu.file.message.search"), //$NON-NLS-1$
				"Ctrl+Shift+D", SWT.MOD1 | SWT.MOD2 | 'D',//$NON-NLS-1$
				getImage("loupe.gif"), actions, //$NON-NLS-1$
				"doMenuFileSearchMessage", showKey); //$NON-NLS-1$
	}

	/**
	 * j[EBhE
	 * 
	 * @param actions CxgXi[
	 * @param menu ẽj[
	 * @param showKey V[gJbgL[\ꍇtrueB
	 * @param hideMainWindow ځuCEBhEvBꍇtrue
	 * @param createSeparator 擪ɃZp[^ǉꍇtrue
	 */
	void createMenuWindow(Menu menu, boolean showKey, boolean hideMainWindow,
			boolean createSeparator) {

		if (createSeparator) {
			SWTMenuUtils.addSeparator(menu);
		}

		Menu menuFileWindow = SWTMenuUtils.addCascade(menu,//
				mm.get("SWTMainWindow.menu.window")); //$NON-NLS-1$

		if (!hideMainWindow) {
			SWTMenuUtils.addMenuItem(menuFileWindow,//
					mm.get("SWTMainWindow.menu.window.mainWindow"), //$NON-NLS-1$
					"Ctrl+J", SWT.MOD1 | 'J', //$NON-NLS-1$
					null, actions, "doMenuWindowFocusMainWindow", showKey); //$NON-NLS-1$

			SWTMenuUtils.addSeparator(menuFileWindow);
		}

		SWTMenuUtils.addMenuItem(menuFileWindow,//
				mm.get("SWTMainWindow.menu.window.minimizeAll"), //$NON-NLS-1$
				"Ctrl+B", SWT.MOD1 | 'B', null, actions, //$NON-NLS-1$
				"doMenuFileMinimize", showKey); //$NON-NLS-1$

		SWTMenuUtils.addSeparator(menuFileWindow);

		SWTMenuUtils.addMenuItem(menuFileWindow,//
				mm.get("SWTMainWindow.menu.window.nextInList"), //$NON-NLS-1$
				"Ctrl+F6", SWT.MOD1 | SWT.F6, null, actions, //$NON-NLS-1$
				"doMenuWindowListNext", showKey); //$NON-NLS-1$

		SWTMenuUtils.addMenuItem(menuFileWindow,//
				mm.get("SWTMainWindow.menu.window.previousInList"), //$NON-NLS-1$
				"Ctrl+Shift+F6", SWT.MOD1 | SWT.MOD2 | SWT.F6, //$NON-NLS-1$
				null, actions, "doMenuWindowListPrevious", showKey); //$NON-NLS-1$

		if (Os.isMacOSX()) {
			// Mac OS XŏILabelGraphic is disposedɂȂ邽ߖɂ
			return;
		}
		SWTMenuUtils.addMenuItem(menuFileWindow,//
				mm.get("SWTMainWindow.menu.window.next"), //$NON-NLS-1$
				"Ctrl+PageDown", SWT.MOD1 | SWT.PAGE_DOWN, //$NON-NLS-1$
				null, actions, "doMenuWindowNext", showKey); //$NON-NLS-1$

		SWTMenuUtils.addMenuItem(menuFileWindow,//
				mm.get("SWTMainWindow.menu.window.previous"), //$NON-NLS-1$
				"Ctrl+PageUp", SWT.MOD1 | SWT.PAGE_UP, //$NON-NLS-1$
				null, actions, "doMenuWindowPrevious", showKey); //$NON-NLS-1$
	}

	/**
	 * t@CċN
	 * @param actions CxgXi[
	 * @param menu ẽj[
	 * @param showKey V[gJbgL[\ꍇtrueB
	 */
	private void createMenuFileRestart(Actions actions, Menu menu,
			boolean showKey) {
		SWTMenuUtils.addMenuItem(menu,//
				mm.get("SWTMainWindow.menu.file.restart"), //$NON-NLS-1$
				"", 0, //$NON-NLS-1$
				getImage("quit.gif"), actions, //$NON-NLS-1$
				"doMenuFileRestart", showKey); //$NON-NLS-1$
	}

	/**
	 * j[t@CI
	 * @param actions CxgXi[
	 * @param menu ẽj[
	 * @param showKey V[gJbgL[\ꍇtrueB
	 */
	private void createMenuFileQuit(Actions actions, Menu menu, boolean showKey) {
		SWTMenuUtils.addMenuItem(menu,//
				mm.get("SWTMainWindow.menu.file.quit"), //$NON-NLS-1$
				"Ctrl+Q", SWT.MOD1 | 'Q', //$NON-NLS-1$
				getImage("quit.gif"), actions, //$NON-NLS-1$
				"doMenuFileQuit", showKey); //$NON-NLS-1$
	}

	/**
	 * j[wv
	 * @param actions CxgXi[
	 * @param menu ẽj[
	 * @param showKey V[gJbgL[\ꍇtrueB
	 */
	private void createMenuHelp(Actions actions, Menu menu, boolean showKey) {
		SWTMenuUtils.addMenuItem(menu,//
				mm.get("SWTMainWindow.menu.help.help"), //$NON-NLS-1$
				"Ctrl+?", SWT.MOD1 | '?', //$NON-NLS-1$
				getImage("help.gif"), actions, //$NON-NLS-1$
				"doMenuHelpHelp", showKey); //$NON-NLS-1$

		SWTMenuUtils.addMenuItem(menu,//
				mm.get("SWTMainWindow.menu.help.feedback"), //$NON-NLS-1$
				"Ctrl+Shift+F", SWT.MOD1 | SWT.MOD2 | 'F', //$NON-NLS-1$
				getImage("feedback.gif"), actions, //$NON-NLS-1$
				"doMenuHelpFeedback", showKey); //$NON-NLS-1$

		SWTMenuUtils.addSeparator(menu);

		SWTMenuUtils.addMenuItem(menu,//
				mm.get("SWTMainWindow.menu.help.javaCacheViewer"), //$NON-NLS-1$
				"Ctrl+Shift+Alt+W", SWT.MOD1 | SWT.MOD2 | SWT.MOD3 | 'W', //$NON-NLS-1$
				null, actions, "doMenuHelpJavawsViewer", showKey); //$NON-NLS-1$

		SWTMenuUtils.addMenuItem(menu,//
				mm.get("SWTMainWindow.menu.help.openConfigFolder"), //$NON-NLS-1$
				"Ctrl+Shift+Alt+C", SWT.MOD1 | SWT.MOD2 | SWT.MOD3 | 'C', //$NON-NLS-1$
				null, actions, "doMenuHelpOpenConfigFolder", showKey); //$NON-NLS-1$

		if (Os.isWindows()) {
			SWTMenuUtils.addSeparator(menu);

			if (getAppPreferences().getRuntime().isSandBox()) {
				SWTMenuUtils.addMenuItem(
						menu,
						mm.get("SWTMainWindow.menu.help.registerJwsStable"), //$NON-NLS-1$
						"Ctrl+Shift+U", SWT.MOD1 | SWT.MOD2 | 'U', //$NON-NLS-1$
						null, actions,
						"doMenuHelpRegisterStartupStable", showKey); //$NON-NLS-1$

				SWTMenuUtils.addMenuItem(menu,//
						mm.get("SWTMainWindow.menu.help.registerJwsDev"), //$NON-NLS-1$
						"", 0, null, actions, //$NON-NLS-1$
						"doMenuHelpRegisterStartupDevelop", showKey); //$NON-NLS-1$
			} else {
				SWTMenuUtils.addMenuItem(menu,//
						mm.get("SWTMainWindow.menu.help.registerLocal"), //$NON-NLS-1$
						"Ctrl+Shift+U", SWT.MOD1 | SWT.MOD2 | 'U', null, //$NON-NLS-1$
						actions, "doMenuHelpRegisterStartup", showKey); //$NON-NLS-1$
			}
		}

		SWTMenuUtils.addSeparator(menu);

		SWTMenuUtils.addMenuItem(menu, //
				mm.get("SWTMainWindow.menu.help.openWebSite", //$NON-NLS-1$
						new Object[] { APP_NAME }), //
				"", 0, null, actions, "doMenuHelpOpenWebSite", showKey); //$NON-NLS-1$ //$NON-NLS-2$

		SWTMenuUtils.addMenuItem(menu,//
				mm.get("SWTMainWindow.menu.help.about", //$NON-NLS-1$
						new Object[] { APP_NAME }),//
				"Ctrl+T", SWT.MOD1 | 'T', //$NON-NLS-1$
				getImage("appl16.png"), actions, //$NON-NLS-1$
				"doMenuHelpAbout", showKey); //$NON-NLS-1$
	}

	/**
	 * assertLȂWo͂s܂B
	 */
	private static void checkAssert() {
		boolean asserts = false;
		assert asserts = true;
		if (asserts) {
			System.out.println("assertion enabled."); //$NON-NLS-1$
		}
	}

	/**
	 * Vbg_EtbN𐶐܂B
	 * @param mainWindow CEBhE
	 */
	private static void setupShutdownHook(SWTMainWindow mainWindow) {
		SWTShutdown sd = new SWTShutdown(mainWindow, mainWindow
				.getAppPreferences().getRuntime());
		Thread sdThread = new Thread(sd);
		sdThread.setName("Shutdown hook"); //$NON-NLS-1$
		Runtime.getRuntime().addShutdownHook(sdThread);
	}

	/**
	 * CEBhËʒuۑ܂B
	 */
	void saveWindowStatus() {
		logger.debug("saveWindowStatus() called."); //$NON-NLS-1$

		Rectangle rect = shell.getBounds();
		AppPreferences pref = getAppPreferences();
		pref.setMainWindowLeft(rect.x);
		pref.setMainWindowTop(rect.y);
		pref.setMainWindowWidth(rect.width);
		pref.setMainWindowHeoght(rect.height);
		// TODO EBhETCY̕ۑ
	}

	/**
	 * Returns the shell.
	 * @return Shell
	 */
	Shell getShell() {
		return shell;
	}

	/**
	 * SWTMessageWindowManager擾܂B
	 * @return SWTMessageWindowManager
	 */
	public SWTMessageWindowManager getSWTMessageWindowManager() {
		return this.desktopMessages;
	}

	/**
	 * O[oV[gJbgnh擾܂B
	 * 
	 * @return O[oV[gJbgnh
	 */
	public SWTAppGlobalShortcut getAppGlobalShortcut() {
		return appGlobalShortcut;
	}

	/**
	 * EBhE؂ւnh擾܂B
	 * 
	 * @return EBhE؂ւnh
	 */
	public SWTWindowSwitcher getWindowSwitcher() {
		return switcher;
	}

	/**
	 * SWTTransparencyCanceller擾܂B
	 * @return ShellListener
	 */
	public SWTTransparencyCanceller getTransparencyCanceller() {
		return transparencyCanceller;
	}

	/**
	 * G[_CAO\ă[U[ɍĎs邩ǂq˂܂B
	 * 
	 * @param message bZ[W
	 * @param e O
	 * @return ĎsȂꍇtrue
	 */
	private boolean showErrorAndAskToRetry(String message, Exception e) {
		logger.error(message, e);

		return !SWTUtils.showErrorAndConfirm(shell,//
				mm.get("SWTMainWindow.failedToSave"), //$NON-NLS-1$
				message + '\n' + e.getMessage());
	}

	/**
	 * AvP[VI
	 */
	private void shutdown() {
		logger.info("shutdown start."); //$NON-NLS-1$

		AppPreferences pref = super.getAppPreferences();
		RuntimeProperties rp = pref.getRuntime();
		Policy policy = pref.getPolicy();

		if (rp.isQuit()) {
			logger.info("Already quitted."); //$NON-NLS-1$
			return;
		}
		logger.debug("Switch to termination mode."); //$NON-NLS-1$
		rp.setQuit(true);

		// EBhEbN
		shell.setEnabled(false);
		setWindowTitle(mm.get("SWTMainWindow.title.quit.1")); //$NON-NLS-1$

		if (hotkeyEventBridge != null) {
			hotkeyEventBridge.dispose();
		}

		// T[oXi[~
		this.serverListener.setEventTransmission(false);

		if (scheduler != null) {
			scheduler.shutdownNow();
		}

		// T[oI
		logger.debug("Quitting messengers."); //$NON-NLS-1$
		super.messengerFacade.quit();

		boolean retry = true;

		// ҏW\ȂN\ۑ
		if (policy.canEditLaunchConfiguration()) {
			logger.debug("Saving launch configurations."); //$NON-NLS-1$
			LaunchConfigurations confs = super.getLaunchConfigurations();

			while (retry) {
				try {
					confs.saveLaunchConfigurations();
					retry = false;
				} catch (IOException e) {
					String s = mm.get("SWTMainWindow.failedToSaveLaunchConfig"); //$NON-NLS-1$
					if (showErrorAndAskToRetry(s, e)) {
						break;
					}
				}
			}
		}

		// CEBhËʒu擾
		logger.debug("Saving main window location."); //$NON-NLS-1$
		saveWindowStatus();

		// M_CAȌ
		sendDialog.saveWindowStatus();

		// bZ[W_CAȌ
		messageSearch.saveWindowStatus();

		// ۊǔ̏
		savedMessageWindow.saveWindowStatus();
		savedMessageWindow.setVisible(false); // ʂȉʍXV΍

		// \̃bZ[WEBhEo^
		logger.debug("Saving desktop message windows."); //$NON-NLS-1$
		for (int i = 0; i < this.desktopMessages.size(); i++) {
			SWTMessageWindow swtmw = this.desktopMessages.get(i);
			swtmw.synchMessage();
			Message mw = swtmw.getMessage();
			mw.desktop = true;
			this.savedMessageWindow.addDesktop(mw);
		}

		setWindowTitle(mm.get("SWTMainWindow.title.quit.2")); //$NON-NLS-1$
		// ۊǔfBXNɏ
		logger.debug("Saving message storage."); //$NON-NLS-1$

		retry = true;
		while (retry) {
			try {
				savedMessageWindow.write();
				retry = false;
			} catch (MessageStorageIOException e) {
				String s = mm.get("SWTMainWindow.failedToSaveStorageBox"); //$NON-NLS-1$
				if (showErrorAndAskToRetry(s, e)) {
					break;
				}
			}
		}

		setWindowTitle(mm.get("SWTMainWindow.title.quit.3")); //$NON-NLS-1$
		// AvP[Vݒۑ
		retry = true;
		while (retry) {
			try {
				logger.debug("Saving Application preferences."); //$NON-NLS-1$
				pref.save();
				retry = false;
			} catch (IOException e) {
				String s = mm.get("SWTMainWindow.failedToSavePreference"); //$NON-NLS-1$
				if (showErrorAndAskToRetry(s, e)) {
					break;
				}
			}
		}

		// ^XNgC
		tray.dispose();

		// Image
		icons.release();

		if (getAppPreferences().getRuntime().isSandBox()) {
			DeathGazer death = new DeathGazer();
			death.start();
		}

		// main()ɂǂSystem.exit(0)
		logger.debug("shutdown end."); //$NON-NLS-1$
	}

	/**
	 * itB[hobNJ܂B
	 */
	private void openFeedbackWindow() {
		SWTQualityFeedback feedback = new SWTQualityFeedback();
		feedback.initialize(true);

		feedback.getShell().setText(mm.get("SWTMainWindow.qualityFeedback")); //$NON-NLS-1$
		feedback.setButtonText(//
				mm.get("SWTMainWindow.qualityFeedback.send"), //$NON-NLS-1$
				mm.get("SWTMainWindow.qualityFeedback.cancel")); //$NON-NLS-1$
		feedback.setExplanation(mm.get("SWTMainWindow.feedback.note")); //$NON-NLS-1$
		// Rg
		feedback.addElement(//
				mm.get("SWTMainWindow.feedback.comment"), //$NON-NLS-1$
				"", false); //$NON-NLS-1$

		// AvP[V
		Version v = Version.loadFromResource();
		feedback.addElement("Application", "Petaclone"); //$NON-NLS-1$ //$NON-NLS-2$
		feedback.addElement("Build number", v == null ? "unknown" : v.getBuildNumberStr()); //$NON-NLS-1$ //$NON-NLS-2$
		feedback.addElement("Version", v == null ? "unknown" : v.getVersionText()); //$NON-NLS-1$ //$NON-NLS-2$
		feedback.addElement("Stage", v == null ? "unknown" : v.getStage()); //$NON-NLS-1$ //$NON-NLS-2$

		// SWT
		feedback.addElement("SWT Version", Integer.toString(SWT.getVersion())); //$NON-NLS-1$
		feedback.addElement("SWT Platform", SWT.getPlatform()); //$NON-NLS-1$

		// System
		feedback.addSystemPropertyElement("os.name"); //$NON-NLS-1$
		feedback.addSystemPropertyElement("os.version"); //$NON-NLS-1$
		feedback.addSystemPropertyElement("os.arch"); //$NON-NLS-1$
		feedback.addSystemPropertyElement("sun.os.patch.level"); //$NON-NLS-1$
		feedback.addSystemPropertyElement("sun.arch.data.model"); //$NON-NLS-1$

		// Locale
		feedback.addSystemPropertyElement("user.language"); //$NON-NLS-1$
		feedback.addSystemPropertyElement("user.country"); //$NON-NLS-1$
		feedback.addSystemPropertyElement("user.variant"); //$NON-NLS-1$
		feedback.addSystemPropertyElement("user.dir"); //$NON-NLS-1$

		//Java
		feedback.addSystemPropertyElement("java.version"); //$NON-NLS-1$
		feedback.addSystemPropertyElement("java.runtime.name"); //$NON-NLS-1$
		feedback.addSystemPropertyElement("java.runtime.version"); //$NON-NLS-1$
		feedback.addSystemPropertyElement("java.vm.vendor"); //$NON-NLS-1$
		feedback.addSystemPropertyElement("java.vm.name"); //$NON-NLS-1$
		feedback.addSystemPropertyElement("java.vm.info"); //$NON-NLS-1$
		feedback.addSystemPropertyElement("java.home"); //$NON-NLS-1$
		feedback.addSystemPropertyElement("java.library.path"); //$NON-NLS-1$

		// Environment
		feedback.addElement("PATH", System.getenv("PATH")); //$NON-NLS-1$ //$NON-NLS-2$

		feedback.addSendButtonListener(new SWTQualityFeedbackSendAction(
				feedback, false));
		feedback.open();
	}

	/**
	 * ShellAdapter: N[Y{bNXтׂĂshell.close()
	 * 
	 */
	private class ShellAction extends ShellAdapter implements DisposeListener {

		/**
		 * WindowsłWM_QUERYENDSESSIONɔăVbg_Es܂B
		 */
		public void widgetDisposed(DisposeEvent event) {
			logger.debug("Shell SWT.Dispose"); //$NON-NLS-1$
			shellDisposed = true;
		}

		/**
		 * shell.close()̃Xi[B
		 * Vbg_EdisposeCxgōs܂B
		 */
		public void shellClosed(ShellEvent event) {
			logger.debug("Shell SWT.Close start."); //$NON-NLS-1$
			if (explicitlyQuit) {
				shutdown();
			} else {
				event.doit = false;
				shell.setMinimized(true);
			}
			logger.debug("Shell SWT.Close end."); //$NON-NLS-1$
		}

		/**
		 * ŏ
		 */
		public void shellIconified(ShellEvent event) {
			AppPreferences pref = getAppPreferences();
			if (pref.hideOnIconified()) {
				event.doit = false;
				hide();
			}
		}
	}

	/**
	 * EBhE\畜A܂B
	 * @param activate EBhEőOʂɂ悤Ƃꍇtrue
	 */
	public void show(boolean activate) {
		// ŏŏƂ͕KvB
		shell.setMinimized(false);
		shell.setVisible(true);
		if (activate) {
			shell.forceActive();
		}
	}

	/**
	 * EBhE̕\ԂԂ܂B
	 * @return \ĂčŏĂȂƂtrue
	 */
	public boolean getVisible() {
		return shell.getVisible() && !shell.getMinimized();
	}

	/**
	 * EBhE\ɂ܂B
	 */
	private void hide() {
		shell.setVisible(false);
	}

	/**
	 * CEBhEɃtH[JX܂B
	 * <p>
	 * ŏA\Ԃ̏ꍇ͉܂B
	 */
	public void setFocus() {
		if (getVisible()) {
			this.shell.setActive();
		} else {
			show(true);
		}
	}

	/**
	 * EBhEŏ܂B
	 */
	public void minimize() {
		shell.setMinimized(true);
	}

	/**
	 * Or[AJ܂B
	 */
	public void openLogViewer() {
		actions.doMenuFileOpenLog(null);
	}

	/**
	 * migemo̎QƃJE^
	 */
	private int migemoReference;

	/**
	 * migemogp邱Ƃ錾܂B
	 * <p>
	 * KvɉĎf[^Ƀ[h܂B
	 * 
	 * @param background [hKvȏꍇɃobNOEhōsꍇAtruew肷B
	 */
	public void attachMigemoDictionaries(boolean background) {
		// TODO migemoobNOEhŃ[h
		// TODO [hʒmƃXi[C^[tF[X
		if (migemoReference <= 0) {
			Policy p = getAppPreferences().getPolicy();
			File path = p.getMigemoDictionariesPath();
			if (!path.isDirectory()) {
				return;
			}
			Migemo.clearDictionary();
			File[] files = path.listFiles();
			for (int i = 0; i < files.length; i++) {
				if (files[i].isFile()) {
					Migemo.loadDictionary(files[i]);
				}
			}
		}
		migemoReference++;
	}

	/**
	 * migemogpȂƂ錾܂B
	 * <p>
	 * KvɉĎf[^ォj܂B
	 */
	public void detachMigemoDictionaries() {
		migemoReference--;
		if (migemoReference <= 0) {
			Migemo.clearDictionary();
		}
	}

	/**
	 * migemoŕϊAꂽK\RpC܂B
	 * 
	 * @param str 
	 * @return RpCꂽK\
	 */
	public Matcher compileMigemo(String str) {
		String regex;
		if (Tool.javaVersion1_5_orHigher) {
			regex = Pattern.quote(str) + '|' + Migemo.lookup(str);
		} else {
			regex = Migemo.lookup(str);
		}
		//System.out.println(regex);
		Pattern p = Pattern.compile(regex, Pattern.CASE_INSENSITIVE);
		return p.matcher(""); //$NON-NLS-1$
	}

	/**
	 * G[̃G[_CAO\܂B
	 * @param message \郁bZ[W
	 */
	private static void showInitializingErrorDialog(String message) {
		Display display = Display.getDefault();
		Shell shell = new Shell(display);
		shell.setVisible(false);
		SWTUtils.showError(shell, message);
		shell.dispose();
		display.dispose();
	}

	/**
	 * C\bh
	 * @param args R}hCBgp܂B
	 */
	public static void main(String[] args) {
		long start = System.currentTimeMillis();
		Tool.saveCurrentDir();
		checkAssert();
		setAwtHeadless();
		SWTMainWindow mainWindow = null;
		try {
			checkEncodingSupported();
			// 
			mainWindow = new SWTMainWindow();

			// (ݒ̃[hAo[ǗAMessengerFacadeAN\̏)
			mainWindow.initialize();

			// ̃NXLogger擾
			logger = Logger.getLogger(SWTMainWindow.class);

			// Vbg_EtbN
			setupShutdownHook(mainWindow);
			logger.info("Shutdown hook is registered."); //$NON-NLS-1$

			// vbgtH[̐ݒ
			Os.setPlatform(SWT.getPlatform() + '.'
					+ System.getProperty("sun.arch.data.model")); //$NON-NLS-1$

			// NEBU[h
			mainWindow.startFirstWizard();

			// ʏN
			mainWindow.open();

			logger.info("SWTMainWindow is started: " //$NON-NLS-1$
					+ (System.currentTimeMillis() - start) + " ms."); //$NON-NLS-1$

			/*
			 * MEMO Cxg[v1ӏɂǂƎvB
			 * -itB[hobÑCxg[v폜
			 * -Throwableŕߑ
			 */
			// EBhEj܂Ń[v
			while (!mainWindow.shell.isDisposed()) {
				try {
					if (!mainWindow.display.readAndDispatch()) {
						mainWindow.display.sleep();
					}
					/*
					 * xOoƕitB[hobN̕
					 * Cxg[vsAȍ~͕itB[hobN\ȂB
					 */
				} catch (SWTException e) {
					logger.error("Unexpected UI exception.", e); //$NON-NLS-1$
					UnexpectedErrorOccured fb = new UnexpectedErrorOccured();
					fb.listen(Events.UnexpectedErrorOccured, e);
				} catch (SWTError e) {
					logger.error("Unexpected UI error.", e); //$NON-NLS-1$
					UnexpectedErrorOccured fb = new UnexpectedErrorOccured();
					fb.listen(Events.UnexpectedErrorOccured, e);
				}
			}
			mainWindow.display.dispose();

		} catch (StartupException e) {

			if (e.isLoggingEnabled()) {
				Logger logger = Logger.getLogger(SWTMainWindow.class);
				logger.error("Initialization error has occured.", e); //$NON-NLS-1$
			} else {
				e.printStackTrace();
			}
			String message = Tool.getStackTraceString(e);
			showInitializingErrorDialog("Initialization error has occured.\n" //$NON-NLS-1$
					+ message);

		} catch (UnsatisfiedLinkError e) {

			StringBuffer sb = new StringBuffer(
					"Native library(maybe SWT) not found. "); //$NON-NLS-1$
			sb.append("java.library.path="); //$NON-NLS-1$
			sb.append(System.getProperty("java.library.path")); //$NON-NLS-1$
			sb.append("\n").append(System.getProperty("java.class.path"));
			if (logger != null) {
				logger.fatal(sb, e);
			} else {
				e.printStackTrace();
				System.err.println(sb);
			}

		} catch (Throwable th) {

			if (logger != null) {
				logger.fatal("Unexpected error occured.", th); //$NON-NLS-1$
			} else {
				th.printStackTrace();
			}

			// MessengerXbhI
			if (mainWindow != null) {
				if (logger != null)
					logger.debug("Quitting messengers."); //$NON-NLS-1$
				if (mainWindow.messengerFacade != null)
					mainWindow.messengerFacade.quit();
			}

			// itB[hobN̕\(ڌĂяo)
			UnexpectedErrorOccured fb = new UnexpectedErrorOccured();
			fb.listen(Events.UnexpectedErrorOccured, th);

			// zbgL[̃XbhI
			if (mainWindow != null) {
				if (logger != null)
					logger.debug("Quitting hotkey thread."); //$NON-NLS-1$
				if (mainWindow.hotkeyEventBridge != null) {
					mainWindow.hotkeyEventBridge.dispose();
				}
			}

			Display.getDefault().dispose();

		} finally {
			if (logger != null) {
				logger.info("Set a SWTMainWindow.exit true and exit."); //$NON-NLS-1$
			}
			exit = true;
		}
	}

	/**
	 * NEBU[h
	 */
	private void startFirstWizard() {
		AppPreferences pref = getAppPreferences();
		if (!pref.isFirsttime()
				|| !getLaunchConfigurations().getConfigurations().isEmpty()) {
			return;
		}
		SWTLaunchConfigEditDialog dialog = new SWTLaunchConfigEditDialog(pref,
				messengerFacade, this, shell, null);

		Display display = Display.getDefault();
		while (!dialog.getShell().isDisposed()) {
			if (!display.readAndDispatch())
				display.sleep();
		}

		if (getLaunchConfigurations().getConfigurations().isEmpty()) {
			Shell sh = new Shell(display);
			// 0: ftHg, 1: Mac OS X
			int os = Os.isMacOSX() ? 1 : 0;

			SWTUtils.showInformation(sh,
					mm.get("SWTLaunchConfigEditDialog.wizard"), //$NON-NLS-1$
					mm.get("SWTLaunchConfigEditDialog.canceledMessage", //$NON-NLS-1$
							new Object[] { new Integer(os) }));
			sh.dispose();
		}

		// ʏN
	}

	/**
	 * NEBU[h(Nݒ)R[obN܂B
	 * 
	 * @see petaclone.swt.SWTLaunchConfigEditDialog.LaunchConfigCallback#addLaunchConfiguration(petaclone.LaunchConfiguration)
	 */
	public void addLaunchConfiguration(LaunchConfiguration conf) {

		List confs = getLaunchConfigurations().getConfigurations();
		confs.add(conf);

		while (true) {
			int index = messengerFacade.getConflictConfiguration(confs);
			if (index == -1) {
				break;
			}
			LaunchConfiguration config = (LaunchConfiguration) confs.get(index);
			config.setIsValid(false);

			SWTUtils.showWarning(
					shell, //
					mm.get("SWTMainWindow.conflictDetected", //$NON-NLS-1$
							new Object[] { config.getName(),
									config.getProtocol() }));
			return;
		}

		try {
			// ۑăX^[gAbvč\
			getLaunchConfigurations().saveLaunchConfigurations();
			getLaunchConfigurations().reloadStartups();

		} catch (IOException e) {
			logger.error("Failed to save launch configrations.", e); //$NON-NLS-1$
			SWTUtils.showError(shell,
					mm.get("SWTMainWindow.failedToSaveLaunchConf")); //$NON-NLS-1$
		}
	}

	public void updateLaunchConfiguration(LaunchConfiguration conf) {
		//ĂяoȂ
	}

	/**
	 * AvP[VI܂B
	 */
	public void quit() {
		explicitlyQuit = true;
		shell.close();
	}

	/**
	 * GR[fBOT|[gĂ邩ǂ܂B
	 */
	private static void checkEncodingSupported() {
		if (!Charset.isSupported("Windows-31J")) { //$NON-NLS-1$
			String s = mm.get("SWTMainWindow.encodingError"); //$NON-NLS-1$
			throw new StartupException(false, s);
		}
	}

	/**
	 * AWT̃[U[C^[tF[Xgpsɂ܂B
	 */
	private static void setAwtHeadless() {
		System.setProperty("java.awt.headless", "true"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * SelectionListener̃Zbg
	 */
	public class Actions extends ListenerDelegate {

		/**
		 * O
		 * @see jp.takayuki.swt.util.ListenerDelegate#handleException(java.lang.Exception)
		 */
		protected void handleException(Exception e) {
			getEventBroker().notifyEvent(Events.UnexpectedErrorOccured, e);
		}

		/**
		 * j[t@CVbZ[W
		 * @param event Cxg
		 */
		public void doMenuFileNew(SelectionEvent event) {
			if (!getAppPreferences().getUseIndependentMessageWindow()
					&& !getVisible()) {
				show(false);
			}
			SWTMessageWindow.openNewMessageWindow(SWTMainWindow.this,
					getAppPreferences(), messengerFacade, getMessageLogger(),
					null, false, false);
		}

		/**
		 * j[t@CVtⳎZbg
		 * @param event Cxg
		 */
		public void doMenuFileNewStickiesSet(SelectionEvent event) {
			if (!getAppPreferences().getUseIndependentMessageWindow()
					&& !getVisible()) {
				show(false);
			}
			SWTMessageWindow.openNewStickiesSet(SWTMainWindow.this,
					getAppPreferences(), messengerFacade, getMessageLogger());
		}

		/**
		 * j[t@CbZ[WM
		 * @param event Cxg
		 */
		public void doMenuFileDisplay(SelectionEvent event) {
			EventBroker broker = getEventBroker();
			broker.notifyEvent(Events.DisplayWindow, null);
		}

		/**
		 * j[t@CbZ[W
		 * 
		 * @param event Cxg
		 */
		public void doMenuFileSearchMessage(SelectionEvent event) {
			EventBroker broker = getEventBroker();
			broker.notifyEvent(Events.ShowDesktopSearch, null);
		}

		/**
		 * j[t@CMJ
		 * @param event Cxg
		 */
		public void doMenuFileOpenStorage(SelectionEvent event) {
			savedMessageWindow.show(true);
		}

		/**
		 * j[t@COJ
		 * @param event Cxg
		 */
		public void doMenuFileOpenLog(SelectionEvent event) {
			AppPreferences pref = getAppPreferences();

			// Ot@C̈ʒu擾
			String pathToLog = pref.getMessageLogPath();

			// vO擾
			String pathToViewer = pref.getLogViewerPath();

			// w肳ĂȂꍇ̓G[
			if (pathToViewer == null || pathToViewer.length() == 0
					|| pathToLog == null || pathToLog.length() == 0) {

				SWTUtils.showInformation(SWTMainWindow.this.getShell(),//
						mm.get("SWTMainWindow.openLog.error.title"), //$NON-NLS-1$
						mm.get("SWTMainWindow.openLog.error.message")); //$NON-NLS-1$
				return;
			}

			// r[ɊJ
			Runtime runtime = Runtime.getRuntime();
			try {
				runtime.exec(new String[] { pathToViewer, pathToLog });
			} catch (IOException e) {
				e.printStackTrace();
			}
		}

		/**
		 * j[t@Cs݃[h
		 * @param event Cxg
		 */
		public void doMenuFileAbsence(SelectionEvent event) {
			//SWTAbsenceDialog dialog =
			new SWTAbsenceDialog(SWTMainWindow.this.shell, messengerFacade,
					getAppPreferences(), false);
		}

		/**
		 * j[t@Cݒ...
		 * <p>
		 * ݒ_CAO̕\B
		 * @param event Cxg
		 */
		public void doMenuFileConfig(SelectionEvent event) {
			SWTConfigurationDialog dialog = new SWTConfigurationDialog(
					SWTMainWindow.this, SWTMainWindow.this,
					SWTMainWindow.this.messengerFacade, getAppPreferences(),
					getLaunchConfigurations());
			dialog.open();
		}

		/**
		 * j[t@Cŏ
		 * <p>
		 * ׂẴEBhEŏ܂B
		 * @param event Cxg
		 */
		public void doMenuFileMinimize(SelectionEvent event) {
			getEventBroker().notifyEvent(Events.MinimizeAllWindows, null);
		}

		/**
		 * j[t@CċN
		 * 
		 * @param event Cxg
		 */
		public void doMenuFileRestart(SelectionEvent event) {
			logger.info("Restarting..."); //$NON-NLS-1$

			RestartHelper restartHelper = createRestartHelper();
			if (restartHelper == null) {
				SWTUtils.showError(getShell(),
						mm.get("SWTMainWindow.restart.notSupported")); //$NON-NLS-1$
				return;
			}
			try {
				restartHelper.restart();

				// ʏ̏IV[PX
				doMenuFileQuit(event);
			} catch (IOException e) {
				logger.error("Restart failed.", e); //$NON-NLS-1$
				SWTUtils.showError(getShell(),
						mm.get("SWTMainWindow.restart.failed")); //$NON-NLS-1$
			}
		}

		/**
		 * j[t@CI
		 * <p>
		 * CEBhEăAvP[VI܂B
		 * @param event Cxg
		 */
		public void doMenuFileQuit(SelectionEvent event) {
			quit();
		}

		/**
		 * j[wvwv
		 * @param event Cxg
		 */
		public void doMenuHelpHelp(SelectionEvent event) {
			SWTUtils.showInformation(
					shell,//
					mm.get("SWTMainWindow.help.title", //$NON-NLS-1$
							new Object[] { APP_NAME }),
					mm.get("SWTMainWindow.help.message")); //$NON-NLS-1$
		}

		/**
		 * j[wvPetaclone Web SiteJ
		 * @param event
		 */
		public void doMenuHelpOpenWebSite(SelectionEvent event) {
			Program.launch("http://hp.vector.co.jp/authors/VA035113/petaclone/index.html"); //$NON-NLS-1$
		}

		/**
		 * j[wvy^N[ɂ...
		 * @param event Cxg
		 */
		public void doMenuHelpAbout(SelectionEvent event) {
			// 擾

			long totalMemory = Runtime.getRuntime().totalMemory();
			long freeMemory = Runtime.getRuntime().freeMemory();
			long committedMemory = totalMemory - freeMemory;
			// bZ[W_CAO\
			Version v = Version.loadFromResource();
			String msg = mm
					.get("SWTMainWindow.about.message", //$NON-NLS-1$
							new Object[] {
									APP_NAME,
									"Copyright 2003-2015, Takayuki Shoji.\n" //$NON-NLS-1$
											+ "http://hp.vector.co.jp/authors/VA035113/\n", //$NON-NLS-1$
									v.getVersionText(), v.getBuildNumberStr(),
									new Long(totalMemory),
									new Long(freeMemory),
									new Long(committedMemory) });
			StringBuffer sb = new StringBuffer();
			sb.append(msg);

			RuntimeProperties rp = getAppPreferences().getRuntime();

			// OC@zXg
			// MEMO AhX\
			if (false) {
				sb.append("username@private"); //$NON-NLS-1$
			} else {
				sb.append(rp.getLoginName()).append('@')
						.append(InetAddressUtils.getHostName());
			}
			sb.append('\n');

			// IPAhX
			String[] addrs = InetAddressUtils.getLocalAddressesString();
			for (int i = 0; i < addrs.length; i++) {
				sb.append(addrs[i]);
				if (i != addrs.length)
					sb.append('\n');
			}

			// VXe
			sb.append('\n');
			sb.append(System.getProperty("os.name")); //$NON-NLS-1$
			sb.append('(');
			sb.append(System.getProperty("os.version")); //$NON-NLS-1$
			sb.append(", "); //$NON-NLS-1$
			sb.append(System.getProperty("os.arch")); //$NON-NLS-1$
			sb.append(", "); //$NON-NLS-1$
			sb.append(System.getProperty("sun.os.patch.level")); //$NON-NLS-1$
			sb.append(")\n"); //$NON-NLS-1$

			// Java
			sb.append(System.getProperty("java.runtime.name")); //$NON-NLS-1$
			sb.append("\n("); //$NON-NLS-1$
			sb.append(System.getProperty("java.runtime.version")); //$NON-NLS-1$
			sb.append(", "); //$NON-NLS-1$
			sb.append(System.getProperty("java.vm.vendor")); //$NON-NLS-1$
			sb.append(")\n"); //$NON-NLS-1$

			sb.append(System.getProperty("java.vm.name")); //$NON-NLS-1$
			sb.append(", "); //$NON-NLS-1$
			sb.append(System.getProperty("java.vm.info")); //$NON-NLS-1$
			sb.append('\n');

			sb.append("java.home=").append(System.getProperty("java.home")); //$NON-NLS-1$ //$NON-NLS-2$

			SWTUtils.showInformation(shell, //
					mm.get("SWTMainWindow.about.title", //$NON-NLS-1$
							new Object[] { APP_NAME }), sb.toString());
		}

		/**
		 * 
		 * @param event Cxg
		 */
		public void doMenuHelpFeedback(SelectionEvent event) {
			openFeedbackWindow();
		}

		/**
		 * j[wvJava Web Start̃LbVr[A
		 * @param event Cxg
		 */
		public void doMenuHelpJavawsViewer(SelectionEvent event) {
			try {
				Runtime.getRuntime().exec("javaws -viewer"); //$NON-NLS-1$
			} catch (IOException ex) {
				logger.error("Exception occurred " //$NON-NLS-1$
						+ "during execution of Java Cache Viewer.", ex); //$NON-NLS-1$
			}
		}

		/**
		 * j[wvX^[gAbvɓo^/
		 * @param event Cxg
		 */
		public void doMenuHelpRegisterStartup(SelectionEvent event) {
			registerStartup("0"); //$NON-NLS-1$
		}

		/**
		 * j[wvłX^[gAbvɓo^/
		 * @param event Cxg
		 */
		public void doMenuHelpRegisterStartupStable(SelectionEvent event) {
			registerStartup("1"); //$NON-NLS-1$
		}

		/**
		 * j[wvJłX^[gAbvɓo^/
		 * @param event Cxg
		 */
		public void doMenuHelpRegisterStartupDevelop(SelectionEvent event) {
			registerStartup("2"); //$NON-NLS-1$
		}

		/**
		 * X^[gAbvɓo^܂B
		 * 
		 * @param option 쐬(0: exep, 1: Java Web Startp, 2: JJava Web Startp)
		 */
		private void registerStartup(String option) {
			File f = null;

			try {
				f = copyScript();
			} catch (IOException e) {
				logger.error("Could not create shortcut.", e); //$NON-NLS-1$
				return;
			}

			try {
				String[] args = { "wscript", f.getAbsolutePath(), option }; //$NON-NLS-1$
				Runtime.getRuntime().exec(args);
			} catch (IOException e) {
				logger.error("Could not execute creating shortcut.", e); //$NON-NLS-1$
			}
		}

		/**
		 * shortcut.vbs\[Xe|t@CɃRs[܂B
		 * <p>
		 * rɃP[ɉݒ菈}܂B
		 * 
		 * @return e|t@C
		 * @throws IOException
		 */
		private File copyScript() throws IOException {

			File f = null;

			FileOutputStream out = null;
			OutputStreamWriter osw = null;
			BufferedWriter w = null;

			try {
				InputStream in = null;
				InputStreamReader isr = null;
				BufferedReader r = null;

				f = File.createTempFile("shortcut", ".vbs"); //$NON-NLS-1$//$NON-NLS-2$
				out = new FileOutputStream(f);
				osw = new OutputStreamWriter(out);
				w = new BufferedWriter(osw);

				try {
					final String resourcePath = "/shortcut.vbs"; //$NON-NLS-1$
					in = SWTMainWindow.class.getResourceAsStream(resourcePath);
					if (in == null) {
						throw new IOException(
								"No resource found:" + resourcePath); //$NON-NLS-1$
					}
					isr = new InputStreamReader(in);
					r = new BufferedReader(isr);
					String line = null;
					while ((line = r.readLine()) != null) {
						if (line.equals("'include")) { //$NON-NLS-1$
							for (int i = 1; i <= 8; i++) {
								String str = "str" + i + " = " + '"' //$NON-NLS-1$ //$NON-NLS-2$
										+ mm.get("shortcut.vbs.str" + i) + '"'; //$NON-NLS-1$
								w.write(str);
								w.newLine();
							}
						} else {
							w.write(line);
							w.newLine();
						}
					}

				} finally {
					if (r != null) {
						r.close();
					}
				}

			} finally {
				if (w != null) {
					w.close();
				}
			}

			return f;
		}

		/**
		 * j[wvݒtH_J
		 * @param event Cxg
		 */
		public void doMenuHelpOpenConfigFolder(SelectionEvent event) {
			Policy policy = getAppPreferences().getPolicy();
			File path = policy.getPreferenceLocation();
			ToolExecutor exec = new ToolExecutor(getAppPreferences());
			exec.openFile(AppPreferences.FILE_MANAGER_FOLDER, path);
		}

		/**
		 * j[EBhECEBhE
		 * @param event
		 */
		public void doMenuWindowFocusMainWindow(SelectionEvent event) {
			setFocus();
		}

		/**
		 * EBhEꗗ\Ď̍ڂI܂B
		 * @param event Cxg
		 */
		public void doMenuWindowListNext(SelectionEvent event) {
			//
			Event ev = new Event();
			ev.type = SWT.Selection;
			ev.keyCode = SWT.F6;
			ev.stateMask = SWT.MOD1;
			switcher.handleEvent(ev);
		}

		/**
		 * EBhEꗗ\đO̍ڂI܂B
		 * @param event Cxg
		 */
		public void doMenuWindowListPrevious(SelectionEvent event) {
			//
			Event ev = new Event();
			ev.type = SWT.Selection;
			ev.keyCode = SWT.F6;
			ev.stateMask = SWT.MOD1 | SWT.MOD2;
			switcher.handleEvent(ev);
		}

		/**
		 * ̃EBhEɃtH[JX܂B
		 * @param event Cxg
		 */
		public void doMenuWindowNext(SelectionEvent event) {
			//
			Event ev = new Event();
			ev.type = SWT.KeyDown;
			ev.keyCode = SWT.PAGE_DOWN;
			ev.stateMask = SWT.MOD1;
			switcher.handleEvent(ev);
		}

		/**
		 * ÕEBhEɃtH[JX܂B
		 * @param event Cxg
		 */
		public void doMenuWindowPrevious(SelectionEvent event) {
			//
			Event ev = new Event();
			ev.type = SWT.KeyDown;
			ev.keyCode = SWT.PAGE_UP;
			ev.stateMask = SWT.MOD1;
			switcher.handleEvent(ev);
		}
	}
}
