package petaclone.swt;

import java.util.Arrays;
import java.util.List;
import java.util.regex.Matcher;

import jp.takayuki.logging.Logger;
import jp.takayuki.swt.util.ListenerDelegate;
import jp.takayuki.swt.util.SWTMenuUtils;
import jp.takayuki.swt.util.SWTUtils;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Os;

import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.SashForm;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.ShellAdapter;
import org.eclipse.swt.events.ShellEvent;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.Monitor;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.TabFolder;
import org.eclipse.swt.widgets.TabItem;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.widgets.Text;

import petaclone.AppPreferences;
import petaclone.EventBroker;
import petaclone.Events;
import petaclone.Message;
import petaclone.MessageComparator;
import petaclone.MessageLogger;
import petaclone.MessageStorageIOException;
import petaclone.MessengerFacade;
import petaclone.TableComparator;
import petaclone.TableViewable;
import petaclone.XmlMessageStorage;

/**
 * ۑbZ[W\EBhE
 * @author Takayuki Shoji
 *
 */
public class SWTMessageStorage extends XmlMessageStorage {

	/**
	 * logger
	 */
	private static final Logger logger = Logger
			.getLogger(SWTMessageStorage.class);

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/** ݒ */
	private AppPreferences pref;

	/** Shell */
	private Shell shell;

	/** CEBhE */
	private SWTMainWindow main;

	/** MessengerFacade */
	private MessengerFacade facade;

	/** MTabFolder̓e */
	private TabFolderContent sentTabContent;

	/** MTabFolder̓e */
	private TabFolderContent receivedTabContent;

	/** EventBroker */
	private EventBroker broker;

	/**
	 * MessageLoggerBbZ[WEBhEɕKvB
	 */
	private MessageLogger msgLogger;

	/** [}(migemo)[h */
	private boolean useMigemo;

	/**
	 * ^utH_BTabItemdataTabFolderContent܂B
	 */
	private TabFolder tabFolder;

	/**
	 * CxgXi[
	 */
	private Actions actions;

	/**
	 * RXgN^
	 * @param main CEBhE
	 * @param pref AvP[Vݒ
	 * @param facade bZW[t@T[h
	 * @param msgLogger bZ[WL^
	 * @param broker EventBroker
	 */
	SWTMessageStorage(SWTMainWindow main, AppPreferences pref,
			MessengerFacade facade, MessageLogger msgLogger, EventBroker broker) {
		super(pref.getPolicy().getPreferenceLocation());

		this.main = main;
		this.pref = pref;
		this.facade = facade;
		this.msgLogger = msgLogger;
		this.broker = broker;

		useMigemo = pref.getUseMigemoInMessageStorage();

		if (useMigemo) {
			main.attachMigemoDictionaries(true);
		}

		initWidgets();

		// \EBhE̕
		try {
			readDesktop();
		} catch (MessageStorageIOException e) {
			logger.error("Exception occurred while reading storage box.", e); //$NON-NLS-1$
			alert(e);
		}

		// bZ[WEBhE̕\
		Display display = Display.getDefault();
		List desktop = super.getDesktop();
		for (int i = 0; i < desktop.size(); i++) {
			Message msg = (Message) desktop.get(i);
			SWTMessageWindow win = new SWTMessageWindow(main, msg, pref,
					facade, msgLogger);
			// ƗIvV:fȂ
			win.open();
			win.setEnabled(false);
			display.readAndDispatch();
		}
		SWTUtils.enableAllShells();
	}

	/**
	 * EBWFbgBEBhE͔\ɂĂB
	 */
	private void initWidgets() {
		shell = new Shell(this.main.getShell(), SWT.MAX | SWT.CLOSE
				| SWT.RESIZE);
		shell.setText(mm.get("SWTMessageStorage.storageBox")); //$NON-NLS-1$
		if (!Os.isMacOSX())
			shell.setImage(main.getImage("messagebox.gif")); //$NON-NLS-1$
		shell.addShellListener(new ShellClosed());

		shell.setLayout(new FillLayout());

		// TCY
		shell.setSize(pref.getStorageWindowWidth(), pref
				.getStorageWindowHeight());

		// ʒu
		int x = pref.getStorageWindowLeft();
		int y = pref.getStorageWindowTop();
		if (x != AppPreferences.IMPOSSIBLE && y != AppPreferences.IMPOSSIBLE) {
			Monitor[] monitors = Display.getDefault().getMonitors();
			if (SWTUtils.getMonitorIndexByCoordinate(monitors, x, y) != -1) {
				shell.setLocation(x, y);
			}
		}

		// j[o[\z
		createMenuBar();
		createTabFolder();

		// ۊǔ̕\
		boolean shown = pref.getStorageWindowShown();
		if (shown) {
			this.show(false);
		} else {
			shell.setVisible(false);
		}
	}

	/**
	 * x\܂B
	 * @param e ƂȂO
	 */
	private void alert(MessageStorageIOException e) {
		SWTUtils.showError(shell, //
				mm.get("SWTMessageStorage.coultNotReadMessage", //$NON-NLS-1$
						new Object[] { e.getCause().getMessage() }));
	}

	/**
	 * EBhE\܂B
	 * @param activate ANeBx[gꍇtrue
	 */
	void show(boolean activate) {
		boolean read;
		try {
			read = super.read();
		} catch (MessageStorageIOException e) {
			logger.error("Exception occurred while reading storage box.", e); //$NON-NLS-1$
			alert(e);
			return;
		}
		// e[uXV
		if (read && !shell.isDisposed()) {
			updateTable(this.receivedTabContent, this.received);
			updateTable(this.sentTabContent, this.sent);
		}
		shell.setVisible(true);
		if (activate) {
			shell.setActive();
		}
	}

	/**
	 * \E\𐧌䂵܂B
	 * @param visibility \ꍇtrue
	 */
	public void setVisible(boolean visibility) {
		shell.setVisible(visibility);
	}

	/**
	 * t@Cɏo͂܂B
	 * @throws MessageStorageIOException
	 */
	protected void write() {
		super.write();
	}

	/**
	 * e[uXV܂B
	 * @param target XVΏ
	 * @param list bZ[W̃Xg(l: Message)
	 */
	private void updateTable(TabFolderContent target, List list) {
		target.table.removeAll();
		for (int i = 0; i < list.size(); i++) {
			Message msg = (Message) list.get(i);
			target.addMessage(msg);
		}
	}

	/**
	 * EBhEB܂B
	 */
	private void hide() {
		shell.setVisible(false);
	}

	/**
	 * EBhËʒuvpeBɕۑ܂B
	 */
	public void saveWindowStatus() {
		logger.debug("saveWindowStatus() called."); //$NON-NLS-1$

		pref.setUseMigemoInMessageStorage(useMigemo);

		Rectangle rect = shell.getBounds();
		pref.setStorageWindowLeft(rect.x);
		pref.setStorageWindowTop(rect.y);
		pref.setStorageWindowWidth(rect.width);
		pref.setStorageWindowHeight(rect.height);

		pref.setStorageWindowShown(shell.getVisible());
	}

	/**
	 * f[^ƃe[uɒǉ܂B
	 * @see petaclone.XmlMessageStorage#add(Message)
	 */
	public void add(Message msg) {
		super.add(msg);
		if (msg.desktop || msg.messageClass == Message.PENDING) {
			return;
		}

		// \
		TabFolderContent target = null;
		switch (msg.messageClass) {
		case Message.RECEIVED:
			target = receivedTabContent;
			break;
		default:
			target = sentTabContent;
			break;
		}
		target.addMessage(msg);
	}

	/**
	 * j[o[\z܂B
	 */
	private void createMenuBar() {

		actions = new Actions();

		Menu menuBar = SWTMenuUtils.addMenuBar(shell);

		Menu menuFile = SWTMenuUtils.addCascade(menuBar, //
				mm.get("SWTMessageStorage.file")); //$NON-NLS-1$

		main.createMenuFileMessage(menuFile, true, false, false);

		SWTMenuUtils.addMenuItem(menuFile, //
				mm.get("SWTMessageStorage.close"), //$NON-NLS-1$
				"Ctrl+W", SWT.MOD1 | 'W', //$NON-NLS-1$
				null, actions, "doMenuFileClose"); //$NON-NLS-1$

		// ҏWj[Ɍ̍ڂǉ
		Menu menuEdit = SWTMenuUtils.addCascade(menuBar, //
				mm.get("SWTMessageStorage.edit")); //$NON-NLS-1$

		FindUtils.createMenu(menuEdit, actions);

		SWTMenuUtils.addSeparator(menuEdit);

		SWTMenuUtils.addMenuItemCheckbox(menuEdit,//
				mm.get("SWTMessageStorage.migemo"), //$NON-NLS-1$
				"Ctrl+K", SWT.MOD1 | 'K', null, actions, "doCheckMigemo", pref //$NON-NLS-1$ //$NON-NLS-2$
						.getUseMigemoInMessageStorage());

		// EBhEj[
		main.createMenuWindow(menuBar, true, false, false);
	}

	/**
	 * ^u\z܂B
	 */
	private void createTabFolder() {
		tabFolder = new TabFolder(shell, SWT.NULL);

		TabItem receivedTab = new TabItem(tabFolder, SWT.BORDER);
		receivedTab.setText(mm.get("SWTMessageStorage.receive")); //$NON-NLS-1$

		TabItem sentTab = new TabItem(tabFolder, SWT.BORDER);
		sentTab.setText(mm.get("SWTMessageStorage.send")); //$NON-NLS-1$

		// e\z
		this.sentTabContent = new TabFolderContent(tabFolder, sentTab);
		this.receivedTabContent = new TabFolderContent(tabFolder, receivedTab);
	}

	/**
	 * ݂̃^uԂ܂B
	 * @return TabFolderContent
	 */
	private TabFolderContent getCurrentTabFolderContent() {
		int index = tabFolder.getSelectionIndex();
		TabItem item = tabFolder.getItem(index);
		return (TabFolderContent) item.getData();
	}

	/**
	 * peLXg{bNXɓ͂ꂽ
	 * v郁o[ׂđI܂B
	 */
	private void selectAllMatchedMessages() {
		TabFolderContent current = getCurrentTabFolderContent();
		Table table = current.table;
		if (table.getItemCount() == 0) {
			return;
		}
		String key = current.searchText.getText();
		if (key == null || key.length() == 0) {
			table.deselectAll();
			//setButtonStatus();
			//updateUserCount();
			return;
		}
		Matcher regex = null;
		if (useMigemo) {
			regex = main.compileMigemo(key);
		}
		table.deselectAll();
		final int count = table.getItemCount();
		int index = 0;
		boolean first = true;
		while ((index = current.find(key, regex, index, 1)) != -1) {
			if (first) {
				table.setSelection(index);
				first = false;
			} else {
				table.select(index);
			}
			index++;
			if (index >= count) {
				break;
			}
		}
		//setButtonStatus();
		//updateUserCount();
		table.setFocus();
	}

	/**
	 * N[Y{bNXshell.close()
	 */
	class ShellClosed extends ShellAdapter {
		public void shellClosed(ShellEvent event) {
			event.doit = false;
			hide();
		}
	}

	/**
	 * TabFolder̓e\zNX
	 */
	private class TabFolderContent implements Findable {
		/**
		 * ۑbZ[Ẅꗗ
		 */
		private Table table;

		/**
		 * From̕\
		 */
		private Label from;

		/**
		 * bZ[W{
		 */
		private Text text;

		/**
		 * {bNX
		 */
		private Text searchText;

		/**
		 * RXgN^
		 * @param parent ^utH_
		 * @param item ^uACe
		 */
		TabFolderContent(TabFolder parent, TabItem item) {
			item.setData(this);

			SashForm sash = new SashForm(parent, SWT.VERTICAL);
			item.setControl(sash);

			Composite tableArea = new Composite(sash, SWT.NONE);

			GridLayout layout = new GridLayout(3, false);
			layout.verticalSpacing = 0;
			layout.marginHeight = 0;
			layout.horizontalSpacing = 3;
			tableArea.setLayout(layout);

			TabFolderActions tabActions = new TabFolderActions();
			// {bNX
			createSearchBox(tableArea, actions);
			// e[u
			createTable(tableArea, tabActions);

			searchText.addKeyListener(new ArrowDownBehavior(table));

			Composite detailArea = new Composite(sash, SWT.NULL);
			GridData rightData = new GridData();
			rightData.verticalAlignment = GridData.FILL;
			rightData.horizontalAlignment = GridData.FILL;
			rightData.grabExcessVerticalSpace = true;
			rightData.grabExcessHorizontalSpace = true;
			detailArea.setLayoutData(rightData);

			GridLayout rightLayout = new GridLayout();
			rightLayout.numColumns = 2;
			rightLayout.makeColumnsEqualWidth = true;
			detailArea.setLayout(rightLayout);
			// From
			from = new Label(detailArea, SWT.WRAP);
			from.setText(mm.get("SWTMessageStorage.from")); //$NON-NLS-1$
			GridData fromData = new GridData();
			fromData.horizontalSpan = 2;
			fromData.horizontalAlignment = SWT.FILL;
			fromData.grabExcessHorizontalSpace = true;
			from.setLayoutData(fromData);
			// {
			text = new Text(detailArea, SWT.MULTI | SWT.BORDER | SWT.V_SCROLL
					| SWT.WRAP | SWT.READ_ONLY);
			text.setBackground(main.getShell().getDisplay().getSystemColor(
					SWT.COLOR_LIST_BACKGROUND));
			GridData textData = new GridData();
			textData.verticalAlignment = SWT.FILL;
			textData.horizontalAlignment = SWT.FILL;
			textData.grabExcessVerticalSpace = true;
			textData.grabExcessHorizontalSpace = true;
			textData.horizontalSpan = 2;
			text.setLayoutData(textData);

			// 폜{^
			Button delete = new Button(detailArea, SWT.PUSH);
			delete.setText(mm.get("SWTMessageStorage.delete")); //$NON-NLS-1$
			GridData deleteGridData = new GridData();
			deleteGridData.horizontalAlignment = SWT.RIGHT;
			delete.setLayoutData(deleteGridData);
			tabActions.registerAndAddListener(delete, "doDeleteButton"); //$NON-NLS-1$

			// J{^
			Button open = new Button(detailArea, SWT.PUSH);
			open.setText(mm.get("SWTMessageStorage.open")); //$NON-NLS-1$
			GridData openGridData = new GridData();
			openGridData.horizontalAlignment = SWT.LEFT;
			open.setLayoutData(openGridData);
			tabActions.registerAndAddListener(open, "doOpenButton"); //$NON-NLS-1$
		}

		/**
		 * {bNX\z܂B
		 * @param parent eEBWFbg
		 * @param actions CxgXi[
		 */
		private void createSearchBox(Composite parent, Actions actions) {
			searchText = FindUtils.createSearchBox(parent, actions, true, this);
		}

		public void findItem(String search, int start, boolean reverse) {
			if (search == null || search.length() == 0) {
				table.deselectAll();
				return;
			}
			// migemoŐK\ɕϊ
			Matcher regex = null;
			if (useMigemo) {
				regex = main.compileMigemo(search);
			}
			int index;
			if (reverse) {
				index = find(search, regex, start, -1);
			} else {
				index = find(search, regex, start, 1);
			}
			if (index == -1) {
				table.deselectAll();
				return;
			}
			table.setSelection(index);

		}

		/**
		 * 
		 * @param search 
		 * @param m K\
		 * @param start Jnʒu
		 * @param direction (1)܂͏(-1)
		 * @return ʒuB݂Ȃꍇ-1B
		 */
		private int find(String search, Matcher m, int start, int direction) {
			String lower = search.toLowerCase();

			final int columnCount = table.getColumnCount();
			final int itemCount = table.getItemCount();
			for (int i = start; 0 <= i && i < itemCount; i += direction) {
				TableItem item = table.getItem(i);
				for (int c = 0; c < columnCount; c++) {
					String s = item.getText(c).toLowerCase();
					if (m != null) {
						m.reset(s);
						if (m.find()) {
							return i;
						}
					} else {
						if (s.indexOf(lower) != -1) {
							return i;
						}
					}
				}
			}
			return -1;
		}

		/**
		 * e[u\z܂B
		 * @param parent eEBWFbg
		 * @param actions CxgXi[
		 */
		private void createTable(Composite parent, TabFolderActions actions) {
			table = new Table(parent, SWT.MULTI | SWT.FULL_SELECTION
					| SWT.BORDER | SWT.H_SCROLL | SWT.V_SCROLL);
			table.setHeaderVisible(true);

			TableComparator comparator = new MessageComparator();
			TableSortListener sortListener = new TableSortListener(comparator);

			SWTUtils.createTableColumn(table, sortListener, 80,
					"SWTMessageStorage.name"); //$NON-NLS-1$
			SWTUtils.createTableColumn(table, sortListener, 120,
					"SWTMessageStorage.date"); //$NON-NLS-1$
			SWTUtils.createTableColumn(table, sortListener, 150,
					"SWTMessageStorage.body"); //$NON-NLS-1$

			GridData data = new GridData();
			data.verticalAlignment = SWT.FILL;
			data.horizontalAlignment = SWT.FILL;
			data.grabExcessHorizontalSpace = true;
			data.grabExcessVerticalSpace = true;
			data.horizontalSpan = 3;

			table.setLayoutData(data);

			// Xi[
			actions.registerAndAddListener(table, "doTableSelection"); //$NON-NLS-1$

			table.addMouseListener(new TableMouseListener());

		}

		/**
		 * IMessageɉĉE̕\ς܂B
		 * @param msg \郁bZ[WBnullB
		 */
		void setDetail(Message msg) {
			if (msg == null) {
				setFrom(""); //$NON-NLS-1$
				setText(""); //$NON-NLS-1$
				return;
			}
			setFrom(msg.userName);
			if (msg.isPasswordRequired) {
				setText(mm.get("SWTMessageStorage.thisMessageIsLocked")); //$NON-NLS-1$
			} else {
				setText(msg.message);
			}
		}

		/**
		 * FromɃeLXgw肵܂B
		 * @param label \eLXg
		 */
		private void setFrom(String label) {
			from.setText(mm.get("SWTMessageStorage.from2", //$NON-NLS-1$
					new Object[] { label.replaceAll("&", "&&") })); //$NON-NLS-1$ //$NON-NLS-2$
		}

		/**
		 * {ݒ肵܂B
		 * @param text \eLXg
		 */
		private void setText(String text) {
			this.text.setText(text);
		}

		/**
		 * bZ[We[uɒǉ܂B
		 * @param msg bZ[W
		 */
		void addMessage(Message msg) {
			TableItem item = new TableItem(table, SWT.NULL);
			item.setText(msg.getRowStrings());
			item.setData(msg);
		}

		/**
		 * TabFolder̃CxgXi[
		 */
		public class TabFolderActions extends ListenerDelegate {
			/**
			 * O
			 * @see jp.takayuki.swt.util.ListenerDelegate#handleException(java.lang.Exception)
			 */
			protected void handleException(Exception e) {
				broker.notifyEvent(Events.UnexpectedErrorOccured, e);
			}

			/**
			 * uJv{^
			 * @param event
			 */
			public void doOpenButton(SelectionEvent event) {
				// IꂽbZ[W擾
				int[] indices = table.getSelectionIndices();
				if (indices.length == 0) {
					return;
				}
				// bZ[WSWTMessageWindowɂĕ\
				for (int i = 0; i < indices.length; i++) {
					Message msg = (Message) table.getItem(indices[i]).getData();
					if (msg.isPasswordRequired
							&& msg.messageClass == Message.RECEIVED) {
						// pX[hKvȎMbZ[W͗ԂŊJ
						msg.windowStyle = Message.BACK;
					}
					SWTMessageWindow win = new SWTMessageWindow(main, msg,
							pref, facade, msgLogger);
					// ƗIvV:fȂ
					win.open();
					win.activate();
				}
			}

			/**
			 * u폜v{^
			 * @param event
			 */
			public void doDeleteButton(SelectionEvent event) {
				// IꂽbZ[W擾
				int[] indices = table.getSelectionIndices();
				if (indices.length == 0) {
					return;
				}

				TableItem[] items = table.getSelection();
				for (int i = 0; i < items.length; i++) {
					Message msg = (Message) items[i].getData();
					SWTMessageStorage.super.delete(msg);
				}

				// \̍XV
				table.remove(indices);

				// \
				setDetail(null);
			}

			/**
			 * e[uACȇI
			 * @param event
			 */
			public void doTableSelection(SelectionEvent event) {
				int s = table.getSelectionIndex();
				if (s == -1) {
					setDetail(null); // Zbg
					return;
				}

				Message msg = (Message) table.getItem(s).getData();
				setDetail(msg);
			}

		}

		/**
		 * MouseAdapter: e[ũ}EXCxg
		 */
		private class TableMouseListener extends MouseAdapter {
			public void mouseDown(MouseEvent event) {
				TableItem item = table.getItem(new Point(event.x, event.y));
				if (item != null) {
					return;
				}
				table.deselectAll();
				setDetail(null);
			}

		}

		/**
		 * e[u\[g郊Xi[
		 */
		private class TableSortListener extends SelectionAdapter {
			/**
			 * rC^[tF[X
			 */
			TableComparator comparator;

			/**
			 * RXgN^
			 * @param comparator rC^[tF[X
			 */
			TableSortListener(TableComparator comparator) {
				this.comparator = comparator;
			}

			public void widgetSelected(SelectionEvent event) {

				// \[g悤Ƃ肷
				int index = SWTUtils.getColumnIndex((TableColumn) event.widget);

				// IԂ
				table.deselectAll();
				// f[^oTableViewable̔zɂ
				TableItem[] items = table.getItems();
				TableViewable[] tvs = new TableViewable[items.length];
				for (int i = 0; i < items.length; i++) {
					tvs[i] = (TableViewable) items[i].getData();
				}

				// e[ȕԂɉă\[g߂
				comparator.setSortColumn(index);
				// \[g
				Arrays.sort(tvs, comparator);

				// e[ũeLXgXV
				for (int i = 0; i < items.length; i++) {
					TableItem item = table.getItem(i);
					item.setText(tvs[i].getRowStrings());
					item.setData(tvs[i]);
				}
			}
		}
	}

	/**
	 * CxgXi[
	 */
	public class Actions extends ListenerDelegate {

		/**
		 * O
		 * @see jp.takayuki.swt.util.ListenerDelegate#handleException(java.lang.Exception)
		 */
		protected void handleException(Exception e) {
			broker.notifyEvent(Events.UnexpectedErrorOccured, e);
		}

		/**
		 * j[t@C
		 * @param event
		 */
		public void doMenuFileClose(SelectionEvent event) {
			hide();
		}

		/**
		 * 
		 * @param event Cxg
		 */
		public void doFindNext(SelectionEvent event) {
			TabFolderContent current = getCurrentTabFolderContent();
			Table table = current.table;
			int count = table.getItemCount();
			if (count == 0) {
				return;
			}
			int index = table.getSelectionIndex();
			if (0 > index || index >= count - 1) {
				index = 0;
			} else {
				index++;
			}
			current.findItem(current.searchText.getText(), index, false);
		}

		/**
		 * O
		 * @param event Cxg
		 */
		public void doFindPrevious(SelectionEvent event) {
			TabFolderContent current = getCurrentTabFolderContent();
			Table table = current.table;
			int count = table.getItemCount();
			if (count == 0) {
				return;
			}
			int index = table.getSelectionIndex();
			if (0 >= index || index >= count) {
				index = count - 1;
			} else {
				index--;
			}
			current.findItem(current.searchText.getText(), index, true);
		}

		/**
		 * [}(migemo)
		 * @param event Cxg
		 */
		public void doCheckMigemo(SelectionEvent event) {
			MenuItem item = (MenuItem) event.widget;
			useMigemo = item.getSelection();
			if (useMigemo) {
				main.attachMigemoDictionaries(false);
			} else {
				main.detachMigemoDictionaries();
			}
		}

		/**
		 * I{^
		 * <p>
		 * }b`郁bZ[WׂđI܂B
		 * @param event Cxg
		 */
		public void doSelectAllMatchedItems(SelectionEvent event) {
			selectAllMatchedMessages();
		}

	}
}
