package petaclone.swt;

import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLEncoder;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;

import jp.takayuki.logging.Logger;
import jp.takayuki.swt.util.SWTUtils;

import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.events.ControlAdapter;
import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.FocusAdapter;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Link;
import org.eclipse.swt.widgets.ScrollBar;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;

/**
 * QualityFeedback
 * @author Takayuki Shoji
 */
public class SWTQualityFeedback {

	/**
	 * logger
	 */
	private static final Logger logger = Logger
			.getLogger(SWTQualityFeedback.class);

	/**
	 * Shell
	 */
	private Shell shell = null; //  @jve:decl-index=0:visual-constraint="10,10"

	/**
	 * Explanation
	 */
	private Link link = null;

	/**
	 * ScrolledComposite
	 */
	private ScrolledComposite scrolledComposite = null;

	/**
	 * ScrolledComposite content composite
	 */
	private Composite content = null;

	/**
	 * Button area composite
	 */
	private Composite buttonArea = null;

	/**
	 * Send button
	 */
	private Button sendButton = null;

	/**
	 * Cancel button
	 */
	private Button cancelButton = null;

	/**
	 * tab focus behavior
	 */
	private FocusListener scroller = new Scroller();

	/**
	 * explanation area padding 
	 */
	private Composite explanationArea = null;

	/**
	 * Transmission data<br>
	 * key: String name<br>
	 * val: Text value
	 */
	private Map map = new LinkedHashMap();

	/**
	 * HTTPX|X
	 */
	private String response;

	/**
	 * This method initializes scrolledComposite	
	 * @param fixWidth ̒ŌŒ肷ꍇtrue
	 */
	private void createScrolledComposite(boolean fixWidth) {
		GridData gridData = new GridData();
		gridData.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridData.verticalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridData.grabExcessHorizontalSpace = true;
		gridData.grabExcessVerticalSpace = true;
		gridData.heightHint = 300;
		if (fixWidth)
			gridData.widthHint = explanationArea.getSize().x;
		scrolledComposite = new ScrolledComposite(shell, SWT.V_SCROLL
				| SWT.H_SCROLL | SWT.BORDER);
		scrolledComposite.setAlwaysShowScrollBars(false);
		scrolledComposite.setLayoutData(gridData);
		scrolledComposite.setLayout(new FillLayout());
		scrolledComposite.setExpandHorizontal(true);
		scrolledComposite.setExpandVertical(true);

		createComposite();
		scrolledComposite.setContent(content);
		scrolledComposite.addControlListener(new ControlAdapter() {
			public void controlResized(ControlEvent arg0) {
				Rectangle rect = scrolledComposite.getClientArea();
				scrolledComposite.setMinSize(content.computeSize(rect.width,
						SWT.DEFAULT));
				ScrollBar bar = scrolledComposite.getVerticalBar();
				bar.setIncrement(8);
				bar.setPageIncrement(rect.height);
			}
		});
	}

	/**
	 * This method initializes composite	
	 *
	 */
	private void createComposite() {

		GridLayout gridLayout = new GridLayout();
		gridLayout.numColumns = 2;

		content = new Composite(scrolledComposite, SWT.NONE);
		content.setLayout(gridLayout);

	}

	/**
	 * Add transmission element.
	 * @param labelText Label text
	 * @param value Content value
	 * @param readOnly set false to create editable text box
	 */
	public void addElement(String labelText, String value, boolean readOnly) {
		if (value == null) {
			value = ""; //$NON-NLS-1$
		}

		GridData gridData = new GridData();
		gridData.horizontalAlignment = GridData.FILL;
		gridData.grabExcessHorizontalSpace = true;

		Label label = new Label(content, SWT.NONE);
		label.setText(labelText);

		int ro = readOnly ? SWT.READ_ONLY : 0;

		Text text = new Text(content, SWT.BORDER | ro | SWT.MULTI);

		if (!readOnly) {
			gridData.heightHint = text.getLineHeight() * 3;
		}

		text.setLayoutData(gridData);
		text.setText(value);
		Color color = Display.getDefault().getSystemColor(
				SWT.COLOR_LIST_BACKGROUND);
		text.setBackground(color);

		text.addFocusListener(scroller);

		// f[^ɒǉ
		map.put(labelText, text);
	}

	/**
	 * Add read-only transmission element.
	 * @param labelText Label text
	 * @param value Content value
	 */
	public void addElement(String labelText, String value) {
		addElement(labelText, value, true);
	}

	/**
	 * Add read-only element that its value is gotten from system property.
	 * @param key key for system property.
	 */
	public void addSystemPropertyElement(String key) {
		addElement(key, getMaskedProperty(key));
	}

	/**
	 * Get url encoded string.
	 * @param encoding encoding
	 * @return QUERY_STRING
	 * @throws UnsupportedEncodingException 
	 */
	public String getURLEncoded(String encoding)
			throws UnsupportedEncodingException {

		StringBuffer sb = new StringBuffer();

		Iterator iter = map.entrySet().iterator();
		while (iter.hasNext()) {
			Entry entry = (Entry) iter.next();
			String key = (String) entry.getKey();
			Text text = (Text) entry.getValue();
			String val = text.getText();
			String encoded = URLEncoder.encode(key, encoding);
			sb.append(encoded);
			sb.append('=');
			encoded = URLEncoder.encode(val, encoding);
			sb.append(encoded);
			if (iter.hasNext()) {
				sb.append('&');
			}
		}

		return sb.toString();
	}

	/**
	 * Tab focus scroll action
	 */
	private class Scroller extends FocusAdapter {
		public void focusGained(org.eclipse.swt.events.FocusEvent e) {
			Control control = (Control) e.widget;

			int textY = control.getLocation().y;
			int textH = control.getSize().y;

			int height = scrolledComposite.getSize().y;
			int top = scrolledComposite.getOrigin().y;
			/*
			 * Border is drawn at inside of height.
			 */
			int parentBw = scrolledComposite.getBorderWidth();

			if (textY + parentBw < top) {
				scrolledComposite.setOrigin(0, textY - parentBw);
			} else if (textY + textH + parentBw > top + height) {
				scrolledComposite.setOrigin(0, textY + textH - height
						+ parentBw * 2);
			}
		}
	}

	/**
	 * This method initializes button area composite.
	 */
	private void createButtonArea() {
		GridData gridData = new GridData();
		gridData.horizontalAlignment = org.eclipse.swt.layout.GridData.CENTER;
		gridData.grabExcessHorizontalSpace = true;
		buttonArea = new Composite(shell, SWT.NONE);
		buttonArea.setLayout(new RowLayout());
		buttonArea.setLayoutData(gridData);
		sendButton = new Button(buttonArea, SWT.NONE);
		sendButton.setText("&Send"); //$NON-NLS-1$
		cancelButton = new Button(buttonArea, SWT.NONE);
		cancelButton.setText("&Cancel"); //$NON-NLS-1$
		cancelButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				shell.close();
			}
		});
		shell.setDefaultButton(sendButton);
	}

	/**
	 * Create explanation composite.
	 */
	private void createExplanationArea() {
		FillLayout fillLayout = new FillLayout();
		fillLayout.marginWidth = 10;
		fillLayout.marginHeight = 10;
		GridData gridData3 = new GridData();
		gridData3.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		explanationArea = new Composite(shell, SWT.NONE);
		explanationArea.setLayoutData(gridData3);
		explanationArea.setLayout(fillLayout);
		Color color = Display.getDefault().getSystemColor(
				SWT.COLOR_LIST_BACKGROUND);
		explanationArea.setBackground(color);

		link = new Link(explanationArea, SWT.NONE);
		link.setText("Sorry, Application error has detected."); //$NON-NLS-1$
		link.setBackground(color);
	}

	/**
	 * Sets an explanation text.
	 * Specified text is set to Link widget.
	 * @param text explanation text(link available)
	 */
	public void setExplanation(String text) {
		checkLink();
		link.setText(text);
	}

	/**
	 * ̃N̑ICxg
	 * CxgXi[ǉ܂B
	 * 
	 * @param listener Xi[
	 */
	public void addLinkSelectionListener(SelectionListener listener) {
		checkLink();
		link.addSelectionListener(listener);
	}

	/**
	 * EBWFbg̐mF܂B
	 */
	private void checkLink() {
		if (link == null) {
			throw new IllegalStateException("Link widget is not initialized."); //$NON-NLS-1$
		}
	}

	/**
	 * M{^̑ICxgCxgXi[ǉ܂B
	 * 
	 * @param listener Xi[
	 */
	public void addSendButtonListener(SelectionListener listener) {
		if (sendButton == null) {
			throw new IllegalStateException("Button widget is not initialized."); //$NON-NLS-1$
		}
		sendButton.addSelectionListener(listener);
	}

	/**
	 * Returns the Shell.
	 * @return Shell
	 */
	public Shell getShell() {
		return shell;
	}

	/**
	 * Close window.
	 */
	public void close() {
		shell.close();
	}

	/**
	 * Sets the button text.
	 * @param sendButton send button text
	 * @param cancelButton cancel button text
	 */
	public void setButtonText(String sendButton, String cancelButton) {
		this.sendButton.setText(sendButton);
		this.cancelButton.setText(cancelButton);
	}

	/**
	 * open the window.
	 */
	public void open() {
		shell.pack();
		shell.setMinimumSize(shell.getSize().x, -1);
		shell.open();
	}

	/**
	 * event loop utility method.
	 */
	public void doEventLoop() {
		Display display = Display.getDefault();
		while (!shell.isDisposed()) {
			try {
				if (!display.readAndDispatch())
					display.sleep();
			} catch (Throwable th) {
				logger
						.error(
								"Exception occurred in event loop at Quality Feedback.", th); //$NON-NLS-1$
			}
		}
	}

	/**
	 * Initialize the widget.
	 * @param fixWidth Set true to fix a window minimum width 
	 * by explain area width.
	 */
	public void initialize(boolean fixWidth) {

		GridLayout layout = new GridLayout();
		layout.marginWidth = 0;
		layout.marginHeight = 0;

		shell = new Shell(SWT.TITLE | SWT.CLOSE | SWT.RESIZE);
		shell.setText("Quality feedback"); //$NON-NLS-1$
		shell.setLayout(layout);

		createExplanationArea();
		createScrolledComposite(fixWidth);
		createButtonArea();
	}

	/**
	 * Enable or disable send button.
	 * @param enabled Set true to enable the send button.
	 */
	public void setSendButtonEnabled(boolean enabled) {
		this.sendButton.setEnabled(enabled);
	}

	/**
	 * M
	 * @param feedbackUrl MURL 
	 * @throws IOException 
	 */
	public void send(String feedbackUrl) throws IOException {

		String data = getURLEncoded("UTF-8"); //$NON-NLS-1$

		OutputStream os = null;
		HttpURLConnection conn = null;
		try {
			URL url = new URL(feedbackUrl);
			conn = (HttpURLConnection) url.openConnection();
			conn.setAllowUserInteraction(false);
			conn.setDoOutput(true);
			conn.setDoInput(true);
			conn.setRequestMethod("POST"); //$NON-NLS-1$
			os = conn.getOutputStream();
			os.write(data.getBytes());
			os.flush();
			this.response = "Response:" + conn.getResponseCode() + " " //$NON-NLS-1$ //$NON-NLS-2$
					+ conn.getResponseMessage();
		} finally {
			try {
				if (os != null) {
					os.close();
				}
			} finally {
				if (conn != null) {
					conn.disconnect();
				}
			}
		}

	}

	/**
	 * Mʂ擾܂B
	 * @return HTTPX|XR[hƃbZ[W
	 */
	public String getSendResponse() {
		return response;
	}

	/**
	 * Gets a property value and mask username with *.
	 * @param key property key
	 * @return masked string
	 */
	public static String getMaskedProperty(String key) {
		String value = System.getProperty(key);
		if (value == null) {
			return null;
		}
		String name = System.getProperty("user.name"); //$NON-NLS-1$
		if (name == null) {
			return value;
		}
		String mask = "*"; //$NON-NLS-1$
		for (int i = 1; i < name.length(); i++) {
			mask += "*"; //$NON-NLS-1$
		}
		return value.replaceAll(name, mask);
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		Display display = Display.getDefault();

		// initialize
		final SWTQualityFeedback thisClass = new SWTQualityFeedback();
		thisClass.initialize(true);
		thisClass.getShell().setText("itB[hobN"); //$NON-NLS-1$
		thisClass.setButtonText("M(&S)", "LZ(&C)"); //$NON-NLS-1$ //$NON-NLS-2$

		// customize
		thisClass.setExplanation("AvP[Vُ̈o܂Bꕔ̋@\~Ă\܂B\n" //$NON-NLS-1$
				+ "sւĐ\󂠂܂B\n" //$NON-NLS-1$
				+ "Jł́Aُ̋NvÕX^bNg[XWĂ܂B\n" //$NON-NLS-1$
				+ "L̏𑗐MĂƂɂAƕiPsƂł܂B\n" //$NON-NLS-1$
				+ "͂낵肢܂B\n" //$NON-NLS-1$
				+ "HTTPgpĕőM܂"); //$NON-NLS-1$

		thisClass.addElement("Rg", "Ȏ쓙LB", false); //$NON-NLS-1$ //$NON-NLS-2$
		thisClass.addElement("Version", "1.0.0"); //$NON-NLS-1$//$NON-NLS-2$
		thisClass
				.addElement(
						"Stack trace", //$NON-NLS-1$
						"java.net.BindException: Address already in use: JVM_Bind\n" //$NON-NLS-1$
								+ "	at java.net.PlainSocketImpl.socketBind(Native Method)\n" //$NON-NLS-1$
								+ "	at java.net.PlainSocketImpl.bind(Unknown Source)\n" //$NON-NLS-1$
								+ "	at java.net.ServerSocket.bind(Unknown Source)\n" //$NON-NLS-1$
								+ "	at java.net.ServerSocket.<init>(Unknown Source)\n" //$NON-NLS-1$
								+ "	at petaclone.ipmsg.IPMTCPServer.bind(IPMTCPServer.java:107)\n" //$NON-NLS-1$
								+ "	at petaclone.ipmsg.IPMessenger.begin(IPMessenger.java:260)\n" //$NON-NLS-1$
								+ "	at petaclone.MessengerFacade.setup(MessengerFacade.java:155)\n" //$NON-NLS-1$
								+ "	at petaclone.MessengerFacade.start(MessengerFacade.java:133)\n" //$NON-NLS-1$
								+ "	at petaclone.BaseMain.startMessengers(BaseMain.java:173)\n" //$NON-NLS-1$
								+ "	at petaclone.swt.SWTMainWindow.open(SWTMainWindow.java:201)\n" //$NON-NLS-1$
								+ "	at petaclone.swt.SWTMainWindow.main(SWTMainWindow.java:918)\n"); //$NON-NLS-1$
		thisClass
				.addElement("java.version", System.getProperty("java.version")); //$NON-NLS-1$ //$NON-NLS-2$
		thisClass.addElement("SWT Version", Integer.toString(SWT.getVersion())); //$NON-NLS-1$
		thisClass.addElement("SWT Platform", SWT.getPlatform()); //$NON-NLS-1$

		// Memory
		long totalMemory = Runtime.getRuntime().totalMemory();
		long freeMemory = Runtime.getRuntime().freeMemory();
		long committedMemory = totalMemory - freeMemory;
		thisClass.addElement("Total Memory", Long.toString(totalMemory)); //$NON-NLS-1$
		thisClass.addElement("Free Memory", Long.toString(freeMemory)); //$NON-NLS-1$
		thisClass
				.addElement("Committed Memory", Long.toString(committedMemory)); //$NON-NLS-1$

		// System
		thisClass.addElement("os.name", System.getProperty("os.name")); //$NON-NLS-1$ //$NON-NLS-2$
		thisClass.addElement("os.version", System.getProperty("os.version")); //$NON-NLS-1$ //$NON-NLS-2$
		thisClass.addElement("os.arch", System.getProperty("os.arch")); //$NON-NLS-1$ //$NON-NLS-2$
		thisClass.addElement("sun.os.patch.level", System //$NON-NLS-1$
				.getProperty("sun.os.patch.level")); //$NON-NLS-1$

		//Java
		thisClass.addElement("java.runtime.name", System //$NON-NLS-1$
				.getProperty("java.runtime.name")); //$NON-NLS-1$
		thisClass.addElement("java.runtime.version", System //$NON-NLS-1$
				.getProperty("java.runtime.version")); //$NON-NLS-1$
		thisClass.addElement("java.vm.vendor", System //$NON-NLS-1$
				.getProperty("java.vm.vendor")); //$NON-NLS-1$
		thisClass
				.addElement("java.vm.name", System.getProperty("java.vm.name")); //$NON-NLS-1$ //$NON-NLS-2$
		thisClass
				.addElement("java.vm.info", System.getProperty("java.vm.info")); //$NON-NLS-1$ //$NON-NLS-2$
		thisClass.addElement("java.home", getMaskedProperty("java.home")); //$NON-NLS-1$ //$NON-NLS-2$
		thisClass.addElement("java.library.path", //$NON-NLS-1$
				getMaskedProperty("java.library.path")); //$NON-NLS-1$

		thisClass.addSendButtonListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				try {
					String encoded = thisClass.getURLEncoded("euc-jp"); //$NON-NLS-1$
					System.out.println(encoded);

					SWTUtils.showInformation(thisClass.getShell(), "itB[hobN", //$NON-NLS-1$
							"͂肪Ƃ܂B"); //$NON-NLS-1$
					thisClass.close();
				} catch (UnsupportedEncodingException e1) {
					e1.printStackTrace();
				}
			}
		});
		thisClass.addLinkSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				System.out.println("NbN=" + e.text); //$NON-NLS-1$
			}
		});

		// open the window
		thisClass.open();

		// event loop
		thisClass.doEventLoop();

		// dispose the display object
		display.dispose();
	}
}
