package petaclone.swt.logic;

import java.util.regex.Matcher;

import jp.takayuki.swt.util.ListenerDelegate;
import jp.takayuki.swt.util.SWTMenuUtils;
import jp.takayuki.swt.util.SWTUtils;
import jp.takayuki.tools.MessageMap;
import jp.takayuki.tools.Os;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.events.KeyAdapter;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.events.ShellAdapter;
import org.eclipse.swt.events.ShellEvent;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.graphics.Region;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.widgets.Widget;

import petaclone.AppPreferences;
import petaclone.EventBroker;
import petaclone.Events;
import petaclone.Message;
import petaclone.MessageFoundListener;
import petaclone.swt.ArrowDownBehavior;
import petaclone.swt.FindUtils;
import petaclone.swt.Findable;
import petaclone.swt.SWTMainWindow;
import petaclone.swt.SWTMessageWindow;
import petaclone.swt.SWTMessageWindowManager;

/**
 * bZ[W_CAO
 * 
 * @author Takayuki Shoji
 */
public class SWTMessageSearch extends AbstractSWTEventListener implements
		Findable, MessageFoundListener {

	/**
	 * bZ[W\[X
	 */
	private static final MessageMap mm = MessageMap.getMessageMap();

	/**
	 * eEBhE
	 */
	private final Shell parent;

	/**
	 * AvP[Vݒ
	 */
	private final AppPreferences pref;

	/**
	 * Ώۂ̃bZ[W
	 */
	private final SWTMessageWindowManager desktopMessages;

	/**
	 * fXNgbṽEBhE
	 */
	private Shell shell;

	/**
	 * eLXg{bNX
	 */
	private Text text;

	/**
	 * ʕ\pe[u
	 */
	private Table table;

	/**
	 * EventBroker
	 */
	private final EventBroker broker;

	/**
	 * CxgXi[
	 */
	private Actions actions = new Actions();

	/** [}(migemo)[h */
	private boolean useMigemo;

	/**
	 * CEBhE
	 */
	private final SWTMainWindow main;

	/**
	 * gShell
	 */
	private Shell borderShell;

	/**
	 * @param main CEBhE
	 * @param parent eEBhE
	 * @param pref AvP[Vݒ
	 * @param desktopMessages 
	 * @param broker EventBroker
	 */
	public SWTMessageSearch(SWTMainWindow main, Shell parent,
			AppPreferences pref, SWTMessageWindowManager desktopMessages,
			EventBroker broker) {
		this.main = main;
		this.parent = parent;
		this.pref = pref;
		this.desktopMessages = desktopMessages;
		this.broker = broker;
		this.useMigemo = pref.getUseMigemoInDesktopSearch();

		if (useMigemo) {
			main.attachMigemoDictionaries(true);
		}

		initWidget();
	}

	/**
	 * EBhȄԂvpeBɕۑ܂B
	 */
	public void saveWindowStatus() {
		// migemo
		pref.setUseMigemoInDesktopSearch(useMigemo);

		// size
		Point size = shell.getSize();
		pref.setMessageSearchWidth(size.x);
		pref.setMessageSearchHeight(size.y);
	}

	/**
	 * EBWFbg܂B
	 */
	private void initWidget() {
		shell = new Shell(parent, SWT.CLOSE | SWT.RESIZE
				| SWT.APPLICATION_MODAL);
		if (!Os.isMacOSX())
			shell.setImage(main.getImage("loupe.gif")); //$NON-NLS-1$
		shell.setText(mm.get("SWTMessageSearch.searchMessage")); //$NON-NLS-1$

		GridLayout layout = new GridLayout();
		layout.numColumns = 1;

		shell.setLayout(layout);

		createMenuBar();

		text = FindUtils.createSearchBox(shell, null, false, this);
		text.addKeyListener(new KeyAdapter() {
			public void keyPressed(KeyEvent e) {
				if (e.keyCode == 0x0A || e.keyCode == 0x0D) {
					int index = table.getSelectionIndex();
					if (index == -1) {
						return;
					}
					TableItem item = table.getItem(index);
					focusMessage(item);
				}
			}
		});

		Label label = new Label(shell, SWT.HORIZONTAL);
		label.setText(mm.get("SWTMessageSearch.matchedMessage")); //$NON-NLS-1$

		GridData gridData = new GridData();
		gridData.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridData.verticalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridData.grabExcessHorizontalSpace = true;
		gridData.grabExcessVerticalSpace = true;

		table = new Table(shell, SWT.FULL_SELECTION | SWT.BORDER);
		table.setLayoutData(gridData);

		table.setHeaderVisible(false);
		table.setLinesVisible(false);
		table.addSelectionListener(new SelectionListener() {
			public void widgetDefaultSelected(SelectionEvent e) {
				// tH[JX
				processEvent(e, true);
			}

			public void widgetSelected(SelectionEvent e) {
				// Ԃg\
				processEvent(e, false);
			}

			private void processEvent(SelectionEvent e, boolean defaultSelected) {
				if (e.item == null) {
					int index = table.getSelectionIndex();
					if (index < 0) {
						return;
					}
					e.item = table.getItem(index);
				}
				if (defaultSelected) {
					focusMessage(e.item);
				} else {
					showBorderShell(e.item);
				}
			}
		});
		table.addFocusListener(new FocusListener() {
			public void focusGained(FocusEvent e) {
				int index = table.getSelectionIndex();
				if (index != -1) {
					TableItem item = table.getItem(index);
					showBorderShell(item);
				}
			}

			public void focusLost(FocusEvent e) {
				if (borderShell != null)
					borderShell.setVisible(false);
			}

		});

		text.addKeyListener(new ArrowDownBehavior(table));

		shell.setSize(pref.getMessageSearchWidth(), pref
				.getMessageSearchHeight());
		SWTUtils.moveToCenter(shell);
		shell.addShellListener(new ShellAdapter() {
			public void shellClosed(ShellEvent e) {

				e.doit = false;
				Shell sh = (Shell) e.widget;
				sh.setVisible(false);

				/*
				 * ɊJƂɃe[uɃtH[JX
				 * borderShellNPEɂȂ̂hB
				 */
				text.setFocus();
				closeBorderShell();
			}
		});
	}

	/**
	 * bZ[WEBhEɃtH[JX܂B
	 * 
	 * @param item e[uACe
	 */
	private void focusMessage(Widget item) {
		Message msg = (Message) item.getData();
		shell.close();
		final SWTMessageWindow mw = desktopMessages.get(msg);
		if (mw != null) {
			mw.activate();
		}
	}

	/**
	 * gShell܂B
	 */
	private void closeBorderShell() {
		if (borderShell != null) {
			borderShell.close();
			borderShell = null;
		}
	}

	/**
	 * j[𐶐܂B
	 */
	private void createMenuBar() {
		Menu menuBar = SWTMenuUtils.addMenuBar(shell);

		Menu menuFile = SWTMenuUtils.addCascade(menuBar, mm
				.get("SWTMessageSearch.menu.file")); //$NON-NLS-1$

		SWTMenuUtils.addMenuItem(menuFile, mm
				.get("SWTMessageSearch.menu.file.close"), //$NON-NLS-1$
				"ESC", 0, null, //$NON-NLS-1$
				actions, "doClose"); //$NON-NLS-1$

		Menu menuEdit = SWTMenuUtils.addCascade(menuBar, mm
				.get("SWTMessageSearch.menu.edit")); //$NON-NLS-1$

		SWTMenuUtils.addMenuItemCheckbox(menuEdit, mm
				.get("SWTMessageSearch.menu.edit.migemo"), //$NON-NLS-1$
				"Ctrl+K", SWT.MOD1 | 'K', null, //$NON-NLS-1$
				actions, "doCheckMigemo", pref.getUseMigemoInDesktopSearch()); //$NON-NLS-1$
	}

	/**
	 * SelectionListener̃Zbg
	 */
	public final class Actions extends ListenerDelegate {

		/**
		 * O
		 * @see jp.takayuki.swt.util.ListenerDelegate#handleException(java.lang.Exception)
		 */
		protected void handleException(Exception e) {
			broker.notifyEvent(Events.UnexpectedErrorOccured, e);
		}

		/**
		 * @param event Cxg
		 */
		public void doClose(SelectionEvent event) {
			shell.close();
		}

		/**
		 * [}(migemo)
		 * @param event Cxg
		 */
		public void doCheckMigemo(SelectionEvent event) {
			MenuItem item = (MenuItem) event.widget;
			useMigemo = item.getSelection();
			if (useMigemo) {
				main.attachMigemoDictionaries(false);
			} else {
				main.detachMigemoDictionaries();
			}
		}
	}

	/**
	 * EBhE\ăeLXg{bNXɃtH[JX܂B
	 */
	public void open() {

		String search = text.getText();
		if (search.length() > 0) {
			findItem(search, 0, false);
		}

		shell.setVisible(true);
		shell.setActive();
		text.setFocus();
		text.selectAll();

		// g
		Display display = Display.getDefault();
		borderShell = new Shell(display, SWT.NO_TRIM | SWT.ON_TOP);
		borderShell.setBackground(display.getSystemColor(SWT.COLOR_RED));
		borderShell.setVisible(false);
	}

	protected void handleEvent(String eventName, Object event) {
		desktopMessages.synchAll();
		open();
	}

	public void findItem(String search, int startIndex, boolean reverse) {

		table.removeAll();
		if (search == null || search.length() == 0) {
			return;
		}
		// migemoŐK\ɕϊ
		Matcher regex = null;
		if (useMigemo) {
			regex = main.compileMigemo(search);
		} else {
			search = search.toLowerCase();
		}

		desktopMessages.find(search, regex, this);
	}

	public void found(Message message) {
		TableItem item = new TableItem(table, SWT.NONE);
		item.setText(new String[] { message.getWindowDescription() });
		item.setData(message);
		item.setImage(main.getImage("message.gif")); //$NON-NLS-1$
		if (table.getItemCount() == 1) {
			table.setSelection(0);
		}
	}

	/**
	 * EBhËʒu܂B
	 * 
	 * FIXME borderShell̏ʉ
	 * 
	 * @param item e[uACe
	 */
	private void showBorderShell(Widget item) {
		if (borderShell == null) {
			return;
		}
		borderShell.setVisible(false);

		Message msg = (Message) item.getData();
		SWTMessageWindow mw = desktopMessages.get(msg);
		if (mw == null) {
			return;
		}
		Rectangle shRect = mw.getCurrentShellRectangle();

		// g̑
		final int w = 5;

		Rectangle outer = new Rectangle(0, 0, shRect.width + w + w,
				shRect.height + w + w);
		Rectangle inner = new Rectangle(w, w, shRect.width, shRect.height);

		Region reg = new Region();
		reg.add(outer);
		reg.subtract(inner);

		borderShell.setRegion(reg);
		borderShell.setSize(outer.width, outer.height);
		borderShell.setLocation(shRect.x - w, shRect.y - w);
		borderShell.setVisible(true);
	}

}
